/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#ifndef _RFUTIL_H_
#define _RFUTIL_H_

#define SQR(a) ((a) == 0.0 ? 0.0 : (a)*(a))

#define ISQR(a) ((a) == 0 ? 0 : (a)*(a))

#define IMAX(a,b) ((a) > (b) ? (a): (b))

#define IMIN(a,b) ((a) < (b) ? (a) : (b))

#define IABS(x) (((x) < 0) ? (-(x)) : (x))

#ifdef __cplusplus
extern "C" {
#endif

int get_num_allocations();
void *rf_malloc(size_t size);
//void *rf_calloc(size_t num, size_t size);
void rf_free(void *memblock);

void rferror(char error_text[]);
float *fvector(long nl, long nh);
int *ivector(long nl, long nh);
unsigned int *uivector(long nl, long nh);
unsigned char *ucvector(long nl, long nh);
unsigned short *usvector(long nl, long nh);
short *svector(long nl, long nh);
unsigned char *ucsubvector(unsigned char *aa,long oldnl,long oldnh,long newnl);
float **fmatrix(long nrl, long nrh, long ncl, long nch);
double **dmatrix(long nrl, long nrh, long ncl, long nch);
int **imatrix(long nrl, long nrh, long ncl, long nch);
unsigned char **ucmatrix(long nrl, long nrh, long ncl, long nch);
unsigned short **usmatrix(long nrl, long nrh, long ncl, long nch);
short **smatrix(long nrl, long nrh, long ncl, long nch);
float **fsubmatrix(float **aa, long oldrl, long oldrh, long oldcl,
                   long oldch, long newrl, long newcl);
int **isubmatrix(int **aa, long oldrl, long oldrh, long oldcl,
                   long oldch, long newrl, long newcl);
void move_isubmatrix(int **m,int **aa,
                     long oldrl, long oldrh,
                     long oldcl,
                     long oldch, long newrl, long newcl);
unsigned char **ucsubmatrix(unsigned char **aa, long oldrl, long oldrh,
                            long oldcl,
                   long oldch, long newrl, long newcl);
void move_ucsubmatrix(unsigned char **m,unsigned char **aa,
                     long oldrl, long oldrh,
                     long oldcl,
                     long oldch, long newrl, long newcl);
void free_fvector(float *v, long nl, long nh);
void free_ivector(int *v, long nl, long nh);
void free_uivector(unsigned int *v, long nl, long nh);
void free_ucvector(unsigned char *v, long nl, long nh);
void free_usvector(unsigned short *v, long nl, long nh);
void free_svector(short *v, long nl, long nh);
void free_ucsubvector(unsigned char *aa,long newnl,long newnh);
void free_fmatrix(float **m, long nrl, long nrh, long ncl, long nch);
void free_dmatrix(double **m, long nrl, long nrh, long ncl, long nch);
void free_imatrix(int **m, long nrl, long nrh, long ncl, long nch);
void free_ucmatrix(unsigned char **m, long nrl, long nrh, long ncl, long nch);
void free_usmatrix(unsigned short **m, long nrl, long nrh, long ncl, long nch);
void free_smatrix(short **m, long nrl, long nrh, long ncl, long nch);
void free_submatrix(float **m, long nrl, long nrh, long ncl, long nch);
void free_isubmatrix(int **m, long nrl, long nrh, long ncl, long nch);
void free_ucsubmatrix(unsigned char **m, long nrl, long nrh, long ncl,
                      long nch);
int get_mem2Ddouble(double ***array2D, int rows, int columns);
int get_mem1Ddouble(double **array1D, int raws);
void free_mem2Ddouble(double **array2D);
void free_mem1Ddouble(double *array1D);
#ifdef __cplusplus
}
#endif

#endif /* _RFUTIL_H_ */
