/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/



#if !defined(_TNTMDEC_H_)
#define _TNTMDEC_H_

#include <stdio.h>
#include "getbits.h"
#include "types.h"

enum {I_PICTURE,P_PICTURE,B_PICTURE};

enum {MODE_SKIP,MODE_INTRA,MODE_16x16,MODE_16x8,MODE_8x16,MODE_8x8,MODE_DIRECT,MODE_LMB32,MODE_LMB64};

extern const int corr_qpc_8x8[42];


#define MAX_B_LEVEL 3
#define MAX_GOP_LENGTH (1<<MAX_B_LEVEL)

#define NUM_EXT_MODES 25
#define max(a, b) (((a) > (b)) ? (a) : (b))
#define min(a, b) (((a) < (b)) ? (a) : (b))

typedef struct
{
    int rows, cols;
    int crows, ccols;
    int **Y, **U, **V;
} diff_frame;

typedef struct
{
    int **Y, **U, **V;
} diff_mb;

typedef struct
{
    int **data[6];
} coeff_block;

typedef struct
{
    int rows, cols;
    coeff_block **frame;
} coeff_frame;



typedef struct
{
    short int x;
    short int y;

	int interp_indx;

} mvpos;

typedef struct
{
    int mode;                   /* 0=P_16x16,1=P_16x8,2=P_8x16 */
    int qp;
    int ref;

     int tr16flag;



    int svt;

    int planar_flag;

} mb_info;

typedef struct
{
    int x;
    int y;
    int mode;
    int coded_y;

	int interp_indx;


	int xBW;
    int yBW;

	int interp_indxBW;


} block_info;

typedef struct
{
  int uvcbpTable[4];
  int modeTable[3][NUM_EXT_MODES];
  int cbpTable6[6];
  int cbpTable63[3][63];
  int LPTableD[10][128];

  // ADAPT_VLC_NUM
  int mode_vlc_index;
  int last_pos_vlc_index[10];
  int cbp_vlc_index[3];

  int LPTableD4[3][32];

} vlc_contexts;


typedef struct
{
    int         width;
    int         height;
    int         origWidth;
    int         origHeight;
    int         extWidth;
    int         extHeight;
    int         widthInMBs;
    int         heightInMBs;    
    int         numBlev;
    int         numBframes;
    image_frame *orig;    
    image_frame **ref_frame_list;
    image_frame *curr;
    image_frame *ref_frame_pad[2];
    image_frame *ref_frame_qp[2];
    stream      *str;
    int svt;

    int enableSwitched;


    int         tr4flag;
    int         resetContextsFlag;
    vlc_contexts *currContext;        //vlc contexts for adaptation

    int         *posMap;


    int         pPeriod;


	int         bi;

} decoder_info;

typedef struct
{
    int         pic_type;
    int         tr;
    int         flag_parallel;
    int         constrainedIntraFlag; 
    int         multi_ref;
    int         qp;
    int         frame_num;
    int         frame_numP;
    int         num_ref;
    int         chroma_rounding;
    int         num_skipped;
    int         bround;
    int         chroma_mv_round_offs;
    
    block_info  **vmap;
    mb_info     *mbmap;
    int         svt_set;



    int		  final_frame;


} frame_info;

typedef struct
{
    int mode;
    int ref;

	int ref_pred;

    unsigned int modelist;
    int specialBmode;
    unsigned int cbp;
    int ycbp;
    int ucbp;
    int vcbp;
    int uvcbp;
    
    int qp;

	int tr16flag;

    int svt;

    int tr4flag;
    int modelist4[16];
    int ycbp4;
    


    int modelist8[4];


    int intraPredAngle8[4];
    int intraPredDirectionChroma;
    int intraPredAngleChroma;



    int planar_flag;
    int planar_deltaY;
    int planar_deltaU;
    int planar_deltaV;


} macroblock_info;


#endif
