/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#include <stdlib.h>
#include "global.h"
#include "tntmenc.h"
//STATIC #include "rfutil.h"
#include "defines.h"


static const int tctable_8x8[46] =
{
    0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,5,5,6,6,7,8,9,9,10,10,11,11,12,12,13,13,14,14
};

static const int betatable_8x8[42] =
{
    0, 0, 0, 0, 0, 0, 6, 7, 8, 9,10,11,12,13,14,15,16,17,
    18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60,62,64
};


int filter_8x8_vertical(unsigned char mb[24][24],int x, int y, int tc,int beta)
{
    int i,d;

    int delta;

    /* First decide if change is too big, resulting in no filtering. */
    d = abs(mb[2+y][1+x]-2*mb[2+y][2+x]+mb[2+y][3+x])+abs(mb[2+y][4+x]-2*mb[2+y][5+x]+mb[2+y][6+x])+
        abs(mb[5+y][1+x]-2*mb[5+y][2+x]+mb[5+y][3+x])+abs(mb[5+y][4+x]-2*mb[5+y][5+x]+mb[5+y][6+x]);
    if (d < beta)
    {
        for (i=0;i<8;i++)
        {

            delta = min(tc,max(-tc,(13*(mb[i+y][4+x]-mb[i+y][3+x]) + 4*(mb[i+y][5+x]-mb[i+y][2+x]) - 5*(mb[i+y][6+x]-mb[i+y][1+x])+16)>>5));                                      
            mb[i+y][3+x] = min(255,max(0,mb[i+y][3+x]+delta));
            mb[i+y][4+x] = min(255,max(0,mb[i+y][4+x]-delta));
            mb[i+y][2+x] = min(255,max(0,mb[i+y][2+x]+delta/2));
            mb[i+y][5+x] = min(255,max(0,mb[i+y][5+x]-delta/2));
        }
    }
    return 0;
}
int filter_8x8_horizontal(unsigned char mb[24][24],int x,int y,int tc,int beta)
{
    int i,d;

    int delta;

    /* First decide if change is too big, resulting in no filtering. */
    d = abs(mb[y+1][x+2]-2*mb[y+2][x+2]+mb[y+3][x+2])+abs(mb[y+4][x+2]-2*mb[y+5][x+2]+mb[y+6][x+2])+
        abs(mb[y+1][x+5]-2*mb[y+2][x+5]+mb[y+3][x+5])+abs(mb[y+4][x+5]-2*mb[y+5][x+5]+mb[y+6][x+5]);
    if (d < beta)
    {
        for (i=0;i<8;i++)
        {

            delta = min(tc,max(-tc,(13*(mb[y+4][x+i]-mb[y+3][x+i]) + 4*(mb[y+5][x+i]-mb[y+2][x+i]) - 5*(mb[y+6][x+i]-mb[y+1][x+i]) + 16)>>5));                                      
            mb[y+3][x+i] = min(255,max(0,mb[y+3][x+i]+delta));
            mb[y+4][x+i] = min(255,max(0,mb[y+4][x+i]-delta));
            mb[y+2][x+i] = min(255,max(0,mb[y+2][x+i]+delta/2));
            mb[y+5][x+i] = min(255,max(0,mb[y+5][x+i]-delta/2));
        }
    }
    return 0;
}

int filter_8x8_vertical_long(unsigned char **mb,int x,int y,int tc,int beta)
{
    int i,d,d_strong;

    int delta;

    
    /* First decide if change is too big, resulting in no filtering. */
    d = abs(mb[y+2][x+1]-2*mb[y+2][x+2]+mb[y+2][x+3])+abs(mb[y+2][x+4]-2*mb[y+2][x+5]+mb[y+2][x+6])+
        abs(mb[y+5][x+1]-2*mb[y+5][x+2]+mb[y+5][x+3])+abs(mb[y+5][x+4]-2*mb[y+5][x+5]+mb[y+5][x+6]);

	if ( d < beta )
	{
		for (i=0;i<8;i++)
		{
            d_strong = abs(mb[y+i][x+0]-mb[y+i][x+3]) + abs(mb[y+i][x+7]-mb[y+i][x+4]);

            if ( (d_strong < (beta>>3)) && (d<(beta>>2)) && ( abs(mb[y+i][x+3]-mb[y+i][x+4] ) < ((tc*5+1)>>1)) ) //strong filtering 
  		 {

                mb[y+i][x+3]=min(255,max(0,(mb[y+i][x+1]+2*mb[y+i][x+2]+2*mb[y+i][x+3]+2*mb[y+i][x+4]+mb[y+i][x+5]+4)>>3)); 
                mb[y+i][x+4]=min(255,max(0,(mb[y+i][x+2]+2*mb[y+i][x+3]+2*mb[y+i][x+4]+2*mb[y+i][x+5]+mb[y+i][x+6]+4)>>3)); 
			      
                mb[y+i][x+2]=min(255,max(0,(mb[y+i][x+1]+mb[y+i][x+2]+mb[y+i][x+3]+mb[y+i][x+4]+2)>>2));
                mb[y+i][x+5]=min(255,max(0,(mb[y+i][x+3]+mb[y+i][x+4]+mb[y+i][x+5]+mb[y+i][x+6]+2)>>2));

                mb[y+i][x+1]=min(255,max(0,(mb[y+i][x+0]*2+mb[y+i][x+1]*3+mb[y+i][x+2]+mb[y+i][x+3]+mb[y+i][x+4]+4)>>3));
                mb[y+i][x+6]=min(255,max(0,(mb[y+i][x+3]+mb[y+i][x+4]+mb[y+i][x+5]+mb[y+i][x+6]*3+mb[y+i][x+7]*2+4)>>3));

				}
				else 
				{


                delta = min(tc,max(-tc,(13*(mb[y+i][x+4]-mb[y+i][x+3]) + 4*(mb[y+i][x+5]-mb[y+i][x+2]) - 5*(mb[y+i][x+6]-mb[y+i][x+1])+16)>>5));                                      

                mb[y+i][x+3] = min(255,max(0,mb[y+i][x+3]+delta));
                mb[y+i][x+4] = min(255,max(0,mb[y+i][x+4]-delta));
                mb[y+i][x+2] = min(255,max(0,mb[y+i][x+2]+delta/2));
                mb[y+i][x+5] = min(255,max(0,mb[y+i][x+5]-delta/2));

				}
		  }		
   }
    return 0;
}

int filter_8x8_horizontal_long(unsigned char **mb,int x,int y,int tc,int beta)
{
    int i,d,d_strong;

    int delta;


    /* First decide if change is too big, resulting in no filtering. */
    d = abs(mb[y+1][x+2]-2*mb[y+2][x+2]+mb[y+3][x+2])+abs(mb[y+4][x+2]-2*mb[y+5][x+2]+mb[y+6][x+2])+
        abs(mb[y+1][x+5]-2*mb[y+2][x+5]+mb[y+3][x+5])+abs(mb[y+4][x+5]-2*mb[y+5][x+5]+mb[y+6][x+5]);

	if ( d < (beta)) 
	{
		for (i=0;i<8;i++)
		{
            d_strong = abs(mb[y+0][x+i]-mb[y+3][x+i]) + abs(mb[y+7][x+i]-mb[y+4][x+i]);

            if ( (d_strong < (beta>>3)) && (d<(beta>>2)) && ( abs(mb[y+3][x+i]-mb[y+4][x+i]) < ((tc*5+1)>>1) ) ) //strong filtering 
				{
                mb[y+3][x+i]=min(255,max(0,(mb[y+1][x+i]+2*mb[y+2][x+i]+2*mb[y+3][x+i]+2*mb[y+4][x+i]+mb[y+5][x+i]+4)>>3)); 
                mb[y+4][x+i]=min(255,max(0,(mb[y+2][x+i]+2*mb[y+3][x+i]+2*mb[y+4][x+i]+2*mb[y+5][x+i]+mb[y+6][x+i]+4)>>3)); 
		      
                mb[y+2][x+i]=min(255,max(0,(mb[y+1][x+i]+mb[y+2][x+i]+mb[y+3][x+i]+mb[y+4][x+i]+2)>>2));
                mb[y+5][x+i]=min(255,max(0,(mb[y+3][x+i]+mb[y+4][x+i]+mb[y+5][x+i]+mb[y+6][x+i]+2)>>2));

                mb[y+1][x+i]=min(255,max(0,(mb[y+0][x+i]*2+mb[y+1][x+i]*3+mb[y+2][x+i]+mb[y+3][x+i]+mb[y+4][x+i]+4)>>3));
                mb[y+6][x+i]=min(255,max(0,(mb[y+3][x+i]+mb[y+4][x+i]+mb[y+5][x+i]+mb[y+6][x+i]*3+mb[y+7][x+i]*2+4)>>3));        

			  }
				else  // weak filtering

				{


                delta = min(tc,max(-tc,(13*(mb[y+4][x+i]-mb[y+3][x+i]) + 4*(mb[y+5][x+i]-mb[y+2][x+i]) - 5*(mb[y+6][x+i]-mb[y+1][x+i]) + 16)>>5));                                      

                mb[y+3][x+i] = min(255,max(0,mb[y+3][x+i]+delta));
                mb[y+4][x+i] = min(255,max(0,mb[y+4][x+i]-delta));
                mb[y+2][x+i] = min(255,max(0,mb[y+2][x+i]+delta/2));
                mb[y+5][x+i] = min(255,max(0,mb[y+5][x+i]-delta/2));

				}
		  
		  }
    }
    return 0;
}




int filter_chroma_vertical_8x8(unsigned char **mb,int x,int y,int tc)
{
    int i,delta;

    for (i=0;i<8;i++)
    {
        delta = min(tc,max(-tc,(((((mb[y+i][x+4]-mb[y+i][x+3])<<2)+mb[y+i][x+2]-mb[y+i][x+5]+4)>>3))));
        mb[y+i][x+3] = min(255,max(0,mb[y+i][x+3]+delta));
        mb[y+i][x+4] = min(255,max(0,mb[y+i][x+4]-delta));
    }
    return 0;
}

int filter_chroma_horizontal_8x8(unsigned char **mb,int x,int y,int tc)
{
    int i,delta;

    for (i=0;i<8;i++)
    {
        delta = min(tc,max(-tc,(((((mb[y+4][x+i]-mb[y+3][x+i])<<2)+mb[y+2][x+i]-mb[y+5][x+i]+4)>>3))));
        mb[y+3][x+i] = min(255,max(0,mb[y+3][x+i]+delta));
        mb[y+4][x+i] = min(255,max(0,mb[y+4][x+i]-delta));
    }
    return 0;
}


int filter_8x8(encoder_info *encoderInfo,frame_info *frameInfo,image_frame *img)
{
    int i,j,k,l;
    int tc;
    //STATIC unsigned char **block = ucsubmatrix(img->Y,0,7,0,7,0,0);
    int mby,mbx;
    int do_filter=0;
    int alt_tc;
    int vy,vx;
    int first;
    int qp;
    int mbxmax = img->cols>>4;
    int mbpos,bpos,mode;

    int tr16flag;

    int svt;
    mb_info *mbmap2 = frameInfo->mbmap;
   
    block_info *mvmap = frameInfo->mvmap;

    /* Y */
    for (i=0;i<img->rows;i+=16){ // for each row of MBs
        for (j=0;j<img->cols;j+=16){ // for each col of MBs
            mby = i>>4;
		    mbx = j>>4;
            mbpos = mby*mbxmax+mbx;
            
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
            
            tr16flag = mbmap2[mbpos].tr16flag;

            svt = mbmap2[mbpos].svt;
			/* Vertical Edges */
			first = (j>0) ? 0 : 8;
            for (l=first;l<16;l+=8){ // for each col of 8x8 blocks			
                for (k=0;k<16;k+=8){ // for each row of 8x8 blocks								
					do_filter = 0;
					alt_tc = 0;
					vy = (i+k)/8;
					vx = (j+l)/8;
                    bpos = vy*2*mbxmax+vx;
					if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA){
						do_filter = 1;						
                        alt_tc = 4;
					}
					/* Check cbp for coded A, B, C or D  */
					if (!do_filter){
						if (mvmap[bpos-1].coded_y || mvmap[bpos].coded_y){
							do_filter = 1;
						}
					}
					/* Check vector differences */
					if (!do_filter){
						if (abs(mvmap[bpos-1].x-mvmap[bpos].x) >= 4 || abs(mvmap[bpos-1].y-mvmap[bpos].y) >= 4){
							do_filter = 1;
						}
					}
                    if ((mode>=MODE_16x16 && mode<MODE_8x8) && tr16flag==0 && svt==1 && (vx&1)){
                        do_filter = 0;
                    }

                   if (mode==MODE_16x16 && tr16flag && (vx&1)){
						do_filter = 0;
					}			

                    if (mode==MODE_INTRA && tr16flag && (vx&1)){
						do_filter = 0;
					}			


                    if (mode==MODE_LMB64 && (vx%(LMBXSIZE/8))){
						do_filter = 0;
					}
                    if (mode==(MODE_LMB32) && (vx%(LMBXSIZE/16))){
						do_filter = 0;
					}

                  if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-1].planar_flag && !(vx&1))
                  {
                    do_filter = 0;  // A fix for Nokia's planar mode filtering
                  }
             
					if (do_filter){
						tc = tctable_8x8[qp+alt_tc];
                        filter_8x8_vertical_long(img->Y,j+l-4,i+k,tc,betatable_8x8[qp]);	      
					}

				}
			}

			/* Horizontal Edges */			
			first = (i>0) ? 0 : 8;
            for (k=first;k<16;k+=8){ // for each row of 8x8 blocks
                for (l=0;l<16;l+=8){ // for each col of 8x8 blocks
					do_filter = 0;
					mby = i>>4;
					mbx = j>>4;					
					alt_tc = 0;
					vy = (i+k)/8;
					vx = (j+l)/8;
                    bpos = vy*2*mbxmax+vx;
					if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA){
						do_filter = 1;
						alt_tc = 4;
					}
					/* Check cbp for coded A, B, C or D  */
					if (!do_filter){
						if (mvmap[bpos-2*mbxmax].coded_y || mvmap[bpos].coded_y){
							do_filter = 1;
						}
					}
					/* Check vector differences */
					if (!do_filter){
						if (abs(mvmap[bpos-2*mbxmax].x-mvmap[bpos].x) >= 4 || abs(mvmap[bpos-2*mbxmax].y-mvmap[bpos].y) >= 4){
							do_filter = 1;
						}
					}
                   if ((mode>=MODE_16x16 && mode<MODE_8x8) && tr16flag==0 && svt==1 && (vy&1)){
                        do_filter = 0;
                    }


                   if (mode==MODE_16x16 && tr16flag && (vy&1)){
						do_filter = 0;
					}

                   if (mode==MODE_INTRA && tr16flag && (vy&1)){
						do_filter = 0;
					}			

                    if (mode==MODE_LMB64 && (vy%(LMBXSIZE/8))){
						do_filter = 0;
					}
                    if (mode==(MODE_LMB32) && (vy%(LMBXSIZE/16))){
						do_filter = 0;
					}

                  if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-mbxmax].planar_flag && !(vy&1))
                  {
                    do_filter = 0;  // A fix for Nokia's planar mode filtering
                  }
              
           
					if (do_filter){
						 tc = tctable_8x8[qp+alt_tc];
                        filter_8x8_horizontal_long(img->Y,j+l,i+k-4,tc,betatable_8x8[qp]);
					}

				}
			}
        }
    }

    /* U */    
    for (i=0;i<img->crows;i+=8){
        for (j=0;j<img->ccols;j+=8){
            mby = i/8;
			mbx = j/8;
            vy = mby*2;
		    vx = mbx*2;
            mbpos = mby*mbxmax+mbx;
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
			/* Vertical edge */
			if (j>0){
                if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA){
                    tc = tctable_8x8[qp+4];
                    filter_chroma_vertical_8x8(img->U,j-4,i,tc);
				}
			}
			/* Horizontal edge */
			if (i>0){
                if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA)
                {
                    tc = tctable_8x8[qp+4];
                    filter_chroma_horizontal_8x8(img->U,j,i-4,tc);
				}
			}
        }
    }

    /* V */    
    for (i=0;i<img->crows;i+=8){
        for (j=0;j<img->ccols;j+=8){
            mby = i/8;
			mbx = j/8;
            vy = mby*2;
			vx = mbx*2;
            mbpos = mby*mbxmax+mbx;
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
			/* Vertical edge */
			if (j>0){
                if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA)
                {
					tc = tctable_8x8[qp+4];
                    filter_chroma_vertical_8x8(img->V,j-4,i,tc);
				}
			}
			/* Horizontal edge */
			if (i>0){
                if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA)
                {
					tc = tctable_8x8[qp+4];
                    filter_chroma_horizontal_8x8(img->V,j,i-4,tc);
				}
			}
        }
    }

    //STATIC free_ucsubmatrix(block,0,7,0,7);
    return 0;
}


#if NEW_PLANAR
int filter_luma_vertical_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[y+ 4][x+k];
      a2 = mb[y+12][x+k];
      for(l=0;l<8;l++)
      {
        mb[y+4+l][x+k] = (a1*(8-l)+a2*l + 4) >> 3;
      }
    }
    return 0;
}

int filter_luma_horizontal_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[y+ k][x+4];
      a2 = mb[y+ k][x+12];
      for(l=0;l<8;l++)
      {
        mb[y+k][x+4+l] = (a1*(8-l)+a2*l + 4) >> 3;
      }
    }
    return 0;
}
#else
int filter_luma_vertical_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
        a1 = mb[y+ 4][x+k];
        a2 = mb[y+11][x+k];
      for(l=0;l<7;l++)
      {
            mb[y+4+l][x+k] = (a1*(7-l)+a2*l + 3)/7;
      }
    }
    return 0;
}

int filter_luma_horizontal_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
        a1 = mb[y+k][x+4];
        a2 = mb[y+k][x+11];
      for(l=0;l<7;l++)
      {
            mb[y+k][x+4+l] = (a1*(7-l)+a2*l + 3)/7;
      }
    }
    return 0;
}
#endif

int filter_chroma_vertical_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<8;k++)
    {
        a1 = mb[y+2][x+k];
        a2 = mb[y+5][x+k];
      for(l=0;l<4;l++)
      {
            mb[y+2+l][x+k] = (a1*(4-l)+a2*l + 2)/4;
      }
    }
    return 0;
}

int filter_chroma_horizontal_planar(unsigned char **mb,int x, int y)
{
    int a1,a2;
    int k,l;

    for(k=0;k<8;k++)
    {
        a1 = mb[y+k][x+2];
        a2 = mb[y+k][x+5];
      for(l=0;l<4;l++)
      {
            mb[y+k][x+l] = (a1*(4-l)+a2*l + 2)/4;
      }
    }
    return 0;
}

int filter_planar(encoder_info *encoderInfo,frame_info *frameInfo,image_frame *img)
{
  int i,j;
  int mby,mbx;
  int mbxmax = img->cols>>4;
  int mbpos;  

  mb_info *mbmap2 = frameInfo->mbmap;
  
  for (i=16;i<img->rows;i+=16){ // for each row of MBs
    for (j=16;j<img->cols;j+=16){ // for each col of MBs
      mby = i>>4;
      mbx = j>>4;
      mbpos = mby*mbxmax+mbx;
      
      if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-mbxmax].planar_flag)
      {
        //filter vertical
#if NEW_PLANAR
        filter_luma_vertical_planar(img->Y,j,i-8);
        filter_chroma_vertical_planar(img->U,j/2,i/2-4);
        filter_chroma_vertical_planar(img->V,j/2,i/2-4);
#else
        filter_luma_vertical_planar(img->Y,j,i-8);
        filter_chroma_vertical_planar(img->Y,j,i-8);
        filter_chroma_vertical_planar(img->Y,j,i-8);
#endif
        
      }
      if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-1].planar_flag)
      {
        //filter horizontal
        filter_luma_horizontal_planar(img->Y,j-7,i);
        filter_chroma_horizontal_planar(img->U,j/2-3,i/2); 
        filter_chroma_horizontal_planar(img->V,j/2-3,i/2); 
      }
    }
  }

  return 0;
}
