/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "tntmenc.h"

#include "transform.h"
#include "defines.h"
#if DBG_COMPLEXITY
#include "rdtsc.h"
#endif


int up_available(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mby, int mbx, int block_y, int block_x, enc_params *params);
int left_available(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mby, int mbx, int block_y, int block_x, enc_params *params);

void get_luma_intra_prediction_4x4(unsigned char *up_row, unsigned char *left_col,int dcval, int dir, unsigned char pblock[4][4])
{
	int k,l;    

	if (dir == PRED_DC){        
		for (k=0;k<4;k++){
			for (l=0;l<4;l++){
				pblock[k][l] = dcval;
			}
		}        
	}
	else if (dir == PRED_VER){        
		for (k=0;k<4;k++){
			for (l=0;l<4;l++){
				pblock[k][l] = up_row[l];
			}
		} 
	}
	else{    
		for (k=0;k<4;k++){
			for (l=0;l<4;l++){
				pblock[k][l] = left_col[k];
			}
		}    
	}
}

void get_luma_intra_prediction_16x16(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int dir, unsigned char pblock[16][16],int dcval16)
{
	int k,l,mby,mbx;
	unsigned char leftpixel;

	image_frame *recon = encoderInfo->recon;
	enc_params *params = encoderInfo->params;
	mbx = macroblockInfo->mbx;
	mby = macroblockInfo->mby;

	if (dir == PRED_DC){
		for (k=0;k<16;k++){            
			memset(&pblock[k][0],dcval16,16*sizeof(unsigned char));
		}    
	}
	else if (dir == PRED_VER && up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params)){
		for (k=0;k<16;k++){
			memcpy(&pblock[k][0],&recon->Y[16*mby-1][16*mbx],16*sizeof(unsigned char)); 
		}
	}
	else if(dir == PRED_HOR && left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params)){
		for (k=0;k<16;k++){            
			leftpixel = recon->Y[16*mby+k][16*mbx-1];
			for (l=0;l<16;l++){
				pblock[k][l] = leftpixel;
			}
		}
	}
	else{
		for (k=0;k<16;k++){                
			memset(&pblock[k][0],128,16*sizeof(unsigned char));
		}
	}
}




void get_chroma_intra_prediction_mb(unsigned char *urow, unsigned char *ucol, int dcvalu,
									int dir, unsigned char pblock[8][8], enc_params *params)
{
	int k,l;

	/* DC prediction */
	if (dir == PRED_DC)
	{
		for (k=0;k<8;k++)
		{
			for (l=0;l<8;l++)
			{
				pblock[k][l] = dcvalu;
			}
		}
	}
	/* Top row predicted from outside the macroblock */
	else if (dir == PRED_VER)
	{
		for (k=0;k<8;k++)
		{
			for (l=0;l<8;l++)
			{
				pblock[k][l] = urow[l];
			}
		}
	}
	/* Left column predicted from outside
	the macroblock */
	else
	{
		
		{
			for (k=0;k<8;k++)
			{
				for (l=0;l<8;l++)
				{
					pblock[k][l] = ucol[k];
				}
			}
		}
	}
}


#if NEW_PLANAR
void get_planar_prediction(int upa, int lefta, int blkSize, int planarDelta, int ypos, int xpos, unsigned char **currFrame, unsigned char *planarBlk)
{
    int k, l, dcval, bottomright, bottomleft, topright;
    int leftcolumn[16], toprow[16], bottomrow[16], rightcolumn[16];
    int blkSizeHalf = blkSize >> 1;
    int firstWeight = blkSize - 1;
    int shift1D     = blkSize == 16 ? 4 : 3;
    int shift2D     = shift1D + 1;
    int offset1D    = blkSize >> 1;
    int offset2D    = blkSize;

    // Calculate DC value identical to the one used in DC prediction
    dcval = 0;
    
    if(upa){
        for (k=0; k<blkSize; k++)           
            dcval += currFrame[ypos-1][xpos+k];
    }
    if(lefta){
        for (k=0; k<blkSize; k++)
            dcval += currFrame[ypos+k][xpos-1];
    }

    if(upa && lefta)
        dcval = (dcval+blkSize)/(2*blkSize);
    else if(upa || lefta)
        dcval = (dcval+blkSize/2)/blkSize;
    else
        dcval = 128;

    // Final reconstruction of the bottom-right sample based on DC and indicated delta
    bottomright = planarDelta + dcval;

    // Get left and above reference column and row
    if (!upa && !lefta) {
        for(k=0;k<blkSize;k++)
            leftcolumn[k] = toprow[k] = 128;
    }
    else {
        if(upa) {
            for(k=0;k<blkSize;k++)
                toprow[k] = currFrame[ypos-1][xpos+k];
        }
        else {
            int leftSample = currFrame[ypos][xpos-1];
            for(k=0;k<blkSize;k++)
                toprow[k] = clip3(((firstWeight-k)*leftSample + (k+1)*bottomright + offset1D) >> shift1D);
        }
        if(lefta) {
            for(k=0;k<blkSize;k++)
                leftcolumn[k] = currFrame[ypos+k][xpos-1];
        }
        else {
            int aboveSample = currFrame[ypos-1][xpos];
            for(k=0;k<blkSize;k++)
                leftcolumn[k] = clip3(((firstWeight-k)*aboveSample + (k+1)*bottomright + offset1D) >> shift1D);
        }
    }

    bottomleft = leftcolumn[blkSize-1];
    topright   = toprow[blkSize-1];

    // Linear interpolation to get the rightmost column and bottom row
    for(k=0;k<blkSize;k++){
        bottomrow[k]   = clip3(((firstWeight-k)*bottomleft + (k+1)*bottomright + offset1D) >> shift1D);
        rightcolumn[k] = clip3(((firstWeight-k)*topright   + (k+1)*bottomright + offset1D) >> shift1D);
    }

    // Bilinear interpolation of the final planar samples
    for (k=0;k<blkSize;k++){
        for (l=0;l<blkSize;l++){
            planarBlk[k*blkSize+l] = clip3(( leftcolumn[k]*(firstWeight-l) + rightcolumn[k]*(l+1) + toprow[l]*(firstWeight-k) + bottomrow[l]*(k+1) + offset2D) >> shift2D);
        }
    }
}
#else
void get_planar_prediction(int upa, int lefta, int blkSize, int planarDelta, int ypos, int xpos, unsigned char **currFrame, unsigned char *planarBlk)
{
    int k, l, dcval, bottomright, bottomleft, topright;
    int leftcolumn[16], toprow[16], bottomrow[16], rightcolumn[16];
    //int blkSizeHalf = blkSize >> 1;
    int firstWeight = blkSize - 1;
    int div1D       = blkSize - 1;
    int div2D       = 2*div1D;
    int offset1D    = blkSize/2 - 1;
    int offset2D    = blkSize - 1;

    // Calculate DC value identical to the one used in DC prediction
    dcval = 0;
    
    if(upa){
        for (k=0; k<blkSize; k++)           
            dcval += currFrame[ypos-1][xpos+k];
    }
    if(lefta){
        for (k=0; k<blkSize; k++)
            dcval += currFrame[ypos+k][xpos-1];
    }

    if(upa && lefta)
        dcval = (dcval+blkSize)/(2*blkSize);
    else if(upa || lefta)
        dcval = (dcval+blkSize/2)/blkSize;
    else
        dcval = 128;

    // Final reconstruction of the bottom-right sample based on DC and indicated delta
    bottomright = planarDelta + dcval;

    // Get left and above reference column and row
    if (!upa && !lefta) {
        for(k=0;k<blkSize;k++)
            leftcolumn[k] = toprow[k] = 128;
    }
    else {
        if(upa) {
            for(k=0;k<blkSize;k++)
                toprow[k] = currFrame[ypos-1][xpos+k];
        }
        else {
            int leftSample = currFrame[ypos][xpos-1];
            for(k=0;k<blkSize;k++)
                toprow[k] = clip3(((firstWeight-k)*leftSample + (k+1)*bottomright + offset1D) / div1D);
        }
        if(lefta) {
            for(k=0;k<blkSize;k++)
                leftcolumn[k] = currFrame[ypos+k][xpos-1];
        }
        else {
            int aboveSample = currFrame[ypos-1][xpos];
            for(k=0;k<blkSize;k++)
                leftcolumn[k] = clip3(((firstWeight-k)*aboveSample + (k+1)*bottomright + offset1D) / div1D);
        }
    }

    bottomleft = leftcolumn[blkSize-1];
    topright   = toprow[blkSize-1];

    // Linear interpolation to get the rightmost column and bottom row
    for(k=0;k<blkSize;k++){
        bottomrow[k]   = clip3(((firstWeight-k)*bottomleft + (k)*bottomright + offset1D) / div1D);
        rightcolumn[k] = clip3(((firstWeight-k)*topright   + (k)*bottomright + offset1D) / div1D);
    }

    // Bilinear interpolation of the final planar samples
    for (k=0;k<blkSize;k++){
        for (l=0;l<blkSize;l++){
            planarBlk[k*blkSize+l] = clip3(( leftcolumn[k]*(firstWeight-l) + rightcolumn[k]*(l) + toprow[l]*(firstWeight-k) + bottomrow[l]*(k) + offset2D) / div2D);
        }
    }
}
#endif

void get_directional_intrapred_8x8(unsigned char *up_row, unsigned char *left_col,int dcval, int dir, int intraPredAngle, unsigned char pblock[8][8])
{
    int k,l;
    int deltaInt, deltaFract, refMainIndex;
    int deltaIntSide, deltaFractSide, refSideIndex;
    int absAng  = abs(intraPredAngle);
    //int signAng = intraPredAngle < 0 ? -1 : 1;
    unsigned char tmp;
    unsigned char *refMain;
    unsigned char *refSide;

    if (dir == PRED_DC){        
        for (k=0;k<8;k++){
            for (l=0;l<8;l++){
                pblock[k][l] = dcval;
            }
        }        
    }
    else if (dir == PRED_VER || dir == PRED_HOR){

        refMain = dir == PRED_VER ? up_row : left_col;
        refSide = dir == PRED_VER ? left_col : up_row;

        if (intraPredAngle == 0){
            for (k=0;k<8;k++){
                for (l=0;l<8;l++){
                    pblock[k][l] = refMain[l];
                }
            } 
        }
        else{

            for (k=0;k<8;k++){

                deltaInt   = ((k+1)*intraPredAngle) >> 3;
                deltaFract = ((k+1)*absAng) % 8;

                if (intraPredAngle < 0){
                    deltaFract = (8 - deltaFract) % 8;
                }

                // If the angle is positive all samples are from the main reference 
                if (intraPredAngle > 0){
                    if (deltaFract){
                        // Do linear filtering
                        for (l=0;l<8;l++){
                            refMainIndex = l+deltaInt+1;
                            pblock[k][l] = (unsigned char) ( ((8-deltaFract)*refMain[refMainIndex]+deltaFract*refMain[refMainIndex+1]+4) >> 3 );
                        }
                    }
                    else{
                        // Just copy the integer samples
                        for (l=0;l<8;l++){
                            pblock[k][l] = refMain[l+deltaInt+1];
                        }
                    }
                }
                else{
                    for (l=0;l<8;l++){
                        refMainIndex = l+deltaInt+1;
                        if (refMainIndex >= 0)
                            pblock[k][l] = (unsigned char) ( ((8-deltaFract)*refMain[refMainIndex]+deltaFract*refMain[refMainIndex+1]+4) >> 3 );
                        else{
                            // Interpolate from the side reference
                            deltaIntSide   = (8*8*(l+1)/absAng) >> 3;
                            deltaFractSide = (8*8*(l+1)/absAng)  % 8;
                            refSideIndex   = k+1-deltaIntSide;
                            if (deltaFractSide)
                                pblock[k][l] = (unsigned char) ( ((8-deltaFractSide)*refSide[refSideIndex]+deltaFractSide*refSide[refSideIndex-1]+4) >> 3 );
                            else
                                pblock[k][l] = refSide[refSideIndex];
                        }
                    }
                }
            } 
        }

        // Flip the block if this is the horizontal mode
        if (dir == PRED_HOR){
            for (k=0;k<7;k++){
                for (l=k+1;l<8;l++){
                    tmp          = pblock[k][l];
                    pblock[k][l] = pblock[l][k];
                    pblock[l][k] = tmp;
                }
            } 
        }
    }
    else{
        fatalerror("Illegal intra prediction mode");
    }
}

void get_intrapred_ref_samples(int lefta, int upa, int uprighta, int downlefta, int ypos, int xpos, unsigned char **recon, unsigned char *col, unsigned char *row)
{
    int k;

    if (lefta && upa){
        col[0] = recon[ypos-1][xpos-1];
    }
    else if (lefta){
        col[0] = recon[ypos][xpos-1];
    }
    else if (upa){
        col[0] = recon[ypos-1][xpos];
    }
    else
        col[0] = 128;

    row[0] = col[0];

    if (lefta){
        for (k=0;k<8;k++){
            col[k+1] = recon[ypos+k][xpos-1];
        }
        if (downlefta){
            for (k=8;k<16;k++){
                col[k+1] = recon[ypos+k][xpos-1];
            }
        }
        else{
            for (k=8;k<16;k++){
                col[k+1] = col[8];
            }
        }
    }
    else{
        for (k=1;k<17;k++){
            col[k] = col[0];
        }
    }

    if (upa){
        for (k=0;k<8;k++){
            row[k+1] = recon[ypos-1][xpos+k];
        }
        if (uprighta){
            for (k=8;k<16;k++){
                row[k+1] = recon[ypos-1][xpos+k];
            }
        }
        else{
            for (k=8;k<16;k++){
                row[k+1] = row[8];
            }
        }
    }
    else{
        for (k=1;k<17;k++){
            row[k] = row[0];
        }
    }
}

void get_filtered_intrapred_ref_samples(int lefta, int upa, int uprighta, int downlefta, int ypos, int xpos, unsigned char **recon, unsigned char *col, unsigned char *row)
{
    int k;

    if (lefta && upa){
        col[0] = (unsigned char)((recon[ypos-1][xpos] + 2*recon[ypos-1][xpos-1] + recon[ypos][xpos-1] + 2) >> 2);
    }
    else if (lefta){
        col[0] = recon[ypos][xpos-1];
    }
    else if (upa){
        col[0] = recon[ypos-1][xpos];
    }
    else
        col[0] = 128;

    row[0] = col[0];

    if (lefta){
        if (upa)
            col[1] = (unsigned char)((recon[ypos-1][xpos-1] + 2*recon[ypos][xpos-1] + recon[ypos+1][xpos-1] + 2) >> 2);
        else
            col[1] = recon[ypos][xpos-1];

        for (k=1;k<7;k++){
            col[k+1] = (unsigned char)((recon[ypos+k-1][xpos-1] + 2*recon[ypos+k][xpos-1] + recon[ypos+k+1][xpos-1] + 2) >> 2);
        }

        if (downlefta){
            for (k=7;k<15;k++){
                col[k+1] = (unsigned char)((recon[ypos+k-1][xpos-1] + 2*recon[ypos+k][xpos-1] + recon[ypos+k+1][xpos-1] + 2) >> 2);
            }
            col[16] = recon[ypos+15][xpos-1];
        }
        else{
            for (k=7;k<16;k++){
                col[k+1] = recon[ypos+7][xpos-1];
            }
        }
    }
    else{
        for (k=1;k<17;k++){
            col[k] = col[0];
        }
    }

    if (upa){
        if (lefta)
            row[1] = (unsigned char)((recon[ypos-1][xpos-1] + 2*recon[ypos-1][xpos] + recon[ypos-1][xpos+1] + 2) >> 2);
        else
            row[1] = recon[ypos-1][xpos];

        for (k=1;k<7;k++){
            row[k+1] = (unsigned char)((recon[ypos-1][xpos+k-1] + 2*recon[ypos-1][xpos+k] + recon[ypos-1][xpos+k+1] + 2) >> 2);
        }
        if (uprighta){
            for (k=7;k<15;k++){
                row[k+1] = (unsigned char)((recon[ypos-1][xpos+k-1] + 2*recon[ypos-1][xpos+k] + recon[ypos-1][xpos+k+1] + 2) >> 2);
            }
            row[16] = recon[ypos-1][xpos+15];
        }
        else{
            for (k=7;k<16;k++){
                row[k+1] = recon[ypos-1][xpos+7];
            }
        }
    }
    else{
        for (k=1;k<17;k++){
            row[k] = row[0];
        }
    }
}
