/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "global.h"
#include "tntmenc.h"
#include "defines.h"
#include "motion_estimate.h"


extern int g_bround;


int get_skip_mv_cand(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,mvpos pred_arr[],int skip_ref_arr[])
{
	int mby = macroblockInfo->mby;
	int mbx = macroblockInfo->mbx;
	int num_pred;
	int i;
	//int ref = 0;

	mvpos pp0[3];
	mvpos pp[3];
	int ref_ret[3];
	int refpp[3];
	int k;

	int vy = mby<<1;
	int vx = mbx<<1;

	pp0[0] = get_mv_candidate(encoderInfo,frameInfo,vy-1,vx-0,&ref_ret[0]);
	pp0[1] = get_mv_candidate(encoderInfo,frameInfo,vy-0,vx-1,&ref_ret[1]);
	pp0[2] = get_mv_candidate(encoderInfo,frameInfo,vy-1,vx+2,&ref_ret[2]);
	i = 0;
	for (k = 0; k < 3; k++)
	{
		if (ref_ret[k] != -1)
		{
			refpp[i] = ref_ret[k];
			pp[i++] = pp0[k];
		}
	}
	if (frameInfo->pic_type == B_PICTURE)
	{
		refpp[0] = refpp[1] = refpp[2] = 0;
		if (i == 3 && ( ( (((pp[0].x+2)>>2) == ((pp[2].x+2)>>2)) && (((pp[0].y+2)>>2) == ((pp[2].y+2)>>2)) ) ||
			( (((pp[1].x+2)>>2) == ((pp[2].x+2)>>2)) && (((pp[1].y+2)>>2) == ((pp[2].y+2)>>2)) ) ))
		{
			i = 2;
		}
		if (i > 1 && ((pp[0].x+2)>>2) == ((pp[1].x+2)>>2) && ((pp[0].y+2)>>2) == ((pp[1].y+2)>>2))
		{
			pp[1] = pp[2];
			i--;
		}
	}
	else
	{
		if (i == 3 && ((pp[0].x == pp[2].x && pp[0].y == pp[2].y && refpp[0] == refpp[2]) || (pp[1].x == pp[2].x && pp[1].y == pp[2].y && refpp[1] == refpp[2])))
		{
			i = 2;
		}
		if (i > 1 && pp[0].x == pp[1].x && pp[0].y == pp[1].y && refpp[0] == refpp[1])
		{
			refpp[1] = refpp[2];
			pp[1] = pp[2];
			i--;
		}
	}
	if (i < 1)
	{
		refpp[0] = 0;
		pp[i].x = pp[i].y = 0;

		pp[i].interp_indx = 0;

		i++;
	}
	num_pred = min(i,2);
	for (k = 0; k < num_pred; k++)
	{
		skip_ref_arr[k] = refpp[k];
		pred_arr[k] = pp[k];
	}

	return num_pred;
}

void find_mv_predictor(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo){

	mvpos cand[3];
	mvpos cand2[3],pred,pred_tmp;
	int ref2[3],ref[3];
	int i,j,r,mv_idx,vy,vx,sign;

	int mby = macroblockInfo->mby;
	int mbx = macroblockInfo->mbx;
	//enc_params *params = encoderInfo->params;
	int pic_type = frameInfo->pic_type;

	vx = 2*mbx;
	vy = 2*mby;

	/* Find the candidates for vector prediction */
	cand[0] = get_mv_candidate(encoderInfo,frameInfo,vy-1,vx-0,&ref[0]);
	cand[1] = get_mv_candidate(encoderInfo,frameInfo,vy-1,vx+2,&ref[1]);
	
	{
		cand[2] = get_mv_candidate(encoderInfo,frameInfo,vy-0,vx-1,&ref[2]);
		cand2[2] = get_mv_candidate(encoderInfo,frameInfo,vy+1,vx-1,&ref2[2]);
	}

	pred = (abs(cand[0].x-cand[1].x)+abs(cand[0].y-cand[1].y)) < (abs(cand[2].x-cand[1].x)+abs(cand[2].y-cand[1].y)) ? cand[0] : cand[2];
	for (r = 0; r < 2; r++)
	{
		mv_idx = -1;
		if (r == ref[0] && r != ref[1] && r != ref[2])
			mv_idx = 0;
		else if (r != ref[0] && r == ref[1] && r != ref[2])
			mv_idx = 1;
		else if (r != ref[0] && r != ref[1] && r == ref[2])
			mv_idx = 2;
		if (mv_idx < 0)
		{
			pred_tmp = pred;
		}
		else
		{
			pred_tmp = cand[mv_idx];
		}
		for (i = 0; i < 4; i++)
		{
			for (j = 0; j < 4; j++)
			{
				macroblockInfo->pred[r][i][j] = pred_tmp;
			}
		}
		if (r == ref[0])
		{
			macroblockInfo->pred[r][1][0] = cand[0];
			macroblockInfo->pred[r][1][1] = cand[0];
		}
		if (r == ref[1])
		{
			macroblockInfo->pred[r][2][1] = cand[1];
			macroblockInfo->pred[r][2][3] = cand[1];
			macroblockInfo->pred[r][3][1] = cand[1];
		}
		if (r == ref[2])
		{
			macroblockInfo->pred[r][2][0] = cand[2];
			macroblockInfo->pred[r][2][2] = cand[2];
			{
				macroblockInfo->pred[r][1][2] = cand2[2];
				macroblockInfo->pred[r][1][3] = cand2[2];
				macroblockInfo->pred[r][3][2] = cand2[2];
			}
		}
	}
	if (pic_type == B_PICTURE){
		for (i = 0; i < 4; i++){
			for (j = 0; j < 4; j++){
				macroblockInfo->pred[1][i][j].x = -macroblockInfo->pred[1][i][j].x;
				macroblockInfo->pred[1][i][j].y = -macroblockInfo->pred[1][i][j].y;
			}
		}
	}
	/* Store candidate vectors for inter motion search. */
	macroblockInfo->candidateMV[0][1].x = (cand[0].x+2)>>2;
	macroblockInfo->candidateMV[0][1].y = (cand[0].y+2)>>2;
	macroblockInfo->candidateMV[0][2].x = (cand[1].x+2)>>2;
	macroblockInfo->candidateMV[0][2].y = (cand[1].y+2)>>2;
	macroblockInfo->candidateMV[0][3].x = (cand[2].x+2)>>2;
	macroblockInfo->candidateMV[0][3].y = (cand[2].y+2)>>2;

	sign = pic_type == B_PICTURE ? -1 : 1;    
	macroblockInfo->candidateMV[1][1].x = (sign*cand[0].x+2)>>2;
	macroblockInfo->candidateMV[1][1].y = (sign*cand[0].y+2)>>2;
	macroblockInfo->candidateMV[1][2].x = (sign*cand[1].x+2)>>2;
	macroblockInfo->candidateMV[1][2].y = (sign*cand[1].y+2)>>2;
	macroblockInfo->candidateMV[1][3].x = (sign*cand[2].x+2)>>2;
	macroblockInfo->candidateMV[1][3].y = (sign*cand[2].y+2)>>2;

}

void ref_predictor(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{
	int ref[3];
	int mby2,mbx2,count[3],ref_pred;
	int mby = macroblockInfo->mby;
	int mbx = macroblockInfo->mbx;
	int mbxmax = encoderInfo->widthInMBs;

	count[0]=count[1]=count[2]=0;
	mby2=mby;
	mbx2=mbx-1;
	if (mby2 < 0 || mbx2 < 0 || mbx2 >= encoderInfo->width/16)
		ref[0] = -1;
	else
	{
		ref[0] = frameInfo->mbmap[mby2*mbxmax+mbx2].ref;
		if(ref[0]!=-1)
			count[ref[0]]++;
	}

	mby2=mby-1;
	mbx2=mbx-1;
	if (mby2 < 0 || mbx2 < 0 || mbx2 >= encoderInfo->width/16)
		ref[1] = -1;
	else
	{
		ref[1] = frameInfo->mbmap[mby2*mbxmax+mbx2].ref;
		if(ref[1]!=-1)
			count[ref[1]]++;
	}

	mby2=mby-1;
	mbx2=mbx;
	if (mby2 < 0 || mbx2 < 0 || mbx2 >= encoderInfo->width/16)
		ref[2] = -1;
	else
	{
		ref[2] = frameInfo->mbmap[mby2*mbxmax+mbx2].ref;
		if(ref[2]!=-1)
			count[ref[2]]++;
	}
	ref_pred=(int)(count[1]>count[0]?1:0);
	macroblockInfo->ref_pred=(int)(count[ref_pred]>=count[2]?ref_pred:2);
}

int interpolate_Y2Y(unsigned char **ip,
                    unsigned char **qp,
                    unsigned char **q2p,
                    const int rows,
                    const int cols)
{
    const int N = 8;
    int i,j,k,ioff,joff;

    /* Copy and pad integer pixels */
    for (i = 16; i < rows-16; i++)
    {
        memcpy(&qp[i][16],&ip[i-16][0],(cols-2*16));
        memcpy(&q2p[i][16],&ip[i-16][0],(cols-2*16));
    }
    /* Left extension */
    for (i = 16; i < rows-16; i++)
    {
        unsigned char tmp = qp[i][16];
        unsigned char tmp2 = q2p[i][16];
        memset(&qp[i][0],tmp,16);
        memset(&q2p[i][0],tmp2,16);
    }
    /* Right extension */
    for (i = 16; i < rows-16; i++)
    {
        unsigned char tmp = qp[i][cols-16-1];
        unsigned char tmp2 = q2p[i][cols-16-1];
        memset(&qp[i][cols-16],tmp,16);
        memset(&q2p[i][cols-16],tmp2,16);
    }
    /* Top extension */
    for (i = 0; i < 16; i++)
    {
        memcpy(&qp[i][0],&qp[16][0],cols);
        memcpy(&q2p[i][0],&q2p[16][0],cols);
    }
    /* Bottom extension */
    for (i = rows-16; i < rows; i++)
    {
        memcpy(&qp[i][0],&qp[rows-16-1][0],cols);
        memcpy(&q2p[i][0],&q2p[rows-16-1][0],cols);
    }

    for (i = N; i < rows-N; i++)
    {
        for (j = 0; j < cols; j+=16)
        {
            unsigned char x1[6][16];
            unsigned char x2[6][16];
            unsigned char x3[6][16];

            short y1[16], y2[16], y3[16];
            short hj[16], vj[16], dj[16];
            //short h11[16], h13[16], h12[16];
            short v11[16], v13[16], v12[16];
            short d11[16], d13[16], d21[16], d23[16];
            short s11[16], s13[16], s21[16], s23[16];
            short d11d21[16], d11d23[16], d13d21[16], d13d23[16];

            const unsigned char *xm20 = &qp[i-2][j];
            const unsigned char *xm10 = &qp[i-1][j];
            const unsigned char *x00  = &qp[i+0][j];
            const unsigned char *xp10 = &qp[i+1][j];
            const unsigned char *xp20 = &qp[i+2][j];
            const unsigned char *xp30 = &qp[i+3][j];

            const unsigned char *xm2m2 = &qp[i-2][j-2];
            const unsigned char *xm2m1 = &qp[i-2][j-1];
            const unsigned char *xm2   = &qp[i-2][j+0];
            const unsigned char *xm2p1 = &qp[i-2][j+1];
            const unsigned char *xm2p2 = &qp[i-2][j+2];
            const unsigned char *xm2p3 = &qp[i-2][j+3];

            const unsigned char *xm1m2 = &qp[i-1][j-2];
            const unsigned char *xm1m1 = &qp[i-1][j-1];
            const unsigned char *xm1   = &qp[i-1][j+0];
            const unsigned char *xm1p1 = &qp[i-1][j+1];
            const unsigned char *xm1p2 = &qp[i-1][j+2];
            const unsigned char *xm1p3 = &qp[i-1][j+3];

            const unsigned char *x0m2 = &qp[i][j-2];
            const unsigned char *x0m1 = &qp[i][j-1];
            const unsigned char *x0   = &qp[i][j+0];
            const unsigned char *x0p1 = &qp[i][j+1];
            const unsigned char *x0p2 = &qp[i][j+2];
            const unsigned char *x0p3 = &qp[i][j+3];

            const unsigned char *xp1m2 = &qp[i+1][j-2];
            const unsigned char *xp1m1 = &qp[i+1][j-1];
            const unsigned char *xp1   = &qp[i+1][j+0];
            const unsigned char *xp1p1 = &qp[i+1][j+1];
            const unsigned char *xp1p2 = &qp[i+1][j+2];
            const unsigned char *xp1p3 = &qp[i+1][j+3];

            const unsigned char *xp2m2 = &qp[i+2][j-2];
            const unsigned char *xp2m1 = &qp[i+2][j-1];
            const unsigned char *xp2   = &qp[i+2][j+0];
            const unsigned char *xp2p1 = &qp[i+2][j+1];
            const unsigned char *xp2p2 = &qp[i+2][j+2];
            const unsigned char *xp2p3 = &qp[i+2][j+3];

            const unsigned char *xp3m2 = &qp[i+3][j-2];
            const unsigned char *xp3m1 = &qp[i+3][j-1];
            const unsigned char *xp3   = &qp[i+3][j+0];
            const unsigned char *xp3p1 = &qp[i+3][j+1];
            const unsigned char *xp3p2 = &qp[i+3][j+2];
            const unsigned char *xp3p3 = &qp[i+3][j+3];

            unsigned char *restrict qa = &qp[0*rows+i][1*cols+j]; 
            unsigned char *restrict qb = &qp[0*rows+i][2*cols+j]; 
            unsigned char *restrict qc = &qp[0*rows+i][3*cols+j]; 
            unsigned char *restrict qd = &qp[1*rows+i][0*cols+j];        
            unsigned char *restrict qe = &qp[1*rows+i][1*cols+j]; 
            unsigned char *restrict qf = &qp[1*rows+i][2*cols+j]; 
            unsigned char *restrict qg = &qp[1*rows+i][3*cols+j]; 
            unsigned char *restrict qh = &qp[2*rows+i][0*cols+j]; 
            unsigned char *restrict qi = &qp[2*rows+i][1*cols+j]; 
            unsigned char *restrict qj = &qp[2*rows+i][2*cols+j]; 
            unsigned char *restrict qk = &qp[2*rows+i][3*cols+j]; 
            unsigned char *restrict ql = &qp[3*rows+i][0*cols+j]; 
            unsigned char *restrict qm = &qp[3*rows+i][1*cols+j]; 
            unsigned char *restrict qn = &qp[3*rows+i][2*cols+j]; 
            unsigned char *restrict qo = &qp[3*rows+i][3*cols+j]; 

            unsigned char *restrict q2e = &q2p[1*rows+i][1*cols+j]; 
            unsigned char *restrict q2f = &q2p[1*rows+i][2*cols+j]; 
            unsigned char *restrict q2g = &q2p[1*rows+i][3*cols+j]; 
            unsigned char *restrict q2i = &q2p[2*rows+i][1*cols+j]; 
            unsigned char *restrict q2j = &q2p[2*rows+i][2*cols+j]; 
            unsigned char *restrict q2k = &q2p[2*rows+i][3*cols+j]; 
            unsigned char *restrict q2m = &q2p[3*rows+i][1*cols+j]; 
            unsigned char *restrict q2n = &q2p[3*rows+i][2*cols+j]; 
            unsigned char *restrict q2o = &q2p[3*rows+i][3*cols+j]; 


            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*xm2m2[k] + (-15)*xm2m1[k] + (-17)*xm2[k] +  ( 37)*xm2p1[k] + (-10)*xm2p2[k] + 2*xm2p3[k] + 64);
                y2[k] = (short)(3*xm2m2[k] + (-17)*xm2m1[k] + ( 78)*xm2[k] +  (-50)*xm2p1[k] + (-17)*xm2p2[k] + 3*xm2p3[k] + 64);
                y3[k] = (short)(2*xm2m2[k] + (-10)*xm2m1[k] + ( 37)*xm2[k] +  (-17)*xm2p1[k] + (-15)*xm2p2[k] + 3*xm2p3[k] + 64);
                y1[k] = (short)(xm2  [k] + (y1[k] >> 7));
                y2[k] = (short)(xm2p1[k] + (y2[k] >> 7));
                y3[k] = (short)(xm2p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                x1[0][k] = (unsigned char) min(max(y1[k], 0), 255);
                x2[0][k] = (unsigned char) min(max(y2[k], 0), 255);
                x3[0][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*xm1m2[k] + (-15)*xm1m1[k] + (-17)*xm1[k] +  ( 37)*xm1p1[k] + (-10)*xm1p2[k] + 2*xm1p3[k] + 64);
                y2[k] = (short)(3*xm1m2[k] + (-17)*xm1m1[k] + ( 78)*xm1[k] +  (-50)*xm1p1[k] + (-17)*xm1p2[k] + 3*xm1p3[k] + 64);
                y3[k] = (short)(2*xm1m2[k] + (-10)*xm1m1[k] + ( 37)*xm1[k] +  (-17)*xm1p1[k] + (-15)*xm1p2[k] + 3*xm1p3[k] + 64);
                y1[k] = (short)(xm1  [k] + (y1[k] >> 7));
                y2[k] = (short)(xm1p1[k] + (y2[k] >> 7));
                y3[k] = (short)(xm1p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                x1[1][k] = (unsigned char) min(max(y1[k], 0), 255);
                x2[1][k] = (unsigned char) min(max(y2[k], 0), 255);
                x3[1][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*x0m2[k] + (-15)*x0m1[k] + (-17)*x0[k] +  ( 37)*x0p1[k] + (-10)*x0p2[k] + 2*x0p3[k] + 64);
                y2[k] = (short)(3*x0m2[k] + (-17)*x0m1[k] + ( 78)*x0[k] +  (-50)*x0p1[k] + (-17)*x0p2[k] + 3*x0p3[k] + 64);
                y3[k] = (short)(2*x0m2[k] + (-10)*x0m1[k] + ( 37)*x0[k] +  (-17)*x0p1[k] + (-15)*x0p2[k] + 3*x0p3[k] + 64);
                y1[k] = (short)(x0  [k] + (y1[k] >> 7));
                y2[k] = (short)(x0p1[k] + (y2[k] >> 7));
                y3[k] = (short)(x0p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                qa[k] = x1[2][k] = (unsigned char) min(max(y1[k], 0), 255);
                qb[k] = x2[2][k] = (unsigned char) min(max(y2[k], 0), 255);
                qc[k] = x3[2][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*xp1m2[k] + (-15)*xp1m1[k] + (-17)*xp1[k] +  ( 37)*xp1p1[k] + (-10)*xp1p2[k] + 2*xp1p3[k] + 64);
                y2[k] = (short)(3*xp1m2[k] + (-17)*xp1m1[k] + ( 78)*xp1[k] +  (-50)*xp1p1[k] + (-17)*xp1p2[k] + 3*xp1p3[k] + 64);
                y3[k] = (short)(2*xp1m2[k] + (-10)*xp1m1[k] + ( 37)*xp1[k] +  (-17)*xp1p1[k] + (-15)*xp1p2[k] + 3*xp1p3[k] + 64);
                y1[k] = (short)(xp1  [k] + (y1[k] >> 7));
                y2[k] = (short)(xp1p1[k] + (y2[k] >> 7));
                y3[k] = (short)(xp1p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                x1[3][k] = (unsigned char) min(max(y1[k], 0), 255);
                x2[3][k] = (unsigned char) min(max(y2[k], 0), 255);
                x3[3][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*xp2m2[k] + (-15)*xp2m1[k] + (-17)*xp2[k] +  ( 37)*xp2p1[k] + (-10)*xp2p2[k] + 2*xp2p3[k] + 64);
                y2[k] = (short)(3*xp2m2[k] + (-17)*xp2m1[k] + ( 78)*xp2[k] +  (-50)*xp2p1[k] + (-17)*xp2p2[k] + 3*xp2p3[k] + 64);
                y3[k] = (short)(2*xp2m2[k] + (-10)*xp2m1[k] + ( 37)*xp2[k] +  (-17)*xp2p1[k] + (-15)*xp2p2[k] + 3*xp2p3[k] + 64);
                y1[k] = (short)(xp2  [k] + (y1[k] >> 7));
                y2[k] = (short)(xp2p1[k] + (y2[k] >> 7));
                y3[k] = (short)(xp2p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                x1[4][k] = (unsigned char) min(max(y1[k], 0), 255);
                x2[4][k] = (unsigned char) min(max(y2[k], 0), 255);
                x3[4][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*xp3m2[k] + (-15)*xp3m1[k] + (-17)*xp3[k] +  ( 37)*xp3p1[k] + (-10)*xp3p2[k] + 2*xp3p3[k] + 64);
                y2[k] = (short)(3*xp3m2[k] + (-17)*xp3m1[k] + ( 78)*xp3[k] +  (-50)*xp3p1[k] + (-17)*xp3p2[k] + 3*xp3p3[k] + 64);
                y3[k] = (short)(2*xp3m2[k] + (-10)*xp3m1[k] + ( 37)*xp3[k] +  (-17)*xp3p1[k] + (-15)*xp3p2[k] + 3*xp3p3[k] + 64);
                y1[k] = (short)(xp3  [k] + (y1[k] >> 7));
                y2[k] = (short)(xp3p1[k] + (y2[k] >> 7));
                y3[k] = (short)(xp3p1[k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                x1[5][k] = (unsigned char) min(max(y1[k], 0), 255);
                x2[5][k] = (unsigned char) min(max(y2[k], 0), 255);
                x3[5][k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*x1[0][k] + (-15)*x1[1][k] + (-17)*x1[2][k] + ( 37)*x1[3][k] + (-10)*x1[4][k] + 2*x1[5][k] + 64);
                y2[k] = (short)(3*x1[0][k] + (-17)*x1[1][k] + ( 78)*x1[2][k] + (-50)*x1[3][k] + (-17)*x1[4][k] + 3*x1[5][k] + 64);
                y3[k] = (short)(2*x1[0][k] + (-10)*x1[1][k] + ( 37)*x1[2][k] + (-17)*x1[3][k] + (-15)*x1[4][k] + 3*x1[5][k] + 64);
                y1[k] = (short)(x1[2][k] + (y1[k] >> 7));
                y2[k] = (short)(x1[3][k] + (y2[k] >> 7));
                y3[k] = (short)(x1[3][k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                q2e[k] = (unsigned char) min(max(y1[k], 0), 255);
                q2i[k] = (unsigned char) min(max(y2[k], 0), 255);
                q2m[k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*x2[0][k] + (-15)*x2[1][k] + (-17)*x2[2][k] + ( 37)*x2[3][k] + (-10)*x2[4][k] + 2*x2[5][k] + 64);
                y2[k] = (short)(3*x2[0][k] + (-17)*x2[1][k] + ( 78)*x2[2][k] + (-50)*x2[3][k] + (-17)*x2[4][k] + 3*x2[5][k] + 64);
                y3[k] = (short)(2*x2[0][k] + (-10)*x2[1][k] + ( 37)*x2[2][k] + (-17)*x2[3][k] + (-15)*x2[4][k] + 3*x2[5][k] + 64);
                y1[k] = (short)(x2[2][k] + (y1[k] >> 7));
                y2[k] = (short)(x2[3][k] + (y2[k] >> 7));
                y3[k] = (short)(x2[3][k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                q2f[k] = (unsigned char) min(max(y1[k], 0), 255);
                q2j[k] = (unsigned char) min(max(y2[k], 0), 255);
                q2n[k] = (unsigned char) min(max(y3[k], 0), 255);
            }

            for (k = 0; k < 16; k++)
            {
                y1[k] = (short)(3*x3[0][k] + (-15)*x3[1][k] + (-17)*x3[2][k] + ( 37)*x3[3][k] + (-10)*x3[4][k] + 2*x3[5][k] + 64);
                y2[k] = (short)(3*x3[0][k] + (-17)*x3[1][k] + ( 78)*x3[2][k] + (-50)*x3[3][k] + (-17)*x3[4][k] + 3*x3[5][k] + 64);
                y3[k] = (short)(2*x3[0][k] + (-10)*x3[1][k] + ( 37)*x3[2][k] + (-17)*x3[3][k] + (-15)*x3[4][k] + 3*x3[5][k] + 64);
                y1[k] = (short)(x3[2][k] + (y1[k] >> 7));
                y2[k] = (short)(x3[3][k] + (y2[k] >> 7));
                y3[k] = (short)(x3[3][k] + (y3[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                q2g[k] = (unsigned char) min(max(y1[k], 0), 255);
                q2k[k] = (unsigned char) min(max(y2[k], 0), 255);
                q2o[k] = (unsigned char) min(max(y3[k], 0), 255);
            }
            //
            //            for (k = 0; k < 16; k++)
            //            {
            //                h11[k] = (short)(3*x0m2[k] + (-15)*x0m1[k] + (-17)*x00[k] + ( 37)*x0p1[k] + (-10)*x0p2[k] + 2*x0p3[k] + 64);
            //                h12[k] = (short)(3*x0m2[k] + (-17)*x0m1[k] + ( 78)*x00[k] + (-50)*x0p1[k] + (-17)*x0p2[k] + 3*x0p3[k] + 64); 
            //                h13[k] = (short)(2*x0m2[k] + (-10)*x0m1[k] + ( 37)*x00[k] + (-17)*x0p1[k] + (-15)*x0p2[k] + 3*x0p3[k] + 64);
            //                h11[k] = (short)(x00 [k] + (h11[k] >> 7));
            //                h12[k] = (short)(x0p1[k] + (h12[k] >> 7));
            //                h13[k] = (short)(x0p1[k] + (h13[k] >> 7));
            //            }
            //            for (k = 0; k < 16; k++)
            //            {
            //                qa[k] = (unsigned char) min(max(h11[k], 0), 255); 
            //                qb[k] = (unsigned char) min(max(h12[k], 0), 255);
            //                qc[k] = (unsigned char) min(max(h13[k], 0), 255);
            //            }

            for (k = 0; k < 16; k++)
            {
                v11[k] = (short)(3*xm20[k] + (-15)*xm10[k] + (-17)*x00[k] + ( 37)*xp10[k] + (-10)*xp20[k] + 2*xp30[k] + 64);
                v12[k] = (short)(3*xm20[k] + (-17)*xm10[k] + ( 78)*x00[k] + (-50)*xp10[k] + (-17)*xp20[k] + 3*xp30[k] + 64); 
                v13[k] = (short)(2*xm20[k] + (-10)*xm10[k] + ( 37)*x00[k] + (-17)*xp10[k] + (-15)*xp20[k] + 3*xp30[k] + 64);  
                v11[k] = (short)(x00 [k] + (v11[k] >> 7));
                v12[k] = (short)(xp10[k] + (v12[k] >> 7));
                v13[k] = (short)(xp10[k] + (v13[k] >> 7));
            }  
            for (k = 0; k < 16; k++)
            {
                qd[k] = (unsigned char) min(max(v11[k], 0), 255);
                qh[k] = (unsigned char) min(max(v12[k], 0), 255);
                ql[k] = (unsigned char) min(max(v13[k], 0), 255);
            } 
 // e = (d11 >> 7)
            for (k = 0; k < 16; k++)
            {
                s11[k] = (short)(3*xm2m2[k] + (-15)*xm1m1[k] + (-17)*x00 [k] +  37*xp1p1[k] + (-10)*xp2p2[k] + 2*xp3p3[k] + 64);
                s21[k] = (short)(3*xm2p3[k] + (-15)*xm1p2[k] + (-17)*x0p1[k] +  37*xp10 [k] + (-10)*xp2m1[k] + 2*xp3m2[k] + 64);
                d11[k] = (short)(x00 [k] + (s11[k] >> 7));
                d21[k] = (short)(x0p1[k] + (s21[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                qe[k] = (unsigned char) min(max(d11[k], 0), 255);
                qg[k] = (unsigned char) min(max(d21[k], 0), 255);
            }

 // f = (d11 + d21 + 128)>>8
            for (k = 0; k < 16; k++)
                d11d21[k] = (short)(min(max(s11[k] + s21[k], -32768), 32767) >> 7);
            for (k = 0; k < 16; k++)
                d11d21[k] = (short)(min(max(d11d21[k] + x00[k], -32768), 32767)); 
            for (k = 0; k < 16; k++)
                d11d21[k] = (short)(min(max(d11d21[k] +  x0p1[k], -32768), 32767) >> 1); 
            for (k = 0; k < 16; k++)
                qf[k] = (unsigned char) min(max(d11d21[k], 0), 255);

 // m = (d23 >> 7)
            for (k = 0; k < 16; k++)
            {
                s13[k] = (short)(2*xm2m2[k] + (-10)*xm1m1[k] +  37*x00[k] + (-17)*xp1p1[k] + (-15)*xp2p2[k] + 3*xp3p3[k] + 64);
                s23[k] = (short)(2*xm2p3[k] + (-10)*xm1p2[k] +  37*x0p1[k] + (-17)*xp10[k] + (-15)*xp2m1[k] + 3*xp3m2[k] + 64);
                d13[k] = (short)(xp1p1[k] + (s13[k] >> 7));
                d23[k] = (short)(xp10[k] + (s23[k] >> 7));
            }
            for (k = 0; k < 16; k++)
            {
                qm[k] = (unsigned char) min(max(d23[k], 0), 255);
                qo[k] = (unsigned char) min(max(d13[k], 0), 255);
            }

 // n = (d13 + d23 + 128)>>8
            for (k = 0; k < 16; k++)
                d13d23[k] = (short)(min(max(s13[k] + s23[k], -32768), 32767) >> 7);
            for (k = 0; k < 16; k++)
                d13d23[k] = (short)(min(max(d13d23[k] + xp1p1[k], -32768), 32767)); 
            for (k = 0; k < 16; k++)
                d13d23[k] = (short)(min(max(d13d23[k] + xp10[k], -32768), 32767) >> 1); 
            for (k = 0; k < 16; k++)
                qn[k] = (unsigned char)(min(max(d13d23[k], 0), 255));

 // i = (d11 + d23 + 128)>>8
            for (k = 0; k < 16; k++)
                d11d23[k] = (short)(min(max(s11[k] + s23[k], -32768), 32767) >> 7);
            for (k = 0; k < 16; k++)
                d11d23[k] = (short)(min(max(d11d23[k] + x00[k], -32768), 32767)); 
            for (k = 0; k < 16; k++)
                d11d23[k] = (short)(min(max(d11d23[k] + xp10[k], -32768), 32767) >> 1); 
            for (k = 0; k < 16; k++)
                qi[k] = (unsigned char) min(max(d11d23[k], 0), 255);

 // k = (d13 + d21 + 128)>>8
            for (k = 0; k < 16; k++)
                d13d21[k] = (short)(min(max(s13[k] + s21[k], -32768), 32767) >> 7);
            for (k = 0; k < 16; k++)
                d13d21[k] = (short)(min(max(d13d21[k] + xp1p1[k], -32768), 32767)); 
            for (k = 0; k < 16; k++)
                d13d21[k] = (short)(min(max(d13d21[k] + x0p1[k], -32768), 32767) >> 1); 
            for (k = 0; k < 16; k++)
                qk[k] = (unsigned char) min(max(d13d21[k], 0), 255);

 // j = (5*vj + 22*hj + 64) >> 7
            for (k = 0; k < 16; k++) 
            {
                vj[k] = (short)(xm10[k]+xm1p1[k]+x0m1[k]+x0p2[k]+xp1m1[k]+xp1p2[k]+xp20[k]+xp2p1[k]);
                vj[k] = (short)(5*vj[k]); 
                hj[k] = (short)(x00[k]+x0p1[k]+xp10[k]+xp1p1[k]);
                hj[k] = (short)(22*hj[k]); 
                dj[k] = (short)(min(max(vj[k] + hj[k], -32768), 32767)); 
                dj[k] = (short)(min(max(dj[k] + 64, -32768), 32767)); 
                dj[k] = (short)(dj[k] >> 7);
            }
            for (k = 0; k < 16; k++)
                qj[k] = (unsigned char) min(max(dj[k], 0), 255);
        }
    }

    for (ioff=0*rows;ioff<=3*rows;ioff+=rows)
    {
        for (joff=(ioff==0)*cols;joff<=3*cols;joff+=cols)
        {
            /* Left extension */
            for (i = N; i < rows-N; i++)
            {
                unsigned char tmp = qp[ioff+i][joff+N];
                unsigned char tmp2 = q2p[ioff+i][joff+N];
                memset(&qp[ioff+i][joff],tmp,N);
                memset(&q2p[ioff+i][joff],tmp2,N);
            }
            /* Right extension */
            for (i = N; i < rows-N; i++)
            {
                unsigned char tmp = qp[ioff+i][joff+cols-N-1];
                unsigned char tmp2 = q2p[ioff+i][joff+cols-N-1];
                memset(&qp[ioff+i][joff+cols-N],tmp,N);
                memset(&q2p[ioff+i][joff+cols-N],tmp2,N);
            }
            /* Top extension */
            for (i = 0; i < N; i++)
            {
                memcpy(&qp[ioff+i][joff],&qp[ioff+N][joff],cols);
                memcpy(&q2p[ioff+i][joff],&q2p[ioff+N][joff],cols);
            }
            /* Bottom extension */
            for (i = rows-N; i < rows; i++)
            {
                memcpy(&qp[ioff+i][joff],&qp[ioff+rows-N-1][joff],cols);
                memcpy(&q2p[ioff+i][joff],&q2p[ioff+rows-N-1][joff],cols);
            }
        }
    }
    return 0;
}




int interpolate_Y(unsigned char **ip,unsigned char **qp,int rows,int cols)
{
	unsigned char tmp;
	int i,j,ioff,joff;
	int N = 8; // Assumes frame extension width is 16, and FilterLength/2 <=8

	int d11,d12,d13,d21,d22,d23;
	int h11,h12,h13;
	int v11,v12,v13;

	int x0m2,x0m1,x00,x0p1,x0p2,x0p3;
	int xm20,xm10,xp10,xp20,xp30;
	int xm2m2,xm2p3, xm1m1,xm1p2, xp1p1, xp2m1,xp2p2, xp3m2,xp3p3;
	int xm1p1,xp1m1,xp1p2,xp2p1;

	/* Copy and pad integer pixels */
	for (i = 16; i < rows-16; i++){
		memcpy(&qp[i][16],&ip[i-16][0],(cols-2*16)*sizeof(unsigned char));
	}    
	/* Left extension */
	for (i = 16; i < rows-16; i++){
		tmp = qp[i][16];                    
		memset(&qp[i][0],tmp,16*sizeof(unsigned char));
	}
	/* Right extension */
	for (i = 16; i < rows-16; i++){
		tmp = qp[i][cols-16-1];                    
		memset(&qp[i][cols-16],tmp,16*sizeof(unsigned char));
	}
	/* Top extension */
	for (i = 0; i < 16; i++){
		memcpy(&qp[i][0],&qp[16][0],cols*sizeof(unsigned char));
	}
	/* Bottom extension */
	for (i = rows-16; i < rows; i++){
		memcpy(&qp[i][0],&qp[rows-16-1][0],cols*sizeof(unsigned char));
	}

	/* Calculate fractional pixels */
	for (i = N; i < rows-N; i++){
		for (j = N; j < cols-N; j++){

			x0m2 = qp[i][j-2];
			x0m1 = qp[i][j-1];
			x00  = qp[i][j+0];
			x0p1 = qp[i][j+1];
			x0p2 = qp[i][j+2];
			x0p3 = qp[i][j+3];

			h11 = 3*x0m2 - 15*x0m1 + 111*x00 +  37*x0p1 - 10*x0p2 + 2*x0p3;
			h12 = 3*x0m2 - 17*x0m1 +  78*x00 +  78*x0p1 - 17*x0p2 + 3*x0p3; 
			h13 = 2*x0m2 - 10*x0m1 +  37*x00 + 111*x0p1 - 15*x0p2 + 3*x0p3;


			xm20 = qp[i-2][j];
			xm10 = qp[i-1][j];                
			xp10 = qp[i+1][j];
			xp20 = qp[i+2][j];
			xp30 = qp[i+3][j];

			v11 = 3*xm20 - 15*xm10 + 111*x00 + 37*xp10 - 10*xp20 + 2*xp30;
			v12 = 3*xm20 - 17*xm10 +  78*x00 + 78*xp10 - 17*xp20 + 3*xp30; 
			v13 = 2*xm20 - 10*xm10 + 37*x00 + 111*xp10 - 15*xp20 + 3*xp30;

			xm2m2 = qp[i-2][j-2];
			xm1m1 = qp[i-1][j-1];
			xp1p1 = qp[i+1][j+1];
			xp2p2 = qp[i+2][j+2];
			xp3p3 = qp[i+3][j+3];

			d11= 3*xm2m2 - 15*xm1m1 + 111*x00 +  37*xp1p1 - 10*xp2p2 + 2*xp3p3;
			d12= 3*xm2m2 - 17*xm1m1 +  78*x00 +  78*xp1p1 - 17*xp2p2 + 3*xp3p3;
			d13= 2*xm2m2 - 10*xm1m1 +  37*x00 + 111*xp1p1 - 15*xp2p2 + 3*xp3p3;   

			xm2p3 = qp[i-2][j+3];
			xm1p2 = qp[i-1][j+2];
			x0p1  = qp[i-0][j+1];
			xp10  = qp[i+1][j+0];
			xp2m1 = qp[i+2][j-1];
			xp3m2 = qp[i+3][j-2];

			d21= 3*xm2p3 - 15*xm1p2 + 111*x0p1 +  37*xp10 - 10*xp2m1 + 2*xp3m2;
			d22= 3*xm2p3 - 17*xm1p2 +  78*x0p1 +  78*xp10 - 17*xp2m1 + 3*xp3m2;
			d23= 2*xm2p3 - 10*xm1p2 +  37*x0p1 + 111*xp10 - 15*xp2m1 + 3*xp3m2;

			xm1p1 = qp[i-1][j+1];
			xp1m1 = qp[i+1][j-1];
			xp1p2 = qp[i+1][j+2];
			xp2p1 = qp[i+2][j+1];

			//a
			qp[(0+0)*rows+i][(0+1)*cols+j] = clip255((  h11 + 64) >> 7);
			//b
			qp[(0+0)*rows+i][(0+2)*cols+j] = clip255((  h12 + 64) >> 7);                
			//c
			qp[(0+0)*rows+i][(0+3)*cols+j] = clip255((  h13 + 64) >> 7);                
			//d
			qp[(0+1)*rows+i][(0+0)*cols+j] = clip255((  v11 + 64) >> 7);
			//e
			qp[(0+1)*rows+i][(0+1)*cols+j] = clip255((d11+64) >> 7);
			//f
			qp[(0+1)*rows+i][(0+2)*cols+j] = clip255((d11+d21+128)>>8);
			//g                
			qp[(0+1)*rows+i][(0+3)*cols+j] = clip255((d21+64) >> 7);
			//h
			qp[(0+2)*rows+i][(0+0)*cols+j] = clip255((  v12 + 64) >> 7);
			//i
			qp[(0+2)*rows+i][(0+1)*cols+j] = clip255((d11+d23+128)>>8);
			//j
			qp[(0+2)*rows+i][(0+2)*cols+j] = clip255((5*(xm10+xm1p1+x0m1+x0p2+xp1m1+xp1p2+xp20+xp2p1) + 22*(x00+x0p1+xp10+xp1p1) + 64)>>7);
			//k
			qp[(0+2)*rows+i][(0+3)*cols+j] = clip255((d13+d21+128)>>8);
			//l
			qp[(0+3)*rows+i][(0+0)*cols+j] = clip255((  v13 + 64) >> 7);
			//m
			qp[(0+3)*rows+i][(0+1)*cols+j] = clip255((d23+64) >> 7);                                              
			//n
			qp[(0+3)*rows+i][(0+2)*cols+j] = clip255((d13+d23+128)>>8);                
			//o
			qp[(0+3)*rows+i][(0+3)*cols+j] = clip255((d13+64) >> 7);                                                                                            
		}
	}

	/* pad fractional pixels */
	for (ioff=0*rows;ioff<=3*rows;ioff+=rows){
		for (joff=(ioff==0)*cols;joff<=3*cols;joff+=cols){            
			/* Left extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+N];                    
				memset(&qp[ioff+i][joff],tmp,N*sizeof(unsigned char));
			}
			/* Right extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+cols-N-1];                    
				memset(&qp[ioff+i][joff+cols-N],tmp,N*sizeof(unsigned char));
			}
			/* Top extension */
			for (i = 0; i < N; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+N][joff],cols*sizeof(unsigned char));
			}
			/* Bottom extension */
			for (i = rows-N; i < rows; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+rows-N-1][joff],cols*sizeof(unsigned char));
			}
		}
	}
	return 0;
}


int interpolate_C(unsigned char **ip,unsigned char **qp,int rows,int cols)
{
#if FAST_MODE
    unsigned char tmp;
    int i,j,k,ioff,joff;
    int N = 4; // Assumes frame extension width is 16, and FilterLength/2 <=8   

    /* Copy and extend integer pixels */
    for (i = 8; i < rows-8; i++){
        memcpy(&qp[i][8],&ip[i-8][0],(cols-2*8)*sizeof(unsigned char));
    }
    /* Left extension */
    for (i = 8; i < rows-8; i++){
        tmp = qp[i][8];                    
        memset(&qp[i][0],tmp,8*sizeof(unsigned char));
    }
    /* Right extension */
    for (i = 8; i < rows-8; i++){
        tmp = qp[i][cols-8-1];                    
        memset(&qp[i][cols-8],tmp,8*sizeof(unsigned char));
    }
    /* Top extension */
    for (i = 0; i < 8; i++){
        memcpy(&qp[i][0],&qp[8][0],cols*sizeof(unsigned char));
    }
    /* Bottom extension */
    for (i = rows-8; i < rows; i++){
        memcpy(&qp[i][0],&qp[rows-8-1][0],cols*sizeof(unsigned char));
    }


    /* Half pixel positions */
    for (i = N; i < rows-N+1; i++)
    {
        for (j = 0; j < cols; j+=16)
        {  
          const unsigned char *x00  = &qp[i][j+0];
          const unsigned char *x0p1 = &qp[i][j+1];
          const unsigned char *xp10 = &qp[i+1][j];
                             
          unsigned char *restrict qa = &qp[0*rows+i][2*cols+j]; 
          unsigned char *restrict qb = &qp[2*rows+i][0*cols+j]; 
          unsigned char *restrict qc = &qp[2*rows+i][2*cols+j]; 

          

          for (k = 0; k < 16; k++)
          {
              const unsigned char c00  = ~x00[k]; 
              const unsigned char c0p1 = ~x0p1[k];
              const unsigned char cp10 = ~xp10[k];
         
              qa[k] = ~((c00  + c0p1 + 1) >> 1); // a = (x00  + x0p1) >> 1
              qb[k] = ~((c00  + cp10 + 1) >> 1); // b = (x00  + xp10) >> 1
              qc[k] = ~((c0p1 + cp10 + 1) >> 1); // c = (x0p1 + xp10) >> 1
           }
        }
    }

    /* Quarter pixel positions */
    for (i = N; i < rows-N; i++)
    {
        for (j = 0; j < cols; j+=16)
        {  
            const unsigned char *x00  = &qp[0*rows+i][0*cols+j];
            const unsigned char *x0p1 = &qp[0*rows+i][0*cols+j+1];
            const unsigned char *x02p = &qp[0*rows+i][2*cols+j];
            const unsigned char *x2p0 = &qp[2*rows+i][0*cols+j];
            const unsigned char *x2p1 = &qp[2*rows+i][0*cols+j+1];
            const unsigned char *x1p0 = &qp[0*rows+i+1][0*cols+j];
            const unsigned char *x12p = &qp[0*rows+i+1][2*cols+j];
            const unsigned char *x1p1 = &qp[0*rows+i+1][0*cols+j+1];
           
            unsigned char *restrict qd = &qp[0*rows+i][1*cols+j];  
            unsigned char *restrict qe = &qp[0*rows+i][3*cols+j];  
            unsigned char *restrict qf = &qp[1*rows+i][0*cols+j];  
            unsigned char *restrict qg = &qp[1*rows+i][1*cols+j];  
            unsigned char *restrict qh = &qp[1*rows+i][2*cols+j];  
            unsigned char *restrict qi = &qp[1*rows+i][3*cols+j];  
            unsigned char *restrict qj = &qp[2*rows+i][1*cols+j];  
            unsigned char *restrict qk = &qp[2*rows+i][3*cols+j];  
            unsigned char *restrict ql = &qp[3*rows+i][0*cols+j];  
            unsigned char *restrict qm = &qp[3*rows+i][1*cols+j];  
            unsigned char *restrict qn = &qp[3*rows+i][2*cols+j];  
            unsigned char *restrict qo = &qp[3*rows+i][3*cols+j];  

            for (k = 0; k < 16; k++)
            {
                qd[k] = ((x00[k]  + x02p[k] + 1) >> 1);
                qe[k] = ((x02p[k] + x0p1[k] + 1) >> 1); 
                qf[k] = ((x00[k]  + x2p0[k] + 1) >> 1);
                qg[k] = ((x02p[k] + x2p0[k] + 1) >> 1);
                qh[k] = ((x00[k]  + x2p1[k] + 1) >> 1);
                qi[k] = ((x02p[k] + x2p1[k] + 1) >> 1); 
                qj[k] = ((x02p[k] + x1p0[k] + 1) >> 1);
                qk[k] = ((x0p1[k] + x12p[k] + 1) >> 1); 
                ql[k] = ((x2p0[k] + x1p0[k] + 1) >> 1);
                qm[k] = ((x2p0[k] + x12p[k] + 1) >> 1);
                qn[k] = ((x2p0[k] + x1p1[k] + 1) >> 1);
                qo[k] = ((x2p1[k] + x12p[k] + 1) >> 1);
            }
        }
    }
            
    for (ioff=0;ioff<=3*rows;ioff+=rows){
        for (joff=(ioff==0)*cols;joff<=3*cols;joff+=cols){
            
            /* Left extension */
            for (i = N; i < rows-N; i++){
                tmp = qp[ioff+i][joff+N];                    
                memset(&qp[ioff+i][joff],tmp,N*sizeof(unsigned char));
            }
            /* Right extension */
            for (i = N; i < rows-N; i++){
                tmp = qp[ioff+i][joff+cols-N-1];                    
                memset(&qp[ioff+i][joff+cols-N],tmp,N*sizeof(unsigned char));
            }
            /* Top extension */
            for (i = 0; i < N; i++){
                memcpy(&qp[ioff+i][joff],&qp[ioff+N][joff],cols*sizeof(unsigned char));
            }
            /* Bottom extension */
            for (i = rows-N; i < rows; i++){
                memcpy(&qp[ioff+i][joff],&qp[ioff+rows-N-1][joff],cols*sizeof(unsigned char));
            }
        }
    }
#else
	unsigned char tmp;
	int i,j,ioff,joff;
	int N = 4; // Assumes frame extension width is 16, and FilterLength/2 <=8


	/* Copy and extend integer pixels */
	for (i = 8; i < rows-8; i++){
		memcpy(&qp[i][8],&ip[i-8][0],(cols-2*8)*sizeof(unsigned char));
	}
	/* Left extension */
	for (i = 8; i < rows-8; i++){
		tmp = qp[i][8];                    
		memset(&qp[i][0],tmp,8*sizeof(unsigned char));
	}
	/* Right extension */
	for (i = 8; i < rows-8; i++){
		tmp = qp[i][cols-8-1];                    
		memset(&qp[i][cols-8],tmp,8*sizeof(unsigned char));
	}
	/* Top extension */
	for (i = 0; i < 8; i++){
		memcpy(&qp[i][0],&qp[8][0],cols*sizeof(unsigned char));
	}
	/* Bottom extension */
	for (i = rows-8; i < rows; i++){
		memcpy(&qp[i][0],&qp[rows-8-1][0],cols*sizeof(unsigned char));
	}


	/* Half pixel positions */
	for (i = N; i < rows-N+1; i++){
		for (j = N; j < cols-N+1; j++)
		{                                
			qp[0*rows+i][2*cols+j] = ((qp[i][j]+qp[i][j+1]) >> 1);
			qp[2*rows+i][0*cols+j] = ((qp[i][j]+qp[i+1][j]) >> 1);
			qp[2*rows+i][2*cols+j] = ((qp[i][j+1]+qp[i+1][j]) >> 1);
		}
	}

	/* Quarter pixel positions */
	for (i = N; i < rows-N; i++){
		for (j = N; j < cols-N; j++){
			qp[0*rows+i][1*cols+j] = ((qp[0*rows+i][0*cols+j]+qp[0*rows+i][2*cols+j]+1) >> 1);
			qp[0*rows+i][3*cols+j] = ((qp[0*rows+i][2*cols+j]+qp[0*rows+i][0*cols+j+1]+1) >> 1); 
			qp[1*rows+i][0*cols+j] = ((qp[0*rows+i][0*cols+j]+qp[2*rows+i][0*cols+j]+1) >> 1);
			qp[1*rows+i][1*cols+j] = ((qp[0*rows+i][2*cols+j]+qp[2*rows+i][0*cols+j]+1) >> 1);
			qp[1*rows+i][2*cols+j] = ((qp[0*rows+i][0*cols+j]+qp[2*rows+i][0*cols+j+1]+1) >> 1);
			qp[1*rows+i][3*cols+j] = ((qp[0*rows+i][2*cols+j]+qp[2*rows+i][0*cols+j+1]+1) >> 1); 
			qp[2*rows+i][1*cols+j] = ((qp[0*rows+i][2*cols+j]+qp[0*rows+i+1][0*cols+j]+1) >> 1);
			qp[2*rows+i][3*cols+j] = ((qp[0*rows+i][0*cols+j+1]+qp[0*rows+i+1][2*cols+j]+1) >> 1); 
			qp[3*rows+i][0*cols+j] = ((qp[2*rows+i][0*cols+j]+qp[0*rows+i+1][0*cols+j]+1) >> 1);
			qp[3*rows+i][1*cols+j] = ((qp[2*rows+i][0*cols+j]+qp[0*rows+i+1][2*cols+j]+1) >> 1);
			qp[3*rows+i][2*cols+j] = ((qp[2*rows+i][0*cols+j]+qp[0*rows+i+1][0*cols+j+1]+1) >> 1);
			qp[3*rows+i][3*cols+j] = ((qp[2*rows+i][0*cols+j+1]+qp[0*rows+i+1][2*cols+j]+1) >> 1);
		}
	}

	for (ioff=0;ioff<=3*rows;ioff+=rows){
		for (joff=(ioff==0)*cols;joff<=3*cols;joff+=cols){

			/* Left extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+N];                    
				memset(&qp[ioff+i][joff],tmp,N*sizeof(unsigned char));
			}
			/* Right extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+cols-N-1];                    
				memset(&qp[ioff+i][joff+cols-N],tmp,N*sizeof(unsigned char));
			}
			/* Top extension */
			for (i = 0; i < N; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+N][joff],cols*sizeof(unsigned char));
			}
			/* Bottom extension */
			for (i = rows-N; i < rows; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+rows-N-1][joff],cols*sizeof(unsigned char));
			}
		}
	}
#endif
	return 0;
}



int interpolate_Y2(unsigned char **ip,unsigned char **qp,int rows,int cols)
{

	unsigned char tmp;
	int i,j,ioff,joff,ii;
	int out1[6];
	int N = 8; // Assumes frame extension width is 16, and FilterLength/2 <=8
	static const int f1[6] = {3,-15,111,37,-10,2};
	static const int f2[6] = {3,-17,78,78,-17,3};
	static const int f3[6] = {2,-10,37,111,-15,3};


	/* Copy and pad integer pixels */
	for (i = 16; i < rows-16; i++){
		memcpy(&qp[i][16],&ip[i-16][0],(cols-2*16)*sizeof(unsigned char));
	}    
	/* Left extension */
	for (i = 16; i < rows-16; i++){
		tmp = qp[i][16];                    
		memset(&qp[i][0],tmp,16*sizeof(unsigned char));
	}
	/* Right extension */
	for (i = 16; i < rows-16; i++){
		tmp = qp[i][cols-16-1];                    
		memset(&qp[i][cols-16],tmp,16*sizeof(unsigned char));
	}
	/* Top extension */
	for (i = 0; i < 16; i++){
		memcpy(&qp[i][0],&qp[16][0],cols*sizeof(unsigned char));
	}
	/* Bottom extension */
	for (i = rows-16; i < rows; i++){
		memcpy(&qp[i][0],&qp[rows-16-1][0],cols*sizeof(unsigned char));
	}

	for (i = N; i < rows-N; i++){
		for (j = N; j < cols-N; j++){


			for(ii=0;ii<6;ii++)
			{
				out1[ii] = clip255((
					+  f1[0]*qp[i-2+ii][j-2]
				+  f1[1]*qp[i-2+ii][j-1]
				+  f1[2]*qp[i-2+ii][j]
				+  f1[3]*qp[i-2+ii][j+1]
				+  f1[4]*qp[i-2+ii][j+2]
				+  f1[5]*qp[i-2+ii][j+3] + 64) >> 7);
			}

			//e
			qp[(0+1)*rows+i][(0+1)*cols+j] = clip255(((f1[0]*out1[0] 
			+ f1[1]*out1[1] 
			+ f1[2]*out1[2] 
			+ f1[3]*out1[3] 
			+ f1[4]*out1[4] 
			+ f1[5]*out1[5] ) + 64 ) >> 7);
      
      //i
			qp[(0+2)*rows+i][(0+1)*cols+j] = clip255(((
				+ f2[0]*out1[0]
			+ f2[1]*out1[1]
			+ f2[2]*out1[2]
			+ f2[3]*out1[3]
			+ f2[4]*out1[4]
			+ f2[5]*out1[5]) + 64) >> 7);

			//m
			qp[(0+3)*rows+i][(0+1)*cols+j] = clip255(((
				+ f3[0]*out1[0]
			+ f3[1]*out1[1]
			+ f3[2]*out1[2]
			+ f3[3]*out1[3]
			+ f3[4]*out1[4]
			+ f3[5]*out1[5] ) + 64) >> 7);

			for(ii=0;ii<6;ii++)
			{
				out1[ii] = clip255((
					+  f2[0]*qp[i-2+ii][j-2]
				+  f2[1]*qp[i-2+ii][j-1]
				+  f2[2]*qp[i-2+ii][j]
				+  f2[3]*qp[i-2+ii][j+1]
				+  f2[4]*qp[i-2+ii][j+2]
				+  f2[5]*qp[i-2+ii][j+3] + 64) >> 7);
			}

			//f
			qp[(0+1)*rows+i][(0+2)*cols+j] = clip255(((
				+ f1[0]*out1[0]
			+ f1[1]*out1[1]
			+ f1[2]*out1[2]
			+ f1[3]*out1[3]
			+ f1[4]*out1[4]
			+ f1[5]*out1[5] ) + 64) >> 7);


			//j
			qp[(0+2)*rows+i][(0+2)*cols+j] = clip255(((
				+ f2[0]*out1[0]
			+ f2[1]*out1[1]
			+ f2[2]*out1[2]
			+ f2[3]*out1[3]
			+ f2[4]*out1[4]
			+ f2[5]*out1[5] ) + 64) >> 7);

			//n
			qp[(0+3)*rows+i][(0+2)*cols+j] = clip255(((
				+ f3[0]*out1[0]
			+ f3[1]*out1[1]
			+ f3[2]*out1[2]
			+ f3[3]*out1[3]
			+ f3[4]*out1[4]
			+ f3[5]*out1[5] ) + 64) >> 7);

			for(ii=0;ii<6;ii++)
			{
				out1[ii] = clip255((
					+  f3[0]*qp[i-2+ii][j-2]
				+  f3[1]*qp[i-2+ii][j-1]
				+  f3[2]*qp[i-2+ii][j]
				+  f3[3]*qp[i-2+ii][j+1]
				+  f3[4]*qp[i-2+ii][j+2]
				+  f3[5]*qp[i-2+ii][j+3] + 64) >> 7);
			}

			//g
			qp[(0+1)*rows+i][(0+3)*cols+j] = clip255(((
				+ f1[0]*out1[0]
			+ f1[1]*out1[1]
			+ f1[2]*out1[2]
			+ f1[3]*out1[3]
			+ f1[4]*out1[4]
			+ f1[5]*out1[5] ) + 64) >> 7);

			//k
			qp[(0+2)*rows+i][(0+3)*cols+j]= clip255(((
				+ f2[0]*out1[0]
			+ f2[1]*out1[1]
			+ f2[2]*out1[2]
			+ f2[3]*out1[3]
			+ f2[4]*out1[4]
			+ f2[5]*out1[5] ) + 64) >> 7);

			//o
			qp[(0+3)*rows+i][(0+3)*cols+j] = clip255(((
				+ f3[0]*out1[0]
			+ f3[1]*out1[1]
			+ f3[2]*out1[2]
			+ f3[3]*out1[3]
			+ f3[4]*out1[4]
			+ f3[5]*out1[5] ) + 64) >> 7);

		}
	}

	for (ioff=0*rows;ioff<=3*rows;ioff+=rows){
		for (joff=(ioff==0)*cols;joff<=3*cols;joff+=cols){

			/* Left extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+N];                    
				memset(&qp[ioff+i][joff],tmp,N*sizeof(unsigned char));
			}
			/* Right extension */
			for (i = N; i < rows-N; i++){
				tmp = qp[ioff+i][joff+cols-N-1];                    
				memset(&qp[ioff+i][joff+cols-N],tmp,N*sizeof(unsigned char));
			}
			/* Top extension */
			for (i = 0; i < N; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+N][joff],cols*sizeof(unsigned char));
			}
			/* Bottom extension */
			for (i = rows-N; i < rows; i++){
				memcpy(&qp[ioff+i][joff],&qp[ioff+rows-N-1][joff],cols*sizeof(unsigned char));
			}
		}
	}

	return 0;

}

void get_luma_inter_prediction_8x8(macroblock_info *macroblockInfo,int vy,int vx,image_frame **iprev,unsigned char pblockr[8][8])
{
	int k,l,ypos,xpos,mbx,mby,r,yoff,xoff,rows,cols,ypos2,xpos2,yp,xp,yp2,xp2,mvx,mvy,mode;    

	k = vy&1;
	l = vx&1;
	mbx = macroblockInfo->mbx;
	mby = macroblockInfo->mby;
	xpos = 16*macroblockInfo->mbx + 8*l;
	ypos = 16*macroblockInfo->mby + 8*k;
	r = macroblockInfo->ref;
	mode = macroblockInfo->mode;
	rows = iprev[0]->rows/4;
	cols = iprev[0]->cols/4;

	/* 8x8 motion compensation */

	if (mode == MODE_DIRECT){

       
		mvy = (macroblockInfo->mvblock[k][l].y+2)&~3;
		mvx = (macroblockInfo->mvblock[k][l].x+2)&~3;
 
		yoff = mvy&3;
		xoff = mvx&3;
		ypos2 = ypos + (mvy>>2) + 16; 
		xpos2 = xpos + (mvx>>2) + 16;
        yp = min(max(6,ypos2),rows-15)+yoff*rows;
		xp = min(max(6,xpos2),cols-15)+xoff*cols; 

		yoff = (-mvy)&3;
		xoff = (-mvx)&3;
		ypos2 = ypos + ((-mvy)>>2) + 16; 
		xpos2 = xpos + ((-mvx)>>2) + 16; 
		yp2 = min(max(6,ypos2),rows-15)+yoff*rows;
		xp2 = min(max(6,xpos2),cols-15)+xoff*cols; 

        if(macroblockInfo->mvblock[k][l].interp_indx == 0) 
        {
            for (k=0; k<8; k++)
                for (l=0; l<8; l++)
                    pblockr[k][l] = (iprev[0]->Y[yp+k][xp+l]+iprev[1]->Y[yp2+k][xp2+l]+g_bround)>>1;
        }
        else
        {
            for (k=0; k<8; k++)
                for (l=0; l<8; l++)
                    pblockr[k][l] = (iprev[0]->Y2[yp+k][xp+l]+iprev[1]->Y2[yp2+k][xp2+l]+g_bround)>>1;
        }
	}
	else{

		if(r==2)
		{
			mvy = macroblockInfo->mvblock[k][l].y;
			mvx = macroblockInfo->mvblock[k][l].x;
			yoff = mvy&3;
			xoff = mvx&3;
			ypos2 = ypos + (mvy>>2) + 16; 
			xpos2 = xpos + (mvx>>2) + 16;
			yp = min(max(6,ypos2),rows-15)+yoff*rows;
			xp = min(max(6,xpos2),cols-15)+xoff*cols;

			mvy = macroblockInfo->mvblockBW[k][l].y;
			mvx = macroblockInfo->mvblockBW[k][l].x;
			yoff = mvy&3;
			xoff = mvx&3;
			ypos2 = ypos + (mvy>>2) + 16; 
			xpos2 = xpos + (mvx>>2) + 16; 
			yp2 = min(max(6,ypos2),rows-15)+yoff*rows;
			xp2 = min(max(6,xpos2),cols-15)+xoff*cols; 

            if(macroblockInfo->mvblockBW[k][l].interp_indx == 0) 
            {
                if(macroblockInfo->mvblock[k][l].interp_indx == 0) 
                {
                    for (k=0; k<8; k++)
                        for (l=0; l<8; l++)
                            pblockr[k][l] = (iprev[0]->Y[yp+k][xp+l]+iprev[1]->Y[yp2+k][xp2+l]+g_bround)>>1;
                }
                else
                {
                    for (k=0; k<8; k++)
                        for (l=0; l<8; l++)
                            pblockr[k][l] = (iprev[0]->Y2[yp+k][xp+l]+iprev[1]->Y[yp2+k][xp2+l]+g_bround)>>1;
                }
            }
            else
            {
                if(macroblockInfo->mvblock[k][l].interp_indx == 0) 
                {
                    for (k=0; k<8; k++)
                        for (l=0; l<8; l++)
                            pblockr[k][l] = (iprev[0]->Y[yp+k][xp+l]+iprev[1]->Y2[yp2+k][xp2+l]+g_bround)>>1;
                }
                else
                {
                    for (k=0; k<8; k++)
                        for (l=0; l<8; l++)
                            pblockr[k][l] = (iprev[0]->Y2[yp+k][xp+l]+iprev[1]->Y2[yp2+k][xp2+l]+g_bround)>>1;
                }
            }
		}
		else
		{
			mvy = macroblockInfo->mvblock[k][l].y;
			mvx = macroblockInfo->mvblock[k][l].x;
			yoff = mvy&3;
			xoff = mvx&3;
			ypos2 = ypos + (mvy>>2) + 16; 
			xpos2 = xpos + (mvx>>2) + 16;
            yp = min(max(6,ypos2),rows-15)+yoff*rows;
            xp = min(max(6,xpos2),cols-15)+xoff*cols;

            if(macroblockInfo->mvblock[k][l].interp_indx == 0) 
            {
                for (k=0; k<8; k++)
                    memcpy(&(pblockr[k][0]),&iprev[r]->Y[yp+k][xp],8*sizeof(unsigned char));
            }
            else
            {
                for (k=0; k<8; k++)
                    memcpy(&(pblockr[k][0]),&iprev[r]->Y2[yp+k][xp],8*sizeof(unsigned char));
            }       
		}
	}
}


void get_chroma_inter_prediction_mb(macroblock_info *macroblockInfo,image_frame **iprev,unsigned char pblockr[8][8], int comp, int rounding)
{
	int i,j,k,l;
	int cypos, cxpos, mbx,mby;    
	int r;
	//STATIC unsigned char *pblock[32];
	//STATIC unsigned char *pblockBW[32];
	int yoff,xoff,rows,cols,ypos2,xpos2,yp,xp,yp2,xp2,mvx,mvy;


	mbx = macroblockInfo->mbx;
	mby = macroblockInfo->mby;
	cxpos = mbx*8;
	cypos = mby*8;


	r = macroblockInfo->ref;

	rows = iprev[0]->crows/4;
	cols = iprev[0]->ccols/4;

	for (i=0;i<2;i++){
		for (j=0;j<2;j++)
        {
			if (macroblockInfo->mode == MODE_DIRECT)
            {
				mvy = ((macroblockInfo->mvblock[i][j].y+4)>>3)<<2;
				mvx = ((macroblockInfo->mvblock[i][j].x+4)>>3)<<2;


				yoff = mvy&3;
				xoff = mvx&3;
				ypos2 = cypos + (mvy>>2) + 8; 
				xpos2 = cxpos + (mvx>>2) + 8;                
				yp = min(max(0,ypos2),rows-8)+yoff*rows;
				xp = min(max(0,xpos2),cols-8)+xoff*cols;                   

				yoff = (-mvy)&3;
				xoff = (-mvx)&3;
				ypos2 = cypos + ((-mvy)>>2) + 8; 
				xpos2 = cxpos + ((-mvx)>>2) + 8;                
				yp2 = min(max(0,ypos2),rows-8)+yoff*rows;
                xp2 = min(max(0,xpos2),cols-8)+xoff*cols;                   

                if (comp==0)
                {
				for (k=i*4;k<i*4+4;k++)
				{
					for (l=j*4;l<j*4+4;l++)
					{
                            pblockr[k][l] = (iprev[0]->U[yp+k][xp+l]+iprev[1]->U[yp2+k][xp2+l]+g_bround)>>1;
                        }
                    }
                }
                else
                {
                    for (k=i*4;k<i*4+4;k++)
                    {
                        for (l=j*4;l<j*4+4;l++)
                        {
                            pblockr[k][l] = (iprev[0]->V[yp+k][xp+l]+iprev[1]->V[yp2+k][xp2+l]+g_bround)>>1;
                        }
					}
				}
			}
			else
			{

				if(r==2)
				{
					mvy = (macroblockInfo->mvblock[i][j].y+1)>>1;
					mvx = (macroblockInfo->mvblock[i][j].x+1)>>1;
					yoff = mvy&3;
					xoff = mvx&3;
					ypos2 = cypos + (mvy>>2) + 8; 
					xpos2 = cxpos + (mvx>>2) + 8;                
					yp = min(max(0,ypos2),rows-8)+yoff*rows;
					xp = min(max(0,xpos2),cols-8)+xoff*cols;                   

					mvy = (macroblockInfo->mvblockBW[i][j].y+1)>>1;
					mvx = (macroblockInfo->mvblockBW[i][j].x+1)>>1;
					yoff = mvy&3;
					xoff = mvx&3;
					ypos2 = cypos + (mvy>>2) + 8; 
					xpos2 = cxpos + (mvx>>2) + 8;                
                    yp2 = min(max(0,ypos2),rows-8)+yoff*rows;
                    xp2 = min(max(0,xpos2),cols-8)+xoff*cols;                   

                    if (comp==0)
                    {
					for (k=i*4;k<i*4+4;k++)
					{
						for (l=j*4;l<j*4+4;l++)
						{
                                pblockr[k][l] = (iprev[0]->U[yp+k][xp+l]+iprev[1]->U[yp2+k][xp2+l]+g_bround)>>1;
                            }
                        }
                    }
                    else
                    {
                        for (k=i*4;k<i*4+4;k++)
                        {
                            for (l=j*4;l<j*4+4;l++)
                            {
                                pblockr[k][l] = (iprev[0]->V[yp+k][xp+l]+iprev[1]->V[yp2+k][xp2+l]+g_bround)>>1;
                            }
						}
					}
				}
				else
				{
					mvy = (macroblockInfo->mvblock[i][j].y+rounding)>>1;
					mvx = (macroblockInfo->mvblock[i][j].x+rounding)>>1;

					yoff = mvy&3;
					xoff = mvx&3;
					ypos2 = cypos + (mvy>>2) + 8; 
					xpos2 = cxpos + (mvx>>2) + 8;                
					yp = min(max(0,ypos2),rows-8)+yoff*rows;
					xp = min(max(0,xpos2),cols-8)+xoff*cols;                   

                    if (comp==0)
                    {
                        for (k=i*4; k<i*4+4; k++)
                        {
                            memcpy(&(pblockr[k][j*4]),&iprev[r]->U[yp+k][xp+j*4],4*sizeof(unsigned char));
                        }
                    }
                    else
                    {
                        for (k=i*4; k<i*4+4; k++)
                        {
                            memcpy(&(pblockr[k][j*4]),&iprev[r]->V[yp+k][xp+j*4],4*sizeof(unsigned char));
                        }
					}
				}



			}
		}
	}
}
