/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/



#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "global.h"
#include "tntmenc.h"
#include "transform.h"
#include "putvlc.h"
//#include "rfutil.h"
#include "defines.h"
#include "cost.h"
                      	
extern int transmitInterpIndx[16];


extern int g_enableSwitched;

static const int num_partitions[] = {1, 2, 2, 4};
static const part_dimensions part_dims[][4] = {
    {{16,16, 0, 0}},
    {{ 8,16, 0, 0}, { 8,16, 8, 0}},
    {{16, 8, 0, 0}, {16, 8, 0, 8}},
    {{ 8, 8, 0, 0}, { 8, 8, 0, 8}, { 8, 8, 8, 0}, { 8, 8, 8, 8}}
};
static const partition part_id[][4] = {
    {PART_ALL},
    {PART_UP, PART_DOWN},
    {PART_LEFT, PART_RIGHT},
    {PART_8x8_0, PART_8x8_1, PART_8x8_2, PART_8x8_3}
};

extern float rd_lambda_acc(int qp);

int diamond_8x8(encoder_info *encoderInfo,macroblock_info *macroblockInfo,image_frame *pad_prev,mvpos start,int start_sad,
                int step,partition part,mvpos *best,int *best_sad,
                mvpos refpos,mvpos pred,int quant,direction dir,int *last_best);
int full_search(encoder_info *encoderInfo,macroblock_info *macroblockInfo,image_frame *pad_prev,mvpos start,int start_sad,
                int range,partition part,mvpos *best,int *best_sad,
                mvpos refpos,mvpos pred,int quant,direction dir,int *last_best);
unsigned int candidate_search(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,
                              int ref,int *rsad,mvpos *vec,mode_t mode);
unsigned int fract_sad(encoder_info *encoderInfo,unsigned char *mblock2,/*unsigned char **mblock,*/int yw,int xw,
                       unsigned int sad_min,int *ydelta,int *xdelta,int yvec,int xvec,
                       image_frame *qpimage,mvpos refpos,mvpos pred,int quant,int pic_type,int *interp_type,mode_t mode);


void calculate_sums(image_frame *pad_prev,int rows,int cols)
{
	int i, j;
	unsigned int sum, sum2, tmp_sum;

	/* Horizontal summation */
	for (i = 0; i < rows; i++)
	{
		sum = 0;
		for (j = 0; j < 8; j++)
		{
			sum += pad_prev->Y[i][j];
		}
		for (j = 0; j < cols-8; j++)
		{
			pad_prev->SEA[i][j] = (unsigned short) sum;
			sum -= pad_prev->Y[i][j];
			sum += pad_prev->Y[i][j+8];
		}
	}

	/* Vertical summation */
	for (j = 0; j < cols-8; j+=2)
	{
		sum  = 0;
		sum2 = 0;
		for (i = 0; i < 8; i++)
		{
			sum  += pad_prev->SEA[i][j  ];
			sum2 += pad_prev->SEA[i][j+1];
		}
		for (i = 0; i < rows-8; i++)
		{
			tmp_sum = sum;
			sum -= pad_prev->SEA[i][j];
			pad_prev->SEA[i][j] = (unsigned short) tmp_sum;
			tmp_sum = sum2;
			sum2 -= pad_prev->SEA[i][j+1];
			pad_prev->SEA[i][j+1] = (unsigned short) tmp_sum;
			sum  += pad_prev->SEA[i+8][j  ];
			sum2 += pad_prev->SEA[i+8][j+1];
		}
	}
}

static inline unsigned int full_search_fine(const unsigned char *org,
                                            const unsigned char *ref,
                                            const int stride,
                                            const encoder_info *encoderInfo,
                                            const image_frame *pad_prev,                       
                                            const unsigned int start_sad,
                                            const int range_y, 
                                            const int range_x,
                                            const int quant,
                                            const mvpos start,
                                            const mvpos refpos,
                                            const mvpos pred,   
                                            mvpos *restrict best                 
                                            )
{
    int i,j,x,y;
    const int min_y = max(start.y - range_y,2);
    const int min_x = max(start.x - range_x,2);
    const int max_y = min(start.y + range_y + 1,encoderInfo->extHeight-17);
    const int max_x = min(start.x + range_x + 1,encoderInfo->extWidth-17);
    unsigned int best_sad = start_sad;
    unsigned char org_[256];
    for (j=0;j<256;j++)
        org_[j] = org[j];    

    best->y = start.y;
    best->x = start.x;

    for(y = min_y; y < max_y; y++)
    {
        for (x = min_x; x < max_x; x++)
        {
            const int dx = 4*(x - refpos.x) - pred.x;
            const int dy = 4*(y - refpos.y) - pred.y;                                           
            unsigned int tmp_sad = rd_cost_lookup(pad_prev,quant,dx,dy);
            for (i=0;i<16;i+=4)
            {
                unsigned char ref_[64];
                unsigned int sad = 0;
                for (j=0;j<16;j++)
                {
                    ref_[0*16+j] = ref[(y + i + 0)*stride + j + x];                                  
                    ref_[1*16+j] = ref[(y + i + 1)*stride + j + x];    
                    ref_[2*16+j] = ref[(y + i + 2)*stride + j + x];                                  
                    ref_[3*16+j] = ref[(y + i + 3)*stride + j + x];   
                }                    
                for (j=0;j<64;j++)
                    sad += abs(org_[i*16+j] - ref_[j]);                                  
                tmp_sad += sad;
                if (tmp_sad > best_sad)
                    break;
            }
            if (tmp_sad < best_sad)
            {
                best_sad = tmp_sad;
                best->y = y;
                best->x = x;
            }
        }
    }
    return best_sad;
}

static inline unsigned int full_search_coarse(const unsigned char *org,
                                              const unsigned char *ref,
                                              const int stride,
                                              const encoder_info *encoderInfo,
                                              const image_frame *pad_prev,                       
                                              const unsigned int start_sad,
                                              const int range_y, 
                                              const int range_x,
                                              const int quant,
                                              const int grid,
                                              const mvpos start,
                                              const mvpos refpos,
                                              const mvpos pred,   
                                              mvpos *restrict best                 
                                              )
{
    int j,x,y;
    const int min_y = max(start.y - range_y,2);
    const int min_x = max(start.x - range_x,2);
    const int max_y = min(start.y + range_y + 1,encoderInfo->extHeight-17);
    const int max_x = min(start.x + range_x + 1,encoderInfo->extWidth-17);
    unsigned int best_sad = start_sad;
    unsigned char org_[256];
    for (j=0;j<256;j++)
        org_[j] = org[j];   

    best->y = start.y;
    best->x = start.x;

    for(y = min_y; y < max_y; y+=grid)
    {
        for (x = min_x; x < max_x; x+=grid)
        {
            int i,j;
            const int dx = 4*(x - refpos.x) - pred.x;
            const int dy = 4*(y - refpos.y) - pred.y;                                           
            unsigned int tmp_sad = rd_cost_lookup(pad_prev,quant,dx,dy);
            for (i=0;i<16;i+=4)
            {
                unsigned char ref_[64];
                unsigned int sad = 0;
                for (j=0;j<16;j++)
                {
                    ref_[0*16+j] = ref[(y + i + 0)*stride + j + x];                                  
                    ref_[1*16+j] = ref[(y + i + 1)*stride + j + x];    
                    ref_[2*16+j] = ref[(y + i + 2)*stride + j + x];                                  
                    ref_[3*16+j] = ref[(y + i + 3)*stride + j + x];   
                }                    
                for (j=0;j<64;j++)
                    sad += abs(org_[i*16+j] - ref_[j]);                                  
                tmp_sad += sad;
                if (tmp_sad > best_sad)
                    break;
            }
            if (tmp_sad < best_sad)
            {
                best_sad = tmp_sad;
                best->y = y;
                best->x = x;
            }
        }
    }
    return best_sad;
}

inline int sad_any(unsigned char **a,unsigned char **b,int row,int col,
                   int rows,int cols,int sadval,int bestsad)
{
    /* Do a SAD calculation, with height of rows and width of cols */
    int i,j;

    for (i=0;i<rows && sadval<bestsad;i++)
    {
        for (j=0;j<cols;j+=8)
        {
            sadval += abs(a[i][j  ]-b[row+i][col+j  ]);
            sadval += abs(a[i][j+1]-b[row+i][col+j+1]);
            sadval += abs(a[i][j+2]-b[row+i][col+j+2]);
            sadval += abs(a[i][j+3]-b[row+i][col+j+3]);
            sadval += abs(a[i][j+4]-b[row+i][col+j+4]);
            sadval += abs(a[i][j+5]-b[row+i][col+j+5]);
            sadval += abs(a[i][j+6]-b[row+i][col+j+6]);
            sadval += abs(a[i][j+7]-b[row+i][col+j+7]);
        }
    }
    return sadval;
}

int diamond_NxN(encoder_info *encoderInfo,unsigned char *org,image_frame *pad_prev, partition part,mvpos *best_vec,int *best_cost,
                mvpos refpos,mvpos pred,int quant,mode_t mode)

{       
    int increment = 1;
    int min_cost;    
    mvpos min_vec;
    mvpos pos;
    mvpos vec;
    int width,height,offx,offy,stride_org,stride_ref;
    int ix,iy; 
    int ylimit = encoderInfo->extHeight-17;
    int xlimit = encoderInfo->extWidth-17;
    int dir,best_dir,new_dir=4;
    int step;
    mvpos center = *best_vec;
    int sad,cost;
    int shift=0;
    int dist=(center.x*4-pred.x)*(center.x*4-pred.x)+(center.y*4-pred.y)*(center.y*4-pred.y);
    
    if (dist>(4*4*2*2)) shift=1;
    if (dist>(4*4*4*4)) shift=2;
    if (dist>(4*4*8*8)) shift=3;

    if (mode<MODE_LMB32){
        width = 8*(part<3) + 8;
        height = 8*(part==0 || part==3 || part==4) + 8;
        offx = 8*(part==4 || part==6 || part==8);
        offy = 8*(part==2 || part==7 || part==8);
        stride_org = 16;
        ylimit = encoderInfo->extHeight-17;
        xlimit = encoderInfo->extWidth-17;
    }
    else if(mode==MODE_LMB32){
        width = height = stride_org = 32;
        offx = offy = 0;
        ylimit = encoderInfo->extHeight-33;
        xlimit = encoderInfo->extWidth-33;
    }
    else{
        width = height = stride_org = 64;
        offx = offy = 0;
        ylimit = encoderInfo->extHeight-65;
        xlimit = encoderInfo->extWidth-65;
    }
    stride_ref = 4*encoderInfo->extWidth;

    for (step=1;step<10;step++){
        increment = (1<<shift);
        min_cost = (1<<30);
        for (dir=0;dir<4;dir++){
            ix = (dir&1) + ((dir>>1)&1) - 1;
            iy = (!(dir&1)) + ((dir>>1)&1) - 1;
            if (new_dir == (dir^3)){
                ix = ix*2;
                iy = iy*2;
            }
            vec.y = center.y + increment*iy;
            vec.x = center.x + increment*ix; 
            pos.y = refpos.y + vec.y;
            pos.x = refpos.x + vec.x;
            if (pos.y <= 1 || pos.y >= ylimit || pos.x <= 1 || pos.x >= xlimit) continue;

            sad = SAD_NxN(&org[offy*stride_org+offx],&pad_prev->Y[pos.y+offy][pos.x+offx],stride_org,stride_ref, mode);
           
            cost = sad + rd_cost_lookup(pad_prev,quant,4*vec.x-pred.x,4*vec.y-pred.y);
            if (cost < min_cost){
                min_cost = cost;
                min_vec = vec;
                best_dir = dir;
            }              
        }
        if (min_cost < *best_cost){
            *best_cost = min_cost;            
            *best_vec = min_vec;
        }
        else{
            if (shift>0)
                shift--;
            else
                break;
        }
        new_dir = best_dir;
        center = min_vec;
    } // for step
    return min_cost;
}



int full_search(encoder_info *encoderInfo,macroblock_info *macroblockInfo/*,unsigned char **mblock*/,image_frame *pad_prev,mvpos start,int start_sad,
                int range,partition part,mvpos *best,int *best_sad,
                mvpos refpos,mvpos pred,int quant,direction dir,int *last_best)

{
    unsigned char *cblock[16];
    int tmp_sad;
    int dy,dx,cost;
    int fsad;
    mvpos f;
    int brows, bcols,row_offs,col_offs;
    int candy,candx;
    int dist,side,pos;
    int i,j,sum,candidate_sum;
    static const int spiral_dir[4][2] = {{1,0},{0,1},{-1,0},{0,-1}};
    static const part_dimensions blk_params[] =
    {
        {16,16,0,0}, { 8,16,0,0}, { 8,16,8,0}, {16, 8,0,0}, 
        {16, 8,0,8}, { 8, 8,0,0}, { 8, 8,0,8}, { 8, 8,8,0}, 
        { 8, 8,8,8}
    };

    brows = blk_params[part].h;
    bcols = blk_params[part].w;
    row_offs = blk_params[part].y;
    col_offs = blk_params[part].x;


    for (i=0;i<brows;i++){
        cblock[i] = &macroblockInfo->orgY[16*(row_offs+i)+col_offs];  
    }

    for (sum = 0, i = 0; i < brows; i++)
    {
        for (j = 0; j < bcols; j++)
        {
            sum += cblock[i][j];
        }
    }

    if (start_sad < *best_sad)
    {
        *best_sad = start_sad;
        best->y = start.y;
        best->x = start.x;
    }

    fsad = *best_sad;
    f.y = best->y;
    f.x = best->x;

    candy = start.y;
    candx = start.x;

    /* Distance from the search center */    
    for (dist = 1; dist <= range; dist++)
    {
        candy--;
        candx--;
        /* Search is peformed in spiral fashion, where spiral has square shape */
        for (side = 0; side < 4; side++)
        {
            /* Evaluate all vectors in the current side */
            for (pos = 0; pos < 2*dist; pos++)
            {                
                if (candy > 1 && candy < encoderInfo->extHeight-17 && candx > 1 && candx < encoderInfo->extWidth-17)
                {
                    dy = candy - refpos.y;
                    dx = candx - refpos.x;

                    cost = rd_cost_lookup(pad_prev,quant,4*dx-pred.x,4*dy-pred.y);

                    /* Sum for candidate block */
                    candidate_sum = pad_prev->SEA[candy+row_offs][candx+col_offs];
                    if (bcols == 16)
                    {
                        candidate_sum += pad_prev->SEA[candy+row_offs][candx+col_offs+8];
                    }
                    if (brows == 16)
                    {
                        candidate_sum += pad_prev->SEA[candy+row_offs+8][candx+col_offs];
                        if (bcols == 16)
                        {
                            candidate_sum += pad_prev->SEA[candy+row_offs+8][candx+col_offs+8];
                        }
                    }                    
                    /* Check Successive Elimination Algorithm condition */
                    if (abs(sum - candidate_sum) + cost <= fsad)
                    {
                        
                        tmp_sad = sad_any(cblock,pad_prev->Y,candy+row_offs,candx+col_offs,brows,bcols,cost,fsad);
                                                
                        if (tmp_sad < fsad)
                        {
                            fsad = tmp_sad;
                            f.y = candy;
                            f.x = candx;
                        }
                    }
                }

                candy += spiral_dir[side][0];
                candx += spiral_dir[side][1];
            }
        }
    }

    if (fsad < *best_sad)
    {
        *best_sad = fsad;
        best->y = f.y;
        best->x = f.x;
    }

    return fsad;
}


unsigned int candidate_search(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int ref,
                              int *rsad,mvpos *vec,mode_t mode)
{
    int xref,yref,i,vdx,vdy,cost;
    //STATIC unsigned char *pblock[16];
    unsigned int sad0,sad1,sad2,sad3,tmpsad;

    //image_frame *curr = encoderInfo->orig;

    image_frame *pad_prev = encoderInfo->ref_frame_qp[ref];


    int xpos = 16*macroblockInfo->mbx;
    int ypos = 16*macroblockInfo->mby;
    mvpos pred[4];
    mvpos c[4];
    int quant = frameInfo->qp;

    c[0] = macroblockInfo->candidateMV[ref][0];
    c[1] = macroblockInfo->candidateMV[ref][1];
    c[2] = macroblockInfo->candidateMV[ref][2];
    c[3] = macroblockInfo->candidateMV[ref][3];

    pred[0] = macroblockInfo->pred[ref][mode - MODE_16x16][0];
    pred[1] = macroblockInfo->pred[ref][mode - MODE_16x16][1];
    pred[2] = macroblockInfo->pred[ref][mode - MODE_16x16][2];
    pred[3] = macroblockInfo->pred[ref][mode - MODE_16x16][3];


    xref = xpos+16;
    yref = ypos+16;
    for (i = 0; i < 4; i++){

        if (xref+c[i].x > 1 && xref+c[i].x < encoderInfo->extWidth-17 && yref+c[i].y > 1 && yref+c[i].y < encoderInfo->extHeight-17)
        {
            const int cols = pad_prev->cols;

            //STATIC move_ucsubmatrix(pblock,pad_prev->Y,yref+c[i].y,yref+c[i].y+15,xref+c[i].x,xref+c[i].x+15,0,0);
            vdx = c[i].x*4-pred[0].x;
            vdy = c[i].y*4-pred[0].y;
            cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
            

            sad0 = SAD_8x8(&macroblockInfo->orgY[0*16+0],&pad_prev->Y[yref+c[i].y+0][xref+c[i].x+0],16,cols);
            sad1 = SAD_8x8(&macroblockInfo->orgY[0*16+8],&pad_prev->Y[yref+c[i].y+0][xref+c[i].x+8],16,cols);
            sad2 = SAD_8x8(&macroblockInfo->orgY[8*16+0],&pad_prev->Y[yref+c[i].y+8][xref+c[i].x+0],16,cols);
            sad3 = SAD_8x8(&macroblockInfo->orgY[8*16+8],&pad_prev->Y[yref+c[i].y+8][xref+c[i].x+8],16,cols);
        
            if (mode == MODE_16x16){
                tmpsad = sad0+sad1+sad2+sad3+cost;
                if ((int)tmpsad < rsad[0]){
                    rsad[0] = rsad[1] = rsad[2] = rsad[3] = tmpsad;
                    vec[0].x = vec[1].x = vec[2].x = vec[3].x = c[i].x;
                    vec[0].y = vec[1].y = vec[2].y = vec[3].y = c[i].y;
                }
            }
            else if (mode == MODE_16x8){
                tmpsad = sad0+sad1+cost;
                if ((int)tmpsad < rsad[0]){
                    rsad[0] = rsad[1] = tmpsad;
                    vec[0].x = vec[1].x = c[i].x;
                    vec[0].y = vec[1].y = c[i].y;
                }
                vdx = c[i].x*4-pred[2].x;
                vdy = c[i].y*4-pred[2].y;
                cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
                tmpsad = sad2+sad3+cost;
                if ((int)tmpsad < rsad[2]){
                    rsad[2] = rsad[3] = tmpsad;
                    vec[2].x = vec[3].x = c[i].x;
                    vec[2].y = vec[3].y = c[i].y;
                }
            }
            else if (mode == MODE_8x16){
                tmpsad = sad0+sad2+cost;
                if ((int)tmpsad < rsad[0]){
                    rsad[0] = rsad[2] = tmpsad;
                    vec[0].x = vec[2].x = c[i].x;
                    vec[0].y = vec[2].y = c[i].y;
                }
                vdx = c[i].x*4-pred[1].x;
                vdy = c[i].y*4-pred[1].y;
                cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
                tmpsad = sad1+sad3+cost;
                if ((int)tmpsad < rsad[1]){
                    rsad[1] = rsad[3] = tmpsad;
                    vec[1].x = vec[3].x = c[i].x;
                    vec[1].y = vec[3].y = c[i].y;
                }
            }
            else if (mode == MODE_8x8){
                tmpsad = sad0+cost;
                if ((int)tmpsad < rsad[0]){
                    rsad[0] = tmpsad;
                    vec[0].x = c[i].x;
                    vec[0].y = c[i].y;
                }
                vdx = c[i].x*4-pred[1].x;
                vdy = c[i].y*4-pred[1].y;
                cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
                tmpsad = sad1+cost;
                if ((int)tmpsad < rsad[1]){
                    rsad[1] = tmpsad;
                    vec[1].x = c[i].x;
                    vec[1].y = c[i].y;
                }
                vdx = c[i].x*4-pred[2].x;
                vdy = c[i].y*4-pred[2].y;
                cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
                tmpsad = sad2+cost;
                if ((int)tmpsad < rsad[2]){
                    rsad[2] = tmpsad;
                    vec[2].x = c[i].x;
                    vec[2].y = c[i].y;
                }
                vdx = c[i].x*4-pred[3].x;
                vdy = c[i].y*4-pred[3].y;
                cost = rd_cost_lookup(pad_prev,quant,vdx,vdy);
                tmpsad = sad3+cost;
                if ((int)tmpsad < rsad[3]){
                    rsad[3] = tmpsad;
                    vec[3].x = c[i].x;
                    vec[3].y = c[i].y;
                }
            }

        }
    }

    return 0;
}




int motion_estimate_coarse(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)                           
{
    mode_t mode = MODE_16x16;
    int ref = macroblockInfo->ref;    
    int ref_sad[4];
    int sad[2][4][4];              
    int xref,yref;
    mvpos refpos;
    mvpos result[4];
    mvpos orig;
    int last_best;
    int i,psad;
    int vy;
    int vx;
    mvpos vec[4];
    int mode_idx,num_parts,parth,partw,party,partx,part,blk_idx;
    image_frame **pad_prev = encoderInfo->ref_frame_qp;
    int ypos = 16*macroblockInfo->mby;
    int xpos = 16*macroblockInfo->mbx;    
    int mby = macroblockInfo->mby;
    int mbx = macroblockInfo->mbx; 
    int quant = frameInfo->qp;
    int grid = encoderInfo->params->coarse_search_grid;
    
    enc_params *params = encoderInfo->params;
    vy = mby*2;
    vx = mbx*2;

    if (mode == MODE_16x16){

        macroblockInfo->tr16flag = 0;    

    }

    if (params->me_type != ME_TYPE_DIAMOND && params->me_type != ME_TYPE_FULL_SEARCH)
    {
        fatalerror("rcdo_motion_estimate_mb(): Unknown motion estimation type.");
    }

    /* Compensate for the padded edge */
    xref = xpos+16;
    yref = ypos+16;
    orig.x = xref;
    orig.y = yref;
    

    mode_idx = mode - MODE_16x16;
    num_parts = num_partitions[mode_idx];

    psad = 0;
    
    macroblockInfo->candidateMV[ref][0].x = 0;
    macroblockInfo->candidateMV[ref][0].y = 0;
    

    for (i=0;i<4;i++) ref_sad[i] = 0xffffff;
    candidate_search(encoderInfo,frameInfo,macroblockInfo,ref,&ref_sad[0],&vec[0],mode);

    
    part = 0;
    partx = part_dims[mode_idx][part].x;
    party = part_dims[mode_idx][part].y;
    partw = part_dims[mode_idx][part].w;
    parth = part_dims[mode_idx][part].h;
    blk_idx = (party*2+partx)/8;

    sad[ref][mode_idx][blk_idx] = 0xfffffff;

    refpos.x = xref;
    refpos.y = yref;
    last_best = 0;

    if (grid == 1)
    {
        sad[ref][mode_idx][blk_idx] = full_search_fine(
            macroblockInfo->orgY,
            &pad_prev[ref]->Y[0][0],
            4*encoderInfo->extWidth,        
            encoderInfo,
            pad_prev[ref],
            ref_sad[blk_idx],
            params->coarse_range_y<<frameInfo->MEfactor,
            params->coarse_range_x<<frameInfo->MEfactor,
            quant,refpos,orig,
            macroblockInfo->pred[ref][mode_idx][blk_idx],
            &result[blk_idx]
        );
    }
    else
    {    
        sad[ref][mode_idx][blk_idx] = full_search_coarse(
            macroblockInfo->orgY,
            &pad_prev[ref]->Y[0][0],
            4*encoderInfo->extWidth,        
            encoderInfo,
            pad_prev[ref],
            ref_sad[blk_idx],
            params->coarse_range_y<<frameInfo->MEfactor,
            params->coarse_range_x<<frameInfo->MEfactor,
            quant,grid,refpos,orig,
            macroblockInfo->pred[ref][mode_idx][blk_idx],
            &result[blk_idx]
        );
    }

    macroblockInfo->candidateMV[ref][0].x = result[0].x-xref;
    macroblockInfo->candidateMV[ref][0].y = result[0].y-yref;

    return psad;
}


static inline
unsigned int quarter_sad_Y(const unsigned char *org,
                           const unsigned char *ref,
                           const int colsOrg,
                           const int colsRef, 
                           image_frame *qpimage,
                           const unsigned int sad_min,
                           int yvec,
                           int xvec,
                           int *restrict ydelta,
                           int *restrict xdelta,
                           const mvpos refpos,
                           const mvpos pred,
                           const int quant,
                           const mode_t mode)
{    
    int i;
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;     
    unsigned int smin = sad_min;

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_NxN(org,&ref[ypos * colsRef + xpos],colsOrg,colsRef,mode);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_hp = m;
                xdelta_hp = n;
            }
        }
    }
    smin = 0xfffffff;
    xvec = xdelta_hp + xvec;
    yvec = ydelta_hp + yvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_NxN(org,&ref[ypos * colsRef + xpos],colsOrg,colsRef,mode);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_qp = m;
                xdelta_qp = n;
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}
static inline
unsigned int quarter_sad_Y2(const unsigned char *org,
                            const unsigned char *ref,
                            const int colsOrg,
                            const int colsRef, 
                            image_frame *qpimage,
                            const unsigned int sad_min,
                            int *restrict ydelta,
                            int *restrict xdelta,
                            int yvec,
                            int xvec,
                            const mvpos refpos,
                            const mvpos pred,
                            const int quant,
                            const mode_t mode)
{    
    int i;
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;     
    unsigned int smin = sad_min;

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;  
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_NxN(org,&ref[ypos * colsRef + xpos],colsOrg,colsRef,mode);

                if(g_enableSwitched)
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    xdelta_hp = n;
                    ydelta_hp = m;                    
                }
            }
        }
    }
    smin = 0xfffffff;
    xvec = xdelta_hp + xvec;
    yvec = ydelta_hp + yvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;  
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_NxN(org,&ref[ypos * colsRef + xpos],colsOrg,colsRef,mode);

                if(g_enableSwitched)
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    xdelta_qp = n;
                    ydelta_qp = m;                    
                }
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}

unsigned int fract_sad(encoder_info *encoderInfo,unsigned char *mblock2,/*unsigned char **mblock,*/int yw,int xw,
                       unsigned int sad_min,int *ydelta,int *xdelta,int yvec,int xvec,
                       image_frame *qpimage,mvpos refpos,mvpos pred,int quant,int pic_type,int *interp_type,mode_t mode)
{
	int sp_pos;
    unsigned int qmin;
	int this_ydelta,this_xdelta;
    int best_ydelta,best_xdelta;
    int best_interp_indx,best_interp_cost;
    const int colsOrg = xw;  
    const int colsRef = qpimage->cols;

    best_interp_cost = (1<<30);

    // Check Y
    {
        qmin = quarter_sad_Y(mblock2,&qpimage->Y[0][0],colsOrg,colsRef,qpimage,sad_min,yvec,xvec,&this_ydelta,&this_xdelta,refpos,pred,quant,mode); 
        if((int)qmin < best_interp_cost)
        {
          best_interp_cost = qmin;
          best_ydelta = this_ydelta;
          best_xdelta = this_xdelta;
          best_interp_indx = 0;
        }    
    }

    //Check Y2
    if(g_enableSwitched != 0)
    {
        qmin = quarter_sad_Y2(mblock2,&qpimage->Y2[0][0],colsOrg,colsRef,qpimage,sad_min,&this_ydelta,&this_xdelta,yvec,xvec,refpos,pred,quant,mode);
        if((int)qmin < best_interp_cost)
        {
            best_interp_cost = qmin;
            best_ydelta = this_ydelta;
            best_xdelta = this_xdelta;
            best_interp_indx = 1;
        }
    }

    sp_pos = ((best_ydelta+4)%4)*4 + ((best_xdelta%4+4)%4);
    if(transmitInterpIndx[sp_pos]== 0)
        best_interp_indx = 0;

    *ydelta = best_ydelta;
    *xdelta = best_xdelta;
    *interp_type = best_interp_indx;
    return best_interp_cost;
}

int motion_estimate(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int ref0,mode_t mode)

{
    /*
      1. Select a candidate vector for the whole macroblock.
      2. Do an integer-pixel search in a diamond pattern from the candidate vector position:
      a) First one step out
      b) Second step out
      c) A double step
      d) A number of quadrouple steps, until max vector or edge
      e) In a-d stop if no improvement over two iterations
      3. Half-pixel search around the best integer position
      4. Quarter-pixel search around the best half-pixel position
    */

    int ref_sad[4];
    int sad[2][4][4];              /* mode, sads */
    int xref,yref;
    mvpos refpos;
    mvpos result[4];
    /* fractional pel adjustment */
    int y0delta=0,x0delta=0;
    mvpos orig;
    int last_best;

    int (* search_func)(encoder_info *encoderInfo,macroblock_info *macroblockInfo,/*unsigned char **,*/image_frame *,mvpos,int,
                        int,partition,mvpos *,int *,
                        mvpos,mvpos,int,direction,int*);

    int range_or_step,range_or_step_init;
    int i,psad;

    int vx,vy;

    mvpos vec[4];
    int mode_idx,num_parts,parth,partw,party,partx,part,blk_idx;    
    int ref;

    int interp_indx;



    //image_frame *curr = encoderInfo->orig;

    image_frame **pad_prev = encoderInfo->ref_frame_qp;

    image_frame **iprev = encoderInfo->ref_frame_qp;
    int ypos = 16*macroblockInfo->mby;
    int xpos = 16*macroblockInfo->mbx;
    
    
    int mby = macroblockInfo->mby;
    int mbx = macroblockInfo->mbx; 
    int quant = frameInfo->qp;
    
    enc_params *params = encoderInfo->params;
    int pic_type = frameInfo->pic_type;

    unsigned char mblock2[256];
        
    vy = mby*2;
    vx = mbx*2;



    result[0].x = result[0].y = 0;
    result[1].x = result[1].y = 0;
    result[2].x = result[2].y = 0;
    result[3].x = result[3].y = 0;

    refpos.x = refpos.y = 0;
    orig.x = orig.y = 0;

    vec[0].x = vec[0].y = 0;
    vec[1].x = vec[1].y = 0;
    vec[2].x = vec[2].y = 0;
    vec[3].x = vec[3].y = 0;

    ref = ref0;

    if (mode == MODE_16x16){
   
        macroblockInfo->tr16flag = 0;

    }                     
    if (params->me_type == ME_TYPE_DIAMOND)
    {

        range_or_step_init = 0;
    }
    else if (params->me_type == ME_TYPE_FULL_SEARCH)
    {
        search_func = full_search;
        range_or_step_init = params->me_range;
        if (pic_type == B_PICTURE)
        {
            range_or_step_init = params->me_rangeb;
        }
    }
    else
    {
        fatalerror("rcdo_motion_estimate_mb(): Unknown motion estimation type.");
    }

    /* Compensate for the padded edge */
    xref = xpos+16;
    yref = ypos+16;
    orig.x = xref;
    orig.y = yref;
  

    mode_idx = mode - MODE_16x16;
    num_parts = num_partitions[mode_idx];


    range_or_step = range_or_step_init;
    if (mode > MODE_16x16)
        range_or_step = range_or_step / 2;
    psad = 0;
    
    for (i=0;i<4;i++) ref_sad[i] = 0xffffff;

    candidate_search(encoderInfo,frameInfo,macroblockInfo,ref,&ref_sad[0],&vec[0],mode);

    for (part = 0; part < num_parts; part++)
    {
        int i,j;

        partx = part_dims[mode_idx][part].x;
        party = part_dims[mode_idx][part].y;
        partw = part_dims[mode_idx][part].w;
        parth = part_dims[mode_idx][part].h;

        blk_idx = (party*2+partx)/8;

        sad[ref][mode_idx][blk_idx] = 0xfffffff;

        refpos.x = xref+vec[blk_idx].x;
        refpos.y = yref+vec[blk_idx].y;
        last_best = 0;


        if (params->me_type == ME_TYPE_DIAMOND){
            mvpos vec0 = vec[blk_idx];
            sad[ref][mode_idx][blk_idx] = ref_sad[blk_idx];
            diamond_NxN(encoderInfo,macroblockInfo->orgY,pad_prev[ref],part_id[mode_idx][part],&vec0,
                        &sad[ref][mode_idx][blk_idx],orig,macroblockInfo->pred[ref][mode_idx][blk_idx],quant,mode);
            result[blk_idx].x = vec0.x + orig.x;
            result[blk_idx].y = vec0.y + orig.y;
        }
        else

        search_func(encoderInfo,macroblockInfo,pad_prev[ref],refpos,ref_sad[blk_idx],range_or_step,part_id[mode_idx][part], 
                    &result[blk_idx],&sad[ref][mode_idx][blk_idx],orig,
                    macroblockInfo->pred[ref][mode_idx][blk_idx],quant,DIR_NONE,&last_best);


        refpos.x = xref*4+partx*4;
        refpos.y = yref*4+party*4;


        for (i=0;i<parth;i++)
            memcpy(&mblock2[i*partw],&macroblockInfo->orgY[16*(party+i)+partx],partw*sizeof(unsigned char));
        psad += fract_sad(encoderInfo,mblock2,parth,partw,sad[ref][mode_idx][blk_idx],&y0delta,&x0delta,
                          party*4+result[blk_idx].y*4,partx*4+result[blk_idx].x*4,iprev[ref],refpos,
                          macroblockInfo->pred[ref][mode_idx][blk_idx],quant,pic_type,&interp_indx,mode);

        i = mode==3 ? part : part/2;
        j = mode==3 ? part/2 : part&1;
        macroblockInfo->mvblock[i][j].x = (result[blk_idx].x-xref)*4+x0delta;
        macroblockInfo->mvblock[i][j].y = (result[blk_idx].y-yref)*4+y0delta;

	    macroblockInfo->mvblock[i][j].interp_indx = interp_indx;


    }
    
    if (mode==MODE_16x16){
        macroblockInfo->mvblock[0][1] = macroblockInfo->mvblock[0][0];
        macroblockInfo->mvblock[1][0] = macroblockInfo->mvblock[0][0];
        macroblockInfo->mvblock[1][1] = macroblockInfo->mvblock[0][0];
    }
    else if(mode==MODE_16x8){
        macroblockInfo->mvblock[0][1] = macroblockInfo->mvblock[0][0];       
        macroblockInfo->mvblock[1][1] = macroblockInfo->mvblock[1][0];
    }
    else if(mode==MODE_8x16){
        macroblockInfo->mvblock[1][0] = macroblockInfo->mvblock[0][0];       
        macroblockInfo->mvblock[1][1] = macroblockInfo->mvblock[0][1];
    }   

    if (mode == MODE_16x16) {                
        macroblockInfo->candidateMV[ref][0].x = result[0].x-xref;
        macroblockInfo->candidateMV[ref][0].y = result[0].y-yref;     
    }
    
    return psad;
}


unsigned int candidate_searchLMB(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int ref,int *rsad,mvpos *vec,mode_t mode)
{

    int xref,yref,i,vdx,vdy;

    //STATIC unsigned char *pblock[LMBYSIZE];
    unsigned int tmpsad;


    image_frame *pad_prev = encoderInfo->ref_frame_qp[ref];

    //STATIC unsigned char **mblock2;
    int xpos = 16*macroblockInfo->mbx;
    int ypos = 16*macroblockInfo->mby;
    mvpos pred[4];
    int quant = frameInfo->qp;

    int offy,offx;    
    mvpos c[4];

    int LMBwidth,LMBheight;
    if (mode==MODE_LMB32){
        LMBwidth = LMBXSIZE/2;
        LMBheight = LMBYSIZE/2;
    }
    else{
        LMBwidth = LMBXSIZE;
        LMBheight = LMBYSIZE;
    }

    offy = LMBheight+1;
    offx = LMBwidth+1;
    
    c[0] = macroblockInfo->candidateMV[ref][0];
    c[1] = macroblockInfo->candidateMV[ref][1];
    c[2] = macroblockInfo->candidateMV[ref][2];
    c[3] = macroblockInfo->candidateMV[ref][3];


    pred[0] = macroblockInfo->pred[ref][0][0];


    //STATIC mblock2 = ucsubmatrix(curr->Y,ypos,ypos+LMBheight-1,xpos,xpos+LMBwidth-1,0,0);

    xref = xpos+16;
    yref = ypos+16;
    for (i = 0; i < 4; i++){

        if (xref+c[i].x > 1 && xref+c[i].x < encoderInfo->extWidth-offx && yref+c[i].y > 1 && yref+c[i].y < encoderInfo->extHeight-offy){

            //STATIC move_ucsubmatrix(pblock,pad_prev->Y,yref+c[i].y,yref+c[i].y+LMBheight-1,xref+c[i].x,xref+c[i].x+LMBwidth-1,0,0);  

            vdx = c[i].x*4-pred[0].x;
            vdy = c[i].y*4-pred[0].y;
            tmpsad = rd_cost_lookup(pad_prev,quant,vdx,vdy);

            tmpsad += SAD_NxN_LMB(macroblockInfo->orgLMBY,&pad_prev->Y[yref+c[i].y][xref+c[i].x],LMBwidth,4*encoderInfo->extWidth, mode);

            if ((int)tmpsad < rsad[0]){
                rsad[0] = tmpsad;
                vec[0].x = c[i].x;
                vec[0].y = c[i].y;
            }            
        }
    }
    
    //STATIC free_ucsubmatrix(mblock2,0,LMBheight-1,0,LMBwidth-1);
    return 0;
}

static inline
unsigned int quarter_sadLMB32_Y(const unsigned char *org,
                                const unsigned char *ref,
                                image_frame *qpimage,
                                const unsigned int sad_min,
                                int yvec,
                                int xvec,
                                int *restrict ydelta,
                                int *restrict xdelta,
                                const mvpos refpos,
                                const mvpos pred,
                                const int quant)
{    
    int i,j;
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;    
    const int colsOrg = cols-32;
    const int colsRef = qpimage->cols;     
    unsigned int smin = sad_min;

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    unsigned char org_[32 * 32];
    for(i=0;i<32;i++)
        for(j=0;j<32;j++)
            org_[32*i+j] = org[colsOrg*i+j];

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_32x32(org_,&ref[ypos * colsRef + xpos],32,colsRef);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_hp = m;
                xdelta_hp = n;
            }
        }
    }
    smin = 0xfffffff;
    xvec = xdelta_hp + xvec;
    yvec = ydelta_hp + yvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_32x32(org_,&ref[ypos * colsRef + xpos],32,colsRef);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_qp = m;
                xdelta_qp = n;
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}

static inline
unsigned int quarter_sadLMB64_Y(const unsigned char *org,
                                const unsigned char *ref,
                                image_frame *qpimage,
                                const unsigned int sad_min,
                                int yvec,
                                int xvec,
                                int *restrict ydelta,
                                int *restrict xdelta,
                                const mvpos refpos,
                                const mvpos pred,
                                const int quant)
{    
    int i,j;
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;    
    const int colsOrg = cols-32;
    const int colsRef = qpimage->cols;     
    unsigned int smin = sad_min;

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    unsigned char org_[64 * 64];
    for(i=0;i<64;i++)
        for(j=0;j<64;j++)
            org_[64*i+j] = org[colsOrg*i+j];

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_64x64(org_,&ref[ypos * colsRef + xpos],64,colsRef);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_hp = m;
                xdelta_hp = n;
            }
        }
    }
    smin = 0xfffffff;
    xvec = xdelta_hp + xvec;
    yvec = ydelta_hp + yvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};

        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];  
            const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
            const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
            unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
            unsigned int stmp = SAD_64x64(org_,&ref[ypos * colsRef + xpos],64,colsRef);

            if(g_enableSwitched)
            {
                int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
                if(transmitInterpIndx[sp_pos] == 1)        
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);
            }
            stmp = stmp+cost;
            if (stmp < smin)
            {
                smin = stmp;
                ydelta_qp = m;
                xdelta_qp = n;
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}

static inline
unsigned int quarter_sadLMB32_Y2(const unsigned char *org,
                                 const unsigned char *ref,
                                 image_frame *qpimage,
                                 const unsigned int sad_min,
                                 int *restrict ydelta,
                                 int *restrict xdelta,
                                 int yvec,
                                 int xvec,
                                 const mvpos refpos,
                                 const mvpos pred,
                                 const int quant)
{    
    int i,j;        
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;  
    const int colsOrg = cols-32;
    const int colsRef = qpimage->cols;    
    unsigned int smin = sad_min;   

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    unsigned char org_[32 * 32];
    for(i=0;i<32;i++)
        for(j=0;j<32;j++)
            org_[32*i+j] = org[colsOrg*i+j];

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];
            const int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_32x32(org_,&ref[ypos * colsRef + xpos],32,colsRef);

                if(g_enableSwitched)
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    ydelta_hp = m;
                    xdelta_hp = n;
                }
            }
        }
    }
    smin = 0xfffffff;
    yvec = ydelta_hp + yvec;
    xvec = xdelta_hp + xvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];
            const int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_32x32(org_,&ref[ypos * colsRef + xpos],32,colsRef);

                if(g_enableSwitched)
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    ydelta_qp = m;
                    xdelta_qp = n;
                }
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}

static inline
unsigned int quarter_sadLMB64_Y2(const unsigned char *org,
                                 const unsigned char *ref,
                                 image_frame *qpimage,
                                 const unsigned int sad_min,
                                 int *restrict ydelta,
                                 int *restrict xdelta,
                                 int yvec,
                                 int xvec,
                                 const mvpos refpos,
                                 const mvpos pred,
                                 const int quant)
{    
    int i,j;
    const int rows = qpimage->rows/4;
    const int cols = qpimage->cols/4;  
    const int colsOrg = cols-32;
    const int colsRef = qpimage->cols;    
    unsigned int smin = sad_min;   

    int ydelta_hp = 0;
    int xdelta_hp = 0; 
    int ydelta_qp = 0;
    int xdelta_qp = 0; 

    unsigned char org_[64 * 64];
    for(i=0;i<64;i++)
        for(j=0;j<64;j++)
            org_[64*i+j] = org[colsOrg*i+j];

    /* Half-pel search */
    {
        const int npos[] = {0,-2,0,0,2,-2,2,-2,2};    
        const int mpos[] = {0,0,-2,2,0,-2,-2,2,2};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];
            const int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_64x64(org_,&ref[ypos * colsRef + xpos],64,colsRef);

                if(g_enableSwitched)
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    ydelta_hp = m;
                    xdelta_hp = n;
                }
            }
        }
    }
    smin = 0xfffffff;
    yvec = ydelta_hp + yvec;
    xvec = xdelta_hp + xvec;

    /* Quarter-pel search */
    {
        const int npos[] = {0,-1,0,0,1,-1,1,-1,1};      
        const int mpos[] = {0,0,-1,1,0,-1,-1,1,1};
        for (i=0;i<9;i++)
        {
            const int m = mpos[i];
            const int n = npos[i];
            const int sp_pos = ((yvec+m+4)%4)*4 + ((xvec+n%4+4)%4);
            if(transmitInterpIndx[sp_pos] != 0)
            {         
                const int ypos = rows*((yvec+m)%4) + (yvec+m)/4;
                const int xpos = cols*((xvec+n)%4) + (xvec+n)/4;                                    
                unsigned int cost = rd_cost_lookup(qpimage,quant,xvec+n-refpos.x-pred.x,yvec+m-refpos.y-pred.y);
                unsigned int stmp = SAD_64x64(org_,&ref[ypos * colsRef + xpos],64,colsRef);

                if(g_enableSwitched) 
                    cost += (int)(qpimage->rd_lambda[quant] + 0.5f);

                stmp = stmp+cost;
                if (stmp < smin)
                {
                    smin = stmp;
                    ydelta_qp = m;
                    xdelta_qp = n;
                }
            }
        }
    }
    *ydelta = ydelta_hp + ydelta_qp;
    *xdelta = xdelta_hp + xdelta_qp;
    return smin;
}


unsigned int fract_sadLMB(encoder_info *encoderInfo,const unsigned char *mblock,int yw,int xw,
                           unsigned int sad_min,int *ydelta,int *xdelta,int yvec,int xvec,
                           image_frame *qpimage,mvpos refpos,mvpos pred,int quant,int pic_type,int *interp_type,mode_t mode)
{
	int sp_pos;
    unsigned int qmin;
	int this_ydelta,this_xdelta;
    int best_ydelta,best_xdelta;
    int best_interp_indx,best_interp_cost;
	
    best_interp_cost = (1<<30);

    // Check Y
    {       
        switch (mode)
        {
        case MODE_LMB32:  
            qmin = quarter_sadLMB32_Y(mblock,&qpimage->Y[0][0],qpimage,sad_min,yvec,xvec,&this_ydelta,&this_xdelta,refpos,pred,quant); 
            break;
        case MODE_LMB64:  
            qmin = quarter_sadLMB64_Y(mblock,&qpimage->Y[0][0],qpimage,sad_min,yvec,xvec,&this_ydelta,&this_xdelta,refpos,pred,quant); 
            break;
        }
        if((int)qmin < best_interp_cost)
        {
          best_interp_cost = qmin;
          best_ydelta = this_ydelta;
          best_xdelta = this_xdelta;
          best_interp_indx = 0;
        }    
    }

    // Check Y2
    if(g_enableSwitched != 0)
    {    
        switch (mode)
        {
        case MODE_LMB32:  
            qmin = quarter_sadLMB32_Y2(mblock,&qpimage->Y2[0][0],qpimage,sad_min,&this_ydelta,&this_xdelta,yvec,xvec,refpos,pred,quant); 
            break;
        case MODE_LMB64:  
            qmin = quarter_sadLMB64_Y2(mblock,&qpimage->Y2[0][0],qpimage,sad_min,&this_ydelta,&this_xdelta,yvec,xvec,refpos,pred,quant); 
            break;
        }               
        if((int)qmin < best_interp_cost)
        {
          best_interp_cost = qmin;
          best_ydelta = this_ydelta;
          best_xdelta = this_xdelta;
          best_interp_indx = 1;
        }    
    }

    sp_pos = ((best_ydelta+4)%4)*4 + ((best_xdelta%4+4)%4);
    if(transmitInterpIndx[sp_pos]== 0)
        best_interp_indx = 0;

    *ydelta = best_ydelta;
    *xdelta = best_xdelta;
    *interp_type = best_interp_indx;
    return best_interp_cost;
}

int motion_estimateLMB(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int ref0,mode_t mode)

{    
    //STATIC unsigned char **mblock_curr;
    int ref_sad[4];
    int sad[2][4][4];              /* mode, sads */
    int xref,yref;
    mvpos refpos;
    mvpos result[4];
    /* fractional pel adjustment */
    int y0delta=0,x0delta=0;
    mvpos orig;
    int last_best;   
    int i,psad;

    int vx,vy;

    mvpos vec[4];
    int mode_idx,num_parts,parth,partw,party,partx,part,blk_idx;    
    int ref;

    int interp_indx;



    image_frame *curr = encoderInfo->orig;

    image_frame **pad_prev = encoderInfo->ref_frame_qp;

    image_frame **iprev = encoderInfo->ref_frame_qp;
    int ypos = 16*macroblockInfo->mby;
    int xpos = 16*macroblockInfo->mbx;
    
    
    int mby = macroblockInfo->mby;
    int mbx = macroblockInfo->mbx; 
    int quant = frameInfo->qp;
    
    int pic_type = frameInfo->pic_type;

    int LMBwidth,LMBheight;
    if (mode==MODE_LMB32){
        LMBwidth = LMBXSIZE/2;
        LMBheight = LMBYSIZE/2;
    }
    else{
        LMBwidth = LMBXSIZE;
        LMBheight = LMBYSIZE;
    }
        
    vy = mby*2;
    vx = mbx*2;



    result[0].x = result[0].y = 0;
    result[1].x = result[1].y = 0;
    result[2].x = result[2].y = 0;
    result[3].x = result[3].y = 0;

    refpos.x = refpos.y = 0;
    orig.x = orig.y = 0;

    vec[0].x = vec[0].y = 0;
    vec[1].x = vec[1].y = 0;
    vec[2].x = vec[2].y = 0;
    vec[3].x = vec[3].y = 0;

    ref = ref0;

    /* Compensate for the padded edge */
    xref = xpos+16;
    yref = ypos+16;
    orig.x = xref;
    orig.y = yref;


    //STATIC mblock_curr = ucsubmatrix(curr->Y,ypos,ypos+LMBheight-1,xpos,xpos+LMBwidth-1,0,0);

    mode_idx = 0;
    num_parts = 1;
    
    psad = 0;
    
    for (i=0;i<4;i++) ref_sad[i] = 0xffffff;
    
    candidate_searchLMB(encoderInfo,frameInfo,macroblockInfo,ref,&ref_sad[0],&vec[0],mode);

    part = 0;
    
        
    partx = 0;
    party = 0;

    partw = LMBwidth;
    parth = LMBheight;


    blk_idx = 0;

    sad[ref][mode_idx][blk_idx] = 0xfffffff;

    refpos.x = xref+vec[blk_idx].x;
    refpos.y = yref+vec[blk_idx].y;
    last_best = 0;
    



    {
        mvpos vec0 = vec[blk_idx];
        sad[ref][mode_idx][blk_idx] = ref_sad[blk_idx];
        diamond_NxN(encoderInfo,macroblockInfo->orgLMBY,pad_prev[ref],part_id[mode_idx][part],&vec0,
                    &sad[ref][mode_idx][blk_idx],orig,macroblockInfo->pred[ref][mode_idx][blk_idx],quant,mode);
        result[blk_idx].x = vec0.x + orig.x;
        result[blk_idx].y = vec0.y + orig.y;
    }


    //STATIC move_ucsubmatrix(mblock_curr,curr->Y,ypos+party,ypos+party+parth-1,xpos+partx,xpos+partx+partw-1,0,0);
    refpos.x = xref*4+partx*4;
    refpos.y = yref*4+party*4;


    psad += fract_sadLMB(encoderInfo,&curr->Y[ypos+party][xpos+partx],parth,partw,sad[ref][mode_idx][blk_idx],&y0delta,&x0delta,
        party*4+result[blk_idx].y*4,partx*4+result[blk_idx].x*4,iprev[ref],refpos,
        macroblockInfo->pred[ref][mode_idx][blk_idx],quant,pic_type,&interp_indx,mode);
   
    macroblockInfo->mvblock[0][0].x = (result[blk_idx].x-xref)*4+x0delta;
    macroblockInfo->mvblock[0][0].y = (result[blk_idx].y-yref)*4+y0delta;

    macroblockInfo->mvblock[0][0].interp_indx = interp_indx;

    macroblockInfo->mvblock[0][1] = macroblockInfo->mvblock[0][0];
    macroblockInfo->mvblock[1][0] = macroblockInfo->mvblock[0][0];
    macroblockInfo->mvblock[1][1] = macroblockInfo->mvblock[0][0];

    //STATIC free_ucsubmatrix(mblock_curr,0,LMBheight-1,0,LMBwidth-1);
    return psad;
}

mvpos get_mv_candidate(encoder_info *encoderInfo,frame_info *frameInfo,int vy2,int vx2,int *ref)
{
	mvpos cand;
	int mbxmax = encoderInfo->widthInMBs;
	int mbx = vx2>>1;
	int mby = vy2>>1;

	if (vy2 < 0 || vx2 < 0 || vx2 >= encoderInfo->width/8){
		cand.x = cand.y = 0;
		*ref = -1;

		cand.interp_indx = 0;

	}
	else{
		cand.x = frameInfo->mvmap[vy2*2*mbxmax+vx2].x;
		cand.y = frameInfo->mvmap[vy2*2*mbxmax+vx2].y;

		cand.interp_indx = frameInfo->mvmap[vy2*2*mbxmax+vx2].interp_indx;

		*ref = frameInfo->mbmap[mby*mbxmax+mbx].ref;

		if(frameInfo->mbmap[mby*mbxmax+mbx].ref == 2)
			*ref = 0;

	}

	return cand;
}

int early_skip_check_8x8(short threshold,unsigned char *org,unsigned char *rec,int stride)
{
#if FAST_MODE
    int i,j,k,ret=0;
    short diff[64],ysum[32],xsum[16];   
    union {
        __int64 l[4];
        int i[8];
        short s[16];
    } sum, dif, flip, flop;

    // Residual
    for (i=0; i<8; i++)
        for (j=0; j<8; j++)
            diff[8*i+j] = org[stride*i+j] - rec[stride*i+j];

    // Sum two and two pixels in same coulmn
    for (i=0; i<4; i++)
        for (j=0; j<8; j++)
            ysum[8*i+j] = diff[16*i+j] + diff[16*i+j+8];

    // Sum two and two pixels in same row
    for (j=0; j<16; j++)
        xsum[j] = ysum[2*j] + ysum[2*j+1];

    // 4x4 hadamard transform
    for(k=0;k<8;k++)
    {
        sum.s[k] = xsum[k] + xsum[k+8];
        dif.s[k] = xsum[k] - xsum[k+8];
    }
    for(k=0;k<2;k++) 
    {
        flip.l[2*k+0] = sum.l[k];
        flip.l[2*k+1] = dif.l[k];
    }
    for(k=0;k<8;k++)
    {
        sum.s[k] = flip.s[k] + flip.s[k+8];
        dif.s[k] = flip.s[k] - flip.s[k+8]; 
    } 
    for(k=0;k<2;k++) 
    {
        flop.l[2*k+0] = sum.l[k];
        flop.l[2*k+1] = dif.l[k];
    }
    for(k=0;k<4;k++) 
    {
        flip.i[k+0] = flop.i[2*k+0];
        flip.i[k+4] = flop.i[2*k+1];
    }
    for(k=0;k<8;k++)
    {
        sum.s[k] = flip.s[k] + flip.s[k+8];
        dif.s[k] = flip.s[k] - flip.s[k+8]; 
    } 
    for(k=0;k<4;k++) 
    {
        flop.i[2*k+0] = sum.i[k];
        flop.i[2*k+1] = dif.i[k];
    }
    for(k=0;k<8;k++)
    {
        xsum[2*k+0] = flop.s[2*k] + flop.s[2*k+1];
        xsum[2*k+1] = flop.s[2*k] - flop.s[2*k+1];                
    }    
    for (k=0; k<16; k++)
        xsum[k] = abs(xsum[k]);
    for (k=0; k<16; k++)
        if (xsum[k] > threshold)
            ret |= 1;
    return(ret);
#else
	int i,j,k,d;
	int diff8[8][8];   
	int tmp[4][4];
	int coeff;

	static const int had[4][4] = {
		{ 1, 1, 1, 1},
		{ 1, 1,-1,-1},
		{ 1,-1,-1, 1},
		{ 1,-1, 1,-1}};

		for (i=0; i<8; i++){
			for (j=0; j<8; j++){
				diff8[i][j] = org[stride*j+i] - rec[stride*j+i];
			}
		}

		for (i=0;i<4;i++){
			for (j=0;j<4;j++){
				tmp[i][j] = 0;
				for (k=0;k<4;k++){
					d = diff8[k*2][j*2]+diff8[k*2][j*2+1]+diff8[k*2+1][j*2]+diff8[k*2+1][j*2+1];
					tmp[i][j] += had[i][k]*d;
				}
			}
		}
		for (i=0;i<4;i++){
			for (j=0;j<4;j++){
				coeff = 0;
				for (k=0;k<4;k++){
					coeff += tmp[i][k]*had[k][j];
				}
				if (abs(coeff) > (int)threshold){
					return 1;
				}            
			}
		}
		return 0;
#endif
}
