/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include "global.h"
#include "putbits.h"
//STATIC #include "rfutil.h"

/* No. of bytes to expand bitstream with when necessary*/
/* #define BCHUNK 2048       */
#define BCHUNK 2000000


static unsigned int mask[33] = {
    0x00000000,0x00000001,0x00000003,0x00000007,
    0x0000000f,0x0000001f,0x0000003f,0x0000007f,
    0x000000ff,0x000001ff,0x000003ff,0x000007ff,
    0x00000fff,0x00001fff,0x00003fff,0x00007fff,
    0x0000ffff,0x0001ffff,0x0003ffff,0x0007ffff,
    0x000fffff,0x001fffff,0x003fffff,0x007fffff,
    0x00ffffff,0x01ffffff,0x03ffffff,0x07ffffff,
    0x0fffffff,0x1fffffff,0x3fffffff,0x7fffffff,
    0xffffffff};

/* initialize buffer, call once before first putbits or alignbits */
stream *initbits(FILE *outfile)
{
    stream *str;

    if ((str = (stream *)malloc((size_t)sizeof(stream))) == NULL)
    {
        fatalerror("Problem allocating bitstream buffer.");
    }

    if ((str->bitstream = (unsigned char *)malloc((size_t)(BCHUNK*sizeof(unsigned char)))) == NULL)
    {
        fatalerror("Problem allocating bitstream buffer.");
    }
    str->bytesize = BCHUNK;
    str->bytepos = 0;
    str->bytecount = 0;
    str->bitbuf = 0;
    str->bitrest = 32;
    str->outfile = outfile;
    return str;
}

void flush_all_bits(stream *str)
{
    int i;
    int bytes = 4 - str->bitrest/8;

    if ((str->bytepos+bytes) > str->bytesize)
    {
        flush_bytebuf(str);
    }
    for (i = 0; i < bytes; i++)
    {
        str->bitstream[str->bytepos++] = (str->bitbuf >> (24-i*8)) & 0xff;
    }
    /* Dropping the following 3 lines, they would often wrongly
       introduce a few additional bits to the count. Keeping the
       (32-bitrest) instead.  This function should only be called at
       the very end of a program. */
/*     bytecount += bytes; */
/*     bitbuf = 0; */
/*     bitrest = 32; */
    if (str->outfile)
    {
        if (fwrite(str->bitstream,sizeof(unsigned char),str->bytepos,str->outfile) != str->bytepos)
        {
            fatalerror("Problem writing bitstream to file.");
        }
    }
    str->bytepos = 0;
}    
                    

void flush_bytebuf(stream *str)
{
    if (str->outfile)
    {
        if (fwrite(str->bitstream,sizeof(unsigned char),str->bytepos,str->outfile) != str->bytepos)
        {
            fatalerror("Problem writing bitstream to file.");
        }
    }
    str->bytepos = 0;
}

void flush_bitbuf(stream *str)
{
    if ((str->bytepos+4) > str->bytesize)
    {
        //flush_bytebuf(str);
        fatalerror("Run out of bits in stream buffer.");
    }
    str->bitstream[str->bytepos++] = (str->bitbuf >> 24) & 0xff;
    str->bitstream[str->bytepos++] = (str->bitbuf >> 16) & 0xff;
    str->bitstream[str->bytepos++] = (str->bitbuf >> 8) & 0xff;
    str->bitstream[str->bytepos++] = str->bitbuf & 0xff;
    str->bytecount += 4;
    str->bitbuf = 0;
    str->bitrest = 32;
}

void putbits(unsigned int n, unsigned int val, stream *str)
{
    unsigned int rest;

    if (n <= str->bitrest)
    {
        str->bitbuf |= ((val & mask[n]) << (str->bitrest-n));
        str->bitrest -= n;
    }
    else
    {
        rest = n-str->bitrest;
        str->bitbuf |= (val >> rest) & mask[n-rest];
        flush_bitbuf(str);
        str->bitbuf |= (val & mask[rest]) << (32-rest);
        str->bitrest -= rest;
    }
}

int bitcount(stream *str)
{
  return 8*str->bytecount + (32-str->bitrest);
}

void closebits(stream *str)
{
    free(str->bitstream);
    free(str);
}


/* copy streams */
void copyCodedStream(stream * destStream, stream *srcStream)
{
	memcpy(destStream->bitstream,srcStream->bitstream, BCHUNK*sizeof(unsigned char));
    destStream->bytesize = srcStream->bytesize;
    destStream->bytepos = srcStream->bytepos;
    destStream->bytecount = srcStream->bytecount;
    destStream->bitbuf = srcStream->bitbuf;
    destStream->bitrest = srcStream->bitrest;

}

