/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#include <stdio.h>
#include <string.h>
#include "global.h"
#include "rdtsc.h"

#if DBG_COMPLEXITY

#ifdef TSC_ROUTINE_TIMING


struct tsc_s tsc;


void tscLoad()
{
  memset(&tsc, 0, sizeof(struct tsc_s));
}

void tscStart(tscTimer_s *timer)
{
  __asm {
    mov     ecx,[timer]
    rdtsc
    mov     [ecx]timer.s.startLo,eax
    mov     [ecx]timer.s.startHi,edx
  }
}

void tscCumul(tscTimer_s *timer)
{
  __asm {
    mov     ecx,[timer]
    rdtsc                             ;read <current TSC>
    sub     eax,[ecx]timer.s.startLo  ;elapsed = <current TSC> - tsc->start (lo 32 bits)
    sbb     edx,[ecx]timer.s.startHi  ;elapsed = <current TSC> - tsc->start (hi 32 bits)
    mov     [ecx]timer.s.onerunLo,eax  ;
    mov     [ecx]timer.s.onerunHi,edx  ;
    add     eax,[ecx]timer.s.cumulLo  ;tsc->cumul += elapsed (lo 32 bits)
    adc     edx,[ecx]timer.s.cumulHi  ;tsc->cumul += elapsed (hi 32 bits)
    mov     [ecx]timer.s.cumulLo,eax  ;
    mov     [ecx]timer.s.cumulHi,edx  ;
  }
}

double tscCalcP(tscTimer_s *funcTime, tscTimer_s *totalTime)
{
  return 100.0*( ((double)(funcTime->l.cumul)) / ((double)(totalTime->l.cumul)) );
  
}
double tscCalcPerFrame()
{
  return ( (double)(tsc.TSC_PER_FRAME_STAT.l.cumul-tsc.TSC_PER_FRAME_STAT.l.start));
}

void tscPrint()
{
  fprintf(stdout, "Complexity of each function (percentage of total time)\n");
  fprintf(stdout, "Loop_filter             %.2f%%\n", tscCalcP(&tsc.TSC_FILTER, &tsc.TSC_TOTAL));
  fprintf(stdout, "Interpolation           %.2f%%\n", tscCalcP(&tsc.TSC_INTERP, &tsc.TSC_TOTAL));
  fprintf(stdout, "RDO                     %.2f%%\n", tscCalcP(&tsc.TSC_RDO, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tSkip RDO              %.2f%%\n", tscCalcP(&tsc.TSC_SKIP_RDO, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tIntra RDO             %.2f%%\n", tscCalcP(&tsc.TSC_INTRA_RDO, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tInter RDO             %.2f%%\n", tscCalcP(&tsc.TSC_INTER_RDO, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tME                    %.2f%%\n", tscCalcP(&tsc.TSC_ME, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tME_COARSE             %.2f%%\n", tscCalcP(&tsc.TSC_ME_COARSE, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tLMB RDO               %.2f%%\n", tscCalcP(&tsc.TSC_LMB_RDO, &tsc.TSC_TOTAL));
  fprintf(stdout, "\tSVT                   %.2f%%\n", tscCalcP(&tsc.TSC_SVT_RDO, &tsc.TSC_TOTAL));

}


#endif  // TSC_ROUTINE_TIMING

#endif //DBG_COMPLEXITY
