/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#ifndef TSC_H
#define TSC_H
//STATIC #include "global.h"

#if DBG_COMPLEXITY

#if !defined _DEBUG  && defined _MSC_VER && _MSC_VER >= 1200 && defined _M_IX86 && _M_IX86 >= 500
#define TSC_ROUTINE_TIMING 1
#endif


#ifdef TSC_ROUTINE_TIMING


#define CLOCK_FREQ 2260000000.0


#define DO_TSC_TOTAL          1
#define DO_TSC_INTERP         1
#define DO_TSC_ME             1
#define DO_TSC_FILTER         1
#define DO_TSC_RDO            1
#define DO_TSC_SKIP_RDO       1
#define DO_TSC_INTRA_RDO      1
#define DO_TSC_INTER_RDO      1
#define DO_TSC_ME_COARSE      1
#define DO_TSC_LMB_RDO        1
#define DO_TSC_SVT_RDO        1
#define DO_TSC_FREE_AGENT     1
#define DO_TSC_PER_FRAME_STAT 1


typedef union _tscTimer_s
{
  struct {
  __int64 start;
  __int64 cumul;
  __int64 onerun;
  } l;
  struct {
  __int32 startLo, startHi;
  __int32 cumulLo, cumulHi;
  __int32 onerunLo,onerunHi;
  } s;
} tscTimer_s;


struct tsc_s {
  tscTimer_s
    TSC_TOTAL,
    TSC_INTERP,
    TSC_ME,
    TSC_FILTER,
    TSC_RDO,
    TSC_SKIP_RDO,      
    TSC_INTRA_RDO,      
    TSC_INTER_RDO,
    TSC_LMB_RDO,
    TSC_SVT_RDO,
    TSC_ME_COARSE,
    TSC_FREE_AGENT,
    TSC_PER_FRAME_STAT;
    
};


extern struct tsc_s tsc;


void tscLoad();
void tscStart(tscTimer_s *timer);
void tscCumul(tscTimer_s *timer);
double tscCalcP(tscTimer_s *funcTime, tscTimer_s *totalTime);
void tscPrint();
double tscCalcPerFrame();
void tscPrintFrameStat();

#define TSC_LOAD() tscLoad()
#define TSC_START(timer_id)  if (DO_ ## timer_id) tscStart(&tsc. ## timer_id);
#define TSC_CUMUL(timer_id)  if (DO_ ## timer_id) tscCumul(&tsc. ## timer_id);
#define TSC_START_TAB(timer_id, idx)  if (DO_ ## timer_id) tscStart(&tsc. ## timer_id ## [(idx)]);
#define TSC_CUMUL_TAB(timer_id, idx)  if (DO_ ## timer_id) tscCumul(&tsc. ## timer_id ## [(idx)]);
#define TSC_PRINT() tscPrint()

#else   // TSC_ROUTINE_TIMING


#define TSC_LOAD()
#define TSC_START(timer_id)
#define TSC_CUMUL(timer_id)
#define TSC_START_TAB(timer_id, idx)
#define TSC_CUMUL_TAB(timer_id, idx)
#define TSC_PRINT()
#define TSC_PRINT_FRAME_STAT()

#endif  // TSC_ROUTINE_TIMING

#endif //DBG_COMPLEXITY

#endif
