/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/

#if !defined(_TRANSFORM_H_)
#define _TRANSFORM_H_

#include "tntmenc.h"

int transform_8x8_satd(short block[8][8]);
int transform_svt(short block[4][16],int coeff[4][16]);
int inverse_transform_svt(int coeff[4][16],int block[4][16],int quant,int chroma);

static inline void transform_4x4  (short block[4][4],  short coeff[8][8]);
static inline void transform_8x8  (short block[8][8],  short coeff[8][8]);
static inline void transform_16x16(short block[16][16],short coeff[8][8]);
static inline void transform_32x32(short block[32][32],short coeff[8][8]);
static inline void transform_64x64(short block[64][64],short coeff[8][8]);

static inline void inverse_transform_4x4  (short coeff[4][4],short block[4][4]  ,int quant);
static inline void inverse_transform_8x8  (short coeff[8][8],short block[8][8]  ,int quant,int chroma);
static inline void inverse_transform_16x16(short coeff[8][8],short block[16][16],int quant,int chroma);
static inline void inverse_transform_32x32(short coeff[8][8],short block[32][32],int quant,int chroma);
static inline void inverse_transform_64x64(short coeff[8][8],short block[64][64],int quant,int chroma);

int quantize_8x8(short coeff[8][8],int quant,int pic_type,int chroma,int *sum_of_levels);



int quantize_NxN_RDOQ(encoder_info *encoderInfo,short coeff[8][8],int quant,int block_type,float lambda_coeff,trsizes trsize);

int quantize_svt(int coeff[4][16],int quant,int pic_type,int chroma,int *no_of_coeffs,int *sum_of_levels);

int dequantize_8x8(short coeff[8][8],short rcoeff[8][8],int quant,int chroma);
int dequantize_svt(int coeff[4][16],int rcoeff[4][16],int quant,int chroma);
int dequantize_NxN(short coeff[8][8],short rcoeff[4][4],int quant,int chroma,int trsize);

int bitCount(encoder_info *encoderInfo,int k,int pos,int n,int lpflag,int levelMode,int run,int maxrun,int vlc_adaptive);
int bitCountN(encoder_info *encoderInfo,int k,int pos,int n,int lpflag,int levelMode,int run,int maxrun,int vlc_adaptive,int N);

static const short g1mat[4][4] = {
    {13, 13, 13, 13},
    {17,  7, -7,-17},
    {13,-13,-13, 13},
    { 7,-17, 17, -7}
};

static inline void transform_4x4(short block[4][4], short coeff[8][8])
{
    int i,j,k;
    short tmp[4][4];
    
    /* Horizontal transform */
    for (i=0; i<4; i++) 
    {			
        int sum[4] = {0,};
        for (j=0; j<4; j++)  
            for (k=0; k<4; k++) 
                sum[j] += g1mat[i][k]*block[j][k];
        for (j=0; j<4; j++) 
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }  
    /* Vertical transform */
    for (i=0; i<4; i++) 
    {                    
        int sum[4] = {0,};
        for (j=0; j<4; j++) 
            for (k=0; k<4; k++) 
                sum[j] += g1mat[i][k]*tmp[j][k];
        for (j=0; j<4; j++)
            sum[j] = (sum[j]+16)>>5;            
        for (j=0; j<4; j++)
            coeff[i][j] = min(max(sum[j], -32768), 32767);
    }
}

static const short g2mat[8][8] = {
    { 13, 13, 13, 13, 13, 13, 13, 13},
    { 17, 16, 11,  3, -3,-11,-16,-17},
    { 17,  7, -7,-17,-17, -7,  7, 17},
    { 16, -3,-17,-11, 11, 17,  3,-16},
    { 13,-13,-13, 13, 13,-13,-13, 13},
    { 11,-17,  3, 16,-16, -3, 17,-11},
    {  7,-17, 17, -7, -7, 17,-17,  7},
    {  3,-11, 16,-17, 17,-16, 11, -3}
};

#if 1
static inline void transform_8x8 (short block[8][8], short coeff[8][8])
{
    int i,j,k;
    short tmp[8][8];
    
    /* Horizontal transform */
    for (i=0; i<8; i++) 
    {
        int sum[8] = {0,};
        for (j=0; j<8; j++) 
            for (k=0; k<8; k++) 
                sum[j] += g2mat[i][k]*block[j][k];
        for (j=0; j<8; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }
    /* Vertical transform */
    for (i=0; i<8; i++) 
    {
        int sum[8] = {0,}; 
        for (j=0; j<8; j++) 
            for (k=0; k<8; k++) 
                sum[j] += g2mat[i][k]*tmp[j][k]; 
        for (j=0; j<8; j++)
            sum[j] = (sum[j]+64)>>7;            
        for (j=0; j<8; j++)
            coeff[i][j] = min(max(sum[j], -32768), 32767);
    }
}
#else
static inline void transform_8x8(short block[8][8], short coeff[8][8])
{
    int i;
    int tmp[8][8];
    int a, b, c, d, a_, b_, c_, d_;
    int aa, bb, aa_, bb_;

    for (i=0; i<8; i++)
    {
        a  = block[i][0] + block[i][7];
        a_ = block[i][0] - block[i][7];
        b  = block[i][1] + block[i][6];
        b_ = block[i][1] - block[i][6];
        c  = block[i][2] + block[i][5];
        c_ = block[i][2] - block[i][5];
        d  = block[i][3] + block[i][4];
        d_ = block[i][3] - block[i][4];
        aa = a + d;
        aa_ = a - d;
        bb = b + c;
        bb_ = b - c;

        tmp[0][i] = ((aa + bb)*g2mat[0][0] + 0) >> 0;
        tmp[4][i] = ((aa - bb)*g2mat[4][0] + 0) >> 0;
        tmp[2][i] = (aa_*g2mat[2][0] + bb_*g2mat[2][1] + 0) >> 0;
        tmp[6][i] = (aa_*g2mat[6][0] + bb_*g2mat[6][1] + 0) >> 0;
        tmp[1][i] = (a_*g2mat[1][0] + b_*g2mat[1][1] + c_*g2mat[1][2] + d_*g2mat[1][3] + 0) >> 0;
        tmp[3][i] = (a_*g2mat[3][0] + b_*g2mat[3][1] + c_*g2mat[3][2] + d_*g2mat[3][3] + 0) >> 0;
        tmp[5][i] = (a_*g2mat[5][0] + b_*g2mat[5][1] + c_*g2mat[5][2] + d_*g2mat[5][3] + 0) >> 0;
        tmp[7][i] = (a_*g2mat[7][0] + b_*g2mat[7][1] + c_*g2mat[7][2] + d_*g2mat[7][3] + 0) >> 0;

    }
    for (i=0; i<8; i++)
    {
        a  = tmp[i][0] + tmp[i][7];
        a_ = tmp[i][0] - tmp[i][7];
        b  = tmp[i][1] + tmp[i][6];
        b_ = tmp[i][1] - tmp[i][6];
        c  = tmp[i][2] + tmp[i][5];
        c_ = tmp[i][2] - tmp[i][5];
        d  = tmp[i][3] + tmp[i][4];
        d_ = tmp[i][3] - tmp[i][4];
        aa = a + d;
        aa_ = a - d;
        bb = b + c;
        bb_ = b - c;


        coeff[0][i] = (((aa + bb)*g2mat[0][0] + 64) >> 7);
        coeff[4][i] = (((aa - bb)*g2mat[4][0] + 64) >> 7);
        coeff[2][i] = ((aa_*g2mat[2][0] + bb_*g2mat[2][1] + 64) >> 7);
        coeff[6][i] = ((aa_*g2mat[6][0] + bb_*g2mat[6][1] + 64) >> 7);
        coeff[1][i] = ((a_*g2mat[1][0] + b_*g2mat[1][1] + c_*g2mat[1][2] + d_*g2mat[1][3] + 64) >> 7);
        coeff[3][i] = ((a_*g2mat[3][0] + b_*g2mat[3][1] + c_*g2mat[3][2] + d_*g2mat[3][3] + 64) >> 7);
        coeff[5][i] = ((a_*g2mat[5][0] + b_*g2mat[5][1] + c_*g2mat[5][2] + d_*g2mat[5][3] + 64) >> 7);
        coeff[7][i] = ((a_*g2mat[7][0] + b_*g2mat[7][1] + c_*g2mat[7][2] + d_*g2mat[7][3] + 64) >> 7);

    }    
}
#endif

static const short g3mat[8][16] = {
    {13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13},
    {18, 18, 16, 14, 12,  9,  5,  2, -2, -5, -9,-12,-14,-16,-18,-18},
    {17, 16, 11,  3, -3,-11,-16,-17,-17,-16,-11, -3,  3, 11, 16, 17},
    {18, 12,  2, -9,-16,-18,-14, -5,  5, 14, 18, 16,  9, -2,-12,-18},
    {17,  7, -7,-17,-17, -7,  7, 17, 17,  7, -7,-17,-17, -7,  7, 17},
    {16,  2,-14,-18, -5, 12, 18,  9, -9,-18,-12,  5, 18, 14, -2,-16},
    {16, -3,-17,-11, 11, 17,  3,-16,-16,  3, 17, 11,-11,-17, -3, 16},
    {14, -9,-18,  2, 18,  5,-16,-12, 12, 16, -5,-18, -2, 18,  9,-14}
};

static inline void transform_16x16(short block[16][16], short coeff[8][8])
{
    int i,j,k;
    short tmp[8][16];
    
    /* Horizontal transform */
    for (i=0; i<8; i++) 
    {           
        int sum[16] = {0,};
        for (j=0; j<16; j++) 
            for (k=0; k<16; k++) 
                sum[j] += g3mat[i][k]*block[j][k];
        for (j=0; j<16; j++)
            sum[j] = (sum[j]+1)>>1;
        for (j=0; j<16; j++) 
            tmp[i][j] = min(max(sum[j], -32768), 32767); 
    }
    /* Vertical transform */
    for (i=0; i<8; i++) 
    {                    
        int sum[8] = {0,};
        for (j=0; j<8; j++) 
            for (k=0; k<16; k++) 
                sum[j] += g3mat[i][k]*tmp[j][k];
        for (j=0; j<8; j++)
            sum[j] = (sum[j]+128)>>8;
        for (j=0; j<8; j++) 
            coeff[i][j] = min(max(sum[j], -32768), 32767); 
    }
}

static const short g4mat[8][32] = {
    {13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13},
    {18, 18, 18, 17, 17, 16, 15, 14, 12, 11,  9,  8,  6,  4,  3,  1, -1, -3, -4, -6, -8, -9,-11,-12,-14,-15,-16,-17,-17,-18,-18,-18},
    {18, 18, 16, 14, 12,  9,  5,  2, -2, -5, -9,-12,-14,-16,-18,-18,-18,-18,-16,-14,-12, -9, -5, -2,  2,  5,  9, 12, 14, 16, 18, 18},
    {18, 17, 14,  9,  4, -1, -6,-11,-15,-17,-18,-18,-16,-12, -8, -3,  3,  8, 12, 16, 18, 18, 17, 15, 11,  6,  1, -4, -9,-14,-17,-18},
    {18, 15, 10,  5, -5,-10,-15,-18,-18,-15,-10, -5,  5, 10, 15, 18, 18, 15, 10,  5, -5,-10,-15,-18,-18,-15,-10, -5,  5, 10, 15, 18},
    {18, 14,  6, -3,-11,-17,-18,-16, -9, -1,  8, 15, 18, 17, 12,  4, -4,-12,-17,-18,-15, -8,  1,  9, 16, 18, 17, 11,  3, -6,-14,-18},
    {18, 12,  2, -9,-16,-18,-14, -5,  5, 14, 18, 16,  9, -2,-12,-18,-18,-12, -2,  9, 16, 18, 14,  5, -5,-14,-18,-16, -9,  2, 12, 18},
    {17,  9, -3,-14,-18,-15, -4,  8, 17, 18, 11, -1,-12,-18,-16, -6,  6, 16, 18, 12,  1,-11,-18,-17, -8,  4, 15, 18, 14,  3, -9,-17}
};

static inline void transform_32x32(short block[32][32], short coeff[8][8]) 
{
    int i,j,k;
    short tmp[8][32];
    
    /* Horizontal transform */
    for (i=0; i<8; i++) 
    {			
        int sum[32] = {0,};
        for (j=0; j<32; j++) 
            for (k=0; k<32; k++) 
                sum[j] += g4mat[i][k]*block[j][k];
        for (j=0; j<32; j++)
            sum[j] = (sum[j]+2)>>2;
        for (j=0; j<32; j++) 
            tmp[i][j] = min(max(sum[j], -32768), 32767); 
    }
    /* Vertical transform */
    for (i=0; i<8; i++) 
    {                    
        int sum[8] = {0,};
        for (j=0; j<8; j++) 
            for (k=0; k<32; k++) 
                sum[j] += g4mat[i][k]*tmp[j][k];
        for (j=0; j<8; j++)
            sum[j] = (sum[j]+256)>>9;
        for (j=0; j<8; j++) 
            coeff[i][j] = min(max(sum[j], -32768), 32767); 
    }
}

static const short g5mat[8][64] = {
    {13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13, 13},
    {18, 18, 18, 18, 18, 18, 18, 17, 17, 16, 16, 16, 15, 15, 14, 13, 13, 12, 11, 11, 10,  9,  8,  7,  7,  6,  5,  4,  3,  2,  1,  0,  0, -1, -2, -3, -4, -5, -6, -7, -7, -8, -9,-10,-11,-11,-12,-13,-13,-14,-15,-15,-16,-16,-16,-17,-17,-18,-18,-18,-18,-18,-18,-18},
    {18, 18, 18, 17, 17, 16, 15, 14, 12, 11,  9,  8,  6,  4,  3,  1, -1, -3, -4, -6, -8, -9,-11,-12,-14,-15,-16,-17,-17,-18,-18,-18,-18,-18,-18,-17,-17,-16,-15,-14,-12,-11, -9, -8, -6, -4, -3, -1,  1,  3,  4,  6,  8,  9, 11, 12, 14, 15, 16, 17, 17, 18, 18, 18},
    {18, 18, 17, 16, 15, 13, 11,  8,  6,  3,  0, -2, -5, -7,-10,-12,-14,-16,-17,-18,-18,-18,-18,-18,-16,-15,-13,-11, -9, -7, -4, -1,  1,  4,  7,  9, 11, 13, 15, 16, 18, 18, 18, 18, 18, 17, 16, 14, 12, 10,  7,  5,  2,  0, -3, -6, -8,-11,-13,-15,-16,-17,-18,-18},
    {18, 18, 16, 14, 12,  9,  5,  2, -2, -5, -9,-12,-14,-16,-18,-18,-18,-18,-16,-14,-12, -9, -5, -2,  2,  5,  9, 12, 14, 16, 18, 18, 18, 18, 16, 14, 12,  9,  5,  2, -2, -5, -9,-12,-14,-16,-18,-18,-18,-18,-16,-14,-12, -9, -5, -2,  2,  5,  9, 12, 14, 16, 18, 18},
    {18, 17, 15, 12,  8,  4,  0, -5, -9,-13,-16,-18,-18,-18,-17,-15,-11, -7, -3,  1,  6, 10, 13, 16, 18, 18, 18, 16, 14, 11,  7,  2, -2, -7,-11,-14,-16,-18,-18,-18,-16,-13,-10, -6, -1,  3,  7, 11, 15, 17, 18, 18, 18, 16, 13,  9,  5,  0, -4, -8,-12,-15,-17,-18},
    {18, 17, 14,  9,  4, -1, -6,-11,-15,-17,-18,-18,-16,-12, -8, -3,  3,  8, 12, 16, 18, 18, 17, 15, 11,  6,  1, -4, -9,-14,-17,-18,-18,-17,-14, -9, -4,  1,  6, 11, 15, 17, 18, 18, 16, 12,  8,  3, -3, -8,-12,-16,-18,-18,-17,-15,-11, -6, -1,  4,  9, 14, 17, 18},
    {18, 16, 12,  7,  0, -6,-11,-16,-18,-18,-16,-13, -7, -1,  5, 11, 15, 18, 18, 17, 13,  8,  2, -4,-10,-15,-18,-18,-17,-14, -9, -3,  3,  9, 14, 17, 18, 18, 15, 10,  4, -2, -8,-13,-17,-18,-18,-15,-11, -5,  1,  7, 13, 16, 18, 18, 16, 11,  6,  0, -7,-12,-16,-18}
};

static inline void transform_64x64(short block[64][64], short coeff[8][8])
{
    int i,j,k;
    short tmp[8][64];
    
    /* Horizontal transform */
    for (i=0; i<8; i++) 
    {			
        int sum[64] = {0,};
        for (j=0; j<64; j++) 
            for (k=0; k<64; k++) 
                sum[j] += g5mat[i][k]*block[j][k];
        for (j=0; j<64; j++)
            sum[j] = (sum[j]+4)>>3;
        for (j=0; j<64; j++) 
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }
    /* Vertical transform */
    for (i=0; i<8; i++) 
    {                    
        int sum[8] = {0,};
        for (j=0; j<8; j++) 
            for (k=0; k<64; k++) 
                sum[j] += g5mat[i][k]*tmp[j][k];
        for (j=0; j<8; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<8; j++) 
            coeff[i][j] = min(max(sum[j], -32768), 32767);
    }
}

static const short g1mat_tp[4][4] = {
    { 13, 17, 13,  7},
    { 13,  7,-13,-17},
    { 13, -7,-13, 17},
    { 13,-17, 13, -7}
};

static inline void inverse_transform_4x4(short coeff[4][4],short block[4][4],int quant)
{
    int i,j,k;
    const int qp0 = quant/6;
    short tmp[4][4];
    
    /* Horizontal transform */
    for (i=0; i<4; i++)
    {
        int sum[4] = {0,};
        for (j=0; j<4; j++)
            for (k=0; k<4; k++)
                sum[j] += g1mat_tp[i][k]*coeff[k][j];
        for (j=0; j<4; j++)
            sum[j] = ((sum[j]<<qp0)+1)>>1;
        for (j=0; j<4; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }
    /* Vertical transform */
    for (i=0; i<4; i++)
    {
        int sum[4] = {0,};
        for (j=0; j<4; j++)
            for (k=0; k<4; k++)
                sum[j] += g1mat_tp[j][k]*tmp[i][k];
        for (j=0; j<4; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<4; j++)
            block[i][j] = min(max(sum[j], -32768), 32767);
    }
}

static const short g2mat_tp[8][8] = {
    { 13, 17, 17, 16, 13, 11,  7,  3},
    { 13, 16,  7, -3,-13,-17,-17,-11},
    { 13, 11, -7,-17,-13,  3, 17, 16},
    { 13,  3,-17,-11, 13, 16, -7,-17},
    { 13, -3,-17, 11, 13,-16, -7, 17},
    { 13,-11, -7, 17,-13, -3, 17,-16},
    { 13,-16,  7,  3,-13, 17,-17, 11},
    { 13,-17, 17,-16, 13,-11,  7, -3}
};

#if 1
static inline void inverse_transform_8x8(short coeff[8][8],short block[8][8],int quant,int chroma)
{
    int i,j,k;
    const int qp = (chroma) ? (quant-corr_qpc_8x8[quant]) : quant;
    const int qp0 = qp/6;
    short tmp[8][8];
    
    /* Horizontal transform */
    for (i=0; i<8; i++)
    {
        int sum[8] = {0,};                       
        for (j=0; j<8; j++)
            for (k=0; k<8; k++)                
                sum[j] += g2mat_tp[i][k]*coeff[k][j];
        for (j=0; j<8; j++)
            sum[j] = ((sum[j]<<qp0)+2)>>2;
        for (j=0; j<8; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }
    /* Vertical transform */
    for (i=0; i<8; i++)
    {
        int sum[8] = {0,};
        for (j=0; j<8; j++)
            for (k=0; k<8; k++)
                sum[j] += g2mat_tp[j][k]*tmp[i][k]; 
        for (j=0; j<8; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<8; j++)
            block[i][j] = min(max(sum[j], -32768), 32767);
    }
}
#else
static inline void inverse_transform_8x8(short coeff[8][8],short block[8][8],int quant,int chroma)
{
    int i;
    int qp0;
    int tmp[8][8];
    int round0;
    int shift0;
    int a, b, c, d, e, f, g, h;
    int aa, bb, aa_, bb_;
    
    if (chroma) quant = quant-corr_qpc_8x8[quant];
    qp0 = quant/6;
    shift0 = min(4,6-qp0);
    //round0 = 1<<(shift0-1);
    round0 = shift0 ? 1<<(shift0-1) : 0; //bugfix rev159
    for (i=0; i<8; i++)
    {

        a = (coeff[0][i] + coeff[4][i]) * g2mat_tp[0][0];
        b = (coeff[0][i] - coeff[4][i]) * g2mat_tp[0][0];
        c = coeff[2][i]*g2mat_tp[0][2] + coeff[6][i]*g2mat_tp[0][6];
        d = coeff[2][i]*g2mat_tp[1][2] + coeff[6][i]*g2mat_tp[1][6];
        e = coeff[1][i]*g2mat_tp[0][1] + coeff[3][i]*g2mat_tp[0][3] + 
            coeff[5][i]*g2mat_tp[0][5] + coeff[7][i]*g2mat_tp[0][7];
        f = coeff[1][i]*g2mat_tp[1][1] + coeff[3][i]*g2mat_tp[1][3] + 
            coeff[5][i]*g2mat_tp[1][5] + coeff[7][i]*g2mat_tp[1][7];
        g = coeff[1][i]*g2mat_tp[2][1] + coeff[3][i]*g2mat_tp[2][3] + 
            coeff[5][i]*g2mat_tp[2][5] + coeff[7][i]*g2mat_tp[2][7];
        h = coeff[1][i]*g2mat_tp[3][1] + coeff[3][i]*g2mat_tp[3][3] + 
            coeff[5][i]*g2mat_tp[3][5] + coeff[7][i]*g2mat_tp[3][7];

        aa = a + c;
        aa_ = a - c;
        bb = b + d;
        bb_ = b - d;

        tmp[0][i] = (((aa + e)<<qp0) + 2) >> 2;
        tmp[7][i] = (((aa - e)<<qp0) + 2) >> 2;
        tmp[1][i] = (((bb + f)<<qp0) + 2) >> 2;
        tmp[6][i] = (((bb - f)<<qp0) + 2) >> 2;
        tmp[2][i] = (((bb_ + g)<<qp0) + 2) >> 2;
        tmp[5][i] = (((bb_ - g)<<qp0) + 2) >> 2;
        tmp[3][i] = (((aa_ + h)<<qp0) + 2) >> 2;
        tmp[4][i] = (((aa_ - h)<<qp0) + 2) >> 2;

    }
    for (i=0; i<8; i++)
    {
        a = (tmp[i][0] + tmp[i][4]) * g2mat_tp[0][0];
        b = (tmp[i][0] - tmp[i][4]) * g2mat_tp[0][0];
        c = tmp[i][2]*g2mat_tp[0][2] + tmp[i][6]*g2mat_tp[0][6];
        d = tmp[i][2]*g2mat_tp[1][2] + tmp[i][6]*g2mat_tp[1][6];
        e = tmp[i][1]*g2mat_tp[0][1] + tmp[i][3]*g2mat_tp[0][3] + 
            tmp[i][5]*g2mat_tp[0][5] + tmp[i][7]*g2mat_tp[0][7];
        f = tmp[i][1]*g2mat_tp[1][1] + tmp[i][3]*g2mat_tp[1][3] + 
            tmp[i][5]*g2mat_tp[1][5] + tmp[i][7]*g2mat_tp[1][7];
        g = tmp[i][1]*g2mat_tp[2][1] + tmp[i][3]*g2mat_tp[2][3] + 
            tmp[i][5]*g2mat_tp[2][5] + tmp[i][7]*g2mat_tp[2][7];
        h = tmp[i][1]*g2mat_tp[3][1] + tmp[i][3]*g2mat_tp[3][3] + 
            tmp[i][5]*g2mat_tp[3][5] + tmp[i][7]*g2mat_tp[3][7];
        aa = a + c;
        aa_ = a - c;
        bb = b + d;
        bb_ = b - d;

        block[i][0] = (aa + e + 512) >> 10;
        block[i][7] = (aa - e + 512) >> 10;
        block[i][1] = (bb + f + 512) >> 10;
        block[i][6] = (bb - f + 512) >> 10;
        block[i][2] = (bb_ + g + 512) >> 10;
        block[i][5] = (bb_ - g + 512) >> 10;
        block[i][3] = (aa_ + h + 512) >> 10;
        block[i][4] = (aa_ - h + 512) >> 10;

    }
    
}
#endif

static const short g3mat_tp[16][8] = {
    { 13, 18, 17, 18, 17, 16, 16, 14},
    { 13, 18, 16, 12,  7,  2, -3, -9},
    { 13, 16, 11,  2, -7,-14,-17,-18},
    { 13, 14,  3, -9,-17,-18,-11,  2},
    { 13, 12, -3,-16,-17, -5, 11, 18},
    { 13,  9,-11,-18, -7, 12, 17,  5},
    { 13,  5,-16,-14,  7, 18,  3,-16},
    { 13,  2,-17, -5, 17,  9,-16,-12},
    { 13, -2,-17,  5, 17, -9,-16, 12},
    { 13, -5,-16, 14,  7,-18,  3, 16},
    { 13, -9,-11, 18, -7,-12, 17, -5},
    { 13,-12, -3, 16,-17,  5, 11,-18},
    { 13,-14,  3,  9,-17, 18,-11, -2},
    { 13,-16, 11, -2, -7, 14,-17, 18},
    { 13,-18, 16,-12,  7, -2, -3,  9},
    { 13,-18, 17,-18, 17,-16, 16,-14}
};

static inline void inverse_transform_16x16(short coeff[8][8],short block[16][16],int quant,int chroma)
{
    int i,j,k;
    const int qp = (chroma) ? (quant-corr_qpc_8x8[quant]) : quant;
    const int qp0 = qp/6;
    short tmp[16][8];
    
    /* Horizontal transform */
    for (i=0; i<16; i++)
    {
        int sum[8] = {0,};
        for (j=0; j<8; j++)
            for (k=0; k<8; k++)
                sum[j] += g3mat_tp[i][k]*coeff[k][j]; 
        for (j=0; j<8; j++)
                sum[j] = ((sum[j]<<qp0)+4) >> 3;
        for (j=0; j<8; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }   
    /* Vertical transform */
    for (i=0; i<16; i++)
    {
        int sum[16] = {0,};
        for (j=0; j<16; j++)
            for (k=0; k<8; k++)
                sum[j] += g3mat_tp[j][k]*tmp[i][k];
        for (j=0; j<16; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<16; j++)
            block[i][j] = min(max(sum[j], -32768), 32767);
    }
}

static const short g4mat_tp[32][8] = {
    { 13, 18, 18, 18, 18, 18, 18, 17},
    { 13, 18, 18, 17, 15, 14, 12,  9},
    { 13, 18, 16, 14, 10,  6,  2, -3},
    { 13, 17, 14,  9,  5, -3, -9,-14},
    { 13, 17, 12,  4, -5,-11,-16,-18},
    { 13, 16,  9, -1,-10,-17,-18,-15},
    { 13, 15,  5, -6,-15,-18,-14, -4},
    { 13, 14,  2,-11,-18,-16, -5,  8},
    { 13, 12, -2,-15,-18, -9,  5, 17},
    { 13, 11, -5,-17,-15, -1, 14, 18},
    { 13,  9, -9,-18,-10,  8, 18, 11},
    { 13,  8,-12,-18, -5, 15, 16, -1},
    { 13,  6,-14,-16,  5, 18,  9,-12},
    { 13,  4,-16,-12, 10, 17, -2,-18},
    { 13,  3,-18, -8, 15, 12,-12,-16},
    { 13,  1,-18, -3, 18,  4,-18, -6},
    { 13, -1,-18,  3, 18, -4,-18,  6},
    { 13, -3,-18,  8, 15,-12,-12, 16},
    { 13, -4,-16, 12, 10,-17, -2, 18},
    { 13, -6,-14, 16,  5,-18,  9, 12},
    { 13, -8,-12, 18, -5,-15, 16,  1},
    { 13, -9, -9, 18,-10, -8, 18,-11},
    { 13,-11, -5, 17,-15,  1, 14,-18},
    { 13,-12, -2, 15,-18,  9,  5,-17},
    { 13,-14,  2, 11,-18, 16, -5, -8},
    { 13,-15,  5,  6,-15, 18,-14,  4},
    { 13,-16,  9,  1,-10, 17,-18, 15},
    { 13,-17, 12, -4, -5, 11,-16, 18},
    { 13,-17, 14, -9,  5,  3, -9, 14},
    { 13,-18, 16,-14, 10, -6,  2,  3},
    { 13,-18, 18,-17, 15,-14, 12, -9},
    { 13,-18, 18,-18, 18,-18, 18,-17}
};

static inline void inverse_transform_32x32(short coeff[8][8],short block[32][32],int quant,int chroma)
{
    int i,j,k;
    const int qp = (chroma) ? (quant-corr_qpc_8x8[quant]) : quant;
    const int qp0 = qp/6;
    short tmp[32][8];
    
    /* Horizontal transform */
    for (i=0; i<32; i++)
    {
        int sum[8] = {0,};
        for (j=0; j<8; j++)
            for (k=0; k<8; k++)
                sum[j] += g4mat_tp[i][k]*coeff[k][j];
        for (j=0; j<8; j++)
            sum[j] = ((sum[j]<<qp0)+8) >> 4;
        for (j=0; j<8; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }    
    /* Vertical transform */
    for (i=0; i<32; i++)
    {
        int sum[32] = {0,};
        for (j=0; j<32; j++)
            for (k=0; k<8; k++)
                sum[j] += g4mat_tp[j][k]*tmp[i][k]; 
        for (j=0; j<32; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<32; j++)
            block[i][j] = min(max(sum[j], -32768), 32767);
    }
}

static const short g5mat_tp[64][8] = {
    { 13, 18, 18, 18, 18, 18, 18, 18},
    { 13, 18, 18, 18, 18, 17, 17, 16},
    { 13, 18, 18, 17, 16, 15, 14, 12},
    { 13, 18, 17, 16, 14, 12,  9,  7},
    { 13, 18, 17, 15, 12,  8,  4,  0},
    { 13, 18, 16, 13,  9,  4, -1, -6},
    { 13, 18, 15, 11,  5,  0, -6,-11},
    { 13, 17, 14,  8,  2, -5,-11,-16},
    { 13, 17, 12,  6, -2, -9,-15,-18},
    { 13, 16, 11,  3, -5,-13,-17,-18},
    { 13, 16,  9,  0, -9,-16,-18,-16},
    { 13, 16,  8, -2,-12,-18,-18,-13},
    { 13, 15,  6, -5,-14,-18,-16, -7},
    { 13, 15,  4, -7,-16,-18,-12, -1},
    { 13, 14,  3,-10,-18,-17, -8,  5},
    { 13, 13,  1,-12,-18,-15, -3, 11},
    { 13, 13, -1,-14,-18,-11,  3, 15},
    { 13, 12, -3,-16,-18, -7,  8, 18},
    { 13, 11, -4,-17,-16, -3, 12, 18},
    { 13, 11, -6,-18,-14,  1, 16, 17},
    { 13, 10, -8,-18,-12,  6, 18, 13},
    { 13,  9, -9,-18, -9, 10, 18,  8},
    { 13,  8,-11,-18, -5, 13, 17,  2},
    { 13,  7,-12,-18, -2, 16, 15, -4},
    { 13,  7,-14,-16,  2, 18, 11,-10},
    { 13,  6,-15,-15,  5, 18,  6,-15},
    { 13,  5,-16,-13,  9, 18,  1,-18},
    { 13,  4,-17,-11, 12, 16, -4,-18},
    { 13,  3,-17, -9, 14, 14, -9,-17},
    { 13,  2,-18, -7, 16, 11,-14,-14},
    { 13,  1,-18, -4, 18,  7,-17, -9},
    { 13,  0,-18, -1, 18,  2,-18, -3},
    { 13,  0,-18,  1, 18, -2,-18,  3},
    { 13, -1,-18,  4, 18, -7,-17,  9},
    { 13, -2,-18,  7, 16,-11,-14, 14},
    { 13, -3,-17,  9, 14,-14, -9, 17},
    { 13, -4,-17, 11, 12,-16, -4, 18},
    { 13, -5,-16, 13,  9,-18,  1, 18},
    { 13, -6,-15, 15,  5,-18,  6, 15},
    { 13, -7,-14, 16,  2,-18, 11, 10},
    { 13, -7,-12, 18, -2,-16, 15,  4},
    { 13, -8,-11, 18, -5,-13, 17, -2},
    { 13, -9, -9, 18, -9,-10, 18, -8},
    { 13,-10, -8, 18,-12, -6, 18,-13},
    { 13,-11, -6, 18,-14, -1, 16,-17},
    { 13,-11, -4, 17,-16,  3, 12,-18},
    { 13,-12, -3, 16,-18,  7,  8,-18},
    { 13,-13, -1, 14,-18, 11,  3,-15},
    { 13,-13,  1, 12,-18, 15, -3,-11},
    { 13,-14,  3, 10,-18, 17, -8, -5},
    { 13,-15,  4,  7,-16, 18,-12,  1},
    { 13,-15,  6,  5,-14, 18,-16,  7},
    { 13,-16,  8,  2,-12, 18,-18, 13},
    { 13,-16,  9,  0, -9, 16,-18, 16},
    { 13,-16, 11, -3, -5, 13,-17, 18},
    { 13,-17, 12, -6, -2,  9,-15, 18},
    { 13,-17, 14, -8,  2,  5,-11, 16},
    { 13,-18, 15,-11,  5,  0, -6, 11},
    { 13,-18, 16,-13,  9, -4, -1,  6},
    { 13,-18, 17,-15, 12, -8,  4,  0},
    { 13,-18, 17,-16, 14,-12,  9, -7},
    { 13,-18, 18,-17, 16,-15, 14,-12},
    { 13,-18, 18,-18, 18,-17, 17,-16},
    { 13,-18, 18,-18, 18,-18, 18,-18}
};

static void inverse_transform_64x64(short coeff[8][8],short block[64][64],int quant,int chroma)
{
    int i,j,k;
    const int qp = (chroma) ? (quant-corr_qpc_8x8[quant]) : quant;
    const int qp0 = qp/6;
    short tmp[64][8];
    
    /* Horizontal transform */
    for (i=0; i<64; i++)
    {
        int sum[8] = {0,};
        for (j=0; j<8; j++)
            for (k=0; k<8; k++)
                sum[j] += g5mat_tp[i][k]*coeff[k][j]; 
        for (j=0; j<8; j++)
            sum[j] = ((sum[j]<<qp0)+16)>>5;
        for (j=0; j<8; j++)
            tmp[i][j] = min(max(sum[j], -32768), 32767);
    }    
    /* Vertical transform */
    for (i=0; i<64; i++)
    {
        int sum[64] = {0,};
        for (j=0; j<64; j++)
            for (k=0; k<8; k++)
                sum[j] += g5mat_tp[j][k]*tmp[i][k]; 
        for (j=0; j<64; j++)
            sum[j] = (sum[j]+512)>>10;
        for (j=0; j<64; j++)
            block[i][j] = min(max(sum[j], -32768), 32767);
    }
}

#endif
