/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <stdlib.h>
#include "global.h"
#include "tntmdec.h"
#include "rfutil.h"


static const int tctable_8x8[46] =
{
    0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,5,5,6,6,7,8,9,9,10,10,11,11,12,12,13,13,14,14
};

static const int betatable_8x8[42] =
{
     0, 0, 0, 0, 0, 0, 6, 7, 8, 9,10,11,12,13,14,15,16,17,
    18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60,62,64
};


/*-----------------------------------------------------------------------------------------
Function:   filter_8x8_vertical
Purpose:    In-place weak filter deblocking of the vertical boundary between two 8x8 luma blocks
Input:      mb         - 2D pointer to pixel data containing the vertical edge
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
            beta       - QP-dependent threshold for whether to do filtering or not (similar to AVC)
Output:     mb         - Pixel data after fltering of the vertical edge
Parameters: None 
-------------------------------------------------------------------------------------------*/

int filter_8x8_vertical(unsigned char **mb,int tc,int beta)
{
    int i,d;

    int delta;

    /* First decide if change is too big, resulting in no filtering. */
    d = IABS(mb[2][1]-2*mb[2][2]+mb[2][3])+IABS(mb[2][4]-2*mb[2][5]+mb[2][6])+
        IABS(mb[5][1]-2*mb[5][2]+mb[5][3])+IABS(mb[5][4]-2*mb[5][5]+mb[5][6]);
    if (d < beta)
    {
        for (i=0;i<8;i++)
        {


            delta = IMIN(tc,IMAX(-tc,(13*(mb[i][4]-mb[i][3]) + 4*(mb[i][5]-mb[i][2]) - 5*(mb[i][6]-mb[i][1])+16)>>5));                                      

            mb[i][3] = IMIN(255,IMAX(0,mb[i][3]+delta));
            mb[i][4] = IMIN(255,IMAX(0,mb[i][4]-delta));
            mb[i][2] = IMIN(255,IMAX(0,mb[i][2]+delta/2));
            mb[i][5] = IMIN(255,IMAX(0,mb[i][5]-delta/2));

        }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_8x8_horizontal
Purpose:    In-place weak filter deblocking of the horizontal boundary between two 8x8 luma blocks
Input:      mb         - 2D pointer to pixel data containing the horizontal edge
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
            beta       - QP-dependent threshold for whether to do filtering or not (similar to AVC)
Output:     mb         - Pixel data after fltering of the vertical edge
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_8x8_horizontal(unsigned char **mb,int tc,int beta)
{
    int i,d;

    int delta;

    /* First decide if change is too big, resulting in no filtering. */
    d = IABS(mb[1][2]-2*mb[2][2]+mb[3][2])+IABS(mb[4][2]-2*mb[5][2]+mb[6][2])+
        IABS(mb[1][5]-2*mb[2][5]+mb[3][5])+IABS(mb[4][5]-2*mb[5][5]+mb[6][5]);
    if (d < beta)
    {
        for (i=0;i<8;i++)
        {


            delta = IMIN(tc,IMAX(-tc,(13*(mb[4][i]-mb[3][i]) + 4*(mb[5][i]-mb[2][i]) - 5*(mb[6][i]-mb[1][i]) + 16)>>5));                                      

            mb[3][i] = IMIN(255,IMAX(0,mb[3][i]+delta));
            mb[4][i] = IMIN(255,IMAX(0,mb[4][i]-delta));
            mb[2][i] = IMIN(255,IMAX(0,mb[2][i]+delta/2));
            mb[5][i] = IMIN(255,IMAX(0,mb[5][i]-delta/2));

        }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_8x8_vertical_long
Purpose:    In-place weak or strong filter deblocking of the vertical boundary between two 8x8 luma blocks
Input:      **mb       - 2D pointer to pixel data containing the vertical edge 
            x          - Horizontal coordinate of the vertical edge within the input block
            y          - Vertical coordinate of the vertical edge within the input block
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
            beta       - QP-dependent threshold for whether to do filtering or not (similar to AVC)
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_8x8_vertical_long(unsigned char **mb,int tc,int beta)
{
    int i,d,d_strong;

    int delta;

    
    /* First decide if change is too big, resulting in no filtering. */
    d = IABS(mb[2][1]-2*mb[2][2]+mb[2][3])+IABS(mb[2][4]-2*mb[2][5]+mb[2][6])+
        IABS(mb[5][1]-2*mb[5][2]+mb[5][3])+IABS(mb[5][4]-2*mb[5][5]+mb[5][6]);

	if ( d < beta )
	{
		for (i=0;i<8;i++)
		{
			 d_strong = IABS(mb[i][0]-mb[i][3]) + IABS(mb[i][7]-mb[i][4]);

       if ( (d_strong < (beta>>3)) && (d<(beta>>2)) && ( IABS(mb[i][3]-mb[i][4] ) < ((tc*5+1)>>1)) ) //strong filtering 
  		 {

						mb[i][3]=IMIN(255,IMAX(0,(mb[i][1]+2*mb[i][2]+2*mb[i][3]+2*mb[i][4]+mb[i][5]+4)>>3)); 
						mb[i][4]=IMIN(255,IMAX(0,(mb[i][2]+2*mb[i][3]+2*mb[i][4]+2*mb[i][5]+mb[i][6]+4)>>3)); 
			      
						mb[i][2]=IMIN(255,IMAX(0,(mb[i][1]+mb[i][2]+mb[i][3]+mb[i][4]+2)>>2));
						mb[i][5]=IMIN(255,IMAX(0,(mb[i][3]+mb[i][4]+mb[i][5]+mb[i][6]+2)>>2));

						mb[i][1]=IMIN(255,IMAX(0,(mb[i][0]*2+mb[i][1]*3+mb[i][2]+mb[i][3]+mb[i][4]+4)>>3));
						mb[i][6]=IMIN(255,IMAX(0,(mb[i][3]+mb[i][4]+mb[i][5]+mb[i][6]*3+mb[i][7]*2+4)>>3));

				}
				else 
				{


            delta = IMIN(tc,IMAX(-tc,(13*(mb[i][4]-mb[i][3]) + 4*(mb[i][5]-mb[i][2]) - 5*(mb[i][6]-mb[i][1])+16)>>5));                                      

            mb[i][3] = IMIN(255,IMAX(0,mb[i][3]+delta));
            mb[i][4] = IMIN(255,IMAX(0,mb[i][4]-delta));
            mb[i][2] = IMIN(255,IMAX(0,mb[i][2]+delta/2));
            mb[i][5] = IMIN(255,IMAX(0,mb[i][5]-delta/2));

				}
		  }		
   }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_8x8_horizontal_long
Purpose:    In-place weak or strong filter deblocking of the horizontal boundary between two 8x8 luma blocks
Input:      **mb       - 2D pointer to pixel data containing the horizontal edge 
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
            beta       - QP-dependent threshold for whether to do filtering or not (similar to AVC)
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_8x8_horizontal_long(unsigned char **mb,int tc,int beta)
{
    int i,d,d_strong;

    int delta;


    /* First decide if change is too big, resulting in no filtering. */
    d = IABS(mb[1][2]-2*mb[2][2]+mb[3][2])+IABS(mb[4][2]-2*mb[5][2]+mb[6][2])+
        IABS(mb[1][5]-2*mb[2][5]+mb[3][5])+IABS(mb[4][5]-2*mb[5][5]+mb[6][5]);

	if ( d < (beta)) 
	{
		for (i=0;i<8;i++)
		{
  		  d_strong = IABS(mb[0][i]-mb[3][i]) + IABS(mb[7][i]-mb[4][i]);

        if ( (d_strong < (beta>>3)) && (d<(beta>>2)) && ( IABS(mb[3][i]-mb[4][i]) < ((tc*5+1)>>1) ) ) //strong filtering 
				{
				  mb[3][i]=IMIN(255,IMAX(0,(mb[1][i]+2*mb[2][i]+2*mb[3][i]+2*mb[4][i]+mb[5][i]+4)>>3)); 
				  mb[4][i]=IMIN(255,IMAX(0,(mb[2][i]+2*mb[3][i]+2*mb[4][i]+2*mb[5][i]+mb[6][i]+4)>>3)); 
		      
				  mb[2][i]=IMIN(255,IMAX(0,(mb[1][i]+mb[2][i]+mb[3][i]+mb[4][i]+2)>>2));
				  mb[5][i]=IMIN(255,IMAX(0,(mb[3][i]+mb[4][i]+mb[5][i]+mb[6][i]+2)>>2));

				  mb[1][i]=IMIN(255,IMAX(0,(mb[0][i]*2+mb[1][i]*3+mb[2][i]+mb[3][i]+mb[4][i]+4)>>3));
				  mb[6][i]=IMIN(255,IMAX(0,(mb[3][i]+mb[4][i]+mb[5][i]+mb[6][i]*3+mb[7][i]*2+4)>>3));        

			  }
				else  // weak filtering

				{


            delta = IMIN(tc,IMAX(-tc,(13*(mb[4][i]-mb[3][i]) + 4*(mb[5][i]-mb[2][i]) - 5*(mb[6][i]-mb[1][i]) + 16)>>5));                                      

            mb[3][i] = IMIN(255,IMAX(0,mb[3][i]+delta));
            mb[4][i] = IMIN(255,IMAX(0,mb[4][i]-delta));
            mb[2][i] = IMIN(255,IMAX(0,mb[2][i]+delta/2));
            mb[5][i] = IMIN(255,IMAX(0,mb[5][i]-delta/2));

				}
		  
		  }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_chroma_vertical_8x8
Purpose:    In-place deblocking of the vertical boundary between two 8x8 chroma blocks
Input:      **mb       - 2D pointer to pixel data containing the vertical edge 
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
            beta       - QP-dependent threshold for whether to do filtering or not (similar to AVC)
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_chroma_vertical_8x8(unsigned char **mb,int tc)
{
    int i,delta;

    for (i=0;i<8;i++)
    {
        delta = IMIN(tc,IMAX(-tc,(((((mb[i][4]-mb[i][3])<<2)+mb[i][2]-mb[i][5]+4)>>3))));
        mb[i][3] = IMIN(255,IMAX(0,mb[i][3]+delta));
        mb[i][4] = IMIN(255,IMAX(0,mb[i][4]-delta));
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_chroma_horizontal_8x8
Purpose:    In-place deblocking of the horizontal boundary between two 8x8 chroma blocks
Input:      **mb       - 2D pointer to pixel data containing the horizontal edge 
            tc         - QP-dependent threshold value of delta modifier (similar to AVC)
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_chroma_horizontal_8x8(unsigned char **mb,int tc)
{
    int i,delta;

    for (i=0;i<8;i++)
    {
        delta = IMIN(tc,IMAX(-tc,(((((mb[4][i]-mb[3][i])<<2)+mb[2][i]-mb[5][i]+4)>>3))));
        mb[3][i] = IMIN(255,IMAX(0,mb[3][i]+delta));
        mb[4][i] = IMIN(255,IMAX(0,mb[4][i]-delta));
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_8x8
Purpose:    In-place deblocking for one frame (luma and chroma)
Parameters: tctable_8x8[46] - QP-dependent threshold values of delta modifier (similar to AVC)
            betatable_8x8[42] - QP-dependent threshold values for whether to do filtering or not (similar to AVC)
-------------------------------------------------------------------------------------------*/

int filter_8x8(decoder_info *decoderInfo,frame_info *frameInfo,image_frame *img)
{
    int i,j,k,l;
    int tc;
    unsigned char **block = ucsubmatrix(img->Y,0,7,0,7,0,0);
    int mby,mbx;
    int do_filter=0;
    int alt_tc;
    int vy,vx;
    int first;
    int qp;
    int mbxmax = img->cols>>4;
    int mbpos,bpos,mode,tr16flag;
    int svt;


    mb_info *mbmap2 = frameInfo->mbmap;
   
    block_info *mvmap = frameInfo->vmap[0];

    /* Y */
    for (i=0;i<img->rows;i+=16){ // for each row of MBs
        for (j=0;j<img->cols;j+=16){ // for each col of MBs
            mby = i>>4;
		    mbx = j>>4;
            mbpos = mby*mbxmax+mbx;
            
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
            tr16flag = mbmap2[mbpos].tr16flag;
            svt = mbmap2[mbpos].svt;

			/* Vertical Edges */
			first = (j>0) ? 0 : 8;
            for (l=first;l<16;l+=8){ // for each col of 8x8 blocks			
                for (k=0;k<16;k+=8){ // for each row of 8x8 blocks								
					do_filter = 0;
					alt_tc = 0;
					vy = (i+k)/8;
					vx = (j+l)/8;
                    bpos = vy*2*mbxmax+vx;
					if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA){
						do_filter = 1;						
                        alt_tc = 4;
					}
					/* Check cbp for coded A, B, C or D  */
					if (!do_filter){
						if (mvmap[bpos-1].coded_y || mvmap[bpos].coded_y){
							do_filter = 1;
						}
					}
					/* Check vector differences */
					if (!do_filter){
						if (abs(mvmap[bpos-1].x-mvmap[bpos].x) >= 4 || abs(mvmap[bpos-1].y-mvmap[bpos].y) >= 4){
							do_filter = 1;
						}
					}
                    if ((mode>=MODE_16x16 && mode<MODE_8x8) && tr16flag==0 && svt==1 && (vx&1)){
                        do_filter = 0;
                    }

                   if (mode==MODE_16x16 && tr16flag && (vx&1)){
						do_filter = 0;
					}			


                    if (mode==MODE_INTRA && tr16flag && (vx&1)){
						do_filter = 0;
					}			


      if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-1].planar_flag && !(vx&1))
      {
        do_filter = 0;  // A fix for Nokia's planar mode filtering
      }
                  
                    //Fred bugfix
					//if (mode==MODE_LMB32 && (vx%(LMBXSIZE/8))){
					if (mode==MODE_LMB64 && (vx%(LMBXSIZE/8))){
						do_filter = 0;
					}
					//Fred bugfix
                    //if (mode==(MODE_LMB64) && (vx%(LMBXSIZE/16))){
					if (mode==(MODE_LMB32) && (vx%(LMBXSIZE/16))){
						do_filter = 0;
					}			


                    
					if (do_filter){
						tc = tctable_8x8[qp+alt_tc];
						move_ucsubmatrix(block,img->Y,i+k,i+k+7,j+l-4,j+l+3,0,0);

						filter_8x8_vertical_long(block,tc,betatable_8x8[qp]);	

					}

				}
			}

			/* Horizontal Edges */			
			first = (i>0) ? 0 : 8;
            for (k=first;k<16;k+=8){ // for each row of 8x8 blocks
                for (l=0;l<16;l+=8){ // for each col of 8x8 blocks
					do_filter = 0;
					mby = i>>4;
					mbx = j>>4;					
					alt_tc = 0;
					vy = (i+k)/8;
					vx = (j+l)/8;
                    bpos = vy*2*mbxmax+vx;
					if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA){
						do_filter = 1;
						alt_tc = 4;
					}
					/* Check cbp for coded A, B, C or D  */
					if (!do_filter){
						if (mvmap[bpos-2*mbxmax].coded_y || mvmap[bpos].coded_y){
							do_filter = 1;
						}
					}
					/* Check vector differences */
					if (!do_filter){
						if (abs(mvmap[bpos-2*mbxmax].x-mvmap[bpos].x) >= 4 || abs(mvmap[bpos-2*mbxmax].y-mvmap[bpos].y) >= 4){
							do_filter = 1;
						}
					}
                   if ((mode>=MODE_16x16 && mode<MODE_8x8) && tr16flag==0 && svt==1 && (vy&1)){
                        do_filter = 0;
                    }


                   if (mode==MODE_16x16 && tr16flag && (vy&1)){
						do_filter = 0;
					}


                   if (mode==MODE_INTRA && tr16flag && (vy&1)){
						do_filter = 0;
					}			

          if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-mbxmax].planar_flag && !(vy&1))
          {
            do_filter = 0;  // A fix for Nokia's planar mode filtering
          }
                  

				   //Fred bugfix
				   //if (mode==MODE_LMB32 && (vy%(LMBXSIZE/8))){
				   if (mode==MODE_LMB64 && (vy%(LMBXSIZE/8))){
					   do_filter = 0;
				   }
				   //Fred bugfix
				   //if (mode==(MODE_LMB64) && (vy%(LMBXSIZE/16))){
				   if (mode==(MODE_LMB32) && (vy%(LMBXSIZE/16))){
					   do_filter = 0;
				   }
                    
					if (do_filter){
						 tc = tctable_8x8[qp+alt_tc];
						move_ucsubmatrix(block,img->Y,i+k-4,i+k+3,j+l,j+l+7,0,0);	

						filter_8x8_horizontal_long(block,tc,betatable_8x8[qp]);

					}

				}
			}
        }
    }

    /* U */    
    for (i=0;i<img->crows;i+=8){
        for (j=0;j<img->ccols;j+=8){
            mby = i/8;
			mbx = j/8;
            vy = mby*2;
		    vx = mbx*2;
            mbpos = mby*mbxmax+mbx;
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
			/* Vertical edge */
			if (j>0){
                if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA){
                    tc = tctable_8x8[qp+4];
					move_ucsubmatrix(block,img->U,i,i+7,j-4,j+3,0,0);
					filter_chroma_vertical_8x8(block,tc);
				}
			}
			/* Horizontal edge */
			if (i>0){
               if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA){
                    tc = tctable_8x8[qp+4];
					move_ucsubmatrix(block,img->U,i-4,i+3,j,j+7,0,0);
					filter_chroma_horizontal_8x8(block,tc);
				}
			}
        }
    }

    /* V */    
    for (i=0;i<img->crows;i+=8){
        for (j=0;j<img->ccols;j+=8){
            mby = i/8;
			mbx = j/8;
            vy = mby*2;
			vx = mbx*2;
            mbpos = mby*mbxmax+mbx;
            qp = mbmap2[mbpos].qp;
            mode = mbmap2[mbpos].mode;
			/* Vertical edge */
			if (j>0){
				if (mode == MODE_INTRA || mbmap2[(vy/2)*mbxmax+((vx-1)/2)].mode == MODE_INTRA){
					tc = tctable_8x8[qp+4];
					move_ucsubmatrix(block,img->V,i,i+7,j-4,j+3,0,0);
					filter_chroma_vertical_8x8(block,tc);
				}
			}
			/* Horizontal edge */
			if (i>0){
				if (mbmap2[((vy-1)/2)*mbxmax+(vx/2)].mode == MODE_INTRA || mode == MODE_INTRA){
					tc = tctable_8x8[qp+4];
					move_ucsubmatrix(block,img->V,i-4,i+3,j,j+7,0,0);
					filter_chroma_horizontal_8x8(block,tc);
				}
			}
        }
    }

    free_ucsubmatrix(block,0,7,0,7);
    return 0;
}

#if NEW_PLANAR

/*-----------------------------------------------------------------------------------------
Function:   filter_luma_vertical_planar
Purpose:    In-place deblocking of the vertical boundary between two 8x8 luma blocks for planar mode
Input:      **mb       - 2D pointer to pixel data containing the vertical edge 
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_luma_vertical_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[ 4][k];
      a2 = mb[12][k];
      for(l=0;l<8;l++)
      {
        mb[4+l][k] = (a1*(8-l)+a2*l + 4) >> 3;
      }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_luma_horizontal_planar
Purpose:    In-place deblocking of the horizontal boundary between two 8x8 luma blocks for planar mode
Input:      **mb       - 2D pointer to pixel data containing the horizontal edge 
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_luma_horizontal_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[ k][4];
      a2 = mb[ k][12];
      for(l=0;l<8;l++)
      {
        mb[k][4+l] = (a1*(8-l)+a2*l + 4) >> 3;
      }
    }
    return 0;
}

#else

int filter_luma_vertical_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[ 4][k];
      a2 = mb[11][k];
      for(l=0;l<7;l++)
      {
        mb[4+l][k] = (a1*(7-l)+a2*l + 3)/7;
      }
    }
    return 0;
}

int filter_luma_horizontal_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<16;k++)
    {
      a1 = mb[ k][4];
      a2 = mb[ k][11];
      for(l=0;l<7;l++)
      {
        mb[k][4+l] = (a1*(7-l)+a2*l + 3)/7;
      }
    }
    return 0;
}
#endif

/*-----------------------------------------------------------------------------------------
Function:   filter_chroma_vertical_planar
Purpose:    In-place deblocking of the vertical boundary between two 8x8 chroma blocks for planar mode
Input:      **mb       - 2D pointer to pixel data containing the vertical edge 
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_chroma_vertical_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<8;k++)
    {
      a1 = mb[2][k];
      a2 = mb[5][k];
      for(l=0;l<4;l++)
      {
        mb[2+l][k] = (a1*(4-l)+a2*l + 2)/4;
      }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_chroma_horizontal_planar
Purpose:    In-place deblocking of the horizontal boundary between two 8x8 chroma blocks for planar mode
Input:      **mb       - 2D pointer to pixel data containing the horizontal edge 
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_chroma_horizontal_planar(unsigned char **mb)
{
    int a1,a2;
    int k,l;

    for(k=0;k<8;k++)
    {
      a1 = mb[ k][2];
      a2 = mb[ k][5];
      for(l=0;l<4;l++)
      {
        mb[k][l] = (a1*(4-l)+a2*l + 2)/4;
      }
    }
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   filter_planar
Purpose:    In-place planar mode deblocking for one frame (luma and chroma)
Parameters: None
-------------------------------------------------------------------------------------------*/

int filter_planar(decoder_info  *decoder_info,frame_info *frameInfo,image_frame *img)
{
  int i,j;
  unsigned char **block = ucsubmatrix(img->Y,0,15,0,15,0,0);
  unsigned char **block_chroma = ucsubmatrix(img->Y,0,7,0,7,0,0);
  int mby,mbx;
  //int do_filter=0;
  int mbxmax = img->cols>>4;
  int mbpos;
  

  mb_info *mbmap2 = frameInfo->mbmap;
  
  //block_info *mvmap = frameInfo->vmap[0];
  
  
  for (i=16;i<img->rows;i+=16){ // for each row of MBs
    for (j=16;j<img->cols;j+=16){ // for each col of MBs
      mby = i>>4;
      mbx = j>>4;
      mbpos = mby*mbxmax+mbx;
      
      if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-mbxmax].planar_flag)
      {
        //filter vertical
        move_ucsubmatrix(block,img->Y,i-8,i+7,j,j+15,0,0);
        filter_luma_vertical_planar(block);        

#if NEW_PLANAR
        move_ucsubmatrix(block_chroma,img->U,i/2-4,i/2+3,j/2,j/2+7,0,0);
        filter_chroma_vertical_planar(block_chroma);

        move_ucsubmatrix(block_chroma,img->V,i/2-4,i/2+3,j/2,j/2+7,0,0);
        filter_chroma_vertical_planar(block_chroma);
#else
        move_ucsubmatrix(block_chroma,img->U,i/2-4,i/2+3,j/2,j/2+7,0,0);
        filter_chroma_vertical_planar(block);

        move_ucsubmatrix(block_chroma,img->V,i/2-4,i/2+3,j/2,j/2+7,0,0);
        filter_chroma_vertical_planar(block);
#endif
        
      }
      if(mbmap2[mbpos].planar_flag && mbmap2[mbpos-1].planar_flag)
      {
        //filter horizontal
        move_ucsubmatrix(block,img->Y,i,i+15,j-7,j+8,0,0);
        filter_luma_horizontal_planar(block);        

#if NEW_PLANAR
        move_ucsubmatrix(block_chroma,img->U,i/2,i/2+7,j/2-3,j/2+4,0,0);
        filter_chroma_horizontal_planar(block_chroma);        

        move_ucsubmatrix(block_chroma,img->V,i/2,i/2+7,j/2-3,j/2+4,0,0);
        filter_chroma_horizontal_planar(block_chroma);        
#else
        move_ucsubmatrix(block,img->U,i/2,i/2+7,j/2-3,j/2+4,0,0);
        filter_chroma_horizontal_planar(block);        

        move_ucsubmatrix(block,img->V,i/2,i/2+7,j/2-3,j/2+4,0,0);
        filter_chroma_horizontal_planar(block);        
#endif
      }
    }
  }
  free_ucsubmatrix(block,0,15,0,15);
  free_ucsubmatrix(block_chroma,0,7,0,7);
  return 0;
}
