/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include "global.h"
#include "getbits.h"

/* to mask the n least significant bits of an integer */

static const unsigned int msk[33] =
{
  0x00000000,0x00000001,0x00000003,0x00000007,
  0x0000000f,0x0000001f,0x0000003f,0x0000007f,
  0x000000ff,0x000001ff,0x000003ff,0x000007ff,
  0x00000fff,0x00001fff,0x00003fff,0x00007fff,
  0x0000ffff,0x0001ffff,0x0003ffff,0x0007ffff,
  0x000fffff,0x001fffff,0x003fffff,0x007fffff,
  0x00ffffff,0x01ffffff,0x03ffffff,0x07ffffff,
  0x0fffffff,0x1fffffff,0x3fffffff,0x7fffffff,
  0xffffffff
};


#if FAST_GETBITS


/*-----------------------------------------------------------------------------------------
Function:   initbits_dec
Purpose:    Initializes bitsream reading, call once before first getbits or showbits
Input:      *infilr    - Input bistream file handle
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int initbits_dec(FILE *infile, stream *str)
{
    fpos_t fpos[1];
    long pos1,pos2;

    str->incnt = 0;
    str->rdptr = str->rdbfr + 2048;
    str->bitcnt = 0;
    str->infile = infile;
    
    fgetpos(str->infile,fpos);
    pos1 = ftell(str->infile);
    fseek(str->infile,0,SEEK_END);
    pos2 = ftell(str->infile);
    fsetpos(str->infile,fpos);
    str->length = pos2 - pos1;
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   fillbfr
Purpose:    Fills in the 32-bit intermediate buffer (str->inbfr) with bytes from bitstream file
Input:      None
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int fillbfr(stream *str)
{
    int l;

    while (str->incnt <= 24 && (str->rdptr < str->rdbfr + 2048))
    {
        str->inbfr = (str->inbfr << 8) | *str->rdptr++;
        str->incnt += 8;
    }

    if (str->rdptr >= str->rdbfr + 2048)
    {
        l = (int)fread(str->rdbfr,sizeof(unsigned char),2048,str->infile);
        str->rdptr = str->rdbfr;

        while (str->incnt <= 24 && (str->rdptr < str->rdbfr + 2048))
        {
            str->inbfr = (str->inbfr << 8) | *str->rdptr++;
            str->incnt += 8;
        }
    }

    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   getbits
Purpose:    Return next n bits (right adjusted) from the bitstream
Input:      n         - Number of bits to return
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int getbits(stream *str, int n)
{

    if (str->incnt < n)
    {
        fillbfr(str);
        if (str->incnt < n)
        {
            unsigned int l = str->inbfr;
            unsigned int k = *str->rdptr++;
            int shift = n-str->incnt;
            str->inbfr = (str->inbfr << 8) | k;
            str->incnt = str->incnt - n + 8;
            str->bitcnt += n;
            return (((l << shift) | (k >> (8-shift))) & msk[n]);
        }
    }

    str->incnt -= n;
    str->bitcnt += n;
    return ((str->inbfr >> str->incnt) & msk[n]);
}

/*-----------------------------------------------------------------------------------------
Function:   getbits1
Purpose:    Return next bit from the bitstream
Input:      None
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int getbits1(stream *str)
{
    if (str->incnt < 1)
    {
        fillbfr(str);
    }
    str->incnt--;
    str->bitcnt++;
    return ((str->inbfr >> str->incnt) & 1);
}

/*-----------------------------------------------------------------------------------------
Function:   showbits
Purpose:    Return next n bits from the bitstream (right adjusted) without advancing
Input:      n         - Number bits to return
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int showbits(stream *str, int n)
{
    if (str->incnt < n)
    {
        fillbfr(str);
        if (str->incnt < n)
        {
            int shift = n-str->incnt;
            return (((str->inbfr << shift) | (str->rdptr[0] >> (8-shift))) & msk[n]);
        }
    }

    return ((str->inbfr >> (str->incnt-n)) & msk[n]);
}

/*-----------------------------------------------------------------------------------------
Function:   flushbits
Purpose:    Advance by n bits
Input:      n       - Number of bits to advance
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int flushbits(stream *str, int n)
{
    str->incnt -= n;
    str->bitcnt += n;
    return 0;
}


#else


/*-----------------------------------------------------------------------------------------
Function:   initbits_dec
Purpose:    Initializes bitsream reading, call once before first getbits or showbits
Input:      *infilr    - Input bistream file handle
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int initbits_dec(FILE *infile, stream *str)
{
    fpos_t fpos[1];
    long pos1,pos2;

    str->incnt = 0;
    str->rdptr = str->rdbfr + 2048;
    str->bitcnt = 0;
    str->infile = infile;
    
    fgetpos(str->infile,fpos);
    pos1 = ftell(str->infile);
    fseek(str->infile,0,SEEK_END);
    pos2 = ftell(str->infile);
    fsetpos(str->infile,fpos);
    str->length = pos2 - pos1;
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   fillbfr
Purpose:    Fills in the intermediate buffer (str->inbfr) with bytes from bitstream file
Input:      None
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int fillbfr(stream *str)
{
    int l;

    str->inbfr[0] = str->inbfr[8];
    str->inbfr[1] = str->inbfr[9];
    str->inbfr[2] = str->inbfr[10];
    str->inbfr[3] = str->inbfr[11];
    if (str->rdptr >= str->rdbfr + 2048)
    {
        l = (int)fread(str->rdbfr,sizeof(unsigned char),2048,str->infile);
        str->rdptr = str->rdbfr;
        if (l < 2048)
        {
            if (l < 0)
                l = 0;

            while (l < 2048)   /* Add recognizable sequence end code */
            {
                str->rdbfr[l++] = 0;
                str->rdbfr[l++] = 0;
                str->rdbfr[l++] = (1 << 7) | (31 << 2);
            }
        }
    }
    for (l=0; l<8; l++)
    {
        str->inbfr[l+4] = str->rdptr[l];
    }
    str->rdptr+= 8;
    str->incnt+= 64;
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   showbits
Purpose:    Return next n bits from the bitstream (right adjusted) without advancing
Input:      n         - Number bits to return
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int showbits(stream *str, int n)
{
  unsigned char *v;
  unsigned int b,tmp;
  int c;

  if (str->incnt < n)
      fillbfr(str);

  v = str->inbfr + ((96 - str->incnt) >> 3);
  b = (v[0] << 24) | (v[1] << 16) | (v[2] << 8) | v[3];
  c = ((str->incnt - 1) & 7) + 25;
  if (n <= c)
      tmp = (b >> (c-n)) & msk[n];
  else
  {
      tmp = b & msk[c];
      tmp <<= (n-c);
      tmp |= v[4]>>(8-(n-c));
  }
  return tmp;
}

/*-----------------------------------------------------------------------------------------
Function:   getbits1
Purpose:    Return next bit from the bitstream
Input:      None
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int getbits1(stream *str)
{
  unsigned int current_byte;
  if (str->incnt < 1)
      fillbfr(str);
  current_byte = str->inbfr[(96 - str->incnt) >> 3];
  str->bitcnt++;
  str->incnt--;
  return ((current_byte >> (str->incnt & 7)) & 1);
}

/*-----------------------------------------------------------------------------------------
Function:   flushbits
Purpose:    Advance by n bits
Input:      n       - Number of bits to advance
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int flushbits(stream *str, int n)
{

    str->bitcnt+= n;
    str->incnt-= n;
    if (str->incnt < 0)
        fillbfr(str);
    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   getbits
Purpose:    Return next n bits (right adjusted) from the bitstream
Input:      n         - Number of bits to return
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

unsigned int getbits(stream *str, int n)
{
  unsigned int l;

  l = showbits(str, n);
  flushbits(str, n);

  return l;
}


#endif  // FAST_GETBITS


/*-----------------------------------------------------------------------------------------
Function:   getbits
Purpose:    Return number of remaining bytes in bitstream
Input:      None
Output:     None
Parameters: None
-------------------------------------------------------------------------------------------*/

int getbyterest(stream *str)
{
    return (str->length - (str->bitcnt+7)/8);        /* ceiling to
                                                      include byte if partially read */
}
