/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/



#if !defined(_TNTMENC_H_)
#define _TNTMENC_H_

#include <stdio.h>
#include "putbits.h"
#include "global.h"
#include "types.h"

enum {I_PICTURE,P_PICTURE,B_PICTURE};

typedef enum {MODE_SKIP,MODE_INTRA,MODE_16x16,MODE_16x8,MODE_8x16,MODE_8x8,MODE_DIRECT,MODE_LMB32,MODE_LMB64} mb_mode;


typedef enum {TR_4x4=0,TR_8x8=1,TR_16x16=2,TR_32x32=3,TR_64x64=4} trsizes;

#define LAMBDA_QP_SCALE_P 0.01
extern const int corr_qpc_8x8[42];


#define MAX_B_LEVEL 3
#define MAX_GOP_LENGTH (1<<MAX_B_LEVEL)


#define NUM_EXT_MODES 25

typedef struct
{
    int cols;
    int rows;
    int qp;
    int orig_rows;
    int orig_cols;
    int dqpI;
    int dqpB[MAX_B_LEVEL];
    int numBlev;
    int me_type;
    int me_range;
    int coarse_range_y;
    int coarse_range_x;
    int coarse_search_use_dynamic;
    int coarse_search_grid;
    float lambda_coef;
    int me_rangeb;
    int constrainedIntraFlag;
    int irPeriod;
	int frame_num;
	int first_frames_skip;
    float lambda_coeff_SAD;
    float lambda_coeff_modeI;
    float lambda_coeff_modeP;
    float lambda_coeff_modeB;
    float lambda_coeff_quantC;
    float lambda_coeff_quantIntraI;
    float lambda_coeff_quantIntraPB;
    float lambda_coeff_quantInterP;
    float lambda_coeff_quantInterB;
    int SADcost;
    int fast_svt;
    int svt;
    int tr4flag;
    int tr16flag;
    int dqpP;
    int pPeriod;
    int fast_tr4_thr1;
    int fast_svt_thr2;
    int early_skip_thr;
	int bi;
    int qp2;
    int qp2start;
    char *infilestr;
    char *outfilestr;
    char *reconfilestr;
    char *statfilestr;
    int num_frames;
    int num_b_frames;
    int multi_ref;
    int intra_period;
    float ifreq;
    float ofreq;
    int trace;
    int num_pixels;    
    int resetContextsFlag;    
} enc_params;

typedef struct
{
    int x;
    int y;
	int interp_indx;
} mvpos;

typedef struct
{
    int top;    
    int run;
    int ref;
    int mode;
    int qp;
    int vec;
    int skip_cand;
    int cbp_luma;
    int cbp_chroma;
    int coeff_y;
    int coeff_u;
    int coeff_v;
    int coeff_uv;
    int ipred;
    int modes[7];
    int side_intra_y;
    int side_intra_u;
    int side_intra_v;
    int lmb;
    int svt_idx;
    int planar;

} bitcounter;


typedef struct
{
  int uvcbpTableE[4];
  int uvcbpTableD[4];
  int modeTableE[3][NUM_EXT_MODES];
  int modeTableD[3][NUM_EXT_MODES];
  int cbpTable6[6];
  int cbpTable63E[3][63];
  int cbpTable63D[3][63];
  int LPTableE[10][128];
  int LPTableD[10][128];

  // ADAPT_VLC_NUM
  int mode_vlc_index;
  int last_pos_vlc_index[10];
  int cbp_vlc_index[3];

  int LPTableE4[3][32];
  int LPTableD4[3][32];

} vlc_contexts;




typedef struct
{
    int x;
    int y;
    int coded_y;
	int interp_indx;
} block_info;



typedef struct
{          
    int qp;
    int mode;                      
    int ref;
    int num_skip_cand;
    int best_skip_mv_cand;            
    int svt;    
	int tr16flag;




    int codedScanToMBX;
    int codedScanToMBY;

    int planar_flag;

} mb_info;

typedef struct
{
    int h, w, y, x;
} part_dimensions;



typedef struct
{
  mvpos *maxvec;
  bitcounter *bitsI;
  bitcounter *bitsP;
  bitcounter *bitsB;
} statistics;

typedef struct
{
    int         width;
    int         height;
    int         extWidth;
    int         extHeight;
    int         widthInMBs;
    int         heightInMBs;    
    image_frame *orig;                //original frame
    image_frame **ref_frame_list;     //unpadded integer pixel reference frame
    image_frame **ref_frame_listUp;   //Padded and upscaled reference frames
    image_frame *recon;               //reconstructed frame
    
    image_frame *ref_frame_qp[2];     //padded quarter pixel reference frames
    stream      *str;                 //bitstream structure
    statistics  *encStat;             //encoder statistics for printout
    enc_params  *params;              //configuration parameters
    vlc_contexts *currContext;        //vlc contexts for adaptation
#if BUGFIX_ME
    mvpos       cmv[2];   
#endif
} encoder_info;

typedef struct
{
  int         pic_type;
  int         qp;
  int         frame_num;
  int         chroma_rounding;
  int         num_skipped;
  int         num_ref;
  int         MEfactor;
  mb_info     *mbmap;                  // MB level map for inter MB prediction, multipass, and deblocking
  block_info  *mvmap;                  // 8x8 block level map for inter MB prediction, multipass, and deblocking
  int         svt_set;

  double      lambdaMode;  
  int         modeCostNorm;


  int		  final_frame;

} frame_info;

typedef struct
{
    short ycoeff[16][8][8];              // 16 8x8 quantized transform coefficients matrices - Y
    short ucoeff[8][8];                  // 8x8 quantized transform coefficients matrices - U
    short vcoeff[8][8];                  // 8x8 quantized transform coefficients matrices - V
    int ycoeff_svt[4][16]; 

    unsigned char orgY[256];
    unsigned char orgU[64];
    unsigned char orgV[64];
    unsigned char recY[256];
    unsigned char recU[64];
    unsigned char recV[64];

    unsigned char orgLMBY[64*64];
    unsigned char orgLMBU[(64*64)/4];
    unsigned char orgLMBV[(64*64)/4];
    unsigned char recLMBY[64*64];
    unsigned char recLMBU[(64*64)/4];
    unsigned char recLMBV[(64*64)/4];

    int mby;
    int mbx;
    int last_pos;
    int num_bits;
    int sum_of_levels;
    int ycbp;
    int ucbp;
    int vcbp;
    int final_encode_flag;             // =0 during RDO mode selection, =1 for final encode
    mb_mode mode;
    int ref;
    int side_intra_y;                  // intra DC if left and above MB not available - Y
    int side_intra_u;                  // intra DC if left and above MB not available - U
    int side_intra_v;                  // intra DC if left and above MB not available - V
    unsigned int modelist;             // intra prediction modes
    int skip_mode_bit_adjust;
    int svt_idx;
    int svt_idx_num;
    int effective_svt_idx;
    int svt_rd_diff;
    int svt;  
    mvpos skipmv;                      // skip MV
    mvpos dirmv;                       // direct mode MV
    mvpos pred[2][4][4];               // MV prediction 2 ref frames x 4 partitions x 4 8x8 blocks
    int inter_sad;
    int jmr;                           // encode on luma MB using only one 8x8 blocks of coefficients
    int num_skip_cand;
    int best_skip_mv_cand;
    int bestModeChromaI;

	int tr16flag;


    int tr4flag;
    int modelist4[16];
    int ycbp4[4];
    int tr_size;

    mvpos mvblock[4][4];               // chosen MV's for this MV, 4 partitions, 4 8x8 blocks

	mvpos mvblockBW[4][4];             //one more MV used in Bi prediction
	mvpos mvsearch_result[2];          //record forward and backward mv for LMB, motion search is done in processLMB and no MVarray to record like normal MB
	int ref_pred;

    int forceIntra;
    mvpos candidateMV[2][4];           // 4 candidate motion vectors for each reference frame - start of diamond from one of these
    int specialBmode;                  // indicates mode=16x61,CBP=0, DMV=0
    int pos;                           // MB position in coded order

    int LMBposFlag;    


    int modelist8[4];


    int intraPredAngle8[4];
    int intraPredDirectionChroma;
    int intraPredAngleChroma;


    int early_skip_flag;





    int planar_flag;
    int check_planar;
    int planar_pred[3];
    int planar_deltaY;
    int planar_deltaU;
    int planar_deltaV;
} macroblock_info;



#endif
