/***************************************************************/
/* TENTM software accompanying JCTVC-A119                      */
/* Copyright Tandberg, Ericsson, Nokia, 2010                   */
/***************************************************************/


#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "tntmenc.h"

//STATIC #include "rfutil.h"
#include "transform.h"
#include "putvlc.h"
#include "deblocking.h"
#include "defines.h"
#include "cost.h"
#include "motion_estimate.h"
#include "motion_comp.h"
#include "intra_pred.h"
#if DBG_COMPLEXITY
#include "rdtsc.h"
#endif

static const short Q_8x8[6] = { 392,336,294,261,235,214 };
//static const short IQ_8x8[6] = { 6,7,8,9,10,11 };

// ADAPT_VLC_NUM
extern const int mode_vlc_num[2][7];
extern const int cbp_vlc_num[3][8];

static const part_dimensions part_dims[][4] = {
    {{16,16, 0, 0}},
    {{ 8,16, 0, 0}, { 8,16, 8, 0}},
    {{16, 8, 0, 0}, {16, 8, 0, 8}},
    {{ 8, 8, 0, 0}, { 8, 8, 0, 8}, { 8, 8, 8, 0}, { 8, 8, 8, 8}}
};



extern int g_bround;
extern float rd_lambda_acc(int qp);
extern int direction_counter[PRED_COUNT];
extern int trace;
extern FILE *tracefile;
extern int transmitInterpIndx[16];
extern int g_enableSwitched;


#define SVT_TABLE_SIZE 58
int svt_pos[SVT_TABLE_SIZE][2] = {
    {0,0},{1,0},{2,0},{3,0},{4,0},{5,0},{6,0},{7,0},{8,0},
    {0,8},{1,8},{2,8},{3,8},{4,8},{5,8},{6,8},{7,8},{8,8},
    {0,1},{0,2},{0,3},{0,4},{0,5},{0,6},{0,7},
    {8,1},{8,2},{8,3},{8,4},{8,5},{8,6},{8,7},
    {0,0},{0,1},{0,2},{0,3},{0,4},{0,5},{0,6},{0,7},{0,8},{0,9},{0,10},{0,11},{0,12},
    {1,0},{1,1},{1,2},{1,3},{1,4},{1,5},{1,6},{1,7},{1,8},{1,9},{1,10},{1,11},{1,12}
};


int best_ycoeff[3][4][8][8]; //[REF][MB_MODE]
int best_ycoeff_svt[3][4][4][16];
unsigned char best_rblock16[3][4][16][16];
int best_idx_svt[3][4];
int best_svt_idx_num[3][4];
int best_effective_svt_idx[3][4];
int sad_idx_svt[3][4][SVT_TABLE_SIZE];
extern int rdcost_stat[2][3][6]; //[svt][ref][mode:MODE_SKIP,MODE_INTRA,MODE_16x16,MODE_16x8,MODE_8x16,MODE_8x8]


int rdcost1[9][9];
int rdcost2[2][13];
static const int svt_tctable_8x8[42] =
{
    0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,5,5,6,6,7,8,9,9,10,10,11,11,12,12
};

static const int svt_betatable_8x8[42] =
{
    0, 0, 0, 0, 0, 0, 6, 7, 8, 9,10,11,12,13,14,15,16,17,
    18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60,62,64
};

/*-----------------------------------------------------------------------------------------
Function:   up_available
Purpose:    Determine wether the 4x4 or 8x8 block above is available for intra prediction or not
Input:      mby        - Vertical MB coordinate (in units of 16 pixels)
            mbx        - Horizontal MB coordinate (in units of 16 pixels)
            block_y    - Vertical position within an MB (in units of pixels)
            block_x    - Horizontal position within an MB (in units of pixels)           
Return:     1 or 0     - Available or not
Parameters: None 
-------------------------------------------------------------------------------------------*/

int up_available(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mby, int mbx, int block_y, int block_x, enc_params *params)
{
	int mbxmax = encoderInfo->widthInMBs;
	if(mby == 0)
	{
		//First MB row
		if(block_y == 0)
			return 0;
		else
			return 1;
	}
	else
	{
		if(params->constrainedIntraFlag == 0)
			return 1;
		else
		{
			if((frameInfo->mbmap[(mby-1)*mbxmax + mbx].mode != MODE_INTRA) && block_y == 0)
				return 0;
			else
				return 1;
		}

	}

}

/*-----------------------------------------------------------------------------------------
Function:   upright_available
Purpose:    Determine wether the 4x4 or 8x8 block above and to the right is available for intra prediction or not
Input:      mby        - Vertical MB coordinate (in units of 16 pixels)
            mbx        - Horizontal MB coordinate (in units of 16 pixels)
            block_y    - Vertical position within an MB (in units of pixels)
            block_x    - Horizontal position within an MB (in units of pixels)           
Return:     1 or 0     - Available or not
Parameters: None 
-------------------------------------------------------------------------------------------*/

int upright_available(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mby, int mbx, int block_y, int block_x, enc_params *params)
{
	int mbxmax = encoderInfo->widthInMBs;

	if (mby==0 && block_y==0) 
		return 0;
	else if (mbx==(mbxmax-1) && block_x==8)
		return 0;
	else if ((mbx%(LMBXSIZE/32)) == ((LMBXSIZE/32)-1) && block_x==8)
		return 0;
	else if (block_x==8 && block_y==8)
		return 0;
    else if(params->constrainedIntraFlag == 0)
        return 1; 
    else
    {
        if((frameInfo->mbmap[(mby-1)*mbxmax + mbx].mode != MODE_INTRA) && block_y == 0 && block_x == 0)
            return 0;

        if((frameInfo->mbmap[(mby-1)*mbxmax + mbx + 1].mode != MODE_INTRA) && block_y == 0 && block_x == 8)
            return 0;

        if((frameInfo->mbmap[mby*mbxmax + mbx + 1].mode != MODE_INTRA) && block_y == 8 && block_x == 8)
            return 0;

        return 1;
    }
}

/*-----------------------------------------------------------------------------------------
Function:   left_available
Purpose:    Determine wether the 4x4 or 8x8 block to the left is available for intra prediction or not
Input:      mby        - Vertical MB coordinate (in units of 16 pixels)
            mbx        - Horizontal MB coordinate (in units of 16 pixels)
            block_y    - Vertical position within an MB (in units of pixels)
            block_x    - Horizontal position within an MB (in units of pixels)           
Return:     1 or 0     - Available or not
Parameters: None 
-------------------------------------------------------------------------------------------*/

int left_available(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mby, int mbx, int block_y, int block_x, enc_params *params)
{
	int mbxmax = encoderInfo->widthInMBs;
	

	if(mbx == 0)
	{
		//First MB col
		if(block_x == 0)
			return 0;
		else
			return 1;
	}
	else 
	{
		if(params->constrainedIntraFlag == 0)
			return 1;
		else
		{
			if((frameInfo->mbmap[mby*mbxmax + mbx - 1].mode != MODE_INTRA) && block_x == 0)
				return 0;
			else
				return 1;
		}
	}

}

/*-----------------------------------------------------------------------------------------
Function:   rec_compensate_mb
Purpose:    Reconstruction of a skip 16x16 MB (both P and B frames)
            Calculate early skip flag
Parameters: None 
-------------------------------------------------------------------------------------------*/

int rec_compensate_mb(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{


	int yoff,xoff,rows,cols,ypos2,xpos2;
	int cxpos,cypos,m,n;
	mvpos lp,cp;
	int xp,yp;
	int i,j;
	int ref = (frameInfo->pic_type == P_PICTURE ? macroblockInfo->ref : 0);

	image_frame *qpimage = encoderInfo->ref_frame_qp[ref];
	image_frame *qpimageBW = encoderInfo->ref_frame_qp[1];            
	int mbx = macroblockInfo->mbx;
	int mby = macroblockInfo->mby;    
	int ypos = mby*16;
	int xpos = mbx*16;
	int pic_type = frameInfo->pic_type;
	mvpos skipmv = macroblockInfo->skipmv;
	int rounding = encoderInfo->params->frame_num&1;

	cypos = ypos/2;
	cxpos = xpos/2;
	if (pic_type == P_PICTURE)
	{
		lp.x = skipmv.x;
		lp.y = skipmv.y;
		cp.x = (skipmv.x+rounding)>>1;
		cp.y = (skipmv.y+rounding)>>1;
	}
	else
	{

		lp.x = ((skipmv.x+2)>>2)<<2;
		lp.y = ((skipmv.y+2)>>2)<<2;
		cp.x = ((skipmv.x+4)>>3)<<2;
		cp.y = ((skipmv.y+4)>>3)<<2;

	}

	/* Fill in skipped MB */

	yoff = lp.y&3;
	xoff = lp.x&3;
	rows = qpimage->rows/4;
	cols = qpimage->cols/4;         
	ypos2 = ypos + (lp.y>>2) + 16; 
	xpos2 = xpos + (lp.x>>2) + 16; 
	for (i = 0; i < 16; i+=8){
		for (j = 0; j < 16; j+=8){                                   
			yp = min(max(8-2,ypos2+i),rows-16+1)+yoff*rows;
			xp = min(max(8-2,xpos2+j),cols-16+1)+xoff*cols;   

			if(skipmv.interp_indx == 0)
            {
                for (m=0; m<8; m++)               
                    memcpy(&macroblockInfo->recY[16*(i+m)+j],&qpimage->Y[yp+m][xp],8*sizeof(unsigned char));
            }
			else
            {
                for (m=0; m<8; m++)               
                    memcpy(&macroblockInfo->recY[16*(i+m)+j],&qpimage->Y2[yp+m][xp],8*sizeof(unsigned char));

			}
		}
	}

	yoff = cp.y&3;
	xoff = cp.x&3;
	rows = qpimage->crows/4;
	cols = qpimage->ccols/4;         
	ypos2 = cypos + (cp.y>>2) + 8; 
	xpos2 = cxpos + (cp.x>>2) + 8; 
	for (i = 0; i < 8; i+=4){
		for (j = 0; j < 8; j+=4){                                   
			yp = min(max(0,ypos2+i),rows-8)+yoff*rows;
			xp = min(max(0,xpos2+j),cols-8)+xoff*cols;   
			for (m=0; m<4; m++){
				memcpy(&macroblockInfo->recU[8*(i+m)+j],&qpimage->U[yp+m][xp],4*sizeof(unsigned char));
				memcpy(&macroblockInfo->recV[8*(i+m)+j],&qpimage->V[yp+m][xp],4*sizeof(unsigned char));
			}
		}
	}

	if (pic_type == B_PICTURE){
		yoff = (-lp.y)&3;
		xoff = (-lp.x)&3;
		rows = qpimage->rows/4;
		cols = qpimage->cols/4;         
		ypos2 = ypos + ((-lp.y)>>2) + 16; 
		xpos2 = xpos + ((-lp.x)>>2) + 16; 
		for (i = 0; i < 16; i+=8){
			for (j = 0; j < 16; j+=8){                                   
				yp = min(max(8-2,ypos2+i),rows-16+1)+yoff*rows;
				xp = min(max(8-2,xpos2+j),cols-16+1)+xoff*cols;   


				if(skipmv.interp_indx == 0)
                {
                    for (m=0; m<8; m++)
                        for (n=0; n<8; n++)                                                  
                            macroblockInfo->recY[16*(i+m)+j+n] = (macroblockInfo->recY[16*(i+m)+j+n] + qpimageBW->Y[yp+m][xp+n] + g_bround)>>1;
					}
                else
                {
                    for (m=0; m<8; m++)
                        for (n=0; n<8; n++)                                                  
                            macroblockInfo->recY[16*(i+m)+j+n] = (macroblockInfo->recY[16*(i+m)+j+n] + qpimageBW->Y2[yp+m][xp+n] + g_bround)>>1;
				}


			}
		}            

		yoff = (-cp.y)&3;
		xoff = (-cp.x)&3;
		rows = qpimage->crows/4;
		cols = qpimage->ccols/4;         
		ypos2 = cypos + ((-cp.y)>>2) + 8; 
		xpos2 = cxpos + ((-cp.x)>>2) + 8;             
		for (i = 0; i < 8; i+=4){
			for (j = 0; j < 8; j+=4){                                   
				yp = min(max(0,ypos2+i),rows-8)+yoff*rows;
				xp = min(max(0,xpos2+j),cols-8)+xoff*cols;   
				for (m=0; m<4; m++){
					for (n=0; n<4; n++){                      
						macroblockInfo->recU[8*(i+m)+j+n] = (macroblockInfo->recU[8*(i+m)+j+n] + qpimageBW->U[yp+m][xp+n] + g_bround)>>1;
						macroblockInfo->recV[8*(i+m)+j+n] = (macroblockInfo->recV[8*(i+m)+j+n] + qpimageBW->V[yp+m][xp+n] + g_bround)>>1;
					}
				}
			}
		}
	}


	if (encoderInfo->params->early_skip_thr){
		int val=0;
		int qp0 = frameInfo->qp/6;
		int qp1 = frameInfo->qp%6;    
		int qpc = frameInfo->qp-corr_qpc_8x8[frameInfo->qp];
		int qpc0 = qpc/6;
		int qpc1 = qpc%6;
		short Ty = (encoderInfo->params->early_skip_thr*(1<<qp0))/Q_8x8[qp1];
		short Tc = (encoderInfo->params->early_skip_thr*(1<<qpc0))/Q_8x8[qpc1];

		val = early_skip_check_8x8(Ty,&macroblockInfo->orgY[16*0+0],&macroblockInfo->recY[16*0+0],16);
		if (!val) val = early_skip_check_8x8(Ty,&macroblockInfo->orgY[16*0+8],&macroblockInfo->recY[16*0+8],16);
		if (!val) val = early_skip_check_8x8(Ty,&macroblockInfo->orgY[16*8+0],&macroblockInfo->recY[16*8+0],16);
		if (!val) val = early_skip_check_8x8(Ty,&macroblockInfo->orgY[16*8+8],&macroblockInfo->recY[16*8+8],16);
		if (!val) val = early_skip_check_8x8(Tc,macroblockInfo->orgU,macroblockInfo->recU,8);
		if (!val) val = early_skip_check_8x8(Tc,macroblockInfo->orgV,macroblockInfo->recV,8);        
		macroblockInfo->early_skip_flag = !val;
	}
	else
		macroblockInfo->early_skip_flag = 0;

	return 1;
}

/*-----------------------------------------------------------------------------------------
Function:   check_svt_skip
Purpose:    Check whetehr SVT mode should be skipped
Parameters: None 
-------------------------------------------------------------------------------------------*/

int check_svt_skip(int dir,int offset,int edge1,int edge2,int edge3,int edge4,int svt_type,int beginy,int beginx)
{
	int vy_delta1,vx_delta1,vy_delta2,vx_delta2;
	if (svt_type == 0)
	{
		vy_delta1 = beginy/8;
		vx_delta1 = beginx/8;
		vy_delta2 = (beginy+7)/8;
		vx_delta2 = (beginx+7)/8;

		if (vx_delta1==vx_delta2 && vy_delta1!=vy_delta2)
		{
			if (vx_delta1==0 && edge2==1)
				return 1;
			else if (vx_delta1==1 && edge3==1)
				return 1;
		}
		else if (vx_delta1!=vx_delta2 && vy_delta1==vy_delta2)
		{
			if (vy_delta1==0 && edge1==1)
				return 1;
			else if (vy_delta1==1 && edge4==1)
				return 1;
		}
	}
	else
	{

		if (dir==0)
		{
			if (offset>=0 && offset<=4)
			{
				if (edge1)
					return 1;
			}
			else if (offset>=5 && offset<=7)
			{
				if (edge1 || edge2 || edge3 || edge4)
					return 1;
			}
			else
			{
				if (edge4)
					return 1;
			}
		}
		else
		{
			if (offset>=0 && offset<=4)
			{
				if (edge2)
					return 1;
			}
			else if (offset>=5 && offset<=7)
			{
				if (edge1 || edge2 || edge3 || edge4)
					return 1;
			}
			else
			{
				if (edge3)
					return 1;
			}
		}
	}
	return 0;

}

/*-----------------------------------------------------------------------------------------
Function:   sad_svt
Purpose:    Calculate SAD of an SVT block
Parameters: None 
-------------------------------------------------------------------------------------------*/

int sad_svt(int dir,int offset,int beginy,int beginx,int svt_type, short dblock16[16][16])
{
	int sad,y,x;
	sad = 0;
	if (svt_type == 0)
	{
		for (y=0; y<8; y++)
			for (x=0; x<8; x++){
				sad  += abs(dblock16[y+beginy][x+beginx]);
			}


	}
	else
	{
		if (dir==0)
		{
			for (y=0; y<4; y++)
				for (x=0; x<16; x++)
				{
					sad += abs(dblock16[y+offset][x]);
				}
		}
		else //dir==1
		{
			for (x=0; x<4; x++)
				for (y=0; y<16; y++)
				{
					sad += abs(dblock16[y][x+offset]);
				}
		}
	}
	return sad;
}

/*-----------------------------------------------------------------------------------------
Function:   write_run
Purpose:    For every 16x16 MB in a skip MB run, and if more than one skip MV candidate, 
            write one bit signalling which skip MV candidate to use
Parameters: None 
-------------------------------------------------------------------------------------------*/

void write_run(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int mbrun,stream *str)
{  
    bitcounter *bits;
    int pic_type = frameInfo->pic_type;
    int last_pos = macroblockInfo->last_pos;
    int mbxmax = encoderInfo->widthInMBs;
    int maxpos = encoderInfo->widthInMBs * encoderInfo->heightInMBs - 1;
    int i;

    if (pic_type==I_PICTURE)
        bits = encoderInfo->encStat->bitsI;
    else if(pic_type==P_PICTURE)
        bits = encoderInfo->encStat->bitsP;
    else
        bits = encoderInfo->encStat->bitsB;

    for (i = last_pos+1; i <= (last_pos + mbrun); i++){

        int my = frameInfo->mbmap[i].codedScanToMBY;
        int mx = frameInfo->mbmap[i].codedScanToMBX;

        if (frameInfo->mbmap[my*mbxmax+mx].num_skip_cand > 1){
            putbits(1,frameInfo->mbmap[my*mbxmax+mx].best_skip_mv_cand,str);      
            bits->skip_cand += 1;
            if ((trace & 2) && str->outfile){
                fprintf(tracefile, "skip_pred_method = %d\n", frameInfo->mbmap[my*mbxmax+mx].best_skip_mv_cand);
            }
            if (i < maxpos)
                macroblockInfo->skip_mode_bit_adjust -= 1;
        }
    }
}

/*-----------------------------------------------------------------------------------------
Function:   write_mode
Purpose:    Encode and write the following parameters to the bitstream:
            - run of skip MBs (16x16)
            - MB mode
            - CBP
            - transform size (16x16, 8x8, or 4x4)
            - SVT flag
Parameters: None 
-------------------------------------------------------------------------------------------*/

void write_mode(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,bitcounter *bits,stream *str)

{
    int extendedMode,cbp,cn,mbrun,mode,len,tmp,x,y,cx,cy;                                      
    int *modeTablePtrE;
    int *modeTablePtrD;

    int pic_type = frameInfo->pic_type;    
    int last_pos = macroblockInfo->last_pos;  

    int n;

    if (pic_type != I_PICTURE){
        mbrun = macroblockInfo->pos-last_pos-1;
        cn = mbrun;                
        bits->run += put_uvlc(cn,str);

        if (mbrun > 0){    
            write_run(encoderInfo,frameInfo,macroblockInfo,mbrun,str);
        }

        if (macroblockInfo->LMBposFlag&1)

        {
            putbits(1,0,str); // LMB bit = 0 for small MBs
            bits->lmb++;
        }
    }

    mode = macroblockInfo->mode;

    cbp = (macroblockInfo->vcbp<<5) + (macroblockInfo->ucbp<<4) + (macroblockInfo->ycbp);

    if (mode == MODE_INTRA){

        if (macroblockInfo->tr16flag){        
            extendedMode = 0;
            if (macroblockInfo->ycbp) extendedMode=1;
            if (macroblockInfo->vcbp || macroblockInfo->ucbp) extendedMode=2;
        }
        else{
            extendedMode = cbp ? 4 : 3;
        }
    }
    else if (mode == MODE_16x16){


        if (macroblockInfo->tr16flag){


            extendedMode = 5;

            if (macroblockInfo->ycbp) extendedMode = 6;
            if (macroblockInfo->vcbp || macroblockInfo->ucbp) extendedMode = 7;

        }
        else

            extendedMode = cbp ? 8 : 5;		
    }
    else if (mode < MODE_DIRECT){
        extendedMode = cbp ? 2*mode+4 : 2*mode+3;
    }
    else{
        extendedMode = 15;
    }

    /* For B pictures the flag variable is used to indicate MV=MVpred */
    if (extendedMode==5 && pic_type == B_PICTURE && macroblockInfo->specialBmode) extendedMode = 16;

    /* Encode extended mode and adapt tables if final encoding */ 	
    modeTablePtrE = encoderInfo->currContext->modeTableE[pic_type];
    modeTablePtrD = encoderInfo->currContext->modeTableD[pic_type];

    x = extendedMode;

    /* For P pictures the flag variable is used to indicate ref0 or ref1 */
    if (pic_type == P_PICTURE && macroblockInfo->ref && mode != MODE_INTRA) x += 10; 
    cx = modeTablePtrE[x];

    // ADAPT_VLC_NUM
    len = put_vlc(mode_vlc_num[(pic_type == I_PICTURE ? 0 : 1)][min(encoderInfo->currContext->mode_vlc_index,6)],cx,str);
    // len = (pic_type == P_PICTURE) ? put_vlc(1,cx,str) : put_vlc(0,cx,str);

    if (mode == MODE_INTRA && pic_type == I_PICTURE)
    {
        if (extendedMode == 0)         
        {
            putbits(1,macroblockInfo->planar_flag,str);
            len++;

            if(macroblockInfo->planar_flag)
            {
                bits->mode += len;

                if (macroblockInfo->final_encode_flag){ 
                    // ADAPT_VLC_NUM
                    encoderInfo->currContext->mode_vlc_index += cx == encoderInfo->currContext->mode_vlc_index ? 0 : (cx < encoderInfo->currContext->mode_vlc_index ? -1 : 1);

                    cy = max(0,cx-1);
                    y = modeTablePtrD[cy];
                    modeTablePtrD[cy] = x;
                    modeTablePtrD[cx] = y;
                    modeTablePtrE[x] = cy;
                    modeTablePtrE[y] = cx;
                }
                return;
            }
        }
    }


    bits->mode += len;

    if (macroblockInfo->final_encode_flag){ 
        // ADAPT_VLC_NUM
        encoderInfo->currContext->mode_vlc_index += cx == encoderInfo->currContext->mode_vlc_index ? 0 : (cx < encoderInfo->currContext->mode_vlc_index ? -1 : 1);

        cy = max(0,cx-1);
        y = modeTablePtrD[cy];
        modeTablePtrD[cy] = x;
        modeTablePtrD[cx] = y;
        modeTablePtrE[x] = cy;
        modeTablePtrE[y] = cx;
    }

    if (cbp==0 || extendedMode==1 || extendedMode==6)
    {
        if ((mode == MODE_INTRA && macroblockInfo->tr16flag==0 && (encoderInfo->params->tr4flag)&1) || 
            (mode >= MODE_16x16 && mode <= MODE_8x8 && macroblockInfo->tr16flag==0 && (encoderInfo->params->tr4flag)&2 && cbp)){
                putbits(1,macroblockInfo->tr4flag,str);
                bits->mode += 1;
        }
        return;
    }


    /* Encode CBP if present */
    if (extendedMode == 2){

        tmp = macroblockInfo->ycbp ? 1 : 0;

        putbits(1,tmp,str);
        len = 1;
    }
    else if(extendedMode == 7){
        cbp = cbp>>3;
        len = (cbp<4) ? 2 : 3;
        tmp = (cbp<6) ? 2 + (cbp&1) : cbp&1;
        putbits(len,tmp,str);
    }
    else{
        if (macroblockInfo->mode!=MODE_INTRA && macroblockInfo->mode!=MODE_8x8 && encoderInfo->params->svt==1) //cixunzhang
        {
            putbits(1,macroblockInfo->svt,str);
            bits->svt_idx++;
        }
        else
            assert(macroblockInfo->svt==0);

        if (macroblockInfo->svt)
        {
            x = macroblockInfo->ucbp*2+macroblockInfo->vcbp;

            cx = encoderInfo->currContext->uvcbpTableE[x];

            if (cx==0)
            {
                len = 1;
                putbits(1,0,str);
            }
            else if (cx==1)
            {
                len = 2;
                putbits(2,2,str);
            }
            else if (cx==2)
            {
                len = 3;
                putbits(3,6,str);
            }
            else if (cx==3)
            {
                len = 3;
                putbits(3,7,str);
            }

            if (macroblockInfo->final_encode_flag){        

                cy = max(0,cx-1);
                y = encoderInfo->currContext->uvcbpTableD[cy];
                encoderInfo->currContext->uvcbpTableD[cy] = x;
                encoderInfo->currContext->uvcbpTableD[cx] = y;
                encoderInfo->currContext->uvcbpTableE[x] = cy;
                encoderInfo->currContext->uvcbpTableE[y] = cx;
            }
        }
        else
        {
            /* Encode cbp-1 and adapt tables if final encoding */ 
            int vlcNum;
            int *cbpTablePtrE;
            int *cbpTablePtrD;

            cbpTablePtrE = encoderInfo->currContext->cbpTable63E[mode==MODE_INTRA ? 0 : pic_type];
            cbpTablePtrD = encoderInfo->currContext->cbpTable63D[mode==MODE_INTRA ? 0 : pic_type];

            x = cbp-1;
            cx = cbpTablePtrE[x];

            // ADAPT_VLC_NUM
            n = mode==MODE_INTRA ? 0 : pic_type;
            vlcNum = cbp_vlc_num[n][min(encoderInfo->currContext->cbp_vlc_index[n],7)];
            // vlcNum = (frameInfo->qp<16 && mode>=MODE_16x16 && mode<MODE_DIRECT) ? 10 : 6;

            len = put_vlc(vlcNum,cx,str);

            if (macroblockInfo->final_encode_flag){
                // ADAPT_VLC_NUM
                encoderInfo->currContext->cbp_vlc_index[n] += cx == encoderInfo->currContext->cbp_vlc_index[n] ? 0 : (cx < encoderInfo->currContext->cbp_vlc_index[n] ? -1 : 1);

                cy = max(0,cx-1);
                y = cbpTablePtrD[cy];
                cbpTablePtrD[cy] = x;
                cbpTablePtrD[cx] = y;
                cbpTablePtrE[x] = cy;
                cbpTablePtrE[y] = cx;
            }

        }
    }
    bits->cbp_luma += len;

    if (macroblockInfo->svt == 0)
        if ((mode == MODE_INTRA && macroblockInfo->tr16flag==0 && (encoderInfo->params->tr4flag)&1) || 
            (mode >= MODE_16x16 && mode <= MODE_8x8 && macroblockInfo->tr16flag==0 && (encoderInfo->params->tr4flag)&2 && cbp)){
                putbits(1,macroblockInfo->tr4flag,str);
                bits->mode += 1;
        }

        if (macroblockInfo->tr4flag){  
            int i;
            for (i=0;i<4;i++){            
                if (macroblockInfo->ycbp4[i]){ 
                    putbits(4,macroblockInfo->ycbp4[i],str);
                    bits->cbp_luma += 4;
                }
            }        
        }

}

/*-----------------------------------------------------------------------------------------
Function:   put_planar_delta
Purpose:    Encode and write the following parameters to the bitstream:
            - planar delta for planar prediction
Parameters: None 
-------------------------------------------------------------------------------------------*/


int put_planar_delta(macroblock_info *macroblockInfo,stream *str)
{
    /*planar quantization
    Y         qY               cW
    0-3   :  0,1,2,3         0-3
    4-15  :  4,6,8..14       4-9
    16-63 : 18,22,26..62    10-21
    64-.. : 68,76...        22-
    */
    int sign,qdelta,bits;
    bits = 0;
    qdelta = macroblockInfo->planar_deltaY;
    sign = qdelta<0 ? 1 : 0;
    qdelta = abs(qdelta);

    if(qdelta<4)
        bits += put_vlc(5,qdelta,str);    
    else if(qdelta < 16)
        bits += put_vlc(5,(qdelta>>1)+2,str);
    else if(qdelta < 64)
        bits += put_vlc(5,(qdelta>>2)+6,str);    
    else
        bits += put_vlc(5,(qdelta>>3)+14,str);


    if(qdelta > 0)
    {
        putbits(1,sign,str);
        bits++;
    }

    if(macroblockInfo->planar_deltaU == 0 && macroblockInfo->planar_deltaV == 0)
    {
        putbits(1,1,str); // chroma zero flag
        bits++;
    }
    else
    {
        putbits(1,0,str); // chroma zero flag
        bits++;
        qdelta = macroblockInfo->planar_deltaU;
        sign = qdelta<0 ? 1 : 0;
        qdelta = abs(qdelta);

        if(qdelta<4)
            bits += put_vlc(5,qdelta,str);    
        else if(qdelta < 16)
            bits += put_vlc(5,(qdelta>>1)+2,str);
        else if(qdelta < 64)
            bits += put_vlc(5,(qdelta>>2)+6,str);    
        else
            bits += put_vlc(5,(qdelta>>3)+14,str);

        if(qdelta > 0)
        {
            putbits(1,sign,str);
            bits++;
        }

        qdelta = macroblockInfo->planar_deltaV;
        sign = qdelta<0 ? 1 : 0;
        qdelta = abs(qdelta);

        if(qdelta<4)
            bits += put_vlc(5,qdelta,str);    
        else if(qdelta < 16)
            bits += put_vlc(5,(qdelta>>1)+2,str);
        else if(qdelta < 64)
            bits += put_vlc(5,(qdelta>>2)+6,str);    
        else
            bits += put_vlc(5,(qdelta>>3)+14,str);

        if(qdelta > 0)
        {
            putbits(1,sign,str);
            bits++;
        }
    }

    return bits;
}

/*-----------------------------------------------------------------------------------------
Function:   intraPredBits
Purpose:    Gives the number of bits used for intra prediction of luma
Parameters: None 
-------------------------------------------------------------------------------------------*/

void intraPredBits(int dir, int angle, int *len, int *code)
{
    int tmp;

    if (dir == PRED_DC){
        *len  = 2;
        *code = 0;
    }
    else if (angle == -8){
        *len  = 2;
        *code = 1;
    }
    else{
        *len  = 6;
        tmp   = dir == PRED_VER ? (2 << 4) : (3 << 4);
        *code = tmp + 7 + angle;
    }
}

/*-----------------------------------------------------------------------------------------
Function:   intraPredBits
Purpose:    Gives the number of bits used for intra prediction of chroma
Parameters: None 
-------------------------------------------------------------------------------------------*/

void intraPredBitsUV(int dir, int *len, int *code)
{
    *len  = 2;
    *code = dir;
}

/*-----------------------------------------------------------------------------------------
Function:   write_one_mb
Purpose:    Write all elements of a 16x16 MB to the bitstream
Parameters: None 
-------------------------------------------------------------------------------------------*/

void write_one_mb(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,stream *str)
{

    int i,j,k,mbxmax,ypos,xpos,mby,mbx,vy,vx,tmp,len,partw,parth,r,best_dir2;


    bitcounter *bits;

    enc_params *params = encoderInfo->params;
    int pic_type = frameInfo->pic_type;
    int num_ref = frameInfo->num_ref;
    //int last_pos = macroblockInfo->last_pos;
    int upa,lefta;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;
    vy = mby*2;
    vx = mbx*2;
    r = macroblockInfo->ref;

    if (pic_type==I_PICTURE)
        bits = encoderInfo->encStat->bitsI;
    else if(pic_type==P_PICTURE)
        bits = encoderInfo->encStat->bitsP;
    else
        bits = encoderInfo->encStat->bitsB;

    upa = up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);
    lefta = left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);

    if(pic_type!=I_PICTURE)
        if ((trace & 2) && str->outfile)
            fprintf(tracefile, "pos = %d\n", mby*mbxmax+mbx);
    if ((trace & 2) && str->outfile)
        fprintf(tracefile, "mode = %d\n", macroblockInfo->mode);


    macroblockInfo->specialBmode = 0;
    if (pic_type == B_PICTURE && macroblockInfo->mode != MODE_INTRA){ 
        if(encoderInfo->params->bi ==1)
        {
            if(r != 2)
            {
                macroblockInfo->specialBmode = ((macroblockInfo->mvblock[0][0].x == macroblockInfo->pred[r][0][0].x) &&
                    (macroblockInfo->mvblock[0][0].y == macroblockInfo->pred[r][0][0].y));
            }
            else
            {
                macroblockInfo->specialBmode = ((macroblockInfo->mvblock[0][0].x == macroblockInfo->pred[0][0][0].x) &&
                    (macroblockInfo->mvblock[0][0].y == macroblockInfo->pred[0][0][0].y) &&
                    (macroblockInfo->mvblockBW[0][0].x == macroblockInfo->pred[1][0][0].x) &&
                    (macroblockInfo->mvblockBW[0][0].y == macroblockInfo->pred[1][0][0].y));
            }
        }
        else
        {
            macroblockInfo->specialBmode = ((macroblockInfo->mvblock[0][0].x == macroblockInfo->pred[r][0][0].x) &&
                (macroblockInfo->mvblock[0][0].y == macroblockInfo->pred[r][0][0].y));
        }
    }


        write_mode(encoderInfo,frameInfo,macroblockInfo,bits,str);

    if (pic_type != I_PICTURE){
        if (num_ref > 1 && macroblockInfo->mode >= MODE_16x16 && macroblockInfo->mode <= MODE_8x8){
            if ((trace & 2) && str->outfile)
                fprintf(tracefile, "ref = %d\n", macroblockInfo->ref);
            if (pic_type == B_PICTURE){
                    if(encoderInfo->params->bi ==1)
                    {
                        if(macroblockInfo->ref == macroblockInfo->ref_pred)
                        {
                            putbits(1,1,str);
                            bits->ref +=1;
                        }
                        else
                        {
                            putbits(1,0,str);
                            if(macroblockInfo->ref == 0)
                                putbits(1,0,str);
                            else if(macroblockInfo->ref == 2)
                                putbits(1,1,str);
                            else if(macroblockInfo->ref_pred ==0)
                                putbits(1,0,str);
                            else
                                putbits(1,1,str);
                            bits->ref +=2;
                        }
                    }
                    else
                    {
                        putbits(1,macroblockInfo->ref,str);    
                        bits->ref += 1;
                    }

            }
        }

    }

    /* Write modelist for intra prediction to bitstream */    
    if (macroblockInfo->mode == MODE_INTRA){

        if(macroblockInfo->tr16flag == 1 && macroblockInfo->planar_flag == 1)
        {
            bits->planar += put_planar_delta(macroblockInfo,str);        
        }
        else
        {

            if (macroblockInfo->tr16flag == 1){
                if(!(upa == 0 && lefta == 0)){
                    //chroma
                    tmp = (macroblockInfo->modelist>>8) & 0x03;
                    len = put_vlc_8x8(5,tmp,str);
                    bits->ipred += len;

                    //luma
                    tmp = (macroblockInfo->modelist)&0x3;
                    best_dir2 = tmp;
                    len = put_vlc_8x8(5,tmp,str);
                    bits->ipred += len;
                }
            }
            else{

                if (macroblockInfo->tr4flag){
                    if( !(upa == 0 && lefta == 0)){
                        tmp = (macroblockInfo->modelist>>8)&0x3;
                        len = put_vlc_8x8(5,tmp,str);
                        bits->ipred += len;
                    }
                    for (k=0;k<16;k++){                    
                        if( !(k == 0 && upa == 0 && lefta == 0)){
                            len = put_vlc_8x8(5,macroblockInfo->modelist4[k],str);
                            bits->ipred += len;
                        }
                    }
                }
                else
                {

                    {

                        {
                            if (upa || lefta){
                                tmp = (macroblockInfo->modelist>>8)&0x3;

                                intraPredBitsUV(tmp, &len, &tmp);
                                putbits(len,tmp,str);

                                bits->ipred += len;
                            }
                            for (k=(!upa && !lefta);k<4;k++){
                                int code;
                                tmp = macroblockInfo->modelist8[k];

                                intraPredBits(macroblockInfo->modelist8[k], macroblockInfo->intraPredAngle8[k], &len, &code);

                                putbits(len,code,str);
                                bits->ipred += len;

                            }
                        }

                    }
                }
            }

        }// PLANAR

    }

    /* Write vector differences to bitstream */

    if (macroblockInfo->mode >= MODE_16x16 && macroblockInfo->mode <= MODE_8x8){
        int tmp,dmvx,dmvy;
        int cbp = (macroblockInfo->vcbp<<5) + (macroblockInfo->ucbp<<4) + (macroblockInfo->ycbp);
        int mode_idx = macroblockInfo->mode - MODE_16x16;
        partw = part_dims[mode_idx][0].w / 8;
        parth = part_dims[mode_idx][0].h / 8;

        if (macroblockInfo->mode == MODE_16x16 && cbp==0 && pic_type == B_PICTURE){
            /* If the condition above is true, the special B mode is used to signal DMV=0 */
            /* If DMV != 0, we know that DMVy and DMVx can not both be zero */              

            if(r!=2)
            {

                dmvx = macroblockInfo->mvblock[0][0].x-macroblockInfo->pred[r][mode_idx][0].x;
                dmvy = macroblockInfo->mvblock[0][0].y-macroblockInfo->pred[r][mode_idx][0].y;
                if (dmvy != 0 || dmvx !=0){                                                               
                    tmp = abs(dmvy)*2;
                    if (dmvy < 0) tmp--;                                
                        bits->vec += put_vlc(10,tmp,str);
                        if (dmvy==0){
                            tmp = abs(dmvx)-1;                                    
                            bits->vec += put_vlc(10,tmp,str); 
                            tmp = (dmvx < 0);
                            putbits(1,tmp,str);
                            bits->vec++;                                    
                        }
                        else{                                
                            tmp = abs(dmvx)*2;
                            if (dmvx < 0) tmp--;                                    
                            bits->vec += put_vlc(10,tmp,str);                                    
                        }                            
                }

            }
            else
            {
                if(macroblockInfo->specialBmode != 1)
                {
                    dmvx = macroblockInfo->mvblock[0][0].x-macroblockInfo->pred[0][mode_idx][0].x;
                    dmvy = macroblockInfo->mvblock[0][0].y-macroblockInfo->pred[0][mode_idx][0].y;
                    bits->vec += put_level(dmvx,str);
                    bits->vec += put_level(dmvy,str);  
                    dmvx = macroblockInfo->mvblockBW[0][0].x-macroblockInfo->pred[1][mode_idx][0].x;
                    dmvy = macroblockInfo->mvblockBW[0][0].y-macroblockInfo->pred[1][mode_idx][0].y;
                    bits->vec += put_level(dmvx,str);
                    bits->vec += put_level(dmvy,str);  
                }

            }

        }                        
        else{
            for (i = 0; i < 2; i+=parth){
                for (j = 0; j < 2; j+=partw){ 

                    if(r == 2)
                    {
                        dmvx = macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[0][mode_idx][i*2+j].x;
                        dmvy = macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[0][mode_idx][i*2+j].y;
                        bits->vec += put_level(dmvx,str);
                        bits->vec += put_level(dmvy,str);                                                          
                        if ((trace & 2) && str->outfile){
                            fprintf(tracefile, "vec.x = %d\n", macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[0][mode_idx][i*2+j].x);
                            fprintf(tracefile, "vec.y = %d\n", macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[0][mode_idx][i*2+j].y);
                        }
                        dmvx = macroblockInfo->mvblockBW[i][j].x-macroblockInfo->pred[1][mode_idx][i*2+j].x;
                        dmvy = macroblockInfo->mvblockBW[i][j].y-macroblockInfo->pred[1][mode_idx][i*2+j].y;
                        bits->vec += put_level(dmvx,str);
                        bits->vec += put_level(dmvy,str);                                                          
                        if ((trace & 2) && str->outfile){
                            fprintf(tracefile, "vec.x = %d\n", macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[1][mode_idx][i*2+j].x);
                            fprintf(tracefile, "vec.y = %d\n", macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[1][mode_idx][i*2+j].y);
                        }
                    }
                    else
                    {
                        dmvx = macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[r][mode_idx][i*2+j].x;
                        dmvy = macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[r][mode_idx][i*2+j].y;
                        bits->vec += put_level(dmvx,str);
                        bits->vec += put_level(dmvy,str);                                                          
                        if ((trace & 2) && str->outfile){
                            fprintf(tracefile, "vec.x = %d\n", macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[r][mode_idx][i*2+j].x);
                            fprintf(tracefile, "vec.y = %d\n", macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[r][mode_idx][i*2+j].y);
                        }
                    }

                    if ((trace & 2) && str->outfile){
                        fprintf(tracefile, "vec.x = %d\n", macroblockInfo->mvblock[i][j].x-macroblockInfo->pred[r][mode_idx][i*2+j].x);
                        fprintf(tracefile, "vec.y = %d\n", macroblockInfo->mvblock[i][j].y-macroblockInfo->pred[r][mode_idx][i*2+j].y);
                    }
                }           
            }
        }


        if(g_enableSwitched == 1){              
            int pos;              
            for (i = 0; i < 2; i+=parth){
                for (j = 0; j < 2; j+=partw){ 
                    pos = ((macroblockInfo->mvblock[i][j].y%4+4)%4)*4 + ((macroblockInfo->mvblock[i][j].x%4+4)%4);
                    if(transmitInterpIndx[pos]== 1){
                        putbits(1,macroblockInfo->mvblock[i][j].interp_indx,str);
                        bits->vec++; 
                    }

                    if(macroblockInfo->ref == 2)
                    {
                        pos = ((macroblockInfo->mvblockBW[i][j].y%4+4)%4)*4 + ((macroblockInfo->mvblockBW[i][j].x%4+4)%4);
                        if(transmitInterpIndx[pos]== 1){
                            putbits(1,macroblockInfo->mvblockBW[i][j].interp_indx,str);
                            bits->vec++; 
                        }
                    }

                }
            }
        }
    } //if mode...

    if (macroblockInfo->mode != MODE_SKIP){
        if ((trace & 2) && str->outfile){
            fprintf(tracefile,"ycbp = 0x%x\n",macroblockInfo->ycbp);
            fprintf(tracefile,"ucbp = 0x%x\n",macroblockInfo->ucbp);
            fprintf(tracefile,"vcbp = 0x%x\n",macroblockInfo->vcbp);
            fflush(tracefile);
        }
    }

    /* Write coefficients to bitstream */ 

    if(macroblockInfo->planar_flag)
        return;


    if(macroblockInfo->mode == MODE_INTRA && upa == 0 && lefta== 0){
        encode_side_intra(macroblockInfo->side_intra_y ,str ,&(bits->side_intra_y),0); 
    }

    if ((macroblockInfo->mode >= MODE_16x16 && macroblockInfo->mode < MODE_8x8) && macroblockInfo->tr16flag == 0 &&macroblockInfo->svt == 1){
        int base;
        int len;
        if (macroblockInfo->svt_idx_num!=0){
            int basetable[]={0,1,1,2,2,2,2,3,3,3,3,3,3,3,3,
                4,4,4,4,4,4,4,4,4,4,4,4,4,4,4,
                4,5,5,5,5,5,5,5,5,5,5,5,5,5,5,
                5,5,5,5,5,5,5,5,5,5,5,5,5};
            base = basetable[macroblockInfo->svt_idx_num-1];
            if (macroblockInfo->effective_svt_idx<2*(macroblockInfo->svt_idx_num-(int)pow(2,base))){                
                len = base+1;
                 putbits(len,macroblockInfo->effective_svt_idx,str);
            }
            else{
                len = base;
                putbits(len,macroblockInfo->effective_svt_idx-(macroblockInfo->svt_idx_num-(int)pow(2,base)),str);
            }
            bits->svt_idx+=len;
            if ((trace & 2) && str->outfile){
                fprintf(tracefile, "svt_idx_num = %d\n", macroblockInfo->svt_idx_num);
                fprintf(tracefile, "effective_svt_idx = %d\n", macroblockInfo->effective_svt_idx);
            }
        }
        else{
            printf("macroblockInfo->svt_idx_num==0\n");
            return;
        }
    }


    if (macroblockInfo->ycbp != 0){
        int block_type;

        if (macroblockInfo->mode==MODE_INTRA)
            block_type = macroblockInfo->tr16flag ? BLOCK_Y16x16I : BLOCK_Y8x8I;

        else if(macroblockInfo->mode==MODE_16x16 && macroblockInfo->tr16flag)
            block_type = pic_type==P_PICTURE ? BLOCK_Y16x16P : BLOCK_Y16x16B;

        else
            block_type = pic_type==P_PICTURE ? BLOCK_Y8x8P : BLOCK_Y8x8B;
        if ((macroblockInfo->mode>=MODE_16x16 && macroblockInfo->mode<=MODE_DIRECT) && macroblockInfo->svt == 1 && pic_type == P_PICTURE){
            block_type = macroblockInfo->svt_idx < 32 ? 9 : 8;
        }


        if (block_type>=BLOCK_Y16x16I && block_type<=BLOCK_Y16x16B)
            encode_coeffs(encoderInfo,macroblockInfo->ycoeff[0],str,&(bits->coeff_y),block_type,macroblockInfo->final_encode_flag);        

        else if ((macroblockInfo->mode>=MODE_16x16 && macroblockInfo->mode<=MODE_DIRECT) && macroblockInfo->svt==1 && macroblockInfo->svt_idx >= 32)
            encode_coeffs_svt(encoderInfo,macroblockInfo->ycoeff_svt, str, &(bits->coeff_y), block_type, macroblockInfo->final_encode_flag);

        else if (macroblockInfo->tr4flag){
            int m,n,l;
            block_type = macroblockInfo->mode==MODE_INTRA ? 0 : (int)pic_type;            
            for (m=0;m<16;m+=8){
                for (n=0;n<16;n+=8){                                                                    
                    for (k=0;k<8;k+=4){
                        for (l=0;l<8;l+=4){
                            int index_4x4 = m + k + n/4 + l/4;  //convert from 2x2/2x2 scan to 4x4 scan                          
                            if (macroblockInfo->ycbp4[m/4+n/8] & (1 << (3-(k/2 + l/4)))){
                                encode_coeffs4(encoderInfo,macroblockInfo->ycoeff[index_4x4], str, &(bits->coeff_y), block_type, macroblockInfo->final_encode_flag);
                            }
                        }
                    }
                }
            }         
        }
        else{
            for (i=0;i<4;i++){                
                if (macroblockInfo->ycbp & (1<<(3-i))){    
                    encode_coeffs(encoderInfo,macroblockInfo->ycoeff[i], str, &(bits->coeff_y), block_type, macroblockInfo->final_encode_flag);
                }
            }
        }
    }


    if(macroblockInfo->tr16flag == 1 && macroblockInfo->mode == MODE_INTRA  && upa == 0 && lefta== 0){       
        encode_side_intra(macroblockInfo->side_intra_u ,str ,&(bits->side_intra_u),1);   
        encode_side_intra(macroblockInfo->side_intra_v ,str ,&(bits->side_intra_v),1);
        assert(macroblockInfo->ucoeff[0][0] == 0);
        assert(macroblockInfo->vcoeff[0][0] == 0);
    }

    if (macroblockInfo->mode == MODE_INTRA && macroblockInfo->tr16flag == 1 && (macroblockInfo->ucbp != 0 || macroblockInfo->vcbp != 0)){
        encode_coeffs(encoderInfo,macroblockInfo->ucoeff, str, &(bits->coeff_uv), 0, macroblockInfo->final_encode_flag);
    }
    else{
        if(macroblockInfo->tr16flag != 1 && macroblockInfo->mode == MODE_INTRA  &&  upa == 0 && lefta== 0){
            encode_side_intra(macroblockInfo->side_intra_u ,str ,&(bits->side_intra_u),1);          
        }

        if (macroblockInfo->ucbp != 0){
            encode_coeffs(encoderInfo,macroblockInfo->ucoeff, str, &(bits->coeff_u), 0, macroblockInfo->final_encode_flag);
        }


        if(macroblockInfo->tr16flag != 1 && macroblockInfo->mode == MODE_INTRA  &&  upa == 0 && lefta== 0){
            encode_side_intra(macroblockInfo->side_intra_v ,str ,&(bits->side_intra_v),1);
        }

        if (macroblockInfo->vcbp != 0){
            encode_coeffs(encoderInfo,macroblockInfo->vcoeff, str, &(bits->coeff_v), 1, macroblockInfo->final_encode_flag);
        }
    }
}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_SVT_P
Purpose:    Encode an 16x16 luma MB in SVT mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_luma_SVT_P(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{
    int k,l,m,cbpflag;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax;
    //STATIC unsigned char **mblock16;
    //STATIC int           **dblock16;
    short dblock8[8][8]; //STATIC int           **dblock8;

    unsigned char pblock16[16][16];
#if (BUGFIX_SVT_INVTR1 || BUGFIX_SVT_INVTR2) 
    short rdblock16[16][16]; // bugfix lpe 17. mars 2010
#else
    int rdblock16[16][16]; 
#endif
    short rdblock8[8][8]; //STATIC int           **rdblock8;
    //unsigned char rblock16[16][16];
    unsigned char tblock8[8][8];
    short rcoeff[4][8][8]; //STATIC int ***rcoeff;
    int tc;
    int beta;

    int coded_sum = 0;

    int svt_idx;
    int best_svt_idx;
    int dir;
    int offset;
    int best_dir;
    int best_offset;
    short dblock8_svt[4][16];
    int rdblock8_svt[4][16]; //STATIC int **rdblock8_svt;
    int rcoeff_svt[4][16]; //STATIC int **rcoeff_svt;
    int svt_type;

    int svt_idx_num = SVT_TABLE_SIZE;
    //int effective_svt_idx = 0;
    int avail_svt_idx[SVT_TABLE_SIZE];

    int ref_idx;
    int mode_idx;

    int edge1=0, edge2=0, edge3=0, edge4=0;



    //unsigned char **block = ucsubmatrix(encoderInfo->recon->Y,0,7,0,7,0,0);    

    int beginx, beginy, bestx=0, besty=0;
    int x, y;
    int num_bits;
    int n;
    int cand_cost = 0, best_cost = 0xfffffff;
    stream *str_tmp = initbits(0);
    int skip_svt_indx[SVT_TABLE_SIZE]; //based on MV
    int sad_skip_svt_indx[SVT_TABLE_SIZE]; //skip based on SAD
    int largest_sad, largest_sad_indx;
    int i;
    int total_sad,num_avail_pos;


    int sumlev;
    int *sum_of_levels = &sumlev;
    int ncoeff;
    int *no_of_coeffs = &ncoeff;
    int pic_type = P_PICTURE;
    image_frame *curr = encoderInfo->orig;        
    image_frame *recon = encoderInfo->recon;
    image_frame **iprev = encoderInfo->ref_frame_qp;

    enc_params *params = encoderInfo->params;
    int mode = macroblockInfo->mode;


    block_info *mvmap = frameInfo->mvmap;
    int mvxmax = 2*encoderInfo->widthInMBs;
    short dblock16[16][16]; //STATIC dblock16 = imatrix(0,15,0,15);



    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;


    vy = mby*2;
    vx = mbx*2;


    //STATIC mblock16 = ucsubmatrix(curr->Y,0,15,0,15,0,0);
    //short dblock16[16][16]; //STATIC dblock16 = imatrix(0,15,0,15);
    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC pblock16 = ucmatrix(0,15,0,15);

    //STATIC eblock = ucmatrix(0,22,0,22);

    //STATIC rdblock16 = imatrix(0,15,0,15);
    //STATIC rdblock8 = imatrix(0,7,0,7);
    //STATIC dblock8_svt = imatrix(0,3,0,15);
    //STATIC rdblock8_svt = imatrix(0,3,0,15);
    //STATIC rcoeff_svt = imatrix(0,3,0,15);
    //STATIC rblock16 = ucsubmatrix(recon->Y,0,15,0,15,0,0);
    //STATIC tblock8 = ucsubmatrix(pblock16,0,7,0,7,0,0);
    //STATIC rcoeff = (int ***)rf_calloc(4,(size_t)(sizeof(int**)));
    //STATIC for (k = 0; k < 4; k++)
    //STATIC {
    //STATIC    rcoeff[k] = imatrix(0,7,0,7);
    //STATIC }


    macroblockInfo->ycbp = 0;

    ref_idx = macroblockInfo->ref;
    mode_idx = mode - MODE_16x16;
    if (macroblockInfo->final_encode_flag == 1)
    {
        macroblockInfo->svt_idx = best_idx_svt[ref_idx][mode_idx];
        best_svt_idx = best_idx_svt[ref_idx][mode_idx];

        macroblockInfo->svt_idx_num = best_svt_idx_num[ref_idx][mode_idx];
        macroblockInfo->effective_svt_idx = best_effective_svt_idx[ref_idx][mode_idx];

        if (best_svt_idx < 32)
        {
            for (k=0;k<8;k++)
            {
                for (l=0;l<8;l++)
                {
                    macroblockInfo->ycoeff[0][k][l] = best_ycoeff[ref_idx][mode_idx][k][l];
                }
            }
        }
        else
        {
            for (k=0;k<4;k++)
            {
                for (l=0;l<16;l++)
                {
                    macroblockInfo->ycoeff_svt[k][l] = best_ycoeff_svt[ref_idx][mode_idx][k][l];
                }
            }
        }
        //move_ucsubmatrix(rblock16,recon->Y,ypos,ypos+15,xpos,xpos+15,0,0);
        for (k=0;k<16;k++)
        {
            for (l=0;l<16;l++)
            {
                recon->Y[ypos+k][xpos+l] = best_rblock16[ref_idx][mode_idx][k][l];
            }
        }

        cbpflag = 1;

        macroblockInfo->ycbp = 1<<3;

        if (best_svt_idx < 32)
        {
            svt_type = 0;
            besty = svt_pos[best_svt_idx][1];
            bestx = svt_pos[best_svt_idx][0];
            best_dir = 0;
            best_offset = 0;
        }
        else
        {
            svt_type = 1;
            besty = 0;
            bestx = 0;
            best_dir = svt_pos[best_svt_idx][0];
            best_offset = svt_pos[best_svt_idx][1];
        }

        if (str_tmp != NULL)
        {
            closebits(str_tmp);
        }

        goto SPEED_CBP;
    }
    else
    {
        for (k=0; k<9; k++)
            for (l=0; l<9; l++)
            {
                rdcost1[k][l] = -1;
            }

            for (k=0; k<2; k++)
                for (l=0; l<13; l++)
                {
                    rdcost2[k][l] = -1;
                }
    }

    //STATIC move_ucsubmatrix(mblock16,curr->Y,ypos,ypos+15,xpos,xpos+15,0,0);
    for (k = 0; k < 2; k++)
    {
        for (l = 0; l < 2; l++)
        {
            //move_ucsubmatrix(tblock8,pblock16,k*8,k*8+7,l*8,l*8+7,0,0);          
            get_luma_inter_prediction_8x8(macroblockInfo,k, l, iprev, tblock8);
            for (m=0;m<8;m++)
            {
                memcpy(&pblock16[k*8+m][l*8],&tblock8[m][0],8);
            }
        }
    }
    for (k=0;k<16;k++)
    {
        for (l=0;l<16;l++)
        {
            dblock16[k][l] = curr->Y[ypos+k][xpos+l]-pblock16[k][l];
        }
    }

    for (k=0;k<8;k++)
    {
        for (l=0;l<8;l++)
        {
            best_ycoeff[ref_idx][mode_idx][k][l] = 0;
        }
    }
    for (k=0;k<4;k++)
    {
        for (l=0;l<16;l++)
        {
            best_ycoeff_svt[ref_idx][mode_idx][k][l] = 0;
        }
    }
    for (k=0;k<16;k++)
    {
        for (l=0;l<16;l++)
        {
            best_rblock16[ref_idx][mode_idx][k][l] = 0;
        }
    }

    for (k=0; k<SVT_TABLE_SIZE; k++)
        avail_svt_idx[k] = 1;

    tc = svt_tctable_8x8[frameInfo->qp];
    beta = svt_betatable_8x8[frameInfo->qp];

    {
        //int th=frameInfo->qp<<4;

        if (mode != MODE_16x16)
        {

            if (abs(macroblockInfo->mvblock[0][0].x-macroblockInfo->mvblock[0][1].x) >= 4 ||
                abs(macroblockInfo->mvblock[0][0].y-macroblockInfo->mvblock[0][1].y) >= 4)
                edge1=1;
            if (abs(macroblockInfo->mvblock[0][0].x-macroblockInfo->mvblock[1][0].x) >= 4 ||
                abs(macroblockInfo->mvblock[0][0].y-macroblockInfo->mvblock[1][0].y) >= 4)
                edge2=1;
            if (abs(macroblockInfo->mvblock[0][1].x-macroblockInfo->mvblock[1][1].x) >= 4 ||
                abs(macroblockInfo->mvblock[0][1].y-macroblockInfo->mvblock[1][1].y) >= 4)
                edge3=1;
            if (abs(macroblockInfo->mvblock[1][0].x-macroblockInfo->mvblock[1][1].x) >= 4 ||
                abs(macroblockInfo->mvblock[1][0].y-macroblockInfo->mvblock[1][1].y) >= 4)
                edge4=1;


            //reference frame
        }
    }
    //First check based on MV
    total_sad = 0;
    num_avail_pos = 0;
    for (svt_idx=0; svt_idx<SVT_TABLE_SIZE; svt_idx++)
    {      
        sad_skip_svt_indx[svt_idx] = 1;
        if (svt_idx < 32)
            svt_type = 0;
        else
            svt_type = 1;

        if (svt_type == 0)
        {
            beginy = svt_pos[svt_idx][1];
            beginx = svt_pos[svt_idx][0];
            dir = 0;
            offset = 0;
        }
        else
        {
            beginy = 0;
            beginx = 0;
            dir = svt_pos[svt_idx][0];
            offset = svt_pos[svt_idx][1];
        }
        skip_svt_indx[svt_idx] = check_svt_skip(dir,offset,edge1,edge2,edge3,edge4,svt_type,beginy,beginx);

        //only use key location parameters.
        if (frameInfo->svt_set == 1)
        {
            if (svt_idx!=0 && svt_idx!=8 && svt_idx!=9 && svt_idx!=17 &&
                svt_idx!=32 && svt_idx!=36 && svt_idx!=40 && svt_idx!=44 &&
                svt_idx!=45 && svt_idx!=49 && svt_idx!=53 && svt_idx!=57
                )
                skip_svt_indx[svt_idx] = 1;

        }
        //end

        if(skip_svt_indx[svt_idx] == 1)
        {
            sad_idx_svt[ref_idx][mode_idx][svt_idx] = -1;
        }
        else
        {
            sad_idx_svt[ref_idx][mode_idx][svt_idx] = sad_svt(dir,offset,beginy,beginx,svt_type,dblock16);
            total_sad += sad_idx_svt[ref_idx][mode_idx][svt_idx];
            num_avail_pos++;
        }
    }

    if(params->fast_svt == 0 || params->fast_svt >= SVT_TABLE_SIZE)
    {
        for (svt_idx=0; svt_idx<SVT_TABLE_SIZE; svt_idx++)
            sad_skip_svt_indx[svt_idx] = 0;
    }
    else
    {

        for(i=0;i<params->fast_svt;i++)
        {   
            largest_sad = -1;
            for (svt_idx=0; svt_idx<SVT_TABLE_SIZE; svt_idx++)
            {     

                if(sad_idx_svt[ref_idx][mode_idx][svt_idx] > largest_sad && sad_skip_svt_indx[svt_idx] == 1)
                {
                    largest_sad = sad_idx_svt[ref_idx][mode_idx][svt_idx];
                    largest_sad_indx = svt_idx;
                }
            }
            sad_skip_svt_indx[largest_sad_indx] = 0;     
        }
    }


    {
        int x,y,sad;
        sad = 0;
        for (y=0; y<16; y++)
        {
            for (x=0; x<16; x++)
            {
                sad += abs(dblock16[y][x]);
            }
        }
        if(params->fast_svt_thr2 > 0 && sad > params->fast_svt_thr2)
        {
            //If sad is large don't check SVT at all
            for (svt_idx=0; svt_idx<SVT_TABLE_SIZE; svt_idx++)
                sad_skip_svt_indx[svt_idx] = 1;
        }
    }

    for (svt_idx=0; svt_idx<SVT_TABLE_SIZE; svt_idx++)
    {
        if (svt_idx < 32)
            svt_type = 0;
        else
            svt_type = 1;
        if (svt_type == 0)
        {
            beginy = svt_pos[svt_idx][1];
            beginx = svt_pos[svt_idx][0];
            dir = 0;
            offset = 0;
        }
        else
        {
            beginy = 0;
            beginx = 0;
            dir = svt_pos[svt_idx][0];
            offset = svt_pos[svt_idx][1];
        }

        if (skip_svt_indx[svt_idx]==1)
        {
            svt_idx_num--;
            avail_svt_idx[svt_idx] = 0;
            continue;
        }
        else if (sad_skip_svt_indx[svt_idx] == 1)
            continue;           

        coded_sum = 0;
        if (svt_type == 0)
        {
            for (y=0; y<8; y++)
            {
                for (x=0; x<8; x++)
                {
                    dblock8[y][x] = dblock16[y+beginy][x+beginx];
                }
                transform_8x8(dblock8,macroblockInfo->ycoeff[0]);
                cbpflag = quantize_8x8(macroblockInfo->ycoeff[0],frameInfo->qp,pic_type,0,sum_of_levels);
            }
        }
        else
        {
            if (dir==0)
            {
                for (y=0; y<4; y++)
                    for (x=0; x<16; x++)
                    {
                        dblock8_svt[y][x] = dblock16[y+offset][x];
                    }
            }
            else //dir==1
            {
                for (x=0; x<4; x++)
                {
                    for (y=0; y<16; y++)
                    {
                        dblock8_svt[x][y] = dblock16[y][x+offset];
                    }
                }
            }

            transform_svt(dblock8_svt,macroblockInfo->ycoeff_svt);	
            cbpflag = quantize_svt(macroblockInfo->ycoeff_svt,frameInfo->qp,pic_type,0,no_of_coeffs,sum_of_levels);
        }

        if (cbpflag == 0)
        {
            if (svt_type == 0)
            {
                rdcost1[beginy][beginx] = 0;
            }
            else
            {
                rdcost2[dir][offset] = 0;
            }
            continue;
        }

        if (svt_type == 0)
        {
            dequantize_8x8(macroblockInfo->ycoeff[0],rcoeff[0],frameInfo->qp,0);
            inverse_transform_8x8(rcoeff[0],rdblock8,frameInfo->qp,0);
        }
        else
        {
            dequantize_svt(macroblockInfo->ycoeff_svt,rcoeff_svt,frameInfo->qp,0);
            inverse_transform_svt(rcoeff_svt,rdblock8_svt,frameInfo->qp,0);
        }

        for (k=0;k<16;k++)
        {
            for (l=0;l<16;l++)
            {
                rdblock16[k][l] = 0;
            }
        }
        if (svt_type == 0)
        {
            for (k=0;k<8;k++)
            {
                for (l=0;l<8;l++)
                {
                    rdblock16[k+beginy][l+beginx] = rdblock8[k][l];
                }
            }
        }
        else
        {
            if (dir==0)
            {
                for (k=0;k<4;k++)
                {
                    for (l=0;l<16;l++)
                    {
                        rdblock16[k+offset][l] = rdblock8_svt[k][l];
                    }
                }                    
            }
            else //dir==1
            {
                for (l=0;l<16;l++)
                {
                    for (k=0;k<4;k++)
                    {
                        rdblock16[l][k+offset] = rdblock8_svt[k][l];
                    }
                }
            }

        }

        macroblockInfo->ycbp = 1<<3;

        //move_ucsubmatrix(rblock16,recon->Y,ypos,ypos+15,xpos,xpos+15,0,0);

        for (k=0;k<16;k++)
        {
            unsigned char *restrict rblock16 = &recon->Y[ypos+k][xpos]; 
#if (BUGFIX_SVT_INVTR1 || BUGFIX_SVT_INVTR2)
            short tmp[16];
            for (l=0;l<16;l++)
                tmp[l] = rdblock16[k][l]+pblock16[k][l];
            for (l=0;l<16;l++)
                rblock16[l] = min(max(tmp[l], 0), 255);
#else
            for (l=0;l<16;l++)
                rblock16[l] =  min(max(rdblock16[k][l]+pblock16[k][l],0),255);
#endif
        }
        {
            unsigned char eblock[24][24] = {0,};

            for (k=0;k<16;k++)
            {
                const unsigned char *rblock16 = &recon->Y[ypos+k][xpos];
                for (l=0;l<16;l++)
                {
                    eblock[k+3][l+3] = rblock16[l];
                }
            }

            //filling
            for (k=0;k<4;k++)
            {
                const unsigned char *rblock16_0 = &recon->Y[ypos+0][xpos];
                const unsigned char *rblock16_15 = &recon->Y[ypos+15][xpos];
                for (l=0;l<16;l++)
                {
                    eblock[k][l+3] = rblock16_0[l];
                    eblock[k+19][l+3] = rblock16_15[l];
                }
            }

            for (k=0;k<16;k++)
            {
                const unsigned char rblock16_0 = recon->Y[ypos+k][xpos+0];
                const unsigned char rblock16_15 = recon->Y[ypos+k][xpos+15];
                for (l=0;l<4;l++)
                {
                    eblock[k+3][l] = rblock16_0;
                    eblock[k+3][l+19] = rblock16_15;
                }
            }

            //filtering
            if (svt_type == 0)
            {
                //filter left vertical edge
                filter_8x8_vertical(eblock,beginx-1,beginy+3,tc,beta);
                //filter right vertical edge
                filter_8x8_vertical(eblock,beginx+7,beginy+3,tc,beta);
                //filter top horizontal edge
                filter_8x8_horizontal(eblock,beginx+3,beginy-1,tc,beta);
                //filter bottom horizontal edge
                filter_8x8_horizontal(eblock,beginx+3,beginy+7,tc,beta);
            }
            else
            {
                if (dir==0)
                {
                    //filter top horizontal edge
                    if (offset>0)
                    {
                        filter_8x8_horizontal(eblock,3,offset-1,tc,beta);
                        filter_8x8_horizontal(eblock,11,offset-1,tc,beta);
                    }
                    //filter bottom horizontal edge
                    if (offset<12)
                    {
                        filter_8x8_horizontal(eblock,3,offset+3,tc,beta);
                        filter_8x8_horizontal(eblock,11,offset+3,tc,beta);
                    }
                }
                else
                {
                    //filter left vertical edge
                    if (offset>0)
                    {
                        filter_8x8_vertical(eblock,offset-1,3,tc,beta);
                        filter_8x8_vertical(eblock,offset-1,11,tc,beta);
                    }

                    //filter right vertical edge
                    if (offset<12)
                    {
                        filter_8x8_vertical(eblock,offset+3,3,tc,beta);
                        filter_8x8_vertical(eblock,offset+3,11,tc,beta);
                    }
                }
            }
            //end

            //copy back the pixel
            for (k=0;k<16;k++)
            {
                unsigned char *restrict rblock16 = &recon->Y[ypos+k][xpos]; 
                for (l=0;l<16;l++)
                {
                    rblock16[l] = eblock[k+3][l+3];
                }
            }
        }

        num_bits = 0;

        if (macroblockInfo->ycbp)
        {
            n = (pic_type == P_PICTURE) ? 8 : 4; 
            if (svt_type == 0)
            {
                n = (pic_type == P_PICTURE) ? 9 : 4; 
                encode_coeffs(encoderInfo,macroblockInfo->ycoeff[0], str_tmp, &num_bits,n,0);
            }
            else
            {
                encode_coeffs_svt(encoderInfo,macroblockInfo->ycoeff_svt, str_tmp, &num_bits,n,0);
            }
        }


        for (k=0;k<16;k++)
        {
            memcpy(&macroblockInfo->recY[16*k],&recon->Y[ypos+k][xpos],16*sizeof(unsigned char));
        }
        for (k=0;k<8;k++)
        {
            memcpy(&macroblockInfo->recU[8*k],&recon->U[ypos/2+k][xpos/2],8*sizeof(unsigned char));
            memcpy(&macroblockInfo->recV[8*k],&recon->V[ypos/2+k][xpos/2],8*sizeof(unsigned char));
        }            
        cand_cost = cost_MB(macroblockInfo,num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);


        if (svt_type == 0)
        {
            rdcost1[beginy][beginx] = cand_cost;
        }
        else
        {
            rdcost2[dir][offset] = cand_cost;
        }

        if (cand_cost < best_cost)
        {
            besty = beginy;
            bestx = beginx;
            best_cost = cand_cost;
            best_svt_idx = svt_idx;
            best_dir = dir;
            best_offset = offset;
            if (svt_idx < 32)
            {
                for (k=0;k<8;k++)
                {
                    for (l=0;l<8;l++)
                    {
                        best_ycoeff[ref_idx][mode_idx][k][l] = macroblockInfo->ycoeff[0][k][l];
                    }
                }
            }
            else
            {
                for (k=0;k<4;k++)
                {
                    for (l=0;l<16;l++)
                    {
                        best_ycoeff_svt[ref_idx][mode_idx][k][l] = macroblockInfo->ycoeff_svt[k][l];
                    }
                }
            }
            for (k=0;k<16;k++)
            {
                const unsigned char *rblock16 = &recon->Y[ypos+k][xpos];
                for (l=0;l<16;l++)
                {
                    best_rblock16[ref_idx][mode_idx][k][l] = rblock16[l];
                }
            }
        }
    }

    if (str_tmp != NULL)
    {
        closebits(str_tmp);
    }


    if (macroblockInfo->ycbp == 0)
        goto SVT_END;

    macroblockInfo->svt_idx = best_svt_idx;
    best_svt_idx_num[ref_idx][mode_idx] = svt_idx_num;
    macroblockInfo->svt_idx_num = svt_idx_num;

    best_effective_svt_idx[ref_idx][mode_idx] = 0;
    for (k=0; k<best_svt_idx; k++)
    {
        best_effective_svt_idx[ref_idx][mode_idx] += avail_svt_idx[k];
    }
    macroblockInfo->effective_svt_idx = best_effective_svt_idx[ref_idx][mode_idx];

    if (best_svt_idx < 32)
    {
        for (k=0;k<8;k++)
        {
            for (l=0;l<8;l++)
            {
                macroblockInfo->ycoeff[0][k][l] = best_ycoeff[ref_idx][mode_idx][k][l];
            }
        }
    }
    else
    {
        for (k=0;k<4;k++)
        {
            for (l=0;l<16;l++)
            {
                macroblockInfo->ycoeff_svt[k][l] = best_ycoeff_svt[ref_idx][mode_idx][k][l];
            }
        }
    }
    for (k=0;k<16;k++)
    {
        unsigned char *restrict rblock16 = &recon->Y[ypos+k][xpos]; 
        for (l=0;l<16;l++)
        {
            rblock16[l] = best_rblock16[ref_idx][mode_idx][k][l];
        }
    }

    best_idx_svt[ref_idx][mode_idx] = best_svt_idx;

    if (best_svt_idx < 32)
        svt_type = 0;
    else
        svt_type = 1;

SPEED_CBP:
    mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = 0;
    mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = 0;
    mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = 0;
    mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = 0;
    if (svt_type == 0)
    {
        if (bestx==0)
        {
            if (besty==0)
            {                        
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;                        
            }
            else if (besty==8)
            {
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag; 

            }
            else
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;

            }
        }
        else if (bestx==8)
        {
            if (besty==0)
            {
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;

            }
            else if (besty==8)
            {
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
            else
            {
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
        }
        else
        {
            if (besty==0)
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;

            }
            else if (besty==8)
            {
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
            else
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
        }
    }
    else
    {
        if (best_dir==0)
        {
            if (best_offset>=0 && best_offset<=4)
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;

            }
            else if (best_offset>4 && best_offset<8)
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
            else
            {
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
        }
        else //best_dir==1
        {
            if (best_offset>=0 && best_offset<=4)
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;

            }
            else if (best_offset>4 && best_offset<8)
            {
                mvmap[(vy+0)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+0)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
            else
            {
                mvmap[(vy+0)*mvxmax+(vx+1)].coded_y = cbpflag;
                mvmap[(vy+1)*mvxmax+(vx+1)].coded_y = cbpflag;
            }
        }
    }

SVT_END:
    //STATIC free_ucsubmatrix(mblock16,0,15,0,15);
    //STATIC free_imatrix(dblock16,0,15,0,15);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_ucmatrix(pblock16,0,15,0,15);
    //STATIc free_ucmatrix(eblock,0,22,0,22);
    //STATIC free_ucsubmatrix(block,0,7,0,7);
    //STATIC free_imatrix(rdblock16,0,15,0,15);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_ucsubmatrix(rblock16,0,15,0,15);
    //STATIC free_ucsubmatrix(tblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8_svt,0,3,0,15);
    //STATIC free_imatrix(rdblock8_svt,0,3,0,15);
    //STATIC free_imatrix(rcoeff_svt,0,3,0,15);
    //STATIC for (k = 0; k < 4; k++)
    //STATIC {
    //STATIC     free_imatrix(rcoeff[k],0,7,0,7);
    //STATIC }
    //STATIC rf_free(rcoeff);
    return;
}

/*-----------------------------------------------------------------------------------------
Function:   encode_planar
Purpose:    
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_planar(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{  
    int topright,bottomleft,bottomright,qdelta,sign;
    //int best_planar_indx_sad = 0xfffffff;                                       
    unsigned char planar_recY[256],planar_recU[64],planar_recV[64];
    int k,l;
    int uvindex;
    unsigned char *recUV, *recUV2;
    int cypos,cxpos,ypos,xpos;
    int sads[3];
    int is_planar_identical = 1;
    //image_frame *recon = encoderInfo->recon;
    // This function currently supports only checking planar if both above and left are available
    int upa   = 1;
    int lefta = 1;
    unsigned char *orgUV;

    ypos = macroblockInfo->mby*16;
    xpos = macroblockInfo->mbx*16;
    cypos = ypos / 2;
    cxpos = xpos / 2;

    macroblockInfo->planar_flag = 1;

    //
    // Planar prediction for the luminance
    //
    topright    = encoderInfo->recon->Y[16*macroblockInfo->mby-1][16*macroblockInfo->mbx+15];
    bottomleft  = encoderInfo->recon->Y[16*macroblockInfo->mby+15][16*macroblockInfo->mbx-1];
    bottomright = (macroblockInfo->orgY[16*15+15] +
                   macroblockInfo->orgY[16*15+14] + 
                   macroblockInfo->orgY[16*14+15] +
                   macroblockInfo->orgY[16*14+14] + 2) >> 2;

    qdelta = bottomright - macroblockInfo->planar_pred[0];
    sign   = qdelta<0 ? -1 : 1;
    qdelta = abs(qdelta);

    if(qdelta<4)
        qdelta=qdelta; 
    else if(qdelta < 16)
        qdelta = (qdelta>>1)<<1; 
    else if(qdelta < 64)
        qdelta = ((qdelta>>2)<<2)+2; 
    else
        qdelta = ((qdelta>>3)<<3)+4; 

    bottomright = qdelta*sign + macroblockInfo->planar_pred[0];//dcval16;
    macroblockInfo->planar_deltaY = qdelta*sign; 

    get_planar_prediction(upa, lefta, 16, macroblockInfo->planar_deltaY, ypos, xpos, encoderInfo->recon->Y, planar_recY);

    //
    // Planar prediction for the chrominance components
    //
    for(uvindex = 0;uvindex<2;uvindex++)
    {
        recUV = uvindex ? planar_recV : planar_recU;
        orgUV = uvindex ? macroblockInfo->orgV : macroblockInfo->orgU;

        // Calculate a value for the bottom right corner of the block
        bottomright = (orgUV[7*8+7] + orgUV[7*8+6] + orgUV[6*8+7] + orgUV[6*8+6] + 2) >> 2;
        qdelta      = bottomright - macroblockInfo->planar_pred[uvindex+1];
        sign        = qdelta < 0 ? -1 : 1;
        qdelta      = abs(qdelta);

        if(qdelta < 4)
            qdelta = qdelta; 
        else if(qdelta < 16)
            qdelta = (qdelta>>1)<<1; 
        else if(qdelta < 64)
            qdelta = ((qdelta>>2)<<2)+2; 
        else
            qdelta = ((qdelta>>3)<<3)+4; 
      
        bottomright = qdelta*sign + macroblockInfo->planar_pred[uvindex+1];

        // Get the planar prediction for this chrominance component
        if (uvindex){
            macroblockInfo->planar_deltaV = qdelta*sign;
            get_planar_prediction(upa, lefta, 8, macroblockInfo->planar_deltaV, cypos, cxpos, encoderInfo->recon->V, recUV);
        }
        else{
            macroblockInfo->planar_deltaU = qdelta*sign;
            get_planar_prediction(upa, lefta, 8, macroblockInfo->planar_deltaU, cypos, cxpos, encoderInfo->recon->U, recUV);
        }
    }


    sads[0] = sads[1] = sads[2] = 0;

    for (k=0;k<16;k++){
        for (l=0;l<16;l++){
            sads[0] += abs(macroblockInfo->orgY[16*k+l]-planar_recY[16*k+l]);      

        }
    }

    for (k=0;k<8;k++){
        for (l=0;l<8;l++){
            sads[1] += abs(macroblockInfo->orgU[8*k+l]-planar_recU[8*k+l]);
            sads[2] += abs(macroblockInfo->orgV[8*k+l]-planar_recV[8*k+l]);
        }
    }


    is_planar_identical = 0;
    for (k=0;k<16;k++){
        for (l=0;l<16;l++){
            is_planar_identical += abs(macroblockInfo->recY[16*k+l] - planar_recY[16*k+l]);            
        }
    }
    if(is_planar_identical < 1)
        is_planar_identical = 1;
    else
        is_planar_identical = 0;

    if(is_planar_identical)
    {
        macroblockInfo->planar_flag = 0;
        return;
    }


    if( sads[0] < 1024 
        && sads[1] < 256 
        && sads[2] < 256   
        )  
    {
        for (k=0;k<16;k++){
            for (l=0;l<16;l++){
                macroblockInfo->recY[16*k+l] = planar_recY[16*k+l];
            }
        }

        for(uvindex = 0;uvindex<2;uvindex++)
        {
            recUV = uvindex ? macroblockInfo->recV : macroblockInfo->recU;
            recUV2 = uvindex ? planar_recV : planar_recU;

            for (k=0;k<8;k++){
                for (l=0;l<8;l++){
                    recUV[8*k+l] = recUV2[8*k+l];
                }
            }
        }

        macroblockInfo->ycbp = 0;             
        macroblockInfo->ucbp = 0;
        macroblockInfo->vcbp = 0;

    }
    else
        macroblockInfo->planar_flag = 0;

}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_16x16I
Purpose:    Encode an 16x16 luma MB in Intra 16x16 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/
void encode_luma_16x16I(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo){


    int k,l,cbpflag,sad;
    int ypos,xpos,mby,mbx,mbxmax;

    int upa,lefta;
    int best_sad, dcval16=0;
    intrapred best_dir = PRED_VER;         
    unsigned char pblock16[16][16]; //STATIC unsigned char **pblock16;    
    //unsigned char **rblock16;
    short dblock16[16][16]; //STATIC int **dblock16;
    short rdblock16[16][16]; //STATIC int **rdblock16;
    short rblock[8][8]; //STATIC int **rblock;
    //image_frame *curr = encoderInfo->orig;    
    image_frame *recon = encoderInfo->recon;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    int ncoeff;
    int *no_of_coeffs = &ncoeff;
    int sumlev=0;
    int *sum_of_levels = &sumlev;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    //STATIC dblock16 = imatrix(0,15,0,15);
    //STATIC rdblock16 = imatrix(0,15,0,15);
    //STATIC pblock16 = ucmatrix(0,15,0,15);
    //STATIC rblock = imatrix(0,7,0,7);
    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);   

    upa = up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);
    lefta = left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);    

    dcval16 = 0;
    if(upa){         
        for (k=0; k<16; k++){
            dcval16+=recon->Y[ypos-1][xpos+k];
        }
    }
    if(lefta){        
        for (k=0; k<16; k++){
            dcval16+=recon->Y[ypos+k][xpos-1];
        }
    }

    if(upa == 1 && lefta==1)
        dcval16 = (dcval16+16)/32;
    else if(upa==1 || lefta==1)
        dcval16 = (dcval16+8)/16;
    else
        dcval16 = 128;


    sad = 0;
    for (k=0;k<16;k++){
        for (l=0;l<16;l++){                            
            sad += abs(macroblockInfo->orgY[16*k+l]-dcval16);
        }
    }
    best_sad = sad;
    best_dir = PRED_DC;

    if(upa){
        sad = 0;
        for (k=0;k<16;k++){
            for (l=0;l<16;l++){
                sad += abs(macroblockInfo->orgY[16*k+l]-recon->Y[ypos-1][xpos+l]);
            }
        }
        if (sad < best_sad){
            best_sad = sad;
            best_dir = PRED_VER;
        }
    }


    if(lefta){
        sad = 0;
        for (k=0;k<16;k++){
            for (l=0;l<16;l++){
                sad += abs(macroblockInfo->orgY[16*k+l]-recon->Y[ypos+k][xpos-1]);
            }
        }
        if (sad < best_sad){
            best_sad = sad;
            best_dir = PRED_HOR;
        }
    }
    macroblockInfo->modelist = best_dir;
    get_luma_intra_prediction_16x16(encoderInfo,frameInfo,macroblockInfo,best_dir, pblock16, dcval16);


    {
        for (k=0;k<16;k++){
            for (l=0;l<16;l++){
                dblock16[k][l] = macroblockInfo->orgY[16*k+l]-pblock16[k][l];
            }
        }
        transform_16x16(dblock16,macroblockInfo->ycoeff[0]);

        {
            float lambda = pic_type==I_PICTURE ? params->lambda_coeff_quantIntraI : params->lambda_coeff_quantIntraPB;

            cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[0],frameInfo->qp,BLOCK_Y16x16I,lambda,TR_16x16);

        }


        if(upa==0 && lefta==0){
            if (cbpflag){
                macroblockInfo->side_intra_y = macroblockInfo->ycoeff[0][0][0]; 
                macroblockInfo->ycoeff[0][0][0] = 0;
                if(macroblockInfo->side_intra_y != 0){
                    (*no_of_coeffs)--;
                    (*sum_of_levels) -= macroblockInfo->side_intra_y;
                    cbpflag = 0;
                    for (k=0; k<8; k++){
                        for(l=0;l<8; l++){
                            if (macroblockInfo->ycoeff[0][k][l] != 0)
                                cbpflag = 1;
                        }
                    }
                }
                macroblockInfo->ycoeff[0][0][0] = macroblockInfo->side_intra_y ;
                dequantize_8x8(macroblockInfo->ycoeff[0],rblock,frameInfo->qp,0);
                inverse_transform_16x16(rblock,rdblock16,frameInfo->qp,0);
                for (k=0;k<16;k++){
                    for (l=0;l<16;l++){
                        macroblockInfo->recY[16*k+l] = clip3(rdblock16[k][l]+pblock16[k][l]);
                    }
                }
                macroblockInfo->ycoeff[0][0][0] = 0;
            }
            else{
                macroblockInfo->side_intra_y = 0;
                for (k=0;k<16;k++){
                    memcpy(&macroblockInfo->recY[16*k],pblock16[k],16*sizeof(unsigned char));
                }
            }
        }
        else{
            if (cbpflag){                
                dequantize_8x8(macroblockInfo->ycoeff[0],rblock,frameInfo->qp,0);
                inverse_transform_16x16(rblock,rdblock16,frameInfo->qp,0);
                for (k=0;k<16;k++){
                    for (l=0;l<16;l++){
                        macroblockInfo->recY[16*k+l] = clip3(rdblock16[k][l]+pblock16[k][l]);
                    }
                }
            }        
            else{            
                for (k=0;k<16;k++){
                    memcpy(&macroblockInfo->recY[16*k],pblock16[k],16*sizeof(unsigned char));
                }
            }        
        }

        macroblockInfo->ycbp = cbpflag * 0xf;        
        macroblockInfo->sum_of_levels = *sum_of_levels;

        macroblockInfo->check_planar = 0;
        if(macroblockInfo->final_encode_flag == 1 && upa == 1 && lefta == 1 && frameInfo->pic_type==I_PICTURE)
        { 
            int all_zero = 1;

            int ii,jj;

            for(ii=0;ii<8;ii++)
            {
                for(jj=0;jj<8;jj++)
                {
                    if(ii==0 && jj == 0)
                        continue;
                    if(abs(macroblockInfo->ycoeff[0][ii][jj]) > 0)
                    {
                        all_zero = 0;
                        break;
                    }
                }
            }

            if(cbpflag == 0)
                all_zero = 1;

            {
                short dblock16_temp[16][16]; //STATIC int **dblock16_temp;
                short ycoeff_temp[8][8]; //STATIC int **ycoeff_temp;
                double hp_energy, ENERGY_THR_HP;

                ENERGY_THR_HP = 30000;
                //STATIC dblock16_temp = imatrix(0,15,0,15);
                //STATIC ycoeff_temp = imatrix(0,7,0,7);
                for (k=0;k<16;k++){
                    for (l=0;l<16;l++){
                        dblock16_temp[k][l] = macroblockInfo->orgY[16*k+l]-dcval16;
                    }
                }
                transform_16x16(dblock16_temp,ycoeff_temp);

                hp_energy = 0;
                for (k=4;k<8;k++){
                    for (l=4;l<8;l++){
                        hp_energy += ycoeff_temp[k][l]*ycoeff_temp[k][l];              
                    }
                }

                hp_energy = 16.0*hp_energy; //To compensate for new scaling in transform

                if(hp_energy > ENERGY_THR_HP)
                    all_zero = 0;

                //STATIC free_imatrix(ycoeff_temp,0,7,0,7);
                //STATIC free_imatrix(dblock16_temp,0,15,0,15);            
            }

            //Check edges
            if(all_zero)
            {
                for(ii=1;ii<16;ii++)
                {
                    if( abs(encoderInfo->recon->Y[16*macroblockInfo->mby-1][16*macroblockInfo->mbx+ii] - encoderInfo->recon->Y[16*macroblockInfo->mby-1][16*macroblockInfo->mbx+ii-1]) > EDGE_THR)
                        all_zero = 0;

                    if( abs(encoderInfo->recon->Y[16*macroblockInfo->mby+ii][16*macroblockInfo->mbx-1] - encoderInfo->recon->Y[16*macroblockInfo->mby+ii-1][16*macroblockInfo->mbx-1]) > EDGE_THR)
                        all_zero = 0;              
                }            
            }          
            macroblockInfo->check_planar = all_zero;
            macroblockInfo->planar_pred[0] = dcval16;
        }

    }


    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_imatrix(dblock16,0,15,0,15);
    //STATIC free_ucmatrix(pblock16,0,15,0,15);
    //STATIC free_imatrix(rdblock16,0,15,0,15);
    //STATIC free_imatrix(rblock,0,7,0,7);

}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_4x4I
Purpose:    Encode an 16x16 luma MB in Intra 4x4 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/


void encode_luma_4x4I(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{
    unsigned char up_row[8],left_col[8];
    int m,n,k,l,cbpflag;
    int ypos,xpos,mby,mbx,mbxmax;         
    int index_4x4,index_8x8;

    int dcval=0,upa,lefta,best_cost,intra_rdo,cand_cost,side_intra;
    int sad_dc,sad_ver,sad_hor,num_bits;
    intrapred best_dir = PRED_VER;
    intrapred dir = PRED_VER;
    intrapred cand_dir; 
    stream *str_tmp;

    unsigned char pblock[4][4]; //STATIC    unsigned char **rblock8,**pblock;
    short rdblock8[4][4],rcoeff8[4][4],dblock8[4][4]; //STATIC int **rdblock8,**dblock8,**rcoeff8;

    //image_frame *curr = encoderInfo->orig;    
    image_frame *recon = encoderInfo->recon;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    stream *str = encoderInfo->str;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    //STATIC rblock8 = ucsubmatrix(recon->Y,0,3,0,3,0,0);
    //STATIC dblock8 = imatrix(0,3,0,3);
    //STATIC rdblock8 = imatrix(0,3,0,3);
    //STATIC rcoeff8 = imatrix(0,3,0,3);
    //STATIC pblock = ucmatrix(0,3,0,3);


    //disabled in r109v1.03 intra_rdo = params->rdo;
    intra_rdo = 0; 
    str_tmp = NULL;
    if (intra_rdo) str_tmp = initbits(0);   

    for (k=0;k<16;k++)
        macroblockInfo->modelist4[k] = 0;

    for (k=0;k<4;k++)
        macroblockInfo->ycbp4[k] = 0;

    for (m=0;m<16;m+=4){
        for (n=0;n<16;n+=4){
            index_4x4 = m + n/4;
            index_8x8 = 2*(m/8) + (n/8);  
            upa = up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,m,n,params);
            lefta = left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,m,n,params);


            /* Intra DC prediction value */            
            dcval = 0;                
            if (upa){
                for (k=0; k<4; k++){
                    up_row[k] = recon->Y[ypos+m-1][xpos+n+k];
                    dcval+=recon->Y[ypos+m-1][xpos+n+k];
                }
            }
            else{                    
                for (k=0; k<4; k++)
                    up_row[k] = 128;
            }                
            if (lefta){
                for (k=0; k<4; k++){
                    left_col[k] = recon->Y[ypos+m+k][xpos+n-1];
                    dcval+=recon->Y[ypos+m+k][xpos+n-1];
                }
            }
            else{                    
                for (k=0; k<4; k++)
                    left_col[k] = 128;     
            }
            if(upa == 1 && lefta==1)
                dcval = (dcval+4)/8;
            else if(upa==1 || lefta==1)
                dcval = (dcval+2)/4;
            else
                dcval = 128;

            if (!intra_rdo){
                sad_dc = sad_ver = sad_hor = 0;
                for (k=0;k<4;k++){
                    for (l=0;l<4;l++){
                        sad_dc += abs(macroblockInfo->orgY[16*(m+k)+n+l]-dcval);
                    }
                }
                if (mby == 0 && m == 0){
                    sad_ver = 0xfffffff;
                }
                else{
                    for (k=0;k<4;k++){
                        for (l=0;l<4;l++){
                            sad_ver += abs(macroblockInfo->orgY[16*(m+k)+n+l]-up_row[l]);
                        }
                    }
                }
                if (mbx == 0 && n == 0){
                    sad_hor = 0xfffffff;
                }
                else{ 
                    for (k=0;k<4;k++){
                        for (l=0;l<4;l++){
                            sad_hor += abs(macroblockInfo->orgY[16*(m+k)+n+l]-left_col[k]);
                        }
                    }
                }
                dir = (sad_dc < sad_ver && sad_dc < sad_hor) ? PRED_DC : (sad_ver <= sad_hor ? PRED_VER : PRED_HOR);
            } // if !intra_rdo


            best_cost = 0xfffffff;
            /* If intra RDO is enabled:
            *   - Code each intra mode and choose the mode having lowest RD-cost
            *   - Code the chosen intra mode in the final loop iteration
            * else (intra RDO is not enabled):
            *   - Cmopute SAD for all intra modes and code the mode having the lowest SAD
            */
            for (cand_dir = PRED_VER; cand_dir <= (intra_rdo?PRED_COUNT:0); cand_dir++){
                /* Intra prediction */
                if (intra_rdo){

                    if(macroblockInfo->final_encode_flag)
                    {
                        cand_dir = PRED_COUNT;
                        best_dir = macroblockInfo->modelist4[index_4x4];
                    }
                    if ((cand_dir == PRED_VER && upa == 0) || 
                        (cand_dir == PRED_HOR && lefta == 0)){                        
                            continue; /* Prediction direction is not available */
                    }

                    if (cand_dir>PRED_DC && cand_dir<PRED_COUNT) continue; //No diagonal prediction modes for 4x4T


                    /* Code current intra mode or the best intra mode */
                    dir = (cand_dir < PRED_COUNT) ? cand_dir : best_dir;

                }
                get_luma_intra_prediction_4x4(up_row, left_col, dcval, dir, pblock);

                for (k=0;k<4;k++){
                    for (l=0;l<4;l++){
                        dblock8[k][l] = macroblockInfo->orgY[16*(m+k)+n+l]-pblock[k][l];
                    }
                }          
                transform_4x4(dblock8,macroblockInfo->ycoeff[index_4x4]);

                {
                    float lambda = pic_type==I_PICTURE ? params->lambda_coeff_quantIntraI : params->lambda_coeff_quantIntraPB;

                    cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[index_4x4],frameInfo->qp,0,lambda,TR_4x4);

                }

                if (index_4x4 == 0 && upa==0 && lefta==0){
                    side_intra = macroblockInfo->ycoeff[index_4x4][0][0];
                    macroblockInfo->side_intra_y = side_intra;
                    macroblockInfo->ycoeff[index_4x4][0][0] = 0;
                    if(side_intra){
                        cbpflag = 0;
                        for (k=0; k<4; k++){
                            for(l=0;l<4; l++){
                                if (macroblockInfo->ycoeff[index_4x4][k][l] != 0)
                                    cbpflag = 1;
                            }
                        }
                    }
                }



                //STATIC move_ucsubmatrix(rblock8,recon->Y,ypos+m,ypos+3+m,xpos+n,xpos+3+n,0,0);
                if (cbpflag || (upa==0 && lefta==0)){
                    if ((upa==0 && lefta==0)) 
                        macroblockInfo->ycoeff[index_4x4][0][0] = side_intra; 
                    dequantize_NxN(macroblockInfo->ycoeff[index_4x4],rcoeff8,frameInfo->qp,0,4);
                    inverse_transform_4x4(rcoeff8,rdblock8,frameInfo->qp);
                    if ((upa==0 && lefta==0)) 
                        macroblockInfo->ycoeff[index_4x4][0][0] = 0;
                    for (k=0;k<4;k++){
                        for (l=0;l<4;l++){
                            macroblockInfo->recY[16*(m+k)+n+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                        }
                    }
                }
                else{
                    for (k=0;k<4;k++){
                        memcpy(&macroblockInfo->recY[16*(m+k)+n],&pblock[k][0],4*sizeof(unsigned char));
                    }
                }



                if (intra_rdo && cand_dir < PRED_COUNT){
                    num_bits = 0;
                    if (index_4x4 == 0 && upa==0 && lefta==0){
                        encode_side_intra(side_intra,str_tmp,&num_bits,0);
                        macroblockInfo->ycoeff[index_4x4][0][0] = 0;
                    }

                    cbpflag = 0;                    
                    for (k=0; k<4; k++){
                        for (l=0; l<4; l++){                    
                            if (macroblockInfo->ycoeff[index_4x4][k][l] != 0)
                                cbpflag = 1;
                        }
                    }


                    if (cbpflag){                        
                        encode_coeffs4(encoderInfo,macroblockInfo->ycoeff[index_4x4], str_tmp, &num_bits,0,0);                         
                    }

                    {
                        unsigned char m4[4*4],r4[4*4];
                        for (k=0;k<4;k++){ 
                            memcpy(&m4[4*k],&macroblockInfo->orgY[16*(m+k)+n],4*sizeof(unsigned char));
                            memcpy(&m4[4*k],&macroblockInfo->recY[16*(m+k)+n],4*sizeof(unsigned char));                            
                        }

                        cand_cost = cost_block(m4,r4,4,4,num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm,0);                        

                    }                   

                    if (cand_cost < best_cost){                        
                        best_cost = cand_cost;
                        best_dir = cand_dir;
                    }

                }                
            } // for cand_dir

            macroblockInfo->ycbp |= (cbpflag << (3-index_8x8));
            macroblockInfo->ycbp4[index_8x8] |= (cbpflag << (3-( (m&4)/2 + (n&4)/4  )));

            macroblockInfo->modelist4[index_4x4] = dir;            
            if (str->outfile){ /* Collect statistics only when writing to bitsstream file */            
                direction_counter[dir]++;
            }

            /* Store edge pixels for forming predictions at neighbour blocks */                     
            if (m<12){
                memcpy(&recon->Y[ypos+m+3][xpos+n+0],&macroblockInfo->recY[16*(m+3)+n],4*sizeof(unsigned char));
            }           
            if (n<12){
                for (k=0;k<4;k++){               
                    recon->Y[ypos+m+k][xpos+n+3] = macroblockInfo->recY[16*(m+k)+n+3];               
                }
            }
        } //for n
    } // for m


    //STATIC free_ucmatrix(pblock,0,3,0,3);
    //STATIC free_imatrix(rdblock8,0,3,0,3);
    //STATIC free_imatrix(rcoeff8,0,3,0,3);
    //STATIC free_imatrix(dblock8,0,3,0,3);
    //STATIC free_ucsubmatrix(rblock8,0,3,0,3);    

    if (str_tmp != NULL) closebits(str_tmp);    
}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_4x4P
Purpose:    Encode an 16x16 luma MB in Inter 4x4 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_luma_4x4P(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)                        
{
    int m,n,i,j,k,l,cbpflag,cbp8;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax;         
    int vyoff,vxoff,index_8x8,index_4x4,large_level;

    unsigned char pblock[8][8]; //STATIC unsigned char **pblock;
    //STATIC int **rdblock8, **dblock8;
    //STATIC int **rcoeff8;


    //image_frame *curr = encoderInfo->orig;    
    //image_frame *recon = encoderInfo->recon;
    image_frame **iprev = encoderInfo->ref_frame_qp;

    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;

    short dblock4[4][4], rdblock4[4][4],rcoeff4[4][4]; //STATIC int **dblock4,**rdblock4,**rcoeff4;
    //STATIC rdblock4 = imatrix(0,3,0,3);
    //STATIC dblock4 = imatrix(0,3,0,3);
    //STATIC rcoeff4 = imatrix(0,3,0,3);

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    vy = mby*2;
    vx = mbx*2;    

    //STATIC pblock = ucmatrix(0,7,0,7);    


    large_level = 0;


    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rcoeff8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);



    for (m=0;m<16;m+=8){
        vyoff = m ? 1 : 0;
        for (n=0;n<16;n+=8){
            vxoff = n ? 1 : 0;
            index_8x8 = m/4+n/8;            


            get_luma_inter_prediction_8x8(macroblockInfo,vyoff, vxoff, iprev, pblock);            

            cbp8 = 0;
            macroblockInfo->ycbp4[index_8x8] = 0;
            for (k=0;k<8;k+=4){
                for (l=0;l<8;l+=4){

                    for (i=0;i<4;i++){
                        for (j=0;j<4;j++){
                            dblock4[i][j] = macroblockInfo->orgY[16*(m+k+i)+n+l+j]-pblock[k+i][l+j]; 
                        }
                    }
                    index_4x4 = m + k + n/4 + l/4;
                    transform_4x4(dblock4,macroblockInfo->ycoeff[index_4x4]);

                    {
                        float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;
                        int block_type = (int)pic_type;

                        cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[index_4x4],frameInfo->qp,block_type,lambda,TR_4x4);


                    }


                    if (cbpflag){
                        dequantize_NxN(macroblockInfo->ycoeff[index_4x4],rcoeff4,frameInfo->qp,0,4);
                        inverse_transform_4x4(rcoeff4,rdblock4,frameInfo->qp);
                        for (i=0;i<4;i++){
                            for (j=0;j<4;j++){
                                macroblockInfo->recY[16*(m+k+i)+n+l+j] = clip3(rdblock4[i][j]+pblock[k+i][l+j]);
                            }
                        }
                    }
                    else{
                        for (i=0;i<4;i++){
                            memcpy(&macroblockInfo->recY[16*(m+k+i)+n+l],&pblock[k+i][l],4*sizeof(unsigned char));
                        }
                    }
                    macroblockInfo->ycbp4[index_8x8] |=  (cbpflag << (3-(k/2 + l/4)));
                    if (cbpflag) cbp8 = 1;
                } //for l=
            } //for k=
            frameInfo->mvmap[(vy+vyoff)*2*mbxmax+vx+vxoff].coded_y = cbp8;                       
            macroblockInfo->ycbp |= (cbp8 << (3-index_8x8));
        } //for n
    } // for m





    //STATIC free_imatrix(rdblock4,0,3,0,3);
    //STATIC free_imatrix(dblock4,0,3,0,3);
    //STATIC free_imatrix(rcoeff4,0,3,0,3);
    //STATIC free_ucmatrix(pblock,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);    
}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_8x8P
Purpose:    Encode an 16x16 luma MB in Inter 8x8 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_luma_8x8P(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)                        
{
    int m,n,k,l,cbpflag;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax;         
    int vyoff,vxoff,index_8x8,large_level;

    unsigned char pblock[8][8]; //STATIC unsigned char **pblock;
    short rdblock8[8][8],dblock8[8][8]; //STATIC int **rdblock8,**dblock8;
    short rcoeff8[8][8]; //STATIC int **rcoeff8;



    //image_frame *curr = encoderInfo->orig;    
    //image_frame *recon = encoderInfo->recon;
    image_frame **iprev = encoderInfo->ref_frame_qp;

    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;



    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    vy = mby*2;
    vx = mbx*2;    

    //STATIC pblock = ucmatrix(0,7,0,7);    

    large_level = 0;    

    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rcoeff8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);


    for (m=0;m<16;m+=8){
        vyoff = m ? 1 : 0;
        for (n=0;n<16;n+=8){
            vxoff = n ? 1 : 0;
            index_8x8 = m/4+n/8;            
            get_luma_inter_prediction_8x8(macroblockInfo,vyoff, vxoff, iprev, pblock);            

            for (k=0;k<8;k++)
            {
                const unsigned char *orgY = &macroblockInfo->orgY[16*(m+k)+n];
                for (l=0;l<8;l++)
                {                                        
                    dblock8[k][l] = orgY[l]-pblock[k][l];
                }
            }                       
            transform_8x8(dblock8,macroblockInfo->ycoeff[index_8x8]);

            {
                float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;
                int block_type = pic_type==P_PICTURE ? BLOCK_Y8x8P : BLOCK_Y8x8B;

                cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[index_8x8],frameInfo->qp,block_type,lambda,TR_8x8);                

            }


            if (cbpflag){
                dequantize_8x8(macroblockInfo->ycoeff[index_8x8],rcoeff8,frameInfo->qp,0);
                inverse_transform_8x8(rcoeff8,rdblock8,frameInfo->qp,0);
                for (k=0;k<8;k++){
                    for (l=0;l<8;l++){
                        macroblockInfo->recY[16*(m+k)+n+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                    }
                }
            }
            else{
                for (k=0;k<8;k++){
                    memcpy(&macroblockInfo->recY[16*(m+k)+n],&pblock[k][0],8*sizeof(unsigned char));
                }
            }
            frameInfo->mvmap[(vy+vyoff)*2*mbxmax+vx+vxoff].coded_y = cbpflag;

            macroblockInfo->ycbp |= (cbpflag << (3-index_8x8));


        } //for n
    } // for m





    //STATIC free_ucmatrix(pblock,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);   
}

/*-----------------------------------------------------------------------------------------
Function:   encode_chroma_8x8P
Purpose:    Encode an 16x16 chroma MB in Inter 8x8 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_chroma_8x8P(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)                        

{
    int k,l,cbpflag;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax,cypos,cxpos;
    unsigned char pblock[8][8]; //STATIC unsigned char **pblock;
    short rdblock8[8][8],dblock8[8][8],rcoeff8[8][8]; //STATIC int **rdblock8,**dblock8,**rcoeff8;       
    //image_frame *curr = encoderInfo->orig;    
    //image_frame *recon = encoderInfo->recon;
    image_frame **iprev = encoderInfo->ref_frame_qp;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    int rounding,uv;
    //STATIC int **coeff;


    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;
    cypos = ypos / 2;
    cxpos = xpos / 2;

    vy = mby*2;
    vx = mbx*2;    

    //STATIC pblock = ucmatrix(0,7,0,7);    
    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);
    //STATIC rcoeff8 = imatrix(0,7,0,7);

    if (pic_type == B_PICTURE){
        rounding = 1;
    }
    else{
        rounding = (params->frame_num)&1;
    }		

    for (uv=0;uv<2;uv++){
        unsigned char *orgUV = uv ? macroblockInfo->orgV : macroblockInfo->orgU;
        unsigned char *recUV = uv ? macroblockInfo->recV : macroblockInfo->recU;
        short coeff[8][8];
        //STATIC coeff = uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff; 

        get_chroma_inter_prediction_mb(macroblockInfo, iprev, pblock, uv, rounding);

        for (k=0;k<8;k++){
            for (l=0;l<8;l++){                
                dblock8[k][l] = orgUV[8*k+l]-pblock[k][l];
            }
        }
        transform_8x8(dblock8,coeff);


        {
            int qp = frameInfo->qp - corr_qpc_8x8[frameInfo->qp];        
            float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;

            cbpflag = quantize_NxN_RDOQ(encoderInfo,coeff,qp,uv ? BLOCK_V : BLOCK_U,lambda,TR_8x8);   

        }






        if (uv){

            macroblockInfo->vcbp = cbpflag;
        }
        else{

            macroblockInfo->ucbp = cbpflag;
        }
        if (cbpflag){            
            dequantize_8x8(coeff,rcoeff8,frameInfo->qp,1);
            inverse_transform_8x8(rcoeff8,rdblock8,frameInfo->qp,1);
            for (k=0;k<8;k++){
                for (l=0;l<8;l++){
                    recUV[8*k+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                }
            }
        }
        else{
            for (k=0;k<8;k++){
                memcpy(&recUV[8*k],&pblock[k][0],8*sizeof(unsigned char));
            }
        }
        memcpy(uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff, coeff, 8*8*sizeof(short));       
    } //uv

    //STATIC free_ucmatrix(pblock,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);
}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_8x8I
Purpose:    Encode an 16x16 luma MB in Intra 8x8 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_luma_8x8I(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{

    int uprighta,downlefta;

    int intraPredAngle, intraPredAngleBest = 0, intraPredAngleRange, tmpLen, tmpCode;
    unsigned char up_rowF[17],left_colF[17],upleftF;

    //int mbymax = encoderInfo->heightInMBs;

    int m,n,k,l,cbpflag;
    int ypos,xpos,mby,mbx,mbxmax;         
    int index_8x8;

    int dcval=0,upa,lefta,best_cost,intra_rdo,cand_cost,side_intra;
    int num_bits;
    intrapred best_dir = PRED_VER;
    intrapred dir = PRED_VER;
    intrapred cand_dir; 
    stream *str_tmp;

    unsigned char pblock[8][8]; //STATIC unsigned char **rblock8,**pblock;
    short rdblock8[8][8],dblock8[8][8],rcoeff8[8][8]; //STATIC int **rdblock8,**dblock8,**rcoeff8;

    //image_frame *curr = encoderInfo->orig;    
    image_frame *recon = encoderInfo->recon;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    stream *str = encoderInfo->str;


    upleftF = 0;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    //STATIC rblock8 = ucsubmatrix(recon->Y,0,7,0,7,0,0);

    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);
    //STATIC rcoeff8 = imatrix(0,7,0,7);

    //STATIC pblock = ucmatrix(0,7,0,7);


    intra_rdo = (pic_type == I_PICTURE);
    str_tmp = NULL;
    if (intra_rdo) str_tmp = initbits(0);   

    macroblockInfo->modelist = 0;

    for (m=0;m<16;m+=8){       
        for (n=0;n<16;n+=8){           
            index_8x8 = m/4+n/8;

            upa = up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,m,n,params);
            lefta = left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,m,n,params);

            uprighta = upright_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,m,n,params);

            downlefta = (lefta && m==0 && n==0);

            /* Intra DC prediction value */
            dcval = 0;                
            if (upa){
                if (m>0)
                    for (k=0; k<8; k++) 
                        dcval+=macroblockInfo->recY[16*(m-1)+n+k];
                else
                    for (k=0; k<8; k++) 
                        dcval+=recon->Y[ypos+m-1][xpos+n+k];                
            }           
            if (lefta){
                for (k=0; k<8; k++)
                    dcval+=recon->Y[ypos+m+k][xpos+n-1];               
            }            



            get_filtered_intrapred_ref_samples(lefta, upa, uprighta, downlefta, ypos+m, xpos+n, recon->Y, left_colF, up_rowF);


            if(upa == 1 && lefta==1)
                dcval = (dcval+8)/16;
            else if(upa==1 || lefta==1)
                dcval = (dcval+4)/8;
            else
                dcval = 128;


            if (!intra_rdo){

                int sad, sadmin = 0xfffffff;

                for (cand_dir = PRED_VER; cand_dir < PRED_COUNT; cand_dir++){

                    if ((cand_dir == PRED_VER && upa == 0) || 
                        (cand_dir == PRED_HOR && lefta == 0) ||
                        (cand_dir == PRED_ANGULAR)){
                            continue; /* Prediction direction is not available */
                    }

                    intraPredAngleRange = cand_dir == PRED_DC ? 0 : 8;

                    for (intraPredAngle = -intraPredAngleRange; intraPredAngle <= intraPredAngleRange; intraPredAngle++)
                    {
                        get_directional_intrapred_8x8(up_rowF, left_colF, dcval, cand_dir, intraPredAngle, pblock);
                        sad = SAD_8x8(&macroblockInfo->orgY[16*m+n],&pblock[0][0],16,8);
                        if (sad < sadmin){
                            sadmin             = sad;
                            dir                = cand_dir;
                            intraPredAngleBest = intraPredAngle;
                        }
                    }
                }                
            } // if !intra_rdo


            best_cost = 0xfffffff;
            /* If intra RDO is enabled:
            *   - Code each intra mode and choose the mode having lowest RD-cost
            *   - Code the chosen intra mode in the final loop iteration
            * else (intra RDO is not enabled):
            *   - Cmopute SAD for all intra modes and code the mode having the lowest SAD
            */

            for (cand_dir = PRED_VER; cand_dir <= (intra_rdo?PRED_COUNT:0); cand_dir++){


                /* Intra prediction */
                if (intra_rdo){

                    if(macroblockInfo->final_encode_flag)
                    {
                        cand_dir           = PRED_COUNT;
                        best_dir           = macroblockInfo->modelist8[index_8x8];
                        intraPredAngleBest = macroblockInfo->intraPredAngle8[index_8x8];
                    }

                    if ((cand_dir == PRED_VER && upa == 0) || 
                        (cand_dir == PRED_HOR && lefta == 0) ||
                        (cand_dir == PRED_ANGULAR)){
                            continue; /* Prediction direction is not available */
                    }

                    /* Code current intra mode or the best intra mode */
                    dir = (cand_dir < PRED_COUNT) ? cand_dir : best_dir;                    
                }

                if (intra_rdo && cand_dir < PRED_COUNT){
                    intraPredAngleRange = dir == PRED_DC ? 0 : 8;
                    intraPredAngle      = -intraPredAngleRange;
                }
                else{
                    intraPredAngleRange = intraPredAngleBest;
                    intraPredAngle      = intraPredAngleBest;
                }

                for (; intraPredAngle <= intraPredAngleRange; intraPredAngle++)
                {

                    get_directional_intrapred_8x8(up_rowF, left_colF, dcval, dir, intraPredAngle, pblock);

                    for (k=0;k<8;k++){
                        for (l=0;l<8;l++){
                            dblock8[k][l] = macroblockInfo->orgY[16*(m+k)+n+l]-pblock[k][l];
                        }
                    }          
                    transform_8x8(dblock8,macroblockInfo->ycoeff[index_8x8]);

                    {
                        float lambda = pic_type==I_PICTURE ? params->lambda_coeff_quantIntraI : params->lambda_coeff_quantIntraPB;

                        cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[index_8x8],frameInfo->qp,BLOCK_Y8x8I,lambda,TR_8x8);

                    }

                    if (index_8x8 == 0 && upa==0 && lefta==0){
                        if (!cbpflag){
                            for (k=0; k<8; k++)
                                memset(macroblockInfo->ycoeff[index_8x8][k],0,8*sizeof(short));
                        }
                        side_intra = macroblockInfo->ycoeff[index_8x8][0][0];
                        macroblockInfo->side_intra_y = side_intra;
                        macroblockInfo->ycoeff[index_8x8][0][0] = 0;
                        if(side_intra){
                            cbpflag = 0;
                            for (k=0; k<8; k++){
                                for(l=0;l<8; l++){
                                    if (macroblockInfo->ycoeff[index_8x8][k][l] != 0)
                                        cbpflag = 1;
                                }
                            }
                        }
                    }


                    //STATIC move_ucsubmatrix(rblock8,recon->Y,ypos+m,ypos+7+m,xpos+n,xpos+7+n,0,0);
                    if (cbpflag || (upa==0 && lefta==0)){
                        if ((upa==0 && lefta==0)) 
                            macroblockInfo->ycoeff[index_8x8][0][0] = side_intra;                   
                        dequantize_8x8(macroblockInfo->ycoeff[index_8x8],rcoeff8,frameInfo->qp,0);
                        inverse_transform_8x8(rcoeff8,rdblock8,frameInfo->qp,0);
                        if ((upa==0 && lefta==0)) 
                            macroblockInfo->ycoeff[index_8x8][0][0] = 0; 
                        for (k=0;k<8;k++){
                            for (l=0;l<8;l++){                                                                    
                                macroblockInfo->recY[16*(m+k)+n+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                            }
                        }                            
                    }
                    else{
                        for (k=0;k<8;k++){                        
                            memcpy(&macroblockInfo->recY[16*(m+k)+n],&pblock[k][0],8*sizeof(unsigned char));                        
                        }
                    }                
                    if (intra_rdo && cand_dir < PRED_COUNT){
                        num_bits = 0;
                        if (index_8x8 == 0 && upa==0 && lefta==0){
                            encode_side_intra(side_intra,str_tmp,&num_bits,0);
                            macroblockInfo->ycoeff[index_8x8][0][0] = 0;
                        }
                        else{
                            if (!cbpflag){
                                for (k=0; k<8; k++)
                                    memset(macroblockInfo->ycoeff[index_8x8][k],0,8*sizeof(short));
                            }
                        }

                        cbpflag = 0;                    
                        for (k=0; k<8; k++){
                            for (l=0; l<8; l++){                    
                                if (macroblockInfo->ycoeff[index_8x8][k][l] != 0)
                                    cbpflag = 1;
                            }
                        }

                        if (cbpflag){					
                            encode_coeffs(encoderInfo,macroblockInfo->ycoeff[index_8x8], str_tmp, &num_bits,BLOCK_Y8x8I,0);
                        }

                        intraPredBits(dir, intraPredAngle, &tmpLen, &tmpCode);
                        num_bits += tmpLen;

                        {
                            unsigned char m8[8*8],r8[8*8];
                            for (k=0;k<8;k++){ 
                                memcpy(&m8[8*k],&macroblockInfo->orgY[16*(m+k)+n],8*sizeof(unsigned char));
                                memcpy(&r8[8*k],&macroblockInfo->recY[16*(m+k)+n],8*sizeof(unsigned char));                            
                            }

                            cand_cost = cost_block(m8,r8,8,8,num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm,MODE_8x8);                        

                        }  


                        if (cand_cost < best_cost){
                            best_cost          = cand_cost;
                            best_dir           = cand_dir;
                            intraPredAngleBest = intraPredAngle;
                        }                    
                    }
                }
            }

            macroblockInfo->ycbp |= (cbpflag << (3-index_8x8));                                    
            macroblockInfo->modelist8[index_8x8] = dir;

            macroblockInfo->intraPredAngle8[index_8x8] = intraPredAngleBest;


            if (str->outfile){ /* Collect statistics only when writing to bitsstream file */            
                direction_counter[dir]++;
            }

            /* Store edge pixels for forming predictions at neighbour blocks */
            if (m==0){                
                memcpy(&recon->Y[ypos+m+7][xpos+n],&macroblockInfo->recY[16*7+n],8*sizeof(unsigned char));
            }
            if (n==0){                
                for (k=0;k<8;k++)
                    recon->Y[ypos+m+k][xpos+n+7] = macroblockInfo->recY[16*(m+k)+7];
                //rblock8[k][7] = macroblockInfo->recY[16*(m+k)+7];
            }
        } //for n
    } // for m

    //STATIC free_ucmatrix(pblock,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_ucsubmatrix(rblock8,0,7,0,7);    

    if (str_tmp != NULL) closebits(str_tmp);    
}

/*-----------------------------------------------------------------------------------------
Function:   encode_chroma_8x8I
Purpose:    Encode an 16x16 chroma MB in Intra 8x8 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_chroma_8x8I(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{
    unsigned char ucol[8],vcol[8],urow[8],vrow[8];
    int i,j,k,l,cbpflag,upa,lefta,uv,side_intra,ypos,xpos,mby,mbx,vy,vx,mbxmax;         

    int dcval=0,best_cost,intra_rdo,cand_cost=0;
    intrapred cand_dir,best_dir;
    int sad_dc,sad_ver,sad_hor,num_bits;//,dcval16=0;
    int no_of_coeffs=0,sum_of_levels=0;
    int dcvalu,dcvalv,cxpos,cypos;
    int coef_scan_uv[2][64];
    int result_coef[64];

    intrapred dir = PRED_VER;
    stream *str_tmp;

    unsigned char pblock[8][8]; //STATIC unsigned char /***rblock8,*/**pblock;
    short rdblock8[8][8],dblock8[8][8],rcoeff8[8][8]; //STATIC int **rdblock8,**dblock8,**rcoeff8;
    //STATIC int **coeff;
    //STATIC unsigned char **cblock = ucsubmatrix(encoderInfo->orig->U,0,7,0,7,0,0);

    image_frame *recon = encoderInfo->recon;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    int qp = frameInfo->qp - corr_qpc_8x8[frameInfo->qp];   


    int uprighta, intraPredAngle, intraPredDirection, sad_dir;
    unsigned char ucolD[17],vcolD[17],urowD[17],vrowD[17];


    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    vy = mby*2;
    vx = mbx*2;
    cypos = ypos / 2;
    cxpos = xpos / 2;       

    //rblock8 = ucsubmatrix(recon->Y,0,7,0,7,0,0);

    //STATIC dblock8 = imatrix(0,7,0,7);
    //STATIC rdblock8 = imatrix(0,7,0,7);
    //STATIC rcoeff8 = imatrix(0,7,0,7);
    //STATIC pblock = ucmatrix(0,7,0,7);

    intra_rdo = (pic_type == I_PICTURE);
    str_tmp = NULL;
    if (intra_rdo) str_tmp = initbits(0);   


    lefta = left_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);
    upa = up_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,0,params);


    intraPredDirection = -1;
    intraPredAngle     = 0;

    if (macroblockInfo->tr_size==TR_8x8 && (lefta || upa)){
        if (macroblockInfo->modelist8[0] < PRED_DC &&
            macroblockInfo->intraPredAngle8[0] != 0){
                intraPredDirection = macroblockInfo->modelist8[0];
                intraPredAngle     = macroblockInfo->intraPredAngle8[0];
        }
        else if (macroblockInfo->modelist8[1] < PRED_DC &&
            macroblockInfo->intraPredAngle8[1] != 0){
                intraPredDirection = macroblockInfo->modelist8[1];
                intraPredAngle     = macroblockInfo->intraPredAngle8[1];
        }
        else if (macroblockInfo->modelist8[2] < PRED_DC &&
            macroblockInfo->intraPredAngle8[2] != 0){
                intraPredDirection = macroblockInfo->modelist8[2];
                intraPredAngle     = macroblockInfo->intraPredAngle8[2];
        }
        else if (macroblockInfo->modelist8[3] < PRED_DC &&
            macroblockInfo->intraPredAngle8[3] != 0){
                intraPredDirection = macroblockInfo->modelist8[3];
                intraPredAngle     = macroblockInfo->intraPredAngle8[3];
        }

        macroblockInfo->intraPredDirectionChroma = intraPredDirection;
        macroblockInfo->intraPredAngleChroma     = intraPredAngle;
    }

    uprighta = upright_available(encoderInfo,frameInfo,macroblockInfo,mby,mbx,0,8,params);

    get_intrapred_ref_samples(lefta, upa, uprighta, 0, cypos, cxpos, recon->U, ucolD, urowD);
    get_intrapred_ref_samples(lefta, upa, uprighta, 0, cypos, cxpos, recon->V, vcolD, vrowD);


    dcval = dcvalu = dcvalv = 0;
    if(lefta){
        for (k=0;k<8;k++){
            ucol[k] = recon->U[cypos+k][cxpos-1];
            dcvalu += ucol[k];           
            vcol[k] = recon->V[cypos+k][cxpos-1];
            dcvalv += vcol[k];           
        }
    }
    if(upa){
        for (k=0;k<8;k++){
            urow[k] = recon->U[cypos-1][cxpos+k];
            dcvalu += urow[k];
            vrow[k] = recon->V[cypos-1][cxpos+k];
            dcvalv += vrow[k];            
        }
    }

    if(lefta && upa){
        dcvalu=(dcvalu+8)>>4;
        dcvalv=(dcvalv+8)>>4;
    }
    else if(lefta || upa){
        dcvalu=(dcvalu+4)>>3;
        dcvalv=(dcvalv+4)>>3;
    }
    else{
        dcvalu=128;
        dcvalv=128;
    }

    best_cost = 0xfffffff;

    if (!intra_rdo){
        //STATIC move_ucsubmatrix(cblock,curr->U,cypos,cypos+7,cxpos,cxpos+7,0,0);
        sad_dc = sad_ver = sad_hor = 0;
        for (i=0;i<8;i++){
            for (j=0;j<8;j++){
                sad_dc += abs(macroblockInfo->orgU[8*i+j]-dcvalu);
            }
        }
        if (mby){        
            for (i=0;i<8;i++){
                for (j=0;j<8;j++){
                    sad_ver += abs(macroblockInfo->orgU[8*i+j]-urow[j]);
                }
            }
        }
        else{
            sad_ver = 0xfffffff;
        }
        if (mbx){
            {
                for (i=0;i<8;i++){
                    for (j=0;j<8;j++){
                        sad_hor += abs(macroblockInfo->orgU[8*i+j]-ucol[i]);
                    }
                }
            }
        }
        else{
            sad_hor = 0xfffffff;
        }
        //STATIC move_ucsubmatrix(cblock,curr->V,cypos,cypos+7,cxpos,cxpos+7,0,0);
        for (i=0;i<8;i++){
            for (j=0;j<8;j++){
                sad_dc += abs(macroblockInfo->orgV[8*i+j]-dcvalv);
            }
        }
        if (mby){
            for (i=0;i<8;i++){
                for (j=0;j<8;j++){
                    sad_ver += abs(macroblockInfo->orgV[8*i+j]-vrow[j]);
                }
            }
        }
        else{
            sad_ver = 0xfffffff;
        }
        if (mbx){
            {
                for (i=0;i<8;i++){
                    for (j=0;j<8;j++){
                        sad_hor += abs(macroblockInfo->orgV[8*i+j]-vcol[i]);
                    }
                }
            }
        }
        else{
            sad_hor = 0xfffffff;
        }


        if (intraPredDirection != -1){

            sad_dir = 0;

            get_directional_intrapred_8x8(vrowD, vcolD, dcvalv, intraPredDirection, intraPredAngle, pblock);

            for (i=0;i<8;i++){
                for (j=0;j<8;j++){
                    sad_dir += abs(macroblockInfo->orgV[8*i+j]-pblock[i][j]);
                }
            }

            get_directional_intrapred_8x8(urowD, ucolD, dcvalu, intraPredDirection, intraPredAngle, pblock);

            for (i=0;i<8;i++){
                for (j=0;j<8;j++){
                    sad_dir += abs(macroblockInfo->orgU[8*i+j]-pblock[i][j]);
                }
            }

            dir = (sad_dc <= sad_ver && sad_dc <= sad_hor && sad_dc <= sad_dir) ? PRED_DC : (sad_ver <= sad_hor && sad_ver <= sad_dir) ? PRED_VER : (sad_hor <= sad_dir) ? PRED_HOR : PRED_ANGULAR;
        }
        else

            dir = (sad_dc < sad_ver && sad_dc < sad_hor) ? PRED_DC : (sad_ver <= sad_hor ? PRED_VER : PRED_HOR);
    }    

    /* Code all intra chroma modes when RDO is enabled */
    for (cand_dir = PRED_VER; cand_dir <= (intra_rdo?PRED_COUNT:0); cand_dir++){

        if (intra_rdo){
            if ((cand_dir == PRED_VER && upa == 0) || (cand_dir == PRED_HOR && lefta==0)){
                continue; /* Prediction direction is not available */
            }

            // Dont do angular for 16x16 and 4x4 modes, or if 8x8 luma had no angular directions
            if (cand_dir == PRED_ANGULAR && intraPredDirection == -1){
				continue;
            }

            dir = (cand_dir < PRED_COUNT) ? cand_dir : best_dir;
        }

        /* Now do the Cb component of the macroblock */

        cand_cost = 0;
        for (uv=0;uv<2;uv++){
            unsigned char *orgUV = uv ? macroblockInfo->orgV : macroblockInfo->orgU;
            unsigned char *recUV = uv ? macroblockInfo->recV : macroblockInfo->recU;
            short coeff[8][8] = {0,};
            //STATIC coeff = uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff;           

            if (dir == PRED_ANGULAR){
                if (uv)
                    get_directional_intrapred_8x8(vrowD, vcolD, dcvalv, intraPredDirection, intraPredAngle, pblock);
                else
                    get_directional_intrapred_8x8(urowD, ucolD, dcvalu, intraPredDirection, intraPredAngle, pblock);
            }
            else
                if (uv)
                    get_chroma_intra_prediction_mb(vrow, vcol, dcvalv, dir, pblock, params);
                else
                    get_chroma_intra_prediction_mb(urow, ucol, dcvalu, dir, pblock, params);

            for (k=0;k<8;k++){
                for (l=0;l<8;l++){
                    dblock8[k][l] = orgUV[8*k+l]-pblock[k][l];
                }
            }
            transform_8x8(dblock8,coeff);


            {
                float lambda = pic_type==I_PICTURE ? params->lambda_coeff_quantIntraI : params->lambda_coeff_quantIntraPB;

                cbpflag = quantize_NxN_RDOQ(encoderInfo,coeff,qp,uv ? BLOCK_V : BLOCK_U,lambda,TR_8x8);

            }


            if(upa == 0 && lefta == 0){
                if (cbpflag==0){ 
                    for (k=0; k<8; k++)
                        memset(coeff[k],0,8*sizeof(short));
                }
                side_intra = coeff[0][0];               
                if (uv)
                    macroblockInfo->side_intra_v = side_intra;
                else
                    macroblockInfo->side_intra_u = side_intra;   
                coeff[0][0] = 0;

                if(side_intra != 0){
                    no_of_coeffs--;
                    sum_of_levels-=side_intra;
                    cbpflag = 0;

                    for (k=0; k<8; k++){
                        for (l=0; l<8; l++){                    
                            if (coeff[k][l] != 0)
                                cbpflag = 1;
                        }
                    }

                }
            }


            if (macroblockInfo->tr16flag == 1){
                if (!cbpflag) 
                    memset(coef_scan_uv[uv],0,32*sizeof(int));
                else
                    /* Scan */
                    for (k=0;k<8;k++){
                        for (l=0;l<8;l++){
                            coef_scan_uv[uv][zigzag[k*8+l]] = coeff[k][l];
                        }
                    }

                    //truncation
                    memset(&coef_scan_uv[uv][32],0,32*sizeof(int));
                    cbpflag = 0;
                    for (k=0; k<32; k++){
                        if (coef_scan_uv[uv][k] != 0)
                            cbpflag = 1;
                    }

                    // i_Scan
                    for (k=0;k<8;k++){
                        for (l=0;l<8;l++){
                            coeff[k][l] = coef_scan_uv[uv][zigzag[k*8+l]];
                        }
                    }

            }

            if (uv)
                macroblockInfo->vcbp = cbpflag;
            else
                macroblockInfo->ucbp = cbpflag;

            if(upa == 0 && lefta == 0){
                coeff[0][0] = side_intra;
                dequantize_8x8(coeff,rcoeff8,frameInfo->qp,1);
                inverse_transform_8x8(rcoeff8,rdblock8,frameInfo->qp,1);               
                for (k=0;k<8;k++){
                    for (l=0;l<8;l++){
                        recUV[8*k+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                    }
                }                
                coeff[0][0] = 0;
            }
            else{
                if (cbpflag){
                    dequantize_8x8(coeff,rcoeff8,frameInfo->qp,1);
                    inverse_transform_8x8(rcoeff8,rdblock8,frameInfo->qp,1);                   
                    for (k=0;k<8;k++){
                        for (l=0;l<8;l++){
                            recUV[8*k+l] = clip3(rdblock8[k][l]+pblock[k][l]);
                        }
                    }                   
                }
                else{                   
                    for (k=0;k<8;k++){
                        for (l=0;l<8;l++){
                            rdblock8[k][l] = 0;
                        }
                    }                   
                    for (k=0;k<8;k++)
                        memcpy(&recUV[8*k],pblock[k],8*sizeof(unsigned char));
                }
            }

            /* RD-cost for Cb component */
            if (intra_rdo && cand_dir < PRED_COUNT){
                num_bits = 0;

                if(upa == 0 && lefta == 0){
                    encode_side_intra(side_intra,str_tmp,&num_bits,1);
                    coeff[0][0] = 0;
                }

                if (cbpflag){
                    encode_coeffs(encoderInfo,coeff, str_tmp, &num_bits, uv, 0);
                }

                cand_cost += cost_block(orgUV,recUV,8,8,num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm,MODE_8x8);                        

                if (uv){
                    if (cand_cost < best_cost){
                        best_cost = cand_cost;
                        best_dir = cand_dir;
                    }
                }
            }

            if (uv)
                macroblockInfo->vcbp = cbpflag;
            else
                macroblockInfo->ucbp = cbpflag;

            memcpy(uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff, coeff, 8*8*sizeof(short));

        } //for uv


    } // for cand_dir


    macroblockInfo->bestModeChromaI = dir;
    macroblockInfo->modelist |= ((dir&0x3) << 8);

    if (macroblockInfo->tr16flag == 1){
        for (i=0; i<32; i++){
            j = i*2;
            result_coef[j  ] = coef_scan_uv[0][i];
            result_coef[j+1] = coef_scan_uv[1][i];
        }        
        for (k=0;k<8;k++){
            for (l=0;l<8;l++){
                macroblockInfo->ucoeff[k][l] = result_coef[zigzag[k*8+l]];
            }
        }
    }


    if(macroblockInfo->check_planar)
    {

        for (k=0;k<8;k++){
            for (l=0;l<8;l++){
                if(k == 0 && l == 0)
                    continue;
                if(macroblockInfo->ucoeff[k][l] > 0 || macroblockInfo->vcoeff[k][l] > 0)
                    macroblockInfo->check_planar = 0;
            }
        }

        macroblockInfo->planar_pred[1] = dcvalu;
        macroblockInfo->planar_pred[2] = dcvalv;
    }    
    //STATIC free_ucmatrix(pblock,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,7,0,7);
    //STATIC free_imatrix(dblock8,0,7,0,7);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);
    //STATIC free_ucsubmatrix(cblock,0,7,0,7);

    if (str_tmp != NULL) closebits(str_tmp);    
}

/*-----------------------------------------------------------------------------------------
Function:   encode_luma_16x16P
Purpose:    Encode an 16x16 luma MB in Inter 16x16 mode
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original 16x16 blocks
Output:     Reconstructed 16x16 block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/


void encode_luma_16x16P(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)                  
{

    int k,l,m,cbpflag;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax;
    short dblock16[16][16];  //STATIC int           **dblock16;
    unsigned char pblock16[16][16]; //STATIC unsigned char **pblock16;
    short rdblock16[16][16]; //STATIC int           **rdblock16;
    unsigned char tblock8[8][8]; //STATIC unsigned char **tblock8;
    short rcoeff8[8][8]; //STATIC int **rcoeff8;


    //image_frame *curr = encoderInfo->orig;    
    //image_frame *recon = encoderInfo->recon;
    image_frame **iprev = encoderInfo->ref_frame_qp;

    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;

    int sumlev=0;
    int *sum_of_levels = &sumlev;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    vy = mby*2;
    vx = mbx*2;

    //STATIC pblock16 = ucmatrix(0,15,0,15);
    //STATIC tblock8 = ucsubmatrix(pblock16,0,7,0,7,0,0);
    for (k = 0; k < 2; k++){
        for (l = 0; l < 2; l++){           
            //STATIC move_ucsubmatrix(tblock8,pblock16,k*8,k*8+7,l*8,l*8+7,0,0);           
            get_luma_inter_prediction_8x8(macroblockInfo,k, l, iprev, tblock8);
            for (m=0;m<8;m++)
                memcpy(&pblock16[k*8+m][l*8],&tblock8[m][0],8);
        }
    }


    //STATIC dblock16 = imatrix(0,15,0,15);		    
    //STATIC rdblock16 = imatrix(0,15,0,15);
    //STATIC rcoeff8 = imatrix(0,7,0,7);
    for (k=0;k<16;k++){
        for (l=0;l<16;l++){
            dblock16[k][l] = macroblockInfo->orgY[16*k+l]-pblock16[k][l];
        }
    }

    transform_16x16(dblock16,macroblockInfo->ycoeff[0]);

    {
        float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;

        cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[0],frameInfo->qp,pic_type==P_PICTURE ? BLOCK_Y16x16P : BLOCK_Y16x16B,lambda,TR_16x16);

    }


    frameInfo->mvmap[(vy+0)*2*mbxmax+(vx+0)].coded_y = cbpflag;
    frameInfo->mvmap[(vy+0)*2*mbxmax+(vx+1)].coded_y = cbpflag;
    frameInfo->mvmap[(vy+1)*2*mbxmax+(vx+0)].coded_y = cbpflag;
    frameInfo->mvmap[(vy+1)*2*mbxmax+(vx+1)].coded_y = cbpflag;


    if (cbpflag){
        dequantize_8x8(macroblockInfo->ycoeff[0],rcoeff8,frameInfo->qp,0);        
        inverse_transform_16x16(rcoeff8,rdblock16,frameInfo->qp,0);

        for (k=0;k<16;k++)
        {
            unsigned char *restrict recY = &macroblockInfo->recY[16*k]; 
            short tmp[16];
            for (l=0;l<16;l++)
                tmp[l] = rdblock16[k][l]+pblock16[k][l];
            for (l=0;l<16;l++)
                recY[l] = min(max(tmp[l], 0), 255);
        }
        macroblockInfo->ycbp = 1<<3;
    }
    else
    { 
        for (k=0;k<16;k++)
        {
            unsigned char *restrict recY = &macroblockInfo->recY[16*k]; 
            for (l=0;l<16;l++)
                recY[l] = pblock16[k][l];
        }
        macroblockInfo->ycbp = 0x0;
    }   
    //STATIC free_imatrix(dblock16,0,15,0,15);
    //STATIC free_ucmatrix(pblock16,0,15,0,15);
    //STATIC free_imatrix(rdblock16,0,15,0,15);
    //STATIC free_imatrix(rcoeff8,0,8,0,8);
    //STATIc free_ucsubmatrix(tblock8,0,7,0,7);
    macroblockInfo->sum_of_levels = *sum_of_levels;

}

/*-----------------------------------------------------------------------------------------
Function:   encode_lumaLMB
Purpose:    Encode a luma large MB (32x32 or 64x64)
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original block
Output:     Reconstructed block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_lumaLMB(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int LMBwidth,int LMBheight)                  
{

    int k,l,m,cbpflag,ypos,xpos,mby,mbx,vy,vx,mbxmax;
    //STATIC unsigned char **mblock16;    
    unsigned char pblock16[64][64]; //STATIC unsigned char **pblock16;
    unsigned char tblock8[8][8]; //STATIC unsigned char **tblock8;
    //STATIC int           **rdblock16;
    //STATIC int           **dblock16;
    short rcoeff8[8][8]; //STATIC int           **rcoeff8;

    image_frame *curr = encoderInfo->orig;    
    image_frame **iprev = encoderInfo->ref_frame_qp;

    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;

    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;

    vy = mby*2;
    vx = mbx*2;

    //STATIC pblock16 = ucmatrix(0,LMBheight-1,0,LMBwidth-1);
    //STATIC tblock8 = ucsubmatrix(pblock16,0,7,0,7,0,0);

    for (k=0;k<LMBheight;k+=8){
        for (l=0;l<LMBwidth;l+=8){
            //STATIC move_ucsubmatrix(tblock8,pblock16,k,k+7,l,l+7,0,0);
            macroblockInfo->mbx = mbx + (l/16);
            macroblockInfo->mby = mby + (k/16);
            get_luma_inter_prediction_8x8(macroblockInfo,k/8,l/8,iprev, tblock8);
            for (m=0;m<8;m++){
                memcpy(&pblock16[k+m][l],&tblock8[m][0],8);
            }
        }
    }

    macroblockInfo->mbx = mbx;
    macroblockInfo->mby = mby;

    //STATIC mblock16 = ucsubmatrix(curr->Y,0,LMBheight-1,0,LMBwidth-1,0,0);
    //STATIC dblock16 = imatrix(0,LMBheight-1,0,LMBwidth-1);		    
    //STATIC rdblock16 = imatrix(0,LMBheight-1,0,LMBwidth-1);
    //STATIC rcoeff8 = imatrix(0,7,0,7);
    //STATIC move_ucsubmatrix(mblock16,curr->Y,ypos,ypos+LMBheight-1,xpos,xpos+LMBwidth-1,0,0);

    //STATIC for (k=0;k<LMBheight;k++){
    //STATIC     for (l=0;l<LMBwidth;l++){
    //STATIC         dblock16[k][l] = mblock16[k][l]-pblock16[k][l];
    //STATIC     }
    //STATIC }

    if (LMBheight==64 && LMBwidth==64)
    {
        short dblock16[64][64];
        for (k=0;k<64;k++)
        {
            const unsigned char *mblock16 = &curr->Y[k+ypos][xpos];
            for (l=0;l<64;l++)
            {
                dblock16[k][l] = mblock16[l] - pblock16[k][l];
            }
        }
        transform_64x64(dblock16,macroblockInfo->ycoeff[0]);
    }
    else if (LMBheight==32 && LMBwidth==32)
    {
        short dblock16[32][32];
        for (k=0;k<32;k++)
        {
            const unsigned char *mblock16 = &curr->Y[k+ypos][xpos];
            for (l=0;l<32;l++)
            {
                dblock16[k][l] = mblock16[l] - pblock16[k][l];
            }
        }
        transform_32x32(dblock16,macroblockInfo->ycoeff[0]);
    }
    else
        fatalerror("Unsupported LMB size.");

    {
        float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;

        cbpflag = quantize_NxN_RDOQ(encoderInfo,macroblockInfo->ycoeff[0],frameInfo->qp,pic_type==P_PICTURE ? BLOCK_Y16x16P : BLOCK_Y16x16B,lambda,LMBheight==32 ? TR_32x32 : TR_64x64);

    }


    if (cbpflag){
        dequantize_8x8(macroblockInfo->ycoeff[0],rcoeff8,frameInfo->qp,0);
        if (LMBheight==64 && LMBwidth==64)
        {
            short rdblock16[64][64];
            inverse_transform_64x64(rcoeff8,rdblock16,frameInfo->qp,0);
            macroblockInfo->ycbp = 1<<0;
            for (k=0;k<64;k++)
            {
                unsigned char *restrict recLMBY = &macroblockInfo->recLMBY[k*LMBwidth]; 
                short tmp[64];
                for (l=0;l<64;l++)
                    tmp[l] = rdblock16[k][l]+pblock16[k][l];
                for (l=0;l<64;l++)
                    recLMBY[l] = min(max(tmp[l], 0), 255);
            }

        }
        else if (LMBheight==32 && LMBwidth==32)
        {
            short rdblock16[32][32];
            inverse_transform_32x32(rcoeff8,rdblock16,frameInfo->qp,0);
            macroblockInfo->ycbp = 1<<0;
            for (k=0;k<32;k++)
            {
                unsigned char *restrict recLMBY = &macroblockInfo->recLMBY[k*LMBwidth]; 
                short tmp[32];
                for (l=0;l<32;l++)
                    tmp[l] = rdblock16[k][l]+pblock16[k][l];
                for (l=0;l<32;l++)
                    recLMBY[l] = min(max(tmp[l], 0), 255);
            }

        }
        else
            fatalerror("Unsupported LMB size.");

    }
    else{

        for (k=0;k<LMBheight;k++){         
            memcpy(&macroblockInfo->recLMBY[k*LMBwidth],pblock16[k],LMBwidth*sizeof(unsigned char));             
        }

        macroblockInfo->ycbp = 0x0;
    }

    //STATIC free_ucsubmatrix(mblock16,0,LMBheight-1,0,LMBwidth-1);
    //STATIC free_imatrix(dblock16,0,LMBheight-1,0,LMBwidth-1);
    //STATIC free_ucmatrix(pblock16,0,LMBheight-1,0,LMBwidth-1);
    //STATIC free_imatrix(rdblock16,0,LMBheight-1,0,LMBwidth-1);

    //STATIC free_imatrix(rcoeff8,0,7,0,7);
    //STATIC free_ucsubmatrix(tblock8,0,7,0,7);   
}

/*-----------------------------------------------------------------------------------------
Function:   encode_chromaLMB
Purpose:    Encode a chroma large MB (32x32 or 64x64)
            prediction/transform/quantize/inverse transform/reconstruct
Input:      Original block
Output:     Reconstructed block
            Quanized transform coefficients
            CBP flag(s)
Parameters: None 
-------------------------------------------------------------------------------------------*/

void encode_chromaLMB(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,int LMBwidth,int LMBheight)                        

{
    int k,l,m,cbpflag;
    int ypos,xpos,mby,mbx,vy,vx,mbxmax,cypos,cxpos;
    //STATIC unsigned char **mblock8;
    unsigned char pblock[32][32]; //STATIC unsigned char **pblock;
    unsigned char tblock8[8][8]; //STATIC unsigned char **tblock8;
    short rcoeff8[8][8]; //STATIC int **rdblock8,**dblock8,**rcoeff8;        
    image_frame *curr = encoderInfo->orig;    
    image_frame **iprev = encoderInfo->ref_frame_qp;
    int pic_type = frameInfo->pic_type;
    enc_params *params = encoderInfo->params;
    int rounding,uv;
    //STATIC int **coeff;       


    mbxmax = encoderInfo->widthInMBs;
    mbx = macroblockInfo->mbx;
    mby = macroblockInfo->mby;
    ypos = mby*16;
    xpos = mbx*16;
    cypos = ypos / 2;
    cxpos = xpos / 2;

    vy = mby*2;
    vx = mbx*2;    

    //STATIC pblock = ucmatrix(0,LMBheight/2-1,0,LMBwidth/2-1);
    //STATIC tblock8 = ucsubmatrix(pblock,0,7,0,7,0,0);
    //STATIC mblock8 = ucsubmatrix(curr->Y,0,LMBheight/2-1,0,LMBwidth/2-1,0,0);    
    //STATIC dblock8 = imatrix(0,LMBheight/2-1,0,LMBwidth/2-1);    
    //STATIC rdblock8 = imatrix(0,LMBheight/2-1,0,LMBwidth/2-1);
    //STATIC rcoeff8 = imatrix(0,7,0,7);


    if (pic_type == B_PICTURE){
        rounding = 1;
    }
    else{
        rounding = (params->frame_num)&1;
    }		

    for (uv=0;uv<2;uv++)
    {
        short coeff[8][8];

        unsigned char *recUV = uv ? macroblockInfo->recLMBV : macroblockInfo->recLMBU;
        unsigned char **UV = uv ? curr->V : curr->U;
        //STATIC coeff = uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff; 

        //STATICmove_ucsubmatrix(mblock8,uv ? curr->V : curr->U,cypos,cypos+LMBheight/2-1,cxpos,cxpos+LMBwidth/2-1,0,0);
        for (k=0;k<LMBheight/2;k+=8)
        {
            for (l=0;l<LMBwidth/2;l+=8)
            {
                //STATIC move_ucsubmatrix(tblock8,pblock,k,k+7,l,l+7,0,0);
                macroblockInfo->mbx = mbx + (l/8);
                macroblockInfo->mby = mby + (k/8);
                get_chroma_inter_prediction_mb(macroblockInfo, iprev, tblock8, uv, rounding);            
                for (m=0;m<8;m++)
                    memcpy(&pblock[k+m][l],&tblock8[m][0],8);
            }
        }




        if (LMBheight==64 && LMBwidth==64)
        {
            short dblock8[32][32];
            for (k=0;k<32;k++)
            {
                const unsigned char *mblock8 = &UV[cypos+k][cxpos];
                for (l=0;l<32;l++)
                {
                    dblock8[k][l] = mblock8[l]-pblock[k][l];
                }
            }
            transform_32x32(dblock8,coeff); 
        }
        else if (LMBheight==32 && LMBwidth==32)
        {
            short dblock8[16][16];
            for (k=0;k<16;k++)
            {
                const unsigned char *mblock8 = &UV[cypos+k][cxpos];
                for (l=0;l<16;l++)
                {
                    dblock8[k][l] = mblock8[l]-pblock[k][l];
                }
            }
            transform_16x16(dblock8,coeff); 
        }
        else
            fatalerror("Unsupported LMB size.");

        {
            int qp = frameInfo->qp - corr_qpc_8x8[frameInfo->qp];        
            float lambda = pic_type==P_PICTURE ? params->lambda_coeff_quantInterP : params->lambda_coeff_quantInterB;

            cbpflag = quantize_NxN_RDOQ(encoderInfo,coeff,qp,uv ? BLOCK_V : BLOCK_U,lambda,LMBheight==32 ? TR_16x16 : TR_32x32);

        }





        if (uv){            
            macroblockInfo->vcbp = cbpflag;
        }
        else{            
            macroblockInfo->ucbp = cbpflag;
        }


        if (cbpflag){            
            dequantize_8x8(coeff,rcoeff8,frameInfo->qp,1);

            if (LMBheight==64 && LMBheight==64)
            {
                short rdblock8[32][32];
                inverse_transform_32x32(rcoeff8,rdblock8,frameInfo->qp,1); 
                for (k=0;k<32;k++){
                    for (l=0;l<32;l++){

                        recUV[k*LMBwidth/2+l] = clip3(rdblock8[k][l]+pblock[k][l]);

                    }
                }
            }
            else if (LMBheight==32 && LMBheight==32)
            {
                short rdblock8[16][16];
                inverse_transform_16x16(rcoeff8,rdblock8,frameInfo->qp,1); 
                for (k=0;k<16;k++){
                    for (l=0;l<16;l++){

                        recUV[k*LMBwidth/2+l] = clip3(rdblock8[k][l]+pblock[k][l]);

                    }
                }
            }
            else
                fatalerror("Unsupported LMB size.");

        }
        else{

            for (k=0;k<LMBheight/2;k++)
                memcpy(&recUV[k*LMBwidth/2],&pblock[k][0],LMBwidth/2*sizeof(unsigned char));

        }
        memcpy(uv ? macroblockInfo->vcoeff : macroblockInfo->ucoeff, coeff, 8*8*sizeof(short));
    } //uv

    //STATIC free_ucmatrix(pblock,0,LMBheight/2-1,0,LMBwidth/2-1);
    //STATIC free_ucsubmatrix(tblock8,0,7,0,7);
    //STATIC free_imatrix(rdblock8,0,LMBheight/2-1,0,LMBwidth/2-1);
    //STATIC free_imatrix(dblock8,0,LMBheight/2-1,0,LMBwidth/2-1);    
    //STATIC free_ucsubmatrix(mblock8,0,LMBheight/2-1,0,LMBwidth/2-1);
    //STATIC free_imatrix(rcoeff8,0,7,0,7);


}

/*-----------------------------------------------------------------------------------------
Function:   processLMB
Purpose:    Process and LMB (32x32 or 64x64)
            Write side info to bitstream
            prediction/transform/quantize/inverse transform/reconstruct (luma&chroma)
            VLC encoding of coefficients
Parameters: None 
-------------------------------------------------------------------------------------------*/

int processLMB(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,stream *str,int LMBwidth,int LMBheight)
{

    int i,mbrun,cbp;        
    int numbits = bitcount(str);
    bitcounter *bits;

    mvpos predmv,largemv;   

    mvpos predmvBW,largemvBW;

    if (frameInfo->pic_type==I_PICTURE)
        bits = encoderInfo->encStat->bitsI;
    else if(frameInfo->pic_type==P_PICTURE)
        bits = encoderInfo->encStat->bitsP;
    else
        bits = encoderInfo->encStat->bitsB;        

    /* Write run and skip bits preceeding this LMB */    
    mbrun = macroblockInfo->pos - macroblockInfo->last_pos - 1;
    bits->run += put_uvlc(mbrun,str);

    if (mbrun > 0){    
        write_run(encoderInfo,frameInfo,macroblockInfo,mbrun,str);       
    }


    putbits(1,1,str); // Upper left corner of 64x64 or 32x32 LMB
    bits->lmb += 1;
    if (LMBheight==32){
        if (macroblockInfo->LMBposFlag&0x2){
            putbits(1,0,str); // Indicate upper left 32x32 LMB of 64x64 LMB
            bits->lmb += 1;
        }        
    }
    else{
        putbits(1,1,str); // Indicate start of 64x64 LMB
        bits->lmb += 1;
    }


    macroblockInfo->mby -= (LMBheight/16-1);
    macroblockInfo->mbx -= (LMBwidth/16-1);  

    /* Find skip vector and set reference to 0 */

    find_mv_predictor(encoderInfo,frameInfo,macroblockInfo);


    if (!macroblockInfo->final_encode_flag)

    for (i=0;i<LMBheight;i++){            
        memcpy(&macroblockInfo->orgLMBY[LMBwidth*i],&encoderInfo->orig->Y[16*macroblockInfo->mby+i][16*macroblockInfo->mbx],LMBwidth*sizeof(unsigned char));
    }

    if (!macroblockInfo->final_encode_flag)

    for (i=0;i<LMBheight/2;i++){            
        memcpy(&macroblockInfo->orgLMBU[LMBwidth/2*i],&encoderInfo->orig->U[8*macroblockInfo->mby+i][8*macroblockInfo->mbx],(LMBwidth/2)*sizeof(unsigned char));
        memcpy(&macroblockInfo->orgLMBV[LMBwidth/2*i],&encoderInfo->orig->V[8*macroblockInfo->mby+i][8*macroblockInfo->mbx],(LMBwidth/2)*sizeof(unsigned char));
    }


    if(frameInfo->pic_type == 2 && encoderInfo->params->bi==1)
        ref_predictor(encoderInfo,frameInfo,macroblockInfo);
    if(macroblockInfo->ref == 2)
    {
        predmv = macroblockInfo->pred[0][0][0];
        predmvBW = macroblockInfo->pred[1][0][0];
    }
    else
    {
        predmv = macroblockInfo->pred[macroblockInfo->ref][0][0];


        if (macroblockInfo->final_encode_flag){
            macroblockInfo->mvblock[0][0] = macroblockInfo->mvsearch_result[macroblockInfo->ref];
            macroblockInfo->mvblock[0][1] = macroblockInfo->mvblock[0][0];
            macroblockInfo->mvblock[1][0] = macroblockInfo->mvblock[0][0];
            macroblockInfo->mvblock[1][1] = macroblockInfo->mvblock[0][0];
        }
        else

        motion_estimateLMB(encoderInfo,frameInfo,macroblockInfo,macroblockInfo->ref,macroblockInfo->mode);
        macroblockInfo->mvsearch_result[macroblockInfo->ref] = macroblockInfo->mvblock[0][0];
    }
    if (frameInfo->pic_type == B_PICTURE)
    {
        if(encoderInfo->params->bi==1)
        {
            if(macroblockInfo->ref == macroblockInfo->ref_pred)
            {
                putbits(1,1,str);
                bits->ref +=1;
            }
            else
            {
                putbits(1,0,str);
                if(macroblockInfo->ref == 0)
                    putbits(1,0,str);
                else if(macroblockInfo->ref == 2)
                    putbits(1,1,str);
                else if(macroblockInfo->ref_pred ==0)
                    putbits(1,0,str);
                else
                    putbits(1,1,str);
                bits->ref +=2;
            }
        }
        else
        {
            putbits(1,macroblockInfo->ref,str);
            bits->lmb++;
        }
    }

    else{
        putbits(1,macroblockInfo->ref,str);
        bits->lmb++;
    }

    if(macroblockInfo->ref == 2)
    {
        largemv = macroblockInfo->mvsearch_result[0];
        bits->vec += put_level(largemv.x-predmv.x,str);
        bits->vec += put_level(largemv.y-predmv.y,str);
        largemvBW = macroblockInfo->mvsearch_result[1];
        bits->vec += put_level(largemvBW.x-predmvBW.x,str);
        bits->vec += put_level(largemvBW.y-predmvBW.y,str);

        macroblockInfo->mvblockBW[0][0] = largemvBW;
        macroblockInfo->mvblockBW[0][1] = largemvBW;
        macroblockInfo->mvblockBW[1][0] = largemvBW;
        macroblockInfo->mvblockBW[1][1] = largemvBW;

        macroblockInfo->mvblock[0][0] = largemv;
        macroblockInfo->mvblock[0][1] = largemv;
        macroblockInfo->mvblock[1][0] = largemv;
        macroblockInfo->mvblock[1][1] = largemv;

        if(g_enableSwitched == 1){
            int pos = ((largemv.y%4+4)%4)*4 + ((largemv.x%4+4)%4);
            if(transmitInterpIndx[pos]== 1){
                putbits(1,largemv.interp_indx,str);
                bits->lmb++;
            }
            pos = ((largemvBW.y%4+4)%4)*4 + ((largemvBW.x%4+4)%4);
            if(transmitInterpIndx[pos]== 1){
                putbits(1,largemvBW.interp_indx,str);
                bits->lmb++;
            }
        }

    }
    else
    {
        largemv = macroblockInfo->mvblock[0][0];
        bits->vec += put_level(largemv.x-predmv.x,str);
        bits->vec += put_level(largemv.y-predmv.y,str);

        if(g_enableSwitched == 1){
            int pos = ((largemv.y%4+4)%4)*4 + ((largemv.x%4+4)%4);
            if(transmitInterpIndx[pos]== 1){
                putbits(1,largemv.interp_indx,str);
                bits->lmb++;
            }
        }

    }

    encode_lumaLMB(encoderInfo,frameInfo,macroblockInfo,LMBwidth,LMBheight);
    encode_chromaLMB(encoderInfo,frameInfo,macroblockInfo,LMBwidth,LMBheight);

    cbp = ((macroblockInfo->vcbp)<<2) + ((macroblockInfo->ucbp)<<1) + (macroblockInfo->ycbp != 0);
    bits->lmb += put_vlc(0,cbp,str);

    if (macroblockInfo->ycbp){
        int block_type = frameInfo->pic_type==P_PICTURE ? BLOCK_Y16x16P : BLOCK_Y16x16B;
        encode_coeffs(encoderInfo,macroblockInfo->ycoeff[0],str,&(bits->coeff_y),block_type,macroblockInfo->final_encode_flag);        
    }

    if (macroblockInfo->ucbp){               
        encode_coeffs(encoderInfo,macroblockInfo->ucoeff,str,&(bits->coeff_u),0,macroblockInfo->final_encode_flag);
    }  

    if (macroblockInfo->vcbp){                
        encode_coeffs(encoderInfo,macroblockInfo->vcoeff,str,&(bits->coeff_v),1,macroblockInfo->final_encode_flag);
    }

    /* Determine number of bits used */
    macroblockInfo->num_bits = bitcount(str) - numbits;

    if (macroblockInfo->final_encode_flag){
        /* Update MB and MV maps accordingly */
        int my,mx;
        int mbxmax = encoderInfo->widthInMBs;
        block_info lmbBlock;

        int sign = (frameInfo->pic_type==B_PICTURE && (macroblockInfo->ref%2)) ? -1 : 1;

        lmbBlock.x = sign*macroblockInfo->mvblock[0][0].x;
        lmbBlock.y = sign*macroblockInfo->mvblock[0][0].y;
        lmbBlock.coded_y = macroblockInfo->ycbp;

        lmbBlock.interp_indx = macroblockInfo->mvblock[0][0].interp_indx;


        for (my=macroblockInfo->mby-(LMBheight/16)+1;my<=macroblockInfo->mby;my++){
            for (mx=macroblockInfo->mbx-(LMBwidth/16)+1;mx<=macroblockInfo->mbx;mx++){
                frameInfo->mbmap[my*mbxmax+mx].ref = macroblockInfo->ref;
                frameInfo->mbmap[my*mbxmax+mx].mode = macroblockInfo->mode;

                frameInfo->mvmap[(2*my+0)*2*mbxmax+(2*mx+0)] = lmbBlock;
                frameInfo->mvmap[(2*my+0)*2*mbxmax+(2*mx+1)] = lmbBlock;
                frameInfo->mvmap[(2*my+1)*2*mbxmax+(2*mx+0)] = lmbBlock;
                frameInfo->mvmap[(2*my+1)*2*mbxmax+(2*mx+1)] = lmbBlock;
            }
        }

        /* Update pos and last_pos */
        macroblockInfo->pos += ((LMBheight/16)*(LMBwidth/16)-1);
        macroblockInfo->last_pos = macroblockInfo->pos;
    }

    return 0;
}

/*-----------------------------------------------------------------------------------------
Function:   process_mb
Purpose:    Process and MB (16x16)
            Write side info to bitstream
            prediction/transform/quantize/inverse transform/reconstruct (luma&chroma)
            VLC encoding of coefficients
Parameters: None 
-------------------------------------------------------------------------------------------*/

int process_mb(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo,stream *str,bitcounter *bits)
{

    int ypos, xpos, cypos, cxpos,mbx,mby;
    int mbxmax=encoderInfo->widthInMBs;
    int mbymax=encoderInfo->heightInMBs;
    int mbrun;
    int vy,vx;   

    int pic_type = frameInfo->pic_type;

    int numbits = bitcount(str);      

    if (pic_type==I_PICTURE)
        bits = encoderInfo->encStat->bitsI;
    else if(pic_type==P_PICTURE)
        bits = encoderInfo->encStat->bitsP;
    else
        bits = encoderInfo->encStat->bitsB;
    macroblockInfo->skip_mode_bit_adjust = 0;

    mby = macroblockInfo->mby;
    mbx = macroblockInfo->mbx;
    ypos = mby*16;
    xpos = mbx*16; 
    cypos = ypos / 2;
    cxpos = xpos / 2;
    vy = mby*2;
    vx = mbx*2;


    macroblockInfo->tr16flag = (macroblockInfo->tr_size==TR_16x16);
    macroblockInfo->tr16flag = (macroblockInfo->tr_size==TR_16x16);
    macroblockInfo->tr4flag = (macroblockInfo->tr_size==TR_4x4);

    macroblockInfo->planar_flag = 0;


    macroblockInfo->ycbp = 0;
    macroblockInfo->modelist = 0;

    frameInfo->mvmap[(vy+0)*2*mbxmax+(vx+0)].coded_y = 0;
    frameInfo->mvmap[(vy+0)*2*mbxmax+(vx+1)].coded_y = 0;
    frameInfo->mvmap[(vy+1)*2*mbxmax+(vx+0)].coded_y = 0;
    frameInfo->mvmap[(vy+1)*2*mbxmax+(vx+1)].coded_y = 0;

    /* Encoding of non-skipped macroblock */    
    if (macroblockInfo->mode > MODE_SKIP){   
        macroblockInfo->sum_of_levels = 0;
        /* Do the luma component of a macroblock first */        
        if (macroblockInfo->mode == MODE_INTRA){          
            if (macroblockInfo->tr16flag == 1) 
                encode_luma_16x16I(encoderInfo,frameInfo,macroblockInfo);

            else if (macroblockInfo->tr4flag == 1)
                encode_luma_4x4I(encoderInfo,frameInfo,macroblockInfo);

            else
            {

                encode_luma_8x8I(encoderInfo,frameInfo,macroblockInfo);
            }

        }
        else
        {

            if (macroblockInfo->mode == MODE_16x16 && macroblockInfo->tr16flag){
                encode_luma_16x16P(encoderInfo,frameInfo,macroblockInfo);                
            }

            else if (macroblockInfo->mode >= MODE_16x16 && macroblockInfo->mode < MODE_8x8 && macroblockInfo->tr16flag == 0 && macroblockInfo->svt == 1){
                encode_luma_SVT_P(encoderInfo,frameInfo,macroblockInfo);

                if (macroblockInfo->ycbp == 0){                                  
                    macroblockInfo->num_bits = bitcount(str) - numbits + macroblockInfo->skip_mode_bit_adjust;
                    return 0;
                }
            }            

            else if (macroblockInfo->tr4flag){
                encode_luma_4x4P(encoderInfo,frameInfo,macroblockInfo);                
            }


            else{
                encode_luma_8x8P(encoderInfo,frameInfo,macroblockInfo);                
            }

        }

        /* Chroma components */
        if (macroblockInfo->mode > MODE_INTRA){

            if (macroblockInfo->tr_size==TR_8x8 || macroblockInfo->final_encode_flag){            
                macroblockInfo->ucbp = 0;    
                macroblockInfo->vcbp = 0;
                encode_chroma_8x8P(encoderInfo,frameInfo,macroblockInfo);
            }
        }
        else{
            macroblockInfo->ucbp = 0;    
            macroblockInfo->vcbp = 0;          
            encode_chroma_8x8I(encoderInfo,frameInfo,macroblockInfo);
        }

        /* Write encoded macroblocks to bitstream */          

        if(macroblockInfo->final_encode_flag && macroblockInfo->check_planar == 1)        
            encode_planar(encoderInfo,frameInfo,macroblockInfo);

        write_one_mb(encoderInfo,frameInfo,macroblockInfo,str);              

    } // end if (mode > MODE_SKIP)
    else{
        rec_compensate_mb(encoderInfo,frameInfo,macroblockInfo);        
        if (macroblockInfo->pos == mbymax*mbxmax-1){
            mbrun = macroblockInfo->pos - macroblockInfo->last_pos;   
            bits->run += put_uvlc(mbrun,str); 
            write_run(encoderInfo,frameInfo,macroblockInfo,mbrun,str);            
        }
        else{
            if (macroblockInfo->num_skip_cand > 1){
                macroblockInfo->skip_mode_bit_adjust += 1;
            }
        }
    }

    if (macroblockInfo->svt==1){
        int i;        
        for (i=0;i<16;i++){
            memcpy(&macroblockInfo->recY[16*i],&encoderInfo->recon->Y[16*mby+i][16*mbx],16*sizeof(unsigned char));                        
        }                     
    }


    macroblockInfo->num_bits = bitcount(str) - numbits + macroblockInfo->skip_mode_bit_adjust;
    return 0;
}


int compareMVs(mvpos mv1,mvpos mv2){


    if (mv1.x == mv2.x && mv1.y == mv2.y && mv1.interp_indx == mv2.interp_indx)

        return 1;
    else
        return 0;
}



unsigned int mode_selection_rdo(encoder_info *encoderInfo,frame_info *frameInfo,macroblock_info *macroblockInfo)
{
    int ypos, xpos, cypos, cxpos,intra_cost,cost;
    int mbymax,mbxmax;
    mvpos cand[3];
    int vy;
    int vx;
    int skip_ref = 0;
    mvpos skipmv_arr[2];
    int skip_ref_arr[2];
    int best_svt;
    int svt_mode;
    int svt_ref;
    int best_ref_for_svt[4] = {0,0,0,0};
    int best_rdcost_for_svt[4] = {0xfffffff,0xfffffff,0xfffffff,0xfffffff};
    int inter_cost = 0xfffffff;
    int best_mode,best_ref,min_cost,mode,ref;  
    int tr,best_tr_size,tmp_cost;
    trsizes tr_size;
    const trsizes tr_size_list[3] = {TR_8x8, TR_16x16, TR_4x4};
    mvpos MVarray[2][4][2][2];

    enc_params *params = encoderInfo->params;
    int pic_type = frameInfo->pic_type;


    int mask[3][4];

    int num_ref = min(frameInfo->frame_num,frameInfo->num_ref);

    int mby = macroblockInfo->mby;
    int mbx = macroblockInfo->mbx;    
    int skip_cost,skip_cost2;
    int early_skip;
    bitcounter *bits;
    bitcounter *bitsTmp;
    stream *strTmp;
    mbymax = encoderInfo->heightInMBs;
    mbxmax = encoderInfo->widthInMBs;


    bitsTmp = (bitcounter*)malloc(sizeof(bitcounter));
    strTmp = (stream*)malloc(sizeof(stream));
    vy = 2*mby;
    vx = 2*mbx;
    if (pic_type==I_PICTURE)
        bits = encoderInfo->encStat->bitsI;
    else if(pic_type==P_PICTURE)
        bits = encoderInfo->encStat->bitsP;
    else
        bits = encoderInfo->encStat->bitsB;   

    ypos = mby * 16;
    xpos = mbx * 16;
    cypos = ypos/2;
    cxpos = xpos/2;


    macroblockInfo->svt = 0;

    min_cost = (1<<30);


    /****************** EVALUATE INTRA MODES **************/

    if (macroblockInfo->forceIntra){
#if DBG_COMPLEXITY
        TSC_START(TSC_INTRA_RDO);
#endif
        macroblockInfo->mode = MODE_INTRA;
        macroblockInfo->ref = -1;   
        for (tr = 0; tr < 3 ; tr++){
            tr_size = tr_size_list[tr]; // check TR_8x8, TR_16x16, TR_4x4
            if (tr_size==TR_8x8 || (tr_size==TR_16x16 && params->tr16flag&1) || (tr_size==TR_4x4 && params->tr4flag&1)){ 

                {   
                    memcpy(strTmp,encoderInfo->str,sizeof(stream));
                    memcpy(bitsTmp,bits,sizeof(bitcounter));
                    macroblockInfo->tr_size = tr_size;
                    process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);

                    tmp_cost = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

                }
                if (tmp_cost < min_cost){
                    min_cost = tmp_cost;
                    best_tr_size = tr_size;
                }
            }
        }
#if DBG_COMPLEXITY
        TSC_CUMUL(TSC_INTRA_RDO);
#endif              
        macroblockInfo->tr_size = best_tr_size;
        return 0;        
    }
    /* Find motion search candidates, motion vector predictor and motion skip candidates */
    find_mv_predictor(encoderInfo,frameInfo,macroblockInfo);

    if(frameInfo->pic_type == 2&& encoderInfo->params->bi==1)
        ref_predictor(encoderInfo,frameInfo,macroblockInfo);

    cand[0] = macroblockInfo->candidateMV[0][1];
    cand[1] = macroblockInfo->candidateMV[0][2];
    cand[2] = macroblockInfo->candidateMV[0][3];

    macroblockInfo->dirmv = macroblockInfo->pred[0][0][0];

    macroblockInfo->num_skip_cand = get_skip_mv_cand(encoderInfo,frameInfo,macroblockInfo,skipmv_arr,skip_ref_arr);
    frameInfo->mbmap[mby*mbxmax+mbx].num_skip_cand = macroblockInfo->num_skip_cand;    


    /****************** EVALUATE TWO SKIP MODES **************/
#if DBG_COMPLEXITY
    TSC_START(TSC_SKIP_RDO);
#endif                    
    macroblockInfo->mode = MODE_SKIP;

    /********* Find cost for SKIP 0 ******/      
    macroblockInfo->skipmv = skipmv_arr[0];    
    macroblockInfo->ref = skip_ref = skip_ref_arr[0];    
    memcpy(strTmp,encoderInfo->str,sizeof(stream));
    memcpy(bitsTmp,bits,sizeof(bitcounter));       
    process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);

    skip_cost = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

    early_skip = macroblockInfo->early_skip_flag;

    if (macroblockInfo->num_skip_cand > 1){       
        macroblockInfo->skipmv = skipmv_arr[1];
        macroblockInfo->ref = skip_ref = skip_ref_arr[1];        
        memcpy(strTmp,encoderInfo->str,sizeof(stream));
        memcpy(bitsTmp,bits,sizeof(bitcounter));               
        process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);

        skip_cost2 = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

        if (skip_cost2 < skip_cost){
            skip_cost = skip_cost2;
            macroblockInfo->skipmv = skipmv_arr[1];
            frameInfo->mbmap[mby*mbxmax+mbx].best_skip_mv_cand = 1;
            skip_ref = skip_ref_arr[1];
        }
        else{
            macroblockInfo->skipmv = skipmv_arr[0];            
            frameInfo->mbmap[mby*mbxmax+mbx].best_skip_mv_cand = 0;
            skip_ref = skip_ref_arr[0];
        }
        if (!early_skip)
            early_skip = macroblockInfo->early_skip_flag;
    }

    if (early_skip){
        macroblockInfo->mvblock[0][0] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[0][1] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[1][0] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[1][1] = macroblockInfo->skipmv;
        macroblockInfo->ref = skip_ref;
        return 0;
    }

    if (skip_cost < min_cost){
        best_mode = MODE_SKIP;
        min_cost = skip_cost;
    }
#if DBG_COMPLEXITY
    TSC_CUMUL(TSC_SKIP_RDO);
#endif                

    rdcost_stat[0][skip_ref][0] = skip_cost;

    /********* ONE DIRECT MODE ******/

    if (pic_type == B_PICTURE){
        macroblockInfo->mvblock[0][0] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[0][1] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[1][0] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[1][1] = macroblockInfo->dirmv;
        macroblockInfo->mode = MODE_DIRECT;
        macroblockInfo->ref = 0;

        memcpy(strTmp,encoderInfo->str,sizeof(stream));
        memcpy(bitsTmp,bits,sizeof(bitcounter));
        macroblockInfo->tr_size = TR_8x8;
        process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);

        cost = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

        if (cost <= min_cost && (macroblockInfo->ycbp != 0 || macroblockInfo->ucbp != 0 || macroblockInfo->vcbp != 0)){            
            best_mode = MODE_DIRECT;
            min_cost = cost;
            best_tr_size = TR_8x8;
        }
    }

    /********* NUMREF x 4 INTER MODES ******/
    min_cost += 1; //Just for bitexact match during cleanup work. Can be removed
#if DBG_COMPLEXITY
    TSC_START(TSC_ME_COARSE);
#endif
    macroblockInfo->ref = 0;
    cost = motion_estimate_coarse(encoderInfo,frameInfo,macroblockInfo);  
#if DBG_COMPLEXITY
    TSC_CUMUL(TSC_ME_COARSE);
#endif        

    if (num_ref>1){
        macroblockInfo->ref = 1;
        motion_estimate_coarse(encoderInfo,frameInfo,macroblockInfo);
    }
    else{
        if (pic_type == B_PICTURE){
            macroblockInfo->candidateMV[1][0].x = -macroblockInfo->candidateMV[0][0].x;
            macroblockInfo->candidateMV[1][0].y = -macroblockInfo->candidateMV[0][0].y;
        }
        else{
            macroblockInfo->candidateMV[1][0].x = macroblockInfo->candidateMV[0][0].x;
            macroblockInfo->candidateMV[1][0].y = macroblockInfo->candidateMV[0][0].y;
        }
    }

    mask[0][0] = mask[0][1] = mask[0][2] = mask[0][3] = 0;
    mask[1][0] = mask[1][1] = mask[1][2] = mask[1][3] = 0; 

    mask[2][0] = mask[2][1] = mask[2][2] = mask[2][3] = 0;

    for (ref = 0; ref < num_ref; ref++){
        for (mode = MODE_8x8; mode >= MODE_16x16; mode--){
            int mode_idx = mode-MODE_16x16;
            int mode_idx2;
            int flag = 0;

            /* If MV's of smaller sub-blocks are equal:
            1) Skip ME for the larger sub-blocks (flag=1)
            2) Omit encoding of the smaller sub-blocks (mask[][]=1)
            */               
            mode_idx2 = MODE_8x8-MODE_16x16;
            if (mode==MODE_16x8 && 
                compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][0][1]) && 
                compareMVs(MVarray[ref][mode_idx2][1][0],MVarray[ref][mode_idx2][1][1])){                
                    memcpy(MVarray[ref][mode_idx],MVarray[ref][mode_idx2],4*sizeof(mvpos));
                    flag = 1;
            }            
            if (mode==MODE_8x16 &&            
                compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][1][0]) &&
                compareMVs(MVarray[ref][mode_idx2][0][1],MVarray[ref][mode_idx2][1][1])){
                    memcpy(MVarray[ref][mode_idx],MVarray[ref][mode_idx2],4*sizeof(mvpos));
                    flag = 1;                                        
            }            
            if (mode==MODE_16x16 &&
                compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][1][0]) &&
                compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][0][1]) &&
                compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][1][1])){
                    memcpy(MVarray[ref][mode_idx],MVarray[ref][mode_idx2],4*sizeof(mvpos));
                    flag = 1;                            
            }
            mask[ref][mode_idx2] = flag;               

            mode_idx2 = MODE_16x8-MODE_16x16;
            if (mode==MODE_16x16 && compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][1][0])){
                memcpy(MVarray[ref][mode_idx],MVarray[ref][mode_idx2],4*sizeof(mvpos));
                flag = 1;
                mask[ref][mode_idx2] = 1;                                            
            }

            mode_idx2 = MODE_8x16-MODE_16x16;
            if (mode==MODE_16x16 && compareMVs(MVarray[ref][mode_idx2][0][0],MVarray[ref][mode_idx2][0][1])){
                memcpy(MVarray[ref][mode_idx],MVarray[ref][mode_idx2],4*sizeof(mvpos));
                flag = 1;
                mask[ref][mode_idx2] = 1;                  
            }

            if (!flag){
#if DBG_COMPLEXITY
                TSC_START(TSC_ME);
#endif 
                motion_estimate(encoderInfo,frameInfo,macroblockInfo,ref,mode);
#if DBG_COMPLEXITY
                TSC_CUMUL(TSC_ME);
#endif                   
                /* Save resulting for this ref and mode */
                MVarray[ref][mode_idx][0][0] = macroblockInfo->mvblock[0][0];
                MVarray[ref][mode_idx][0][1] = macroblockInfo->mvblock[0][1];
                MVarray[ref][mode_idx][1][0] = macroblockInfo->mvblock[1][0];
                MVarray[ref][mode_idx][1][1] = macroblockInfo->mvblock[1][1];
            }  // if (!flag)

        } //for mode
    } //for ref


    for (mode = MODE_16x16; mode <= MODE_8x8; mode++){        

        for (ref = 0; ref < num_ref+((pic_type == B_PICTURE && encoderInfo->params->bi ==1) ? 1:0); ref++){

            int mode_idx = mode-MODE_16x16;

            if(ref < 2)
            {
                macroblockInfo->mvblock[0][0] = MVarray[ref][mode_idx][0][0];
                macroblockInfo->mvblock[0][1] = MVarray[ref][mode_idx][0][1];
                macroblockInfo->mvblock[1][0] = MVarray[ref][mode_idx][1][0];
                macroblockInfo->mvblock[1][1] = MVarray[ref][mode_idx][1][1];
            }
            else
            {
                macroblockInfo->mvblock[0][0] = MVarray[0][mode_idx][0][0];
                macroblockInfo->mvblock[0][1] = MVarray[0][mode_idx][0][1];
                macroblockInfo->mvblock[1][0] = MVarray[0][mode_idx][1][0];
                macroblockInfo->mvblock[1][1] = MVarray[0][mode_idx][1][1];

                macroblockInfo->mvblockBW[0][0] = MVarray[1][mode_idx][0][0];
                macroblockInfo->mvblockBW[0][1] = MVarray[1][mode_idx][0][1];
                macroblockInfo->mvblockBW[1][0] = MVarray[1][mode_idx][1][0];
                macroblockInfo->mvblockBW[1][1] = MVarray[1][mode_idx][1][1];

            }


            macroblockInfo->mode = mode;
            macroblockInfo->ref = ref;
#if DBG_COMPLEXITY
            TSC_START(TSC_INTER_RDO);
#endif            

            if (!mask[ref][mode_idx]){
                for (tr = 0; tr < 3 ; tr++){
                    tr_size = tr_size_list[tr]; // check TR_8x8, TR_16x16, TR_4x4
                    if (tr_size==TR_8x8 || (tr_size==TR_16x16 && mode==MODE_16x16 && params->tr16flag&2) || (tr_size==TR_4x4 && params->tr4flag&2)){


                        if (frameInfo->modeCostNorm==2)

                            if(tr_size == TR_4x4 && macroblockInfo->svt_rd_diff > 0)
                            {
                                if(min_cost/macroblockInfo->svt_rd_diff < params->fast_tr4_thr1)
                                    continue;
                            }

                            memcpy(strTmp,encoderInfo->str,sizeof(stream));
                            memcpy(bitsTmp,bits,sizeof(bitcounter));
                            macroblockInfo->tr_size = tr_size;
                            process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);

                            tmp_cost = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

                            if (tmp_cost < min_cost){ 
                                best_ref = ref;                
                                best_mode = mode;
                                min_cost = tmp_cost;
                                best_tr_size = tr_size;                       
                                best_svt = macroblockInfo->svt;
                            }
                    }
                    if (tmp_cost<best_rdcost_for_svt[mode_idx]){
                        best_rdcost_for_svt[mode_idx] = tmp_cost;
                        best_ref_for_svt[mode_idx] = ref;
                    }
                    rdcost_stat[0][ref][mode] = tmp_cost;
                    if (tmp_cost < inter_cost){
                        inter_cost = tmp_cost;
                    }
                }
            }
#if DBG_COMPLEXITY
            TSC_CUMUL(TSC_INTER_RDO);
#endif                 
        }
    }


    /****************** EVALUATE INTRA MODES **************/


#if DBG_COMPLEXITY
    TSC_START(TSC_INTRA_RDO);
#endif
    macroblockInfo->mode = MODE_INTRA;
    macroblockInfo->ref = -1;
    intra_cost = (1<<30);
    for (tr = 0; tr < 3 ; tr++){
        tr_size = tr_size_list[tr]; // check TR_8x8, TR_16x16, TR_4x4
        if (tr_size==TR_8x8 || (tr_size==TR_16x16 && params->tr16flag&1) || (tr_size==TR_4x4 && params->tr4flag&1)){

            {  
                memcpy(strTmp,encoderInfo->str,sizeof(stream));
                memcpy(bitsTmp,bits,sizeof(bitcounter));
                macroblockInfo->tr_size = tr_size;
                process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);
                tmp_cost = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);
            }
            if (tmp_cost < min_cost){
                min_cost = tmp_cost;
                best_tr_size = tr_size;
                best_mode = MODE_INTRA;
            }
            if (tmp_cost < intra_cost){
                intra_cost = tmp_cost;                
            }
        }
    }
#if DBG_COMPLEXITY
    TSC_CUMUL(TSC_INTRA_RDO);
#endif
    rdcost_stat[0][0][1] = intra_cost;        


    macroblockInfo->tr_size = TR_8x8;

#if DBG_COMPLEXITY
    TSC_START(TSC_SVT_RDO);
#endif

#if SVT_SAD
    if (1)
#else
    if (frameInfo->modeCostNorm==2)
#endif
    {
        if (params->svt == 1) //cixunzhang
        {
            if ( ((inter_cost<intra_cost || skip_cost<intra_cost) && min_cost<skip_cost+macroblockInfo->svt_rd_diff) || (params->fast_svt==0))
            {
                for (svt_mode=MODE_16x16; svt_mode<MODE_8x8; svt_mode++)
                {
                    if(params->fast_svt > 0)
                    {
                        if (best_rdcost_for_svt[svt_mode-MODE_16x16]>=skip_cost+macroblockInfo->svt_rd_diff)
                            continue;              
                        if (best_rdcost_for_svt[svt_mode-MODE_16x16]>=inter_cost+macroblockInfo->svt_rd_diff)
                            continue;
                    }

                    for (svt_ref = 0; svt_ref < num_ref+((pic_type == B_PICTURE && encoderInfo->params->bi==1) ? 1:0); svt_ref++){

                        int psad_svt;
                        if (params->fast_svt > 0){
                            if (svt_ref != best_ref_for_svt[svt_mode-MODE_16x16] &&
                                rdcost_stat[0][svt_ref][svt_mode]>=best_rdcost_for_svt[svt_mode-MODE_16x16]+macroblockInfo->svt_rd_diff)
                                continue;
                        }

                        /* save MV for this candidate mode */

                        if(svt_ref < 2)
                        {
                            macroblockInfo->mvblock[0][0] = MVarray[svt_ref][svt_mode-MODE_16x16][0][0];
                            macroblockInfo->mvblock[0][1] = MVarray[svt_ref][svt_mode-MODE_16x16][0][1];
                            macroblockInfo->mvblock[1][0] = MVarray[svt_ref][svt_mode-MODE_16x16][1][0];
                            macroblockInfo->mvblock[1][1] = MVarray[svt_ref][svt_mode-MODE_16x16][1][1];
                        }
                        else
                        {
                            macroblockInfo->mvblock[0][0] = MVarray[0][svt_mode-MODE_16x16][0][0];
                            macroblockInfo->mvblock[0][1] = MVarray[0][svt_mode-MODE_16x16][0][1];
                            macroblockInfo->mvblock[1][0] = MVarray[0][svt_mode-MODE_16x16][1][0];
                            macroblockInfo->mvblock[1][1] = MVarray[0][svt_mode-MODE_16x16][1][1];

                            macroblockInfo->mvblockBW[0][0] = MVarray[1][svt_mode-MODE_16x16][0][0];
                            macroblockInfo->mvblockBW[0][1] = MVarray[1][svt_mode-MODE_16x16][0][1];
                            macroblockInfo->mvblockBW[1][0] = MVarray[1][svt_mode-MODE_16x16][1][0];
                            macroblockInfo->mvblockBW[1][1] = MVarray[1][svt_mode-MODE_16x16][1][1];
                        }

                        macroblockInfo->svt = 1;
                        macroblockInfo->mode = svt_mode;
                        macroblockInfo->ref = svt_ref;
                        process_mb(encoderInfo,frameInfo,macroblockInfo,strTmp,bitsTmp);
                        if (macroblockInfo->ycbp!=0){

                            psad_svt = cost_MB(macroblockInfo,macroblockInfo->num_bits,frameInfo->lambdaMode,frameInfo->modeCostNorm);

                            macroblockInfo->svt = (psad_svt < min_cost);
                            if (psad_svt < min_cost){
                                min_cost = psad_svt;
                                best_mode = svt_mode;
                                best_ref = svt_ref;
                                best_svt = 1;

                                best_tr_size = TR_8x8;

                            }
                            rdcost_stat[1][svt_ref][svt_mode] = psad_svt;
                        }
                    }
                } 
            }
        }
    }

#if DBG_COMPLEXITY
    TSC_CUMUL(TSC_SVT_RDO);
#endif

    /* Save final results of RDO mode selection */
    macroblockInfo->mode = best_mode;   
    macroblockInfo->svt = 0;
    if (best_mode==MODE_SKIP){
        macroblockInfo->mvblock[0][0] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[0][1] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[1][0] = macroblockInfo->skipmv;
        macroblockInfo->mvblock[1][1] = macroblockInfo->skipmv;
        macroblockInfo->ref = skip_ref;        
    }
    else if(best_mode==MODE_DIRECT){
        macroblockInfo->mvblock[0][0] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[0][1] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[1][0] = macroblockInfo->dirmv;
        macroblockInfo->mvblock[1][1] = macroblockInfo->dirmv;
        macroblockInfo->ref = 0;
    }
    else if(best_mode==MODE_INTRA){
        macroblockInfo->ref = -1;
        macroblockInfo->tr_size = best_tr_size;
    }
    else{         


        if(best_ref < 2)
        {
            macroblockInfo->mvblock[0][0] = MVarray[best_ref][best_mode-MODE_16x16][0][0];
            macroblockInfo->mvblock[0][1] = MVarray[best_ref][best_mode-MODE_16x16][0][1];
            macroblockInfo->mvblock[1][0] = MVarray[best_ref][best_mode-MODE_16x16][1][0];
            macroblockInfo->mvblock[1][1] = MVarray[best_ref][best_mode-MODE_16x16][1][1];
        }
        else
        {
            macroblockInfo->mvblock[0][0] = MVarray[0][best_mode-MODE_16x16][0][0];
            macroblockInfo->mvblock[0][1] = MVarray[0][best_mode-MODE_16x16][0][1];
            macroblockInfo->mvblock[1][0] = MVarray[0][best_mode-MODE_16x16][1][0];
            macroblockInfo->mvblock[1][1] = MVarray[0][best_mode-MODE_16x16][1][1];

            macroblockInfo->mvblockBW[0][0] = MVarray[1][best_mode-MODE_16x16][0][0];
            macroblockInfo->mvblockBW[0][1] = MVarray[1][best_mode-MODE_16x16][0][1];
            macroblockInfo->mvblockBW[1][0] = MVarray[1][best_mode-MODE_16x16][1][0];
            macroblockInfo->mvblockBW[1][1] = MVarray[1][best_mode-MODE_16x16][1][1];
        }

        macroblockInfo->ref = best_ref;
        macroblockInfo->svt = best_svt;
        macroblockInfo->tr_size = best_tr_size;
    }

    free(strTmp);
    free(bitsTmp);
    return 0;
}


